Package: DTRreg
Type: Package
Title: DTR Estimation and Inference via G-Estimation, Dynamic WOLS,
        Q-Learning, and Dynamic Weighted Survival Modeling (DWSurv)
Version: 2.2
Date: 2024-03-23
Author: Michael Wallace, Erica E M Moodie, David A Stephens, Gabrielle Simoneau, Shannon T. Holloway and Juliana Schulz
Maintainer: Shannon T. Holloway <shannon.t.holloway@gmail.com>
Description: Dynamic treatment regime estimation and inference via G-estimation, 
  dynamic weighted ordinary least squares (dWOLS) and Q-learning. Inference via 
  bootstrap and recursive sandwich estimation. Estimation and 
  inference for survival outcomes via Dynamic Weighted Survival Modeling (DWSurv). 
  Extension to continuous treatment variables. Wallace et al. (2017) 
  <DOI:10.18637/jss.v080.i02>; Simoneau et al. (2020) 
  <DOI:10.1080/00949655.2020.1793341>.
License: GPL-2
Imports: graphics, nnet, R6, stats, utils
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.1
NeedsCompilation: no
Repository: CRAN
Collate: 'treatmentClasses.R' 'Ahat.R' 'interactive.R'
        'inputProcessing.R' 'utils.R' 'weights.R' 'sandwich.R' 'gest.R'
        'dwols.R' 'bootstrapStep.R' 'dtrProcedure.R' 'DTRreg.R'
        'DWsurv.R' 'chooseM.R' 'data_generate.R' 's3-methods.R'
        'varest.R'
Packaged: 2024-03-23 20:05:47 UTC; 19194
Depends: R (>= 3.5.0)
Date/Publication: 2024-03-23 20:20:02 UTC
Built: R 4.2.3; ; 2024-03-23 21:31:24 UTC; unix

Package: DTRreg
Type: Package
Title: DTR Estimation and Inference via G-Estimation, Dynamic WOLS,
        Q-Learning, and Dynamic Weighted Survival Modeling (DWSurv)
Version: 2.4
Date: 2026-05-10
Author: Michael Wallace [aut],
  Erica E M Moodie [aut],
  David A Stephens [aut],
  Gabrielle Simoneau [aut],
  Shannon T. Holloway [aut, cre],
  Juliana Schulz [aut]
Maintainer: Shannon T. Holloway <shannon.t.holloway@gmail.com>
Authors@R: c(person(given = "Michael",
                      family = "Wallace",
                      role = "aut"),
               person(given = c("Erica", "E", "M"),
                      family = "Moodie",
                      role = "aut"),
               person(given = c("David", "A"),
                      family = "Stephens",
                      role = "aut"),
               person(given = "Gabrielle",
                      family = "Simoneau",
                      role = "aut"),
               person(given = c("Shannon", "T."),
                      family = "Holloway",
                      role = c("aut", "cre"),
                      email = "shannon.t.holloway@gmail.com"),
               person(given = "Juliana",
                      family = "Schulz",
                      role = "aut"))
Description: Dynamic treatment regime estimation and inference via G-estimation, 
  dynamic weighted ordinary least squares (dWOLS) and Q-learning. Inference via 
  bootstrap and recursive sandwich estimation. Estimation and 
  inference for survival outcomes via Dynamic Weighted Survival Modeling (DWSurv). 
  Extension to continuous treatment variables. Wallace et al. (2017) 
  <DOI:10.18637/jss.v080.i02>; Simoneau et al. (2020) 
  <DOI:10.1080/00949655.2020.1793341>.
License: GPL-2
Imports: ggplot2, ggplotify, graphics, nnet, R6, stats, utils
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Repository: CRAN
Collate: 'treatmentClasses.R' 'Ahat.R' 'varest.R' 'sandwich.R'
        'utils.R' 'weights.R' 'gest.R' 'dwols.R' 'dtrProcedure.R'
        'bootstrapStep.R' 'variance.R' 'interactive.R'
        'inputProcessing.R' 'DTRreg.R' 'DWsurv.R' 'chooseM.R'
        'data_generate.R' 's3-methods.R'
Depends: R (>= 4.1.0)
Packaged: 2026-05-13 16:04:10 UTC; 19194
Date/Publication: 2026-05-13 16:30:02 UTC
Built: R 4.6.0; ; 2026-05-13 20:09:38 UTC; unix

Package: DTRreg
Type: Package
Title: DTR Estimation and Inference via G-Estimation, Dynamic WOLS,
        Q-Learning, and Dynamic Weighted Survival Modeling (DWSurv)
Version: 2.4
Date: 2026-05-10
Author: Michael Wallace [aut],
  Erica E M Moodie [aut],
  David A Stephens [aut],
  Gabrielle Simoneau [aut],
  Shannon T. Holloway [aut, cre],
  Juliana Schulz [aut]
Maintainer: Shannon T. Holloway <shannon.t.holloway@gmail.com>
Authors@R: c(person(given = "Michael",
                      family = "Wallace",
                      role = "aut"),
               person(given = c("Erica", "E", "M"),
                      family = "Moodie",
                      role = "aut"),
               person(given = c("David", "A"),
                      family = "Stephens",
                      role = "aut"),
               person(given = "Gabrielle",
                      family = "Simoneau",
                      role = "aut"),
               person(given = c("Shannon", "T."),
                      family = "Holloway",
                      role = c("aut", "cre"),
                      email = "shannon.t.holloway@gmail.com"),
               person(given = "Juliana",
                      family = "Schulz",
                      role = "aut"))
Description: Dynamic treatment regime estimation and inference via G-estimation, 
  dynamic weighted ordinary least squares (dWOLS) and Q-learning. Inference via 
  bootstrap and recursive sandwich estimation. Estimation and 
  inference for survival outcomes via Dynamic Weighted Survival Modeling (DWSurv). 
  Extension to continuous treatment variables. Wallace et al. (2017) 
  <DOI:10.18637/jss.v080.i02>; Simoneau et al. (2020) 
  <DOI:10.1080/00949655.2020.1793341>.
License: GPL-2
Imports: ggplot2, ggplotify, graphics, nnet, R6, stats, utils
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Repository: CRAN
Collate: 'treatmentClasses.R' 'Ahat.R' 'varest.R' 'sandwich.R'
        'utils.R' 'weights.R' 'gest.R' 'dwols.R' 'dtrProcedure.R'
        'bootstrapStep.R' 'variance.R' 'interactive.R'
        'inputProcessing.R' 'DTRreg.R' 'DWsurv.R' 'chooseM.R'
        'data_generate.R' 's3-methods.R'
Depends: R (>= 4.1.0)
Packaged: 2026-05-13 16:04:10 UTC; 19194
Date/Publication: 2026-05-13 16:30:02 UTC
Built: R 4.6.0; ; 2026-05-13 17:28:25 UTC; unix

# Generated by roxygen2: do not edit by hand

S3method(autoplot,DTRreg)
S3method(coef,DTRreg)
S3method(confint,DTRreg)
S3method(plot,DTRreg)
S3method(predict,DTRreg)
S3method(print,DTRreg)
S3method(summary,DTRreg)
export(DTRreg)
export(DWSurv)
export(chooseM)
import(R6)
import(ggplot2)
importFrom(R6,is.R6)
importFrom(graphics,hist)
importFrom(nnet,multinom)
importFrom(stats,Gamma)
importFrom(stats,as.formula)
importFrom(stats,coef)
importFrom(stats,complete.cases)
importFrom(stats,confint)
importFrom(stats,family)
importFrom(stats,fitted)
importFrom(stats,gaussian)
importFrom(stats,get_all_vars)
importFrom(stats,glm)
importFrom(stats,lm)
importFrom(stats,model.frame)
importFrom(stats,model.matrix)
importFrom(stats,model.response)
importFrom(stats,na.pass)
importFrom(stats,predict)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,rbinom)
importFrom(stats,rnorm)
importFrom(stats,runif)
importFrom(stats,sd)
importFrom(stats,terms)
importFrom(stats,update.formula)
importFrom(stats,var)
importFrom(utils,capture.output)
importFrom(utils,tail)

#' Process treatment and predicted treatment; setup Binary variable
#' 
#' @noRd
#' @param A A vector. The treatment coded as 0/1.
#' @param data A data.frame object. The complete data for the participants
#'   under analysis in the stage.
#' @param tx.var A character. The treatment variable.
#' @param models A list object. The models.
#' 
#' @return A list object containing 
#'   `$A`, the observed stage treatments, 
#'   `$A.hat`, the predicted treatment probabilities,
#'   `$cts`, a character object indicating binary treatment
#'   `$cts.obj`, an R6 object of class Binary,
#'   `$treat.mod.fitted`, a fitted glm object if user did not provide 
#'     treatment probabilities.
#'   
#' @importFrom stats glm predict
#' @include treatmentClasses.R
#' @keywords internal
.Ahat_binary <- function(A, data, tx.var, models, ...) {
  
  stopifnot(
    "`A` must be a vector" = !missing(A) && is.vector(A) && length(A) > 1L,
    "`data` must be a data.frame" = !missing(data) && is.data.frame(data) &&
      nrow(data) == length(A),
    "`tx.var` must be a character" = !missing(tx.var) && is.character(tx.var),
    "`models` must be a list" = !missing(models) && is.list(models) &&
      "treat" %in% names(models)
  )
  
  # Ensure that A is 0/1; convert if needed
  if (!is.factor(A))  A <- as.factor(A)
  A <- {unclass(A) - 1L} |> as.integer()
  attributes(A) <- NULL
  
  cts_obj <- Binary$new(tf.model = models$tf, 
                        blip.model = models$blip, 
                        tx.var = tx.var)

  data[[tx.var]] <- A
  fit <- tryCatch(stats::glm(models$treat, data, family = "binomial"),
                  error = function(e) {
                    stop("unable to fit treatment model\n\t",
                         e$message, call. = FALSE)
                  })
    
  Ahat <- stats::predict(fit, type = "response")
  if (length(Ahat) != length(A)) {
    stop("A.hat predictions not returned for all cases; NAs likely present\n",
         "contact developer", call. = FALSE)
  }
  
  list("A" = A, 
       "A.hat" = Ahat,
       "cts" = "bin", 
       "cts.obj" = cts_obj, 
       "tx.mod.fitted" = fit)
}

#' Process treatment and predicted treatment; setup MulitNom variable
#' 
#' @noRd
#' @param A A vector. The treatment as given in the original data.frame.
#' @param data A data.frame object. The complete data for the participants
#'   under analysis in the stage.
#' @param tx.var A character. The treatment variable.
#' @param models A list object. The models.
#' 
#' @return A list object containing 
#'   `$A`, the observed stage treatments, 
#'   `$A.hat`, a matrix, the predicted treatment probabilities for all treatments,
#'   `$cts.obj`, an R6 object,
#'   `$cts`, a character object indicating multi-nomial treatment
#'   `$treat.mod.fitted`, a fitted multinom object if user did not provide 
#'     treatment probabilities.
#'   
#' @importFrom nnet multinom
#' @importFrom stats predict
#' @importFrom utils capture.output
#'
#' @include treatmentClasses.R
#'
#' @keywords internal
.Ahat_multinom <- function(A, data, tx.var, models, ...) {

  stopifnot(
    "`A` must be a vector" = !missing(A) && {is.vector(A) || is.factor(A)} && 
      length(A) > 1L,
    "`data` must be a data.frame" = !missing(data) && is.data.frame(data) &&
      nrow(data) == length(A),
    "`tx.var` must be a character" = !missing(tx.var) && is.character(tx.var),
    "`models` must be a list" = !missing(models) && 
      all(c("tf", "blip", "treat") %in% names(models))
    )
  
  # Ensure that A is a factor variable
  if (!is.factor(A)) A <- factor(A)
  
  levs <- levels(A)
  n_levels <- length(levs)
  if (n_levels < 3L) {
    stop("provided treatment variable is not multinomial", call. = FALSE)
  }
  
  cts_obj <- MultiNom$new(tf.model = models$tf,
                          blip.model = models$blip,
                          tx.var = tx.var, 
                          tx.levels = levs)
  
  data[[tx.var]] <- A
  utils::capture.output(fit <- tryCatch(
      suppressMessages(nnet::multinom(models$treat, data)),
      error = function(e) {
        stop("unable to fit treatment model\n\t",
             e$message, call. = FALSE)
        }))
  Ahat <- stats::predict(fit, type = "probs")
    
  if (nrow(Ahat) != length(A)) {
    stop("A.hat predictions not returned for all cases; NAs likely present\n",
          "contact developer", call. = FALSE)
  }
  if (ncol(Ahat) == {n_levels - 1L}) {
    Ahat <- cbind(1.0 - rowSums(Ahat), Ahat)
    colnames(Ahat) <- levs
  } else if (ncol(Ahat) != length(levels(A))) {
    stop("too few A.hat predictions returned; contact developer",
          call. = FALSE)
  }

  list("A" = A, 
       "A.hat" = Ahat,
       "cts" = "multinom", 
       "cts.obj" = cts_obj, 
       "tx.mod.fitted" = fit)
}

#' Process treatment and predicted treatment; setup Linear/Quadratic variable
#' 
#' @noRd
#' @param A A numeric vector object. The treatment as given in the original 
#'   data.frame. It is expected that participants that are not under analysis
#'   are removed.
#' @param data A data.frame object. The complete data for participants under
#'   analysis.
#' @param tx.var A character object. The treatment variable.
#' @param models A list object. The models.
#' @param tx.range A numeric vector. The min/max of the allowed treatment values.
#' @param tx.family A character object. Name of the family function passed to
#'   `glm` to fit treatment model.
#' 
#' @return A list object containing 
#'   `$A`, the observed stage treatments, 
#'   `$A.hat`, the predicted treatments,
#'   `$cts`, a character object indicating linear or quadratic blip
#'   `$cts.obj`, an R6 object,
#'   `$treat.mod.fitted`, a fitted lm object of user did not provide 
#'     treatment predictions.
#'
#' @importFrom stats glm predict
#'
#' @include treatmentClasses.R
#'
#' @keywords internal
.Ahat_cont <- function(A, data, tx.var, models, tx.range, tx.family, ...) {
  stopifnot(
    "`A` must be a vector" = !missing(A) && is.vector(A, mode = "numeric") && length(A) > 1L,
    "`data` must be a data.frame" = !missing(data) && is.data.frame(data) &&
      nrow(data) == length(A),
    "`tx.var` must be a character" = !missing(tx.var) && is.character(tx.var),
    "`models` must be a list" = !missing(models) && is.list(models) &&
      all(c("tf", "blip", "treat") %in% names(models)),
    "`tx.range` must be a numeric vector of length 2" = 
      is.numeric(tx.range) && is.vector(tx.range) && length(tx.range) == 2L,
    "`tx.family` must be a character in {'gaussian', 'Gamma'}" = 
      !missing(tx.family) && 
      {{is.character(tx.family) && tx.family %in% c("gaussian", "Gamma")} ||
      {inherits(tx.family, "family") && tx.family$family %in% c("gaussian", "Gamma")}}
  )
  
  data[[tx.var]] <- A
  fit <- tryCatch(stats::glm(models$treat, data, family = tx.family),
                  error = function(e) {
                    stop("unable to fit treatment model\n\t",
                         e$message, call. = FALSE)
                  })

  Ahat <- stats::predict(fit, type = "response")

  if (any(Ahat > tx.range[2L] | Ahat < tx.range[1L])) {
    warning("one or more predicted treatments are outside of allowed ",
            "treatment range; predicted treatment range: ", 
            format(range(Ahat), digits = 4L), "\n", 
            call. = FALSE)
  }
  
  psi_vars <- all.vars(models$blip)
  
  if (tx.var %in% psi_vars) {
    # continuous with quadratic blip
    cts_obj <- ContQuadraticBlip$new(tf.model = models$tf,
                                     blip.model = models$blip,
                                     treat.range = tx.range,
                                     tx.var = tx.var)
    cts <- "cts.q"
  } else {
    # continuous with linear blip
    cts_obj <- ContLinearBlip$new(tf.model = models$tf,
                                  blip.model = models$blip,
                                  treat.range = tx.range,
                                  tx.var = tx.var)
    cts <- "cts.l"
  }

  list("A" = A,
       "A.hat" = Ahat,
       "cts" = cts,
       "cts.obj" = cts_obj,
       "tx.mod.fitted" = fit)
}

#' Process stage treatment information
#' 
#' @noRd
#' @param stage.models A list object. Contains the 3 stage model named
#'   `$blip`, `$tf`, and `$treat`.
#' @param data A data.frame object. The complete covariate and treatment dataset
#'   for all participants under analysis in current step.
#' @param tx.var A character object. The column header of `data` pertaining to
#'   the treatment variable.
#' @param tx.range A numeric vector object of length 2. For continuous tx,
#'   the min and max of the allowed treatments. Ignored if not continuous tx.
#' @param tx.type A character object. Must be one of {"bin", "multi", "cont"}
#'   indicating if the treatments are binary, multinomial, or continuous,
#'   respectively.
#' 
#' @return A list object containing 
#'   `$A`, the observed tx variable (recoded if required); 
#'   `$A.hat`, the estimated or user specified tx probability; 
#'   `$treat.mod.fitted`, the regression object, 
#'   `$cts.obj` an R6 object
#'   `$cts`, a character object indicating the type of tx (bin/lin/quad); and
#'   Note that all vectors are of the length sum(stage.cases).
#'   
#' @keywords internal
.processTreatment <- function(stage.models, data, tx.var, tx.range, 
                              tx.type, tx.family) {
  
  stopifnot(
    "`stage.models` must be a list of formula" = !missing(stage.models) &&
      is.list(stage.models) && "treat" %in% names(stage.models) && 
      all(sapply(stage.models, inherits, what = "formula")),
    "`data` must be a data.frame" = !missing(data) && is.data.frame(data),
    "`tx.var` must be a character object" = !missing(tx.var) &&
      is.character(tx.var) && is.vector(tx.var) && length(tx.var) == 1L,
    "`tx.range` must be a numeric object or NULL" = !missing(tx.range) &&
      {any({is.null(tx.range) | is.na(tx.range)}) ||
      {is.numeric(tx.range) && is.vector(tx.range) && length(tx.range) == 2L}},
    "`tx.type` must be a character" = !missing(tx.type) && is.character(tx.type) &&
      tx.type %in% c("bin", "multi", "cont")  )
  
  # extract stage treatment variable
  A <- tryCatch(data[, tx.var],
                error = function(e) {
                  stop("unable to extract treatment variable ", 
                       tx.var, "\n\t", e$message, call. = FALSE)
                })
  
  if (tx.type == "bin") {
    
    # ensure that A is in fact binary
    # note that only the complete cases are provided, so NAs should not be
    # an issue here
    if (length(unique(A)) != 2L) {
      stop("treatment does not appear to be binary", call. = FALSE)
    }
    
    tx_info <- .Ahat_binary(A = A, 
                            data = data,
                            tx.var = tx.var,
                            models = stage.models)
    
  } else if (tx.type == "multi") {
    
    # ensure that 3 or more categories are present in the data
    # note that only the complete cases are provided, so NAs should not be
    # an issue here
    if (length(unique(A)) < 3L) {
      stop("treatment does not appear to be multinomial", call. = FALSE)
    }
    
    if (length(unique(A)) > 10L) {
      warning("a large number of treatment options identified; ",
              "verify that tx is multinomial\n", call. = FALSE)
    }
    
    tx_info <- .Ahat_multinom(A = A, 
                              data = data,
                              tx.var = tx.var,
                              models = stage.models)
    
  } else {
    
    tx_info <- .Ahat_cont(A = A, 
                          data = data,
                          tx.var = tx.var,
                          models = stage.models,
                          tx.range = tx.range,
                          tx.family = tx.family)
  } 
  tx_info
}

n <- 50L
set.seed(42L)

dat_bin <- data.frame(
  X  = rnorm(n),
  A  = rbinom(n, 1L, 0.5),
  Y  = rnorm(n)
)

dat_multi <- data.frame(
  X = rnorm(n),
  A = factor(sample(c("low", "med", "high"), n, replace = TRUE),
             levels = c("low", "med", "high")),
  Y = rnorm(n)
)

dat_cont <- data.frame(
  X = cos(seq(0.0, pi, length.out = n)),
  A = seq(0.1, 0.9, length.out = n),
  Y = rnorm(n)
)

# Stage model lists
models_bin <- list(
  blip  = ~X,
  treat = A ~ X,
  tf    = ~X
)

models_multi <- list(
  blip  = ~X,
  treat = A ~ X,
  tf    = ~X
)

# Quadratic blip — tx.var appears in blip model
models_cont_q <- list(
  blip  = ~A + X,
  treat = A ~ X,
  tf    = ~X
)

# Linear blip — tx.var does NOT appear in blip model
models_cont_l <- list(
  blip  = ~X,
  treat = A ~ X,
  tf    = ~X
)


test_that(".Ahat_binary: returns list with required element names", {
  res <- .Ahat_binary(dat_bin$A, dat_bin, "A", models_bin)
  expect_named(res, c("A", "A.hat", "cts", "cts.obj", "tx.mod.fitted"),
               ignore.order = TRUE)
})

test_that(".Ahat_binary: cts is 'bin'", {
  res <- .Ahat_binary(dat_bin$A, dat_bin, "A", models_bin)
  expect_equal(res$cts, "bin")
})

test_that(".Ahat_binary: cts.obj is a Binary R6 object", {
  res <- .Ahat_binary(dat_bin$A, dat_bin, "A", models_bin)
  expect_true(inherits(res$cts.obj, "Binary"))
})

test_that(".Ahat_binary: tx.mod.fitted is a glm object", {
  res <- .Ahat_binary(dat_bin$A, dat_bin, "A", models_bin)
  expect_true(inherits(res$tx.mod.fitted, "glm"))
})

test_that(".Ahat_binary: treatment model formula contains tx.var on LHS", {
  res <- .Ahat_binary(dat_bin$A, dat_bin, "A", models_bin)
  expect_equal(as.character(formula(res$tx.mod.fitted))[2L], "A")
})

test_that(".Ahat_binary: treatment model formula contains covariate on RHS", {
  res <- .Ahat_binary(dat_bin$A, dat_bin, "A", models_bin)
  expect_true("X" %in% names(coef(res$tx.mod.fitted)))
})

test_that(".Ahat_binary: A is recoded to 0/1 integer", {
  res <- .Ahat_binary(dat_bin$A, dat_bin, "A", models_bin)
  expect_true(all(res$A %in% c(0L, 1L)))
  expect_true(is.integer(res$A))
})

test_that(".Ahat_binary: A recoding oracle — non-0/1 binary maps correctly", {
  dat_coded <- dat_bin
  dat_coded$A <- dat_coded$A + 5L  # code as 5/6 instead of 0/1
  res <- .Ahat_binary(dat_coded$A, dat_coded, "A", models_bin)
  expect_true(all(res$A %in% c(0L, 1L)))
})

test_that(".Ahat_binary: A.hat is numeric vector of length n", {
  res <- .Ahat_binary(dat_bin$A, dat_bin, "A", models_bin)
  expect_true(is.numeric(res$A.hat))
  expect_length(res$A.hat, n)
})

test_that(".Ahat_binary: A.hat values are in (0, 1)", {
  res <- .Ahat_binary(dat_bin$A, dat_bin, "A", models_bin)
  expect_true(all(res$A.hat > 0.0 & res$A.hat < 1.0))
})

test_that(".Ahat_binary: stops when A is missing", {
  expect_error(.Ahat_binary(data = dat_bin, tx.var = "A", models = models_bin),
               "`A`")
})

test_that(".Ahat_binary: stops when A has length 1", {
  expect_error(.Ahat_binary(1L, dat_bin[1L, ], "A", models_bin), "`A`")
})

test_that(".Ahat_binary: stops when data is not a data.frame", {
  expect_error(.Ahat_binary(dat_bin$A, as.matrix(dat_bin), "A", models_bin),
               "`data`")
})

test_that(".Ahat_binary: stops when data nrow != length(A)", {
  expect_error(.Ahat_binary(dat_bin$A, dat_bin[1L:10L, ], "A", models_bin),
               "`data`")
})

test_that(".Ahat_binary: stops when tx.var is not a character", {
  expect_error(.Ahat_binary(dat_bin$A, dat_bin, 1L, models_bin), "`tx.var`")
})

test_that(".Ahat_binary: stops when models has no 'treat' element", {
  expect_error(.Ahat_binary(dat_bin$A, dat_bin, "A", list(blip = ~X)),
               "`models`")
})

test_that(".Ahat_binary: stops when treatment model formula references absent column", {
  bad_models <- list(blip = ~X, treat = A ~ Z, tf = ~X)
  expect_error(.Ahat_binary(dat_bin$A, dat_bin, "A", bad_models),
               "unable to fit treatment model")
})


test_that(".Ahat_multinom: returns list with required element names", {
  res <- suppressMessages(
    .Ahat_multinom(dat_multi$A, dat_multi, "A", models_multi)
  )
  expect_named(res, c("A", "A.hat", "cts", "cts.obj", "tx.mod.fitted"),
               ignore.order = TRUE)
})

test_that(".Ahat_multinom: cts is 'multinom'", {
  res <- suppressMessages(
    .Ahat_multinom(dat_multi$A, dat_multi, "A", models_multi)
  )
  expect_equal(res$cts, "multinom")
})

test_that(".Ahat_multinom: cts.obj is a MultiNom R6 object", {
  res <- suppressMessages(
    .Ahat_multinom(dat_multi$A, dat_multi, "A", models_multi)
  )
  expect_true(inherits(res$cts.obj, "MultiNom"))
})

test_that(".Ahat_multinom: treatment model formula contains tx.var on LHS", {
  res <- suppressMessages(
    .Ahat_multinom(dat_multi$A, dat_multi, "A", models_multi)
  )
  expect_equal(as.character(formula(res$tx.mod.fitted))[2L], "A")
})

test_that(".Ahat_multinom: A.hat is a matrix with n rows and n_levels columns", {
  res <- suppressMessages(
    .Ahat_multinom(dat_multi$A, dat_multi, "A", models_multi)
  )
  expect_true(is.matrix(res$A.hat))
  expect_equal(nrow(res$A.hat), n)
  expect_equal(ncol(res$A.hat), 3L)
})

test_that(".Ahat_multinom: A.hat column names match treatment levels", {
  res <- suppressMessages(
    .Ahat_multinom(dat_multi$A, dat_multi, "A", models_multi)
  )
  expect_equal(colnames(res$A.hat), levels(dat_multi$A))
})

test_that(".Ahat_multinom: A.hat rows sum to 1", {
  res <- suppressMessages(
    .Ahat_multinom(dat_multi$A, dat_multi, "A", models_multi)
  )
  expect_equal(unname(rowSums(res$A.hat)), rep(1.0, n), tolerance = 1e-10)
})

test_that(".Ahat_multinom: stops when A has fewer than 3 levels", {
  A_bin <- factor(rbinom(n, 1L, 0.5))
  expect_error(.Ahat_multinom(A_bin, dat_bin, "A", models_multi),
               "not multinomial")
})

test_that(".Ahat_multinom: stops when data nrow != length(A)", {
  expect_error(
    suppressMessages(
      .Ahat_multinom(dat_multi$A, dat_multi[1L:10L, ], "A", models_multi)
    ),
    "`data`"
  )
})

test_that(".Ahat_multinom: stops when tx.var is not a character", {
  expect_error(
    suppressMessages(.Ahat_multinom(dat_multi$A, dat_multi, 1L, models_multi)),
    "`tx.var`"
  )
})

test_that(".Ahat_multinom: stops when models missing required elements", {
  expect_error(
    suppressMessages(
      .Ahat_multinom(dat_multi$A, dat_multi, "A", list(blip = ~X))
    ),
    "`models`"
  )
})

test_that(".Ahat_cont: returns list with required element names", {
  res <- .Ahat_cont(dat_cont$A, dat_cont, "A", models_cont_q,
                    tx.range = c(0.1, 0.9), tx.family = gaussian())
  expect_named(res, c("A", "A.hat", "cts", "cts.obj", "tx.mod.fitted"),
               ignore.order = TRUE)
})

test_that(".Ahat_cont: cts is 'cts.q' when tx.var appears in blip model", {
  res <- .Ahat_cont(dat_cont$A, dat_cont, "A", models_cont_q,
                    tx.range = c(0.1, 0.9), tx.family = gaussian())
  expect_equal(res$cts, "cts.q")
})

test_that(".Ahat_cont: cts.obj is ContQuadraticBlip when tx.var in blip model", {
  res <- .Ahat_cont(dat_cont$A, dat_cont, "A", models_cont_q,
                    tx.range = c(0.1, 0.9), tx.family = gaussian())
  expect_true(inherits(res$cts.obj, "ContQuadraticBlip"))
})

test_that(".Ahat_cont: tx.mod.fitted is a glm object", {
  res <- .Ahat_cont(dat_cont$A, dat_cont, "A", models_cont_q,
                    tx.range = c(0.1, 0.9), tx.family = gaussian())
  expect_true(inherits(res$tx.mod.fitted, "glm"))
})

test_that(".Ahat_cont: treatment model LHS is tx.var", {
  res <- .Ahat_cont(dat_cont$A, dat_cont, "A", models_cont_q,
                    tx.range = c(0.1, 0.9), tx.family = gaussian())
  expect_equal(as.character(formula(res$tx.mod.fitted))[2L], "A")
})

test_that(".Ahat_cont: treatment model RHS contains covariate", {
  res <- .Ahat_cont(dat_cont$A, dat_cont, "A", models_cont_q,
                    tx.range = c(0.1, 0.9), tx.family = gaussian())
  expect_true("X" %in% names(coef(res$tx.mod.fitted)))
})

test_that(".Ahat_cont: A.hat is numeric vector of length n", {
  res <- .Ahat_cont(dat_cont$A, dat_cont, "A", models_cont_q,
                    tx.range = c(0.1, 0.9), tx.family = gaussian())
  expect_true(is.numeric(res$A.hat))
  expect_length(res$A.hat, n)
})

test_that(".Ahat_cont: warns when predicted treatments fall outside tx.range", {
  expect_warning(
    .Ahat_cont(dat_cont$A, dat_cont, "A", models_cont_q,
               tx.range = c(0.5, 0.51), tx.family = gaussian()),
    "outside of allowed"
  )
})

test_that(".Ahat_cont: stops when tx.var not in blip model (linear blip not supported)", {
  expect_error(
    .Ahat_cont(dat_cont$A, dat_cont, "A", models_cont_l,
               tx.range = c(0.1, 0.9), tx.family = gaussian()),
    "not supported"
  )
})

test_that(".Ahat_cont: stops when A is not a vector", {
  expect_error(
    .Ahat_cont(as.list(dat_cont$A), dat_cont, "A", models_cont_q,
               tx.range = c(0.1, 0.9), tx.family = gaussian()),
    "`A`"
  )
})

test_that(".Ahat_cont: stops when tx.range is wrong length", {
  expect_error(
    .Ahat_cont(dat_cont$A, dat_cont, "A", models_cont_q,
               tx.range = c(0.1, 0.5, 0.9), tx.family = gaussian()),
    "tx.range"
  )
})

test_that(".Ahat_cont: stops when tx.family is not gaussian or Gamma", {
  expect_error(
    .Ahat_cont(dat_cont$A, dat_cont, "A", models_cont_q,
               tx.range = c(0.1, 0.9), tx.family = "poisson"),
    "tx.family"
  )
})

test_that(".Ahat_cont: stops when treatment model references absent column", {
  bad_models <- list(blip = ~A + X, treat = A ~ Z, tf = ~X)
  expect_error(
    .Ahat_cont(dat_cont$A, dat_cont, "A", bad_models,
               tx.range = c(0.1, 0.9), tx.family = gaussian()),
    "unable to fit treatment model"
  )
})

pt <- function(stage.models = models_bin,
               data         = dat_bin,
               tx.var       = "A",
               tx.range     = NA_real_,
               tx.type      = "bin",
               tx.family    = gaussian()) {
  .processTreatment(stage.models, data, tx.var, tx.range, tx.type, tx.family)
}

test_that(".processTreatment: dispatches to binary path — cts is 'bin'", {
  res <- pt()
  expect_equal(res$cts, "bin")
})

test_that(".processTreatment: dispatches to multinomial path — cts is 'multinom'", {
  res <- suppressMessages(
    pt(stage.models = models_multi, data = dat_multi, tx.type = "multi")
  )
  expect_equal(res$cts, "multinom")
})

test_that(".processTreatment: dispatches to continuous path — cts is 'cts.q'", {
  res <- pt(stage.models = models_cont_q, data = dat_cont,
            tx.range = c(0.1, 0.9), tx.type = "cont")
  expect_equal(res$cts, "cts.q")
})

test_that(".processTreatment: errors as expected", {
  expect_error(pt(stage.models = list(treat = "A ~ X", blip = ~X, tf = ~X)),
               "stage.models")
  expect_error(pt(data = as.matrix(dat_bin)), "`data`")
  expect_error(pt(tx.var = c("A", "B")), "tx.var")
  expect_error(pt(tx.var = 1L),          "tx.var")
  expect_error(pt(tx.type = "ordered"), "tx.type")
  expect_error(pt(tx.var = "Z"), "treatment variable")
  
  dat_one       <- dat_bin
  dat_one$A     <- 0L
  expect_error(pt(data = dat_one), "binary")

  dat_bin2 <- dat_bin
  dat_bin2$A <- rbinom(n, 1L, 0.5)
  expect_error(
    pt(stage.models = models_multi, data = dat_bin2, tx.type = "multi"),
    "multinomial"
  )

  dat_many   <- dat_bin
  dat_many$A <- factor(sample(letters[1L:12L], n, replace = TRUE))
  models_many <- list(blip = ~X, treat = A ~ X, tf = ~X)
  expect_warning(
    suppressMessages(
      pt(stage.models = models_many, data = dat_many, tx.type = "multi")
    ),
    "large number"
  )
})

#' Estimate parameters using dynamic weighted ordinary least squares
#'   (only appropriate for binary treatments)
#'   
#' It is expected that the dimensions of the inputs are limited to only the
#'   cases included in the stage analysis.
#'   
#' @noRd
#' @param Y A numeric vector object of length n. The outcome of interest.
#' @param A A vector object of length n. The stage treatment
#' @param data A data.frame object. The full covariate set.
#' @param wgts A numeric vector object of length n. The complete case + tx weights.
#' @param cts.obj An R6 object defining treatment + blip specific functions.
#' @param tx.var A character object. The treatment variable name.
#' 
#' @return A list object. Element `$outcome.fit` is the lm object; element
#'   `$beta` is the parameter estimates of the treatment-free component of
#'   the model; and `$psi` is the parameter estimates of the blip component.
#'   
#' @importFrom R6 is.R6
#' @importFrom stats coef lm terms
#' 
#' @include treatmentClasses.R
#' 
#' @keywords internal
.dwols <- function(Y, A, data, wgts, cts.obj, tx.var, ...) {
  
  stopifnot(
    "`Y` must be a numeric vector object" = !missing(Y) &&
      is.numeric(Y) && is.vector(Y) && length(Y) > 1L,
    "`A` must be a vector object" = !missing(A) && {is.vector(A) || is.factor(A)} && 
      length(A) == length(Y),
    "`data` must be a data.frame" = !missing(data) && is.data.frame(data) &&
      nrow(data) == length(Y),
    "`wgts` must be a numeric vector object with length equivalent to that of `Y`" = 
      !missing(wgts) &&
      is.numeric(wgts) && is.vector(wgts) && length(wgts) == length(Y),
    "`cts.obj` must be an R6 object" = !missing(cts.obj) && R6::is.R6(cts.obj),
    "`tx.var` must be a character" = !missing(tx.var) && is.character(tx.var)
  )
  
  y_name <- "Y"
  while (y_name %in% colnames(data)) 
    y_name <- paste("Y", sample(.Machine$integer.max, size = 1L), sep = "_")

  # update the outcome model to include the internally defined outcome
  full_model <- stats::reformulate(attr(stats::terms(cts.obj$full.model), "term.labels"),
                                   response = y_name)

  # add internally defined outcome and stage treatment to analysis data.frame
  data[[y_name]] <- Y
  data[[tx.var]] <- A

  # if a quadratic blip function, update the internal
  # I(a^2) variable with the square of the current treatment variable value
  data <- cts.obj$prep(data, A)
  
  fit <- tryCatch(stats::lm(full_model, data, weights = wgts),
                  error = function(e) {
                    stop("unable to complete DWOLS\n\t",
                         e$message, call. = FALSE)
                    })

  # need blip terms to extract appropriate elements of the parameter estimates
  est <- stats::coef(fit)

  blip_vars <- cts.obj$blip_params(est)

  psi <- est[blip_vars]
  names(psi) <- names(blip_vars)
  
  list("outcome.fit" = fit,
       "beta" = est[-blip_vars],
       "psi" = psi)
}

n <- 30L
set.seed(42L)

dat <- data.frame(
  X = cos(seq(0.0, pi, length.out = n)),
  A = rep(0L:1L, times = n / 2L),
  Y = rnorm(n)
)

wgts <- rep(1.0, n)

models_bin <- list(blip = ~X, treat = A ~ X, tf = ~X)

cts_bin <- Binary$new(tf.model = models_bin$tf,
                      blip.model = models_bin$blip,
                      tx.var = "A")

# Continuous quadratic setup
dat_cont <- data.frame(
  X = cos(seq(0.0, pi, length.out = n)),
  A = seq(0.1, 0.9, length.out = n),
  Y = rnorm(n)
)

models_cont <- list(blip = ~A + X, treat = A ~ X, tf = ~X)

cts_cont <- ContQuadraticBlip$new(tf.model = models_cont$tf,
                                  blip.model = models_cont$blip,
                                  tx.var = "A",
                                  treat.range = c(0.1, 0.9))


local({
  res <- .dwols(Y = dat$Y,
                A = dat$A,
                data = dat,
                wgts = wgts,
                cts.obj = cts_bin,
                tx.var = "A")
  
  test_that(".dwols: returns list with required element names", {
    expect_named(res, c("outcome.fit", "beta", "psi"), ignore.order = TRUE)
  })
  
  test_that(".dwols: outcome.fit is an lm object", {
    expect_true(inherits(res$outcome.fit, "lm"))
  })
  
  test_that(".dwols: psi contains only blip-related terms (contains tx.var)", {
    expect_true(all(grepl("A", names(res$psi))))
  })
  
  test_that(".dwols: beta contains no blip-related terms (no tx.var)", {
    expect_false(any(grepl("^A", names(res$beta))))
  })
  
  test_that(".dwols: psi and beta together account for all coefficients", {
    all_coef_names <- names(coef(res$outcome.fit))
    expect_setequal(c(names(res$psi), names(res$beta)), all_coef_names)
  })
  
  test_that(".dwols: outcome.fit formula LHS is not 'Y' when Y column exists in data", {
    lhs <- as.character(formula(res$outcome.fit))[2L]
    expect_false(lhs == "Y")
  })
  
  test_that(".dwols: outcome.fit formula RHS contains tx.var", {
    rhs_terms <- attr(terms(formula(res$outcome.fit)), "term.labels")
    expect_true(any(grepl("^A", rhs_terms)))
  })
  
  test_that(".dwols: outcome.fit formula RHS contains tf covariate", {
    rhs_terms <- attr(terms(formula(res$outcome.fit)), "term.labels")
    expect_true("X" %in% rhs_terms)
  })
})

test_that(".dwols: handles collision when 'Y' already exists as column in data", {
  dat_collision <- dat
  # "Y" is already in colnames — function must use a different internal name
  expect_no_error(
    .dwols(Y = dat$Y, A = dat$A, data = dat_collision,
           wgts = wgts, cts.obj = cts_bin, tx.var = "A")
  )
})

local({
  res <- .dwols(Y = dat_cont$Y,
                A = dat_cont$A,
                data = dat_cont,
                wgts = wgts,
                cts.obj = cts_cont,
                tx.var = "A")
  
  test_that(".dwols: ContQuadraticBlip — psi contains both linear and quadratic tx terms", {
    expect_true("A" %in% names(res$psi))
    expect_true(length(res$psi) >= 2L)
  })
  
  test_that(".dwols: ContQuadraticBlip — squared tx column appears in outcome.fit terms", {
    fit_terms <- attr(terms(formula(res$outcome.fit)), "term.labels")
    sq_col <- setdiff(names(res$outcome.fit$model),
                      c("X", "A", as.character(formula(res$outcome.fit))[2L]))
    expect_true(length(sq_col) >= 1L)
  })
  
  test_that(".dwols: ContQuadraticBlip — beta contains no tx.var terms", {
    expect_false(any(grepl("^A", names(res$beta))))
  })
})

test_that(".dwols: errors as expected", {
  expect_error(
    .dwols(A = dat$A, data = dat, wgts = wgts, cts.obj = cts_bin, tx.var = "A"),
    "`Y`"
  )
  expect_error(
    .dwols(Y = as.character(dat$Y), A = dat$A, data = dat,
           wgts = wgts, cts.obj = cts_bin, tx.var = "A"),
    "`Y`"
  )
  expect_error(
    .dwols(Y = 1.0, A = dat$A[1L], data = dat[1L, ], wgts = wgts[1L],
           cts.obj = cts_bin, tx.var = "A"),
    "`Y`"
  )
  expect_error(
    .dwols(Y = dat$Y, A = dat$A[1L:5L], data = dat,
           wgts = wgts, cts.obj = cts_bin, tx.var = "A"),
    "`A`"
  )
  expect_error(
    .dwols(Y = dat$Y, A = dat$A, data = as.matrix(dat),
           wgts = wgts, cts.obj = cts_bin, tx.var = "A"),
    "`data`"
  )
  expect_error(
    .dwols(Y = dat$Y, A = dat$A, data = dat[1L:5L, ],
           wgts = wgts, cts.obj = cts_bin, tx.var = "A"),
    "`data`"
  )
  expect_error(
    .dwols(Y = dat$Y, A = dat$A, data = dat,
           wgts = as.character(wgts), cts.obj = cts_bin, tx.var = "A"),
    "`wgts`"
  )
  expect_error(
    .dwols(Y = dat$Y, A = dat$A, data = dat,
           wgts = wgts[1L:5L], cts.obj = cts_bin, tx.var = "A"),
    "`wgts`"
  )
  expect_error(
    .dwols(Y = dat$Y, A = dat$A, data = dat,
           wgts = wgts, cts.obj = list(), tx.var = "A"),
    "`cts.obj`"
  )
  expect_error(
    .dwols(Y = dat$Y, A = dat$A, data = dat,
           wgts = wgts, cts.obj = cts_bin, tx.var = 1L),
    "`tx.var`"
  )
  # Pass a cts.obj whose full.model references a column not in data
  cts_bad <- Binary$new(tf.model = ~Z, blip.model = ~X, tx.var = "A")
  expect_error(
    .dwols(Y = dat$Y, A = dat$A, data = dat,
           wgts = wgts, cts.obj = cts_bad, tx.var = "A"),
    "unable to complete DWOLS"
  )
})

#' Specialized function for Binary treatments
#' 
#' @noRd
#' @import R6
#' @keywords internal
Binary <- R6::R6Class(
  "Binary",
  public = list(
    blip.model = NULL,
    # the full outcome model
    full.model = NULL,
    
    initialize = function(tf.model, blip.model, tx.var, ...) {
        private$tx.var <- tx.var
        
        # add the treatment variable dependence to the blip formula
        blip.model.cov <- attr(stats::terms(blip.model), "term.labels")
        if (length(blip.model.cov) > 0L) {
          int.terms <- paste(tx.var, blip.model.cov, sep = ":", collapse = "+")
          self$blip.model <- paste("~ - 1", tx.var, int.terms, sep = " + ") |> as.formula()
        } else {
          self$blip.model <- paste("~ - 1 + ", tx.var) |> as.formula()
        }

        tf_vars <- paste(attr(stats::terms(tf.model), "term.labels"), collapse = "+")
        blip_vars <- paste(attr(stats::terms(self$blip.model), "term.labels"), 
                           collapse = "+")
        
        self$full.model <- paste("~", tf_vars, "+", blip_vars) |> as.formula()
    },
    
    blip_params = function(coefs) {
      possible_names <- private$tx.var
      nms <- names(coefs) |> strsplit(":")
      nms_list <- lapply(nms,
                         function(x) {
                           any(x %in% possible_names)
                         }) |> unlist()
      blip_vars <- which(nms_list)
      names(blip_vars) <- names(coefs)[blip_vars]
      
      blip_vars
    },
    
    ipw = function(A, A.hat, ...) {
      1.0 / {A * A.hat + {1.0 - A} * {1.0 - A.hat}}
    }, 
    
    .pom = function(...) {
      stop(".pom() is not appropriate for binary treatments", call. = FALSE)
    },
    
    Hd = function(data, A, ...) {
      data[, private$tx.var] <- A
      model.matrix(self$full.model, data)
    },
    
    Hpsi = function(data, A, ...) {
      data[, private$tx.var] <- A
      model.matrix(self$blip.model, data)
    },
    
    Hw = function(data, A, A.hat, wgt, ...) {
        data[, private$tx.var] <- {A - A.hat} * wgt
        model.matrix(self$full.model, data)
      },
    
    # If DTR, add regret function; otherwise, subtract stage blip at observed
    # treatment
    shiftY = function(type, outcome.fit, data, opt, A, ...) {
        if (type == "DTR") {
          self$regret(outcome.fit, data, opt, A)
        } else {
          # setting optimal treatment to 0 is equivalent to stage blip at
          # observed treatment
          self$regret(outcome.fit, data, 0L, A)
        }
      },
    
    regret = function(outcome.fit, data, opt, A) {
      # predicted outcome at optimal treatment
      data[, private$tx.var] <- opt
      at_opt <- predict(outcome.fit, data)
      
      # predicted outcome at observed treatment
      data[, private$tx.var] <- A
      at_A <- predict(outcome.fit, data)
      
      at_opt - at_A
    },
    
    # return optimal treatment
    opt = function(outcome.fit, data, ...) {
        data[, private$tx.var] <- 1L
        y_1 <- predict(outcome.fit, data)
        data[, private$tx.var] <- 0L
        y_0 <- predict(outcome.fit, data)
        as.integer({y_1 - y_0} > 1e-8) |> drop()
      },
    
    prep = function(data, A) data
    
  ),
  private = list(
    tx.var = NULL
  )
)

#' Specialized function for Multinomial treatments
#' 
#' @noRd
#' @keywords internal
MultiNom <- R6::R6Class(
  "MultiNom",
  public = list(
    blip.model = NULL,
    # the full outcome model
    full.model = NULL,

    initialize = function(tf.model, blip.model, tx.var, tx.levels, ...) {
      private$tx.var <- tx.var
      private$tx.levels <- tx.levels
      
      # add the treatment variable dependence to the blip formula
      blip.model.cov <- attr(stats::terms(blip.model), "term.labels")
      if (length(blip.model.cov) > 0L) {
        int.terms <- paste(tx.var, blip.model.cov, sep = ":", collapse = "+")
        self$blip.model <- paste("~ - 1", tx.var, int.terms, sep = " + ") |> as.formula()
      } else {
        self$blip.model <- paste("~ - 1 + ", tx.var) |> as.formula()
      }
      
      tf_vars <- paste(attr(stats::terms(tf.model), "term.labels"), collapse = "+")
      blip_vars <- paste(attr(stats::terms(self$blip.model), "term.labels"), 
                         collapse = "+")
      
      self$full.model <- paste("~", tf_vars, "+", blip_vars) |> as.formula()
    },
    
    blip_params = function(coefs) {
      possible_names <- paste0(private$tx.var, private$tx.levels[-1L])
      nms <- names(coefs) |> strsplit(":")
      nms_list <- lapply(nms,
                         function(x) {
                           any(x %in% possible_names)
                         }) |> unlist()
      blip_vars <- which(nms_list)
      names(blip_vars) <- names(coefs)[blip_vars]

      blip_vars
    },
    
    ipw = function(A, A.hat, ...) {
      idx <- match(A, private$tx.levels)
      deno <- A.hat[cbind(seq_len(length(A)), idx)]
      1.0 / drop(deno)
    }, 
    
    .pom = function(A, tx.mod.fitted, data, m) {
      
      if (all(is.null(tx.mod.fitted)) || all(is.na(tx.mod.fitted))) {
        stop("`tx.weight` cannot be 'qpom' if treatment is not modelled",
             call. = FALSE)
      }
      X_alpha <- stats::model.matrix(tx.mod.fitted, 
                                     stats::model.frame(tx.mod.fitted$terms, data))

      a_cat <- A

      if (ncol(X_alpha) == 1L && attr(terms(tx.mod.fitted), "intercept") == 1L) {
        pom <- MASS::polr(a_cat ~ 1)
      } else{
        pom <- MASS::polr(a_cat ~ X_alpha[, -1L])
      }
      A <- unclass(A)
      
      prob <- pom$fitted.values[cbind(seq_along(A), A)]
      sum_prob <- rowSums(1.0 / pom$fitted.values)
      
      list("prob" = prob, "sum.prob" = sum_prob)
    },
    
    Hd = function(data, A, ...) {
      data[, private$tx.var] <- A
      model.matrix(self$full.model, data)
    },
    
    Hpsi = function(data, A, ...) {
      stop("Multinomial with g-estimation is not yet supported",
           call. = FALSE)
    },
    
    Hw = function(data, A, A.hat, wgt, ...) {
      stop("Multinomial with g-estimation is not yet supported",
           call. = FALSE)
    },
    
    # If DTR, add regret function; otherwise, subtract stage blip at observed
    # treatment
    shiftY = function(type, outcome.fit, data, opt, A, ...) {
      if (type == "DTR") {
        self$regret(outcome.fit, data, opt, A)
      } else {
        # setting optimal treatment to base level is equivalent to stage blip at
        # observed treatment
        self$regret(outcome.fit, data, private$tx.levels[1L], A)
      }
    },
    
    regret = function(outcome.fit, data, opt, A, ...) {
      # predict outcome at optimal level
      data[, private$tx.var] <- factor(opt, levels = private$tx.levels)
      at_opt <- predict(outcome.fit, data)
      
      # predict outcome at observed level
      data[, private$tx.var] <- factor(A, levels = private$tx.levels)
      at_A <- predict(outcome.fit, data)
      
      at_opt - at_A
    },
    
    # estimate optimal treatment
    opt = function(outcome.fit, data, ...) {
      est_Y <- matrix(0.0, nrow = nrow(data), ncol = length(private$tx.levels))
      for (i in seq_along(private$tx.levels)) {
        data[, private$tx.var] <- private$tx.levels[i]
        est_Y[, i] <- predict(outcome.fit, data)
      }
      max_tx <- apply(est_Y, 1L, which.max)
      factor(private$tx.levels[max_tx], levels = private$tx.levels)
    },
    
    prep = function(data, A) data
  ),
  private = list(
    tx.var = NULL,
    tx.levels = NULL
  )
)

#' Specialized function for Continuous treatment with linear blip model
#' 
#' @noRd
#' @keywords internal
ContLinearBlip <- R6::R6Class(
  "ContLinearBlip",
  public = list(
    blip.model = NULL,
    # the full outcome model
    full.model = NULL,
    
    initialize = function(tf.model, blip.model, tx.var, treat.range) {
      
      stop("linear blip functions are not supported", call. = FALSE)
      
    },
    
    blip_params = function(coefs) {
      stop("linear blip functions are not supported", call. = FALSE)
    },
    
    ipw = function(A, tx.mod.fitted, ...) {
      stop("linear blip functions are not supported", call. = FALSE)
    }, 
    
    Hd = function(data, A, ...) {
      stop("linear blip functions are not supported", call. = FALSE)
    },
    
    Hpsi = function(data, A, ...) {
      stop("linear blip functions are not supported", call. = FALSE)
    },
    
    Hw = function(data, A, A.hat, wgt, ...) {
      stop("linear blip functions are not supported", call. = FALSE)
    },
    
    # If DTR, add regret function; otherwise, subtract stage blip at observed
    # treatment
    shiftY = function(type, outcome.fit, data, opt, A, ...) {
      stop("linear blip functions are not supported", call. = FALSE)
    },
    
    regret = function(outcome.fit, data, opt, A) {
      stop("linear blip functions are not supported", call. = FALSE)
    },
    
    # estimate optimal treatment
    opt = function(outcome.fit, data, ...) {
      stop("linear blip functions are not supported", call. = FALSE)
    },
    
    prep = function(data, A) stop("linear blip functions are not supported", call. = FALSE)
  ),
  private = list(
    min.tx = NULL,
    max.tx = NULL,
    tx.var = NULL
  )
) 

#' Specialized function for Continuous treatment with quadratic blip model
#' 
#' @noRd
#' @keywords internal
ContQuadraticBlip <- R6::R6Class(
  "ContQuadraticBlip",
  public = list(
    blip.model = NULL,
    # the full outcome model
    full.model = NULL,
    
    initialize = function(tf.model, blip.model, tx.var, treat.range) {
        private$max.tx <- max(treat.range)
        private$min.tx <- min(treat.range)
        private$tx.var <- tx.var
        private$tx.var.b <- paste0(".__", tx.var, "_sq__.")
        
        # need to identify quad and linear
        # this should never be an intercept only model
        orig_vars <- attr(stats::terms(blip.model), "term.labels")
        
        if (length(orig_vars) == 0L) {
          stop("cannot provide an intercept only model for quadratic blip",
               call. = FALSE)
        }
        
        mod_vars <- sapply(orig_vars,
                           function(x) {
                             elems <- strsplit(x, ":")[[1L]]
                             if (any(elems == tx.var)) {
                               # any terms that includes tx.var is a quadratic
                               # convert I(a^2) to a new variable to facilitate
                               # later methods
                               elems[elems == tx.var] <- private$tx.var.b
                             } else {
                               # add treatment as interaction term
                               elems <- c(tx.var, elems)
                             }
                             paste(elems, collapse = ":")
                           })

        self$blip.model <- paste("~ -1 +", tx.var, "+",
                                 paste(mod_vars, collapse = "+")) |> 
          as.formula()

        tf_vars <- paste(attr(stats::terms(tf.model), "term.labels"), collapse = "+")
        blip_vars <- paste(attr(stats::terms(self$blip.model), "term.labels"), 
                           collapse = "+")
        self$full.model <- paste("~", tf_vars, "+", blip_vars) |> as.formula()
    },
    
    blip_params = function(coefs) {
      possible_names <- c(private$tx.var, private$tx.var.b)
      nms <- names(coefs) |> strsplit(":")
      nms_list <- lapply(nms,
                         function(x) {
                           any(x %in% possible_names)
                         }) |> unlist()
      blip_vars <- which(nms_list)
      names(blip_vars) <- names(coefs)[blip_vars]
      
      blip_vars
    },

    ipw = function(A, tx.mod.fitted, ...) {
      dispersion <- summary(tx.mod.fitted)$dispersion

      if (tx.mod.fitted$family$family == "gaussian") {
        1.0 / dnorm(A, mean = tx.mod.fitted$fitted.values, 
                    sd = sqrt(dispersion))
      } else if (tx.mod.fitted$family$family == "Gamma") {
        1.0 / dgamma(A, shape = 1.0 / dispersion,
                     scale = tx.mod.fitted$fitted.values * dispersion)
      } else {
        stop("unsupported family", call. = FALSE)
      }
    }, 
    
    .pom = function(A, tx.mod.fitted, data, m) {
      if (all(is.null(tx.mod.fitted)) || all(is.na(tx.mod.fitted))) {
        stop("cannot use `weight = 'qpom' or 'overlap' if treatment not modeled",
             call. = FALSE)
      }
      
      if (!inherits(tx.mod.fitted, "glm")) {
        stop("`tx.weight` cannot be 'qpom' if treatment is not modelled",
             call. = FALSE)
      }
      
      X_alpha <- stats::model.matrix(tx.mod.fitted, 
                                     stats::model.frame(tx.mod.fitted, data))
      
      a_binned <- dplyr::ntile(A, m)
      a_cat <- as.factor(a_binned)
      
      if (ncol(X_alpha) == 1L && attr(terms(tx.mod.fitted), "intercept") == 1L) {
        pom <- MASS::polr(a_cat ~ 1)
      } else{
        pom <- MASS::polr(a_cat ~ X_alpha[, -1L])
      }
      
      prob <- pom$fitted.values[cbind(seq_along(a_binned), a_binned)]
      sum_prob <- rowSums(1.0 / pom$fitted.values)
      
      list("prob" = prob, "sum.prob" = sum_prob)
    },
    
    Hd = function(data, A, ...) {
      data[, private$tx.var] <- A
      data[, private$tx.var.b] <- A^2
      model.matrix(self$full.model, data)
    },
    
    Hw = function(data, A, A.hat, wgt, treat.mod.fitted, ...) {
      A2hat <- summary(treat.mod.fitted)$dispersion + A.hat * A.hat

      data[, private$tx.var] <- {A - A.hat} * wgt
      data[, private$tx.var.b] <- {A * A - A2hat} * wgt

      model.matrix(self$full.model, data)
      
     },
    
    shiftY = function(type, outcome.fit, data, opt, A, ...) {
        if (type == "DTR") {
          self$regret(outcome.fit, data, opt, A)
        } else {
          self$regret(outcome.fit, data, 0.0, A)
        }
      },
    
    regret = function(outcome.fit, data, opt, A) {
        data[, private$tx.var] <- opt
        data[, private$tx.var.b] <- opt^2
        at_opt <- predict(outcome.fit, data)

        data[, private$tx.var] <- A
        data[, private$tx.var.b] <- A^2
        at_A <- predict(outcome.fit, data)
      
        at_opt - at_A
      },
    
    opt = function(outcome.fit, data, quiet = FALSE, ...) {
        # both treatment variables to 0 to get the non-blip term
        data[, private$tx.var] <- 0.0
        data[, private$tx.var.b] <- 0.0
        y_no_tx <- predict(outcome.fit, data)
        
        # only linear terms of blip model
        data[, private$tx.var] <- 1.0
        data[, private$tx.var.b] <- 0.0
        y_linear <- predict(outcome.fit, data) - y_no_tx

        # only quadratic terms of blip model
        data[, private$tx.var] <- 0.0
        data[, private$tx.var.b] <- 1.0
        y_quad <- predict(outcome.fit, data) - y_no_tx

        if (!is.infinite(private$min.tx) && !is.infinite(private$max.tx)) {
          data[, private$tx.var] <- private$min.tx
          data[, private$tx.var.b] <- private$min.tx^2
          predict_min <- predict(outcome.fit, data)
          
          data[, private$tx.var] <- private$max.tx
          data[, private$tx.var.b] <- private$max.tx^2
          predict_max <- predict(outcome.fit, data)
          
          opt <- -0.5 * y_linear / y_quad
          opt[y_quad >= 0.0 & predict_min >= predict_max] <- private$min.tx
          opt[y_quad >= 0.0 & predict_min < predict_max] <- private$max.tx
        } else {
          opt <- -0.5 * y_linear / y_quad
          if (any(y_quad > 0.0) && !quiet) {
            warning(" optimal treatment may be a minimum for some observations.\n",
                    call. = FALSE)
          }
          
        }
        pmin(pmax(opt, private$min.tx), private$max.tx)
      },
    
    prep = function(data, A) {
      data[[private$tx.var.b]] <- A^2
      data
    }
    
  ),
  private = list(
    max.tx = NULL,
    min.tx = NULL,
    tx.var = NULL,
    tx.var.b = NULL
  )
)

# Shared minimal data
n_obs <- 20L

dat_bin <- data.frame(
  X = seq(-1.0, 1.0, length.out = n_obs),
  A = rep(0L:1L, times = n_obs / 2L),
  Y = seq_len(n_obs) / n_obs
)

dat_multi <- data.frame(
  X = seq(-1.0, 1.0, length.out = n_obs),
  A = factor(rep(c("low", "med", "high"), length.out = n_obs),
             levels = c("low", "med", "high")),
  Y = seq_len(n_obs) / n_obs
)

dat_cont <- data.frame(
  X = cos(seq(0.0, pi, length.out = n_obs)),  # not linear in A
  A = seq(0.1, 0.9, length.out = n_obs),
  Y = seq_len(n_obs) / n_obs
)

test_that("Binary$initialize: blip.model has no intercept and contains tx.var", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  blip_terms <- attr(terms(obj$blip.model), "term.labels")
  expect_equal(attr(terms(obj$blip.model), "intercept"), 0L)
  expect_true("A" %in% blip_terms)
})

test_that("Binary$initialize: blip.model with covariate adds interaction term A:X", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  blip_terms <- attr(terms(obj$blip.model), "term.labels")
  expect_true("A:X" %in% blip_terms)
})

test_that("Binary$initialize: intercept-only blip produces only tx.var in blip.model", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~1, tx.var = "A")
  blip_terms <- attr(terms(obj$blip.model), "term.labels")
  expect_identical(blip_terms, "A")
})

test_that("Binary$initialize: full.model contains tf and blip terms", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  full_terms <- attr(terms(obj$full.model), "term.labels")
  expect_true("X" %in% full_terms)
  expect_true("A" %in% full_terms)
  expect_true("A:X" %in% full_terms || "X:A" %in% full_terms)
})

test_that("Binary$blip_params: returns named indices for coefficients containing tx.var", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  coefs <- c("(Intercept)" = 1.0, "X" = 2.0, "A" = 3.0, "A:X" = 4.0)
  res <- obj$blip_params(coefs)
  expect_named(res, c("A", "A:X"), ignore.order = FALSE)
  expect_equal(unname(res), c(3L, 4L))
})

test_that("Binary$blip_params: treatment-free coefficients are excluded", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  coefs <- c("(Intercept)" = 1.0, "X" = 2.0, "A" = 3.0, "A:X" = 4.0)
  res <- obj$blip_params(coefs)
  expect_false("(Intercept)" %in% names(res))
  expect_false("X" %in% names(res))
})

test_that("Binary$ipw: matches 1 / (A*Ahat + (1-A)*(1-Ahat))", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  A <- c(1L, 0L, 1L, 0L, 1L)
  Ahat <- c(0.8, 0.3, 0.6, 0.4, 0.9)
  expected <- 1.0 / (A * Ahat + (1.0 - A) * (1.0 - Ahat))
  expect_equal(obj$ipw(A, Ahat), expected)
})

test_that("Binary$.pom: stops - not appropriate for binary treatments", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  expect_error(obj$.pom(), "not appropriate for binary")
})

test_that("Binary$prep: returns data unchanged", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  expect_identical(obj$prep(dat_bin, dat_bin$A), dat_bin)
})

test_that("Binary$Hd: returns matrix with correct dimensions and expected column names", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  mm <- obj$Hd(dat_bin, dat_bin$A)
  expect_true(is.matrix(mm))
  expect_equal(nrow(mm), n_obs)
  expect_true(all(c("X", "A", "A:X") %in% colnames(mm)) ||
              all(c("X", "A", "X:A") %in% colnames(mm)))
})

test_that("Binary$Hpsi: returns no-intercept matrix containing tx.var term", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  mm <- obj$Hpsi(dat_bin, dat_bin$A)
  expect_true(is.matrix(mm))
  expect_false("(Intercept)" %in% colnames(mm))
  expect_true("A" %in% colnames(mm))
})

test_that("Binary$Hw: returns matrix with same dimensions as Hd", {
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  wgt <- rep(1.0, n_obs)
  Ahat <- rep(0.5, n_obs)
  expect_equal(dim(obj$Hw(dat_bin, dat_bin$A, Ahat, wgt)),
               dim(obj$Hd(dat_bin, dat_bin$A)))
})

local({
  obj <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  dat <- dat_bin
  dat$Y_int <- dat$Y
  fit <- lm(update(obj$full.model, Y_int ~ .), dat)
  
  test_that("Binary$regret: returns numeric vector of length n", {
    res <- obj$regret(fit, dat, opt = dat$A, A = dat$A)
    expect_true(is.numeric(res))
    expect_length(res, n_obs)
  })
  
  test_that("Binary$regret: is zero when opt equals observed treatment", {
    res <- obj$regret(fit, dat, opt = dat$A, A = dat$A)
    expect_equal(unname(res), rep(0.0, n_obs))
  })
  
  test_that("Binary$opt: returns integer 0/1 vector of length n", {
    res <- obj$opt(fit, dat)
    expect_true(is.integer(res))
    expect_true(all(res %in% c(0L, 1L)))
    expect_length(res, n_obs)
  })
  
  test_that("Binary$opt: matches oracle - treat if predicted(A=1) - predicted(A=0) > 1e-8", {
    dat0 <- dat; dat0$A <- 0L
    dat1 <- dat; dat1$A <- 1L
    expected <- as.integer((predict(fit, dat1) - predict(fit, dat0)) > 1e-8)
    expect_equal(obj$opt(fit, dat), expected)
  })
  
  test_that("Binary$shiftY: 'DTR' passes opt through to regret", {
    opt_vec <- rep(0L, n_obs)
    expect_equal(obj$shiftY("DTR", fit, dat, opt = opt_vec, A = dat$A),
                 obj$regret(fit, dat, opt = opt_vec, A = dat$A))
  })
  
  test_that("Binary$shiftY: non-DTR type uses 0L as opt (effect parameterisation)", {
    expect_equal(obj$shiftY("effect", fit, dat, opt = rep(1L, n_obs), A = dat$A),
                 obj$regret(fit, dat, opt = 0L, A = dat$A))
  })
})


# MultiNom
tx_levels <- c("low", "med", "high")

test_that("MultiNom$initialize: blip.model has no intercept and contains tx.var terms", {
  obj<- MultiNom$new(tf.model = ~X, blip.model = ~X, tx.var = "A", tx.levels = tx_levels)
  blip_terms <- attr(terms(obj$blip.model), "term.labels")
  expect_equal(attr(terms(obj$blip.model), "intercept"), 0L)
  expect_true(any(grepl("^A", blip_terms)))
})

test_that("MultiNom$initialize: full.model contains tf covariate and blip terms", {
  obj <- MultiNom$new(tf.model = ~X, blip.model = ~X, tx.var = "A", tx.levels = tx_levels)
  full_terms <- attr(terms(obj$full.model), "term.labels")
  expect_true("X" %in% full_terms)
  expect_true(any(grepl("^A", full_terms)))
})

test_that("MultiNom$blip_params: returns indices for non-reference level tx terms", {
  obj <- MultiNom$new(tf.model = ~X, blip.model = ~X, tx.var = "A", tx.levels = tx_levels)
  coefs <- c("(Intercept)" = 1.0, "X" = 2.0,
             "Amed" = 3.0, "Ahigh" = 4.0, "Amed:X" = 5.0, "Ahigh:X" = 6.0)
  res <- obj$blip_params(coefs)
  expect_equal(sort(unname(res)), c(3L, 4L, 5L, 6L))
})

test_that("MultiNom$blip_params: reference level and tf coefficients are excluded", {
  obj <- MultiNom$new(tf.model = ~X, blip.model = ~X, tx.var = "A", tx.levels = tx_levels)
  coefs <- c("(Intercept)" = 1.0, "X" = 2.0,
             "Amed" = 3.0, "Ahigh" = 4.0, "Amed:X" = 5.0, "Ahigh:X" = 6.0)
  res <- obj$blip_params(coefs)
  expect_false("(Intercept)" %in% names(res))
  expect_false("X"           %in% names(res))
  expect_false(any(grepl("Alow", names(res))))
})

test_that("MultiNom$ipw: matches 1 / prob(observed treatment level) for each observation", {
  obj <- MultiNom$new(tf.model = ~X, blip.model = ~1, tx.var = "A", tx.levels = tx_levels)
  A <- factor(c("low", "med", "high", "med", "low"), levels = tx_levels)
  Ahat <- matrix(c(0.5, 0.3, 0.2,
                   0.2, 0.6, 0.2,
                   0.1, 0.2, 0.7,
                   0.3, 0.5, 0.2,
                   0.6, 0.2, 0.2), nrow = 5L, byrow = TRUE)
  expected <- sapply(seq_along(A), function(i) {
    1.0 / Ahat[i, match(as.character(A[i]), tx_levels)]
  })
  expect_equal(obj$ipw(A, Ahat), expected)
})

test_that("MultiNom$Hpsi: stops - g-estimation not supported for multinomial", {
  obj <- MultiNom$new(tf.model = ~X, blip.model = ~1, tx.var = "A",
                      tx.levels = tx_levels)
  expect_error(obj$Hpsi(dat_multi, dat_multi$A), "not yet supported")
})

test_that("MultiNom$Hw: stops - g-estimation not supported for multinomial", {
  obj  <- MultiNom$new(tf.model = ~X, blip.model = ~1, tx.var = "A",
                       tx.levels = tx_levels)
  Ahat <- matrix(1.0 / 3.0, nrow = n_obs, ncol = 3L)
  expect_error(obj$Hw(dat_multi, dat_multi$A, Ahat, rep(1.0, n_obs)),
               "not yet supported")
})

test_that("MultiNom$prep: returns data unchanged", {
  obj <- MultiNom$new(tf.model = ~X, blip.model = ~1, tx.var = "A",
                      tx.levels = tx_levels)
  expect_identical(obj$prep(dat_multi, dat_multi$A), dat_multi)
})

test_that("MultiNom$Hd: returns matrix with n rows", {
  obj <- MultiNom$new(tf.model = ~X, blip.model = ~1, tx.var = "A",
                      tx.levels = tx_levels)
  mm <- obj$Hd(dat_multi, dat_multi$A)
  expect_true(is.matrix(mm))
  expect_equal(nrow(mm), n_obs)
})

local({
  obj <- MultiNom$new(tf.model = ~X, blip.model = ~1, tx.var = "A",
                            tx.levels = tx_levels)
  dat <- dat_multi
  dat$Y_int <- dat$Y
  fit <- lm(update(obj$full.model, Y_int ~ .), dat)
  
  test_that("MultiNom$regret: returns numeric vector of length n", {
    res <- obj$regret(fit, dat, opt = dat$A, A = dat$A)
    expect_true(is.numeric(res))
    expect_length(res, n_obs)
  })
  
  test_that("MultiNom$regret: is zero when opt equals observed treatment", {
    res <- obj$regret(fit, dat, opt = dat$A, A = dat$A)
    expect_equal(unname(res), rep(0.0, n_obs))
  })
  
  test_that("MultiNom$opt: returns values drawn from tx.levels", {
    res <- obj$opt(fit, dat)
    expect_true(all(res %in% tx_levels))
    expect_length(res, n_obs)
  })
  
  test_that("MultiNom$opt: matches oracle - level with highest predicted outcome", {
    pred_mat <- sapply(tx_levels, function(lv) {
      d <- dat
      d$A <- factor(lv, levels = tx_levels)
      predict(fit, d)
    })
    expected <- factor(tx_levels[apply(pred_mat, 1L, which.max)], levels = tx_levels)
    expect_equal(obj$opt(fit, dat), expected)
  })
  
  test_that("MultiNom$shiftY: 'DTR' passes opt through to regret", {
    opt_vec <- factor(rep("low", n_obs), levels = tx_levels)
    expect_equal(obj$shiftY("DTR",fit, dat, opt = opt_vec, A = dat$A),
                 obj$regret(fit, dat, opt = opt_vec, A = dat$A))
  })
  
  test_that("MultiNom$shiftY: non-DTR type uses first tx.level as reference", {
    opt_vec <- factor(rep("med", n_obs), levels = tx_levels)
    expect_equal(obj$shiftY("effect", fit, dat, opt = opt_vec, A = dat$A),
                 obj$regret(fit, dat, opt = tx_levels[1L], A = dat$A))
  })
})

# ContLinearBlip

test_that("ContLinearBlip$initialize: stops immediately - not supported", {
  expect_error(
    ContLinearBlip$new(tf.model = ~X, blip.model = ~X, tx.var = "A",
                       treat.range = c(0.0, 1.0)),
    "not supported"
  )
})


# ContQuadraticBlip

test_that("ContQuadraticBlip$initialize: stops on intercept-only blip model", {
  expect_error(
    ContQuadraticBlip$new(tf.model = ~X, blip.model = ~1, tx.var = "A",
                          treat.range = c(0.0, 1.0)),
    "intercept only"
  )
})

test_that("ContQuadraticBlip$initialize: blip.model has no intercept and contains tx.var", {
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.0, 1.0))
  blip_terms <- attr(terms(obj$blip.model), "term.labels")
  expect_equal(attr(terms(obj$blip.model), "intercept"), 0L)
  expect_true("A" %in% blip_terms)
})

test_that("ContQuadraticBlip$initialize: full.model contains tf covariate and blip terms", {
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.0, 1.0))
  full_terms <- attr(terms(obj$full.model), "term.labels")
  expect_true("X" %in% full_terms)
  expect_true("A" %in% full_terms)
})

test_that("ContQuadraticBlip$blip_params: identifies both linear and quadratic tx terms", {
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.0, 1.0))
  dat <- dat_cont
  prepped <- obj$prep(dat, dat$A)
  prepped$Y_int <- prepped$Y
  fit <- lm(update(obj$full.model, Y_int ~ .), prepped)
  res <- obj$blip_params(coef(fit))
  # Treatment-free terms must be absent; tx terms must be present
  expect_false("(Intercept)" %in% names(res))
  expect_false("X" %in% names(res))
  expect_true(length(res) >= 1L)
  expect_true("A" %in% names(res))
})

test_that("ContQuadraticBlip$ipw gaussian: matches 1 / dnorm(A, fitted, sqrt(dispersion))", {
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.0, 1.0))
  fit <- glm(A ~ X, dat_cont, family = gaussian())
  disp <- summary(fit)$dispersion
  expected <- 1.0 / dnorm(dat_cont$A, mean = fit$fitted.values, sd = sqrt(disp))
  expect_equal(obj$ipw(dat_cont$A, fit), expected)
})

test_that("ContQuadraticBlip$ipw Gamma: matches 1 / dgamma(A, shape, scale)", {
  dat_pos <- data.frame(X = seq(-1.0, 1.0, length.out = n_obs),
                        A = seq(0.1, 1.0, length.out = n_obs))
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.1, 1.0))
  fit <- glm(A ~ X, dat_pos, family = Gamma(link = "log"))
  disp <- summary(fit)$dispersion
  expected <- 1.0 / dgamma(dat_pos$A, shape = 1.0 / disp,
                           scale = fit$fitted.values * disp)
  expect_equal(obj$ipw(dat_pos$A, fit), expected)
})

test_that("ContQuadraticBlip$ipw: stops on unsupported glm family", {
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.0, 1.0))
  fit <- glm(A ~ X, dat_cont, family = gaussian())
  fit$family$family <- "poisson"
  expect_error(obj$ipw(dat_cont$A, fit), "unsupported family")
})

test_that("ContQuadraticBlip$.pom: stops when tx.mod.fitted is NULL", {
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.0, 1.0))
  expect_error(obj$.pom(dat_cont$A, NULL, dat_cont, 4L))
})

test_that("ContQuadraticBlip$.pom: stops when tx.mod.fitted is not a glm", {
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.0, 1.0))
  expect_error(obj$.pom(dat_cont$A, list(not = "a glm"), dat_cont, 4L))
})

test_that("ContQuadraticBlip$prep: adds exactly one new column with values A^2", {
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.0, 1.0))
  A <- dat_cont$A
  result <- obj$prep(dat_cont, A)
  new_col <- setdiff(names(result), names(dat_cont))
  expect_length(new_col, 1L)
  expect_equal(result[[new_col]], A^2)
})

test_that("ContQuadraticBlip$prep: original columns are unchanged", {
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.0, 1.0))
  result <- obj$prep(dat_cont, dat_cont$A)
  expect_equal(result[, names(dat_cont)], dat_cont)
})

test_that("ContQuadraticBlip$Hd: returns matrix with n rows and tx.var column", {
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.0, 1.0))
  mm <- obj$Hd(dat_cont, dat_cont$A)
  expect_true(is.matrix(mm))
  expect_equal(nrow(mm), n_obs)
  expect_true("A" %in% colnames(mm))
})

test_that("ContQuadraticBlip$Hw: returns matrix with same dimensions as Hd", {
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                                treat.range = c(0.0, 1.0))
  fit <- glm(A ~ X, dat_cont, family = gaussian())
  Ahat <- fit$fitted.values
  wgt <- rep(1.0, n_obs)
  expect_equal(dim(obj$Hw(dat_cont, dat_cont$A, Ahat, wgt, treat.mod.fitted = fit)),
               dim(obj$Hd(dat_cont, dat_cont$A)))
})

local({
  obj <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                               treat.range = c(0.0, 1.0))
  dat <- dat_cont
  prepped <- obj$prep(dat, dat$A)
  prepped$Y_int <- prepped$Y
  fit <- lm(update(obj$full.model, Y_int ~ .), prepped)
  
  test_that("ContQuadraticBlip$regret: returns numeric vector of length n", {
    res <- obj$regret(fit, dat, opt = dat$A, A = dat$A)
    expect_true(is.numeric(res))
    expect_length(res, n_obs)
  })
  
  test_that("ContQuadraticBlip$regret: is zero when opt equals observed treatment", {
    res <- obj$regret(fit, dat, opt = dat$A, A = dat$A)
    expect_equal(unname(res), rep(0.0, n_obs))
  })
  
  test_that("ContQuadraticBlip$opt: result is within treat.range for all observations", {
    res <- obj$opt(fit, dat)
    expect_true(is.numeric(res))
    expect_length(res, n_obs)
    expect_true(all(res >= 0.0 & res <= 1.0))
  })
  
  test_that("ContQuadraticBlip$opt: no warning issued when quiet = TRUE", {
    obj_inf <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                                     treat.range = c(-Inf, Inf))
    expect_no_warning(obj_inf$opt(fit, dat, quiet = TRUE))
  })
  
  test_that("ContQuadraticBlip$shiftY: 'DTR' passes opt through to regret", {
    opt_vec <- rep(0.5, n_obs)
    expect_equal(obj$shiftY("DTR", fit, dat, opt = opt_vec, A = dat$A),
                 obj$regret(fit, dat, opt = opt_vec, A = dat$A))
  })
  
  test_that("ContQuadraticBlip$shiftY: non-DTR type uses 0.0 as reference treatment", {
    expect_equal(obj$shiftY("effect", fit, dat, opt = rep(0.5, n_obs), A = dat$A),
                 obj$regret(fit, dat, opt = 0.0, A = dat$A))
  })
})

local({
  # Y = A^2 is perfectly U-shaped: quadratic coefficient >> 0, guaranteeing
  # y_quad > 0 and the warning branch fires when treat.range is infinite
  A_seq <- seq(0.1, 0.9, length.out = n_obs)
  dat_cup <- data.frame(X = rep(0.0, n_obs), A = A_seq, Y = A_seq^2)
  obj_inf <- ContQuadraticBlip$new(tf.model = ~X, blip.model = ~A, tx.var = "A",
                                   treat.range = c(-Inf, Inf))
  prepped <- obj_inf$prep(dat_cup, dat_cup$A)
  prepped$Y_int <- prepped$Y
  fit_cup <- lm(update(obj_inf$full.model, Y_int ~ .), prepped)
  
  test_that("ContQuadraticBlip$opt: warns about potential minimum when quiet = FALSE", {
    expect_warning(obj_inf$opt(fit_cup, dat_cup, quiet = FALSE),
                   "may be a minimum")
  })
})

# printing
#' @param x An object of class `DTRreg`.
#' @rdname DTRreg
#' @importFrom utils tail
#' @export
print.DTRreg <- function(x,...) {

  K <- x$K

  cat(ifelse(x$setup$type == "DTR", 
             paste("DTR estimation over ", K, " stages:\n\n"), 
             ""),
      "Blip parameter estimates",
      ifelse(x$setup$var.estim == "none", 
             " (standard errors not requested)\n", 
             "\n"))
  
  for (k in 1L:x$K) {
    if (x$setup$var.estim == "none" || {is.list(x$covmat) && any(x$covmat[[k]] == "aborted")}) {
      cat("Stage ", k, " (n = ", x$analysis$n[[k]], ")\n", sep = "")
      result <- data.frame("Estimate" = sprintf(x$psi[[k]], fmt = '%#.4f'))
      rownames(result) <- names(x$psi[[k]])
      print(result)
      cat("\n")
    } else {
      cat("Stage ", k, " (n = ", x$analysis$n[[k]], ")\n", sep = "")
      
      se <- utils::tail(sqrt(diag(x$covmat[[k]])), length(x$psi[[k]]))
      ci <- cbind(x$psi[[k]] - 1.96 * se, x$psi[[k]] + 1.96 * se)
      result <- data.frame("Estimate" = sprintf(x$psi[[k]], fmt = '%#.4f'),
                           "Std_Error" = sprintf(se, fmt = '%#.4f'),
                           "95%_CI_Lower" = sprintf(pmin(ci[, 1L], ci[, 2L]), fmt = '%#.4f'), 
                           "95%_CI_Upper" = sprintf(pmax(ci[, 1L], ci[, 2L]), fmt = '%#.4f'))
      rownames(result) <- names(x$psi[[k]])
      print(result)
    }
  }

  # print warnings if non-regularity concerns
  if (x$setup$type == "DTR") {
    if (any(!is.na(x$nonreg))) {
      
      tst <- unlist(x$analysis$cts) == "bin" & x$nonreg > 0.05
      if (any(tst)) {
        use_plural <- sum(tst) > 1L
        cat("\nWarning: possible non-regularity at ",
            ifelse(use_plural, "stages ", "stage "),
            paste(which(tst), collapse = ", "),
            ifelse(use_plural, " (probs = ", "(prob = "),
            paste(format(x$nonreg[tst], digits = 3), collapse = ", "), 
            ifelse(use_plural, ", respectively)\n", ")\n"), sep = "")
      }
    }
    
    # print DTRregs explicitly if bin or cts.l
    if (x$analysis$cts[[1L]] == "bin") cat("Recommended dynamic treatment regimen:\n")
    
    for (j in 1L:K) {
      
      # format parameter estimate to be 3 digits w/ scientific notation
      est <- sprintf(abs(x$psi[[j]]), fmt = '%#.4f')
      
      # identify the appropriate +/- connector
      sg <- c("", rep("+", length(est) - 1L))
      sg[x$psi[[j]] < 0.0] <- "-"
      
      # names of parameters
      nms <- names(x$psi[[j]])
      nms[1L] <- ""
      
      if (is.null(x$optimization)) x$optimization <- "max"
      
      # combine these to form the rule  
      rule <- paste0(paste(sg, est, nms, collapse = " "), " > 0")
      
      if (x$analysis$cts[[j]] == "bin") {
        cat("Stage ", j, ": treat if", rule, "\n", sep = "")
      } else if (x$analysis$cts[[j]] == "cts.l") {
        cat("Stage ", j, ": maximum treatment if",
            rule, ", otherwise minimum treatment.\n", sep = "")
      }
    }
  }
}


#' @param object An object of class `DTRreg`.
#' @param ... Ignored.
#' @rdname DTRreg
#' @export
summary.DTRreg <- function(object,...) {
  print.DTRreg(object)
}

#' Optimal Outcome Prediction for DTRs
#' 
#' Predicted outcome assuming optimal treatment (according to analysis via 
#'   G-estimation or dWOLS) was followed.  Assumes blip and treatment-free 
#'   models correctly specified.
#'  
#' This function may be used in a similar fashion to more traditional modeling 
#'   commands (such as lm).  Users are referred to the primary `DTRreg()` 
#'   and `DTRSurv()` help command
#'   (and associated literature) for information concerning model specification.
#'   In particular, we note that the predict function assumes that the 
#'   treatment-free model has been correctly specified, as the treatment-free 
#'   parameters are used in the prediction process.
#'   
#' @param object A model object generated by the function `DTRreg()` or `DWSurv()`.
#' @param newdata A dataset (usually the data analyzed by DTRreg for which 
#'   predicted outcomes are desired.  If a new dataset is provided, variable 
#'   names should correspond to those presented to `DTRreg()` or `DWSurv()`.
#' @param treat.range If treatment is continuous (rather than binary), a 
#'   vectors of the form c(min,max) which specify the minimum and maximum value 
#'   the treatment may take at stage 1.  
#'   If unspecified, this will be inferred from the 
#'   treat.range provided with use of the original DTRreg command.  As such, if 
#'   no treatment range was specified there either, treat.range will be the 
#'   minimum and maximum observed first stage treatment.
#' @param ... Space for additional arguments (not currently used)
#' 
#' @return An \eqn{n}{n} x 1 matrix of predicted outcome values.
#' 
#' @references
#' Chakraborty, B., Moodie, E. E. M. (2013) \emph{Statistical Methods for 
#'   Dynamic Treatment Regimes}. New York: Springer.
#'   
#' Robins, J. M. (2004) \emph{Optimal structural nested models for optimal 
#'   sequential decisions}. In Proceedings of the Second Seattle Symposium on 
#'   Biostatistics, D. Y. Lin and P. J. Heagerty (eds), 189-326. 
#'   New York: Springer.
#'   
#' Wallace, M. P., Moodie, E. M. (2015) Doubly-Robust Dynamic Treatment Regimen 
#'   Estimation Via Weighted Least Squares. \emph{Biometrics} \bold{71}(3), 
#'   636-644 (doi:10.1111/biom.12306.)
#'   
#' @author Michael Wallace
#' 
#' @examples
#' # example single run of a 2-stage g-estimation analysis
#' 
#' set.seed(1)
#' 
#' # expit function
#' expit <- function(x) { 1.0 / (1.0 + exp(-x)) }
#' 
#' # sample size
#' n <- 10000
#' 
#' # variables (X = patient information, A = treatment)
#' X1 <- rnorm(n)
#' A1 <- rbinom(n, 1, expit(X1))
#' X2 <- rnorm(n)
#' A2 <- rbinom(n, 1, expit(X2))
#' 
#' # blip functions
#' gamma1 <- A1 * (1 + X1)
#' gamma2 <- A2 * (1 + X2)
#' 
#' # observed outcome: treatment-free outcome plus blip functions
#' Y <- exp(X1) + exp(X2) + gamma1 + gamma2 + rnorm(n)
#'  
#' # models to be passed to DTRreg
#' # blip model
#' blip.mod <- list(~ X1, ~ X2)
#' # treatment model (correctly specified)
#' treat.mod <- list(A1 ~ X1, A2 ~ 1)
#' # treatment-free model (incorrectly specified)
#' tf.mod <- list(~ X1, ~ X2)
#' 
#' # perform G-estimation
#' mod1 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
#'               data = twoStageCont, method = "gest")
#'               
#' # predicted Y for optimal treatment
#' dat <- data.frame(X1, X2, A1, A2)
#' predict(mod1, newdata = dat)
#' @concept dynamic treatment regimens
#' @concept adaptive treatment strategies
#' @concept personalized medicine
#' @concept g-estimation
#' @concept dynamic weighted ordinary least squares
#' @export
predict.DTRreg <- function(object, newdata, treat.range = NULL, ...) {
  K <- object$K
  
  stopifnot("`treat.range` must be a list of length `stage`" = is.null(treat.range) ||
              {is.list(treat.range) && all(lengths(treat.range) == 2L)})
  
  tx.range <- if (!is.list(treat.range)) object$setup$tx.range[[1L]] else treat.range[[1L]]
  cts_obj <- switch(object$analysis$cts[[1L]],
                    "bin"      = Binary$new(tf.model   = object$setup$models[[1L]]$tf,
                                            blip.model = object$setup$models[[1L]]$blip,
                                            tx.var     = object$setup$tx.vars[1L]),
                    "multinom" = MultiNom$new(tf.model   = object$setup$models[[1L]]$tf,
                                              blip.model = object$setup$models[[1L]]$blip,
                                              tx.var     = object$setup$tx.vars[1L],
                                              tx.levels  = levels(object$training_data$A[[1L]])),
                    "cts.q"    = ContQuadraticBlip$new(tf.model    = object$setup$models[[1L]]$tf,
                                                       blip.model  = object$setup$models[[1L]]$blip,
                                                       tx.var      = object$setup$tx.vars[1L],
                                                       treat.range = tx.range)
  )
  
  if (!missing(newdata)) {
    if (!all(complete.cases(newdata))) 
      stop("if provided, `newdata` must be complete", call. = FALSE)
    newdata[[object$setup$tx.vars[1L]]] <-
      cts_obj$opt(outcome.fit = object$analysis$outcome.fit[[1L]], data = newdata)
  } else {
    newdata <- object$training_data$data
    newdata[[object$setup$tx.vars[1L]]] <- object$analysis$opt.treat[[1L]]
  }
  
  newdata <- cts_obj$prep(newdata, newdata[[object$setup$tx.vars[1L]]])
  
  pred <- predict(object$analysis$outcome.fit[[1L]], newdata = newdata)
  if ("DWSurv" %in% class(object)) pred <- exp(pred)
  pred
}

#' @rdname DTRreg
#' @export
coef.DTRreg <- function(object,...) {
  K <- object$K
  psi <- object$psi
  names(psi) <- paste0("stage_", 1L:K)
  psi
}

#' @noRd
#' @keywords internal
.ciStep.se <- function(psij, level, covmatj) {
  psi <- matrix(NA, nrow = length(psij), ncol = 2L)
  rownames(psi) <- names(psij)
  colnames(psi) <- c(paste(100.0 * {{1.0 - level} / 2.0},"%"), 
                          paste(100.0 * {{1.0 + level} / 2.0},"%"))
  
  psi[, 1L] <- psij + 
    stats::qnorm({1.0 - level} / 2.0) * sqrt(diag(covmatj))
    
  psi[, 2L] <- psij + 
    stats::qnorm({1.0 + level} / 2.0) * sqrt(diag(covmatj))

  psi    
}

#' @noRd
#' @keywords internal
.ciStep.percentile <- function(psij, level, var.estim, psi.bootj) {
  if (var.estim != "bootstrap") {
    stop("Percentile confidence intervals only available with bootstrap.",
         call. = FALSE)
  }
  
  psi <- matrix(NA, nrow = length(psij), ncol = 2L)
  rownames(psi) <- names(psij)
  colnames(psi) <- c(paste(100.0 * {{1.0 - level} / 2.0},"%"), 
                     paste(100.0 * {{1.0 + level} / 2.0},"%"))
  
  psi[, 1L] <- apply(psi.bootj, 2L, stats::quantile, probs = {1.0 - level} / 2.0)
  psi[, 2L] <- apply(psi.bootj, 2L, stats::quantile, probs = {1.0 + level} / 2.0)
  psi
}

#' Confidence Interval Calculations for DTRs
#' 
#' Confidence intervals for parameters, with the option of 
#'   constructing the confidence intervals using the percentile method 
#'   when bootstrap is used.
#'   
#' @param object A model object generated by the function DTRreg.
#' @param parm Not available for DTRreg objects.
#' @param level The confidence level required.
#' @param type Typical Wald-type confidence interval "se" (default) or 
#'   confidence intervals derived with the percentile method "percentile" 
#'   (bootstrap variance estimates only).
#' @param ... Space for additional arguments (not currently used).
#' 
#' @return A list with columns giving lower and upper confidence limits for each
#'   parameter. These will be labelled as (1-level)/2 and 1 - (1-level)/2 in 
#'   percentage (by default 2.5\% and 97.5\%).
#'   
#' @importFrom stats confint qnorm quantile
#' @export
confint.DTRreg <- function(object, parm = NULL, level = 0.95, 
                           type = c("se", "percentile"), ...) {
  
  type <- match.arg(type)
  
  if (object$setup$var.estim == "none") {
    message("confint() only available when DTRreg called with a variance",
            "estimation option")
    return(invisible(NULL))
  }
  
  if (any(object$covmat[[1]] == "aborted")) {
    message("variance estimation aborted; unable to obtain confidence intervals")
    return(invisible(NULL))
  }
    
  K <- object$K
  psi <- object$psi
  names(psi) <- paste0("stage_", 1L:K)
  
  for (j in 1L:K) {
    psi[[j]] <- if (type == "se") {
      .ciStep.se(object$psi[[j]], level, object$covmat[[j]])
    } else {
      .ciStep.percentile(object$psi[[j]], level, object$setup$var.estim, object$psi.boot[[j]])
    }
  }
  psi
}

#' Diagnostic Plots for DTR Estimation
#' 
#' Diagnostic plots for assessment of treatment, treatment-free, and blip models 
#'   following DTR estimation using DTRreg or DWSurv.
#'   
#' DTR estimation using G-estimation and dWOLS requires the specification of 
#'   three models: the treatment, treatment-free, and blip.  The treatment model 
#'   may be assessed via standard diagnostics, whereas the treatment-free and 
#'   blip models may be simultaneously assessed using diagnostic plots 
#'   introduced by Rich et al.  The plot() function first presents diagnostic 
#'   plots that assess the latter, plotting fitted values against residuals and 
#'   covariates following DTR estimation.  If there is any evidence of a 
#'   relationship between the variables in these plots, this is evidence that at 
#'   least one of the blip or treatment-free models is mis-specified.
#'   
#' Following these plots, the plot() function will present standard diagnostic 
#'   plots for the treatment model.  These are produced directly by the standard 
#'   plot() command applied to the models that were fit.  For example, if 
#'   treatment is binary, the resulting plots are the same as those that are 
#'   generated by the plot() command applied to a glm object for logistic 
#'   regression.
#' 
#' @param x A model object generated by the functions DTRreg and DWSurv.
#' @param ... Space for additional arguments (not currently used)
#' @returns Invisibly, a list of gg objects.
#' 
#' @examples
#' # example single run of a 2-stage g-estimation analysis
#' 
#' set.seed(1)
#' 
#' # expit function
#' expit <- function(x) { 1.0 / (1.0 + exp(-x)) }
#' 
#' # sample size
#' n <- 10000
#' 
#' # variables (X = patient information, A = treatment)
#' X1 <- rnorm(n)
#' A1 <- rbinom(n, 1, expit(X1))
#' X2 <- rnorm(n)
#' A2 <- rbinom(n, 1, expit(X2))
#' 
#' # blip functions
#' gamma1 <- A1 * (1 + X1)
#' gamma2 <- A2 * (1 + X2)
#' 
#' # observed outcome: treatment-free outcome plus blip functions
#' Y <- exp(X1) + exp(X2) + gamma1 + gamma2 + rnorm(n)
#'  
#' # models to be passed to DTRreg
#' # blip model
#' blip.mod <- list(~ X1, ~ X2)
#' # treatment model (correctly specified)
#' treat.mod <- list(A1 ~ X1, A2 ~ 1)
#' # treatment-free model (incorrectly specified)
#' tf.mod <- list(~ X1, ~ X2)
#' 
#' # perform G-estimation
#' mod1 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
#'               data = twoStageCont, method = "gest")
#'               
#' # model diagnostics: note treatment-free model is mis-specified
#' plot(mod1)
#' 
#' @export
plot.DTRreg <- function(x, ...) {
  # call the autoplot method to get the list of ggplots
  plots <- ggplot2::autoplot(x, ...)
  
  # handle the interactive display
  if (interactive() && length(plots) > 1) {
    old_ask <- grDevices::devAskNewPage(TRUE)
    on.exit(grDevices::devAskNewPage(old_ask), add = TRUE)
  }
  
  # print each plot to the active graphics device
  for (p in plots) {
    # if it's a ggplot, print it; if it's a base plot caught in the list, plot it.
    if (inherits(p, "ggplot") || inherits(p, "grob")) {
      print(p)
    } else {
      plot(p)
    }
  }
  
  # Return the list invisibly
  invisible(plots)
}


#' Diagnostic Plots for DTR Estimation via autoplot()
#' 
#' Diagnostic plots for assessment of treatment, treatment-free, and blip models 
#'   following DTR estimation using DTRreg or DWSurv.
#'   
#' @param object A model object generated by the functions DTRreg and DWSurv.
#' @param ... Space for additional arguments (not currently used)
#' @returns A named list of gg objects.
#' 
#' @examples
#' # example single run of a 2-stage g-estimation analysis
#' 
#' set.seed(1)
#' 
#' # expit function
#' expit <- function(x) { 1.0 / (1.0 + exp(-x)) }
#' 
#' # sample size
#' n <- 10000
#' 
#' # variables (X = patient information, A = treatment)
#' X1 <- rnorm(n)
#' A1 <- rbinom(n, 1, expit(X1))
#' X2 <- rnorm(n)
#' A2 <- rbinom(n, 1, expit(X2))
#' 
#' # blip functions
#' gamma1 <- A1 * (1 + X1)
#' gamma2 <- A2 * (1 + X2)
#' 
#' # observed outcome: treatment-free outcome plus blip functions
#' Y <- exp(X1) + exp(X2) + gamma1 + gamma2 + rnorm(n)
#'  
#' # models to be passed to DTRreg
#' # blip model
#' blip.mod <- list(~ X1, ~ X2)
#' # treatment model (correctly specified)
#' treat.mod <- list(A1 ~ X1, A2 ~ 1)
#' # treatment-free model (incorrectly specified)
#' tf.mod <- list(~ X1, ~ X2)
#' 
#' # perform G-estimation
#' mod1 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
#'               data = twoStageCont, method = "gest")
#'               
#' # model diagnostics: note treatment-free model is mis-specified
#' gg_list <- ggplot2::autoplot(mod1)
#' 
#' @import ggplot2
#' @export
autoplot.DTRreg <- function(object, ...) {
  plot_list <- list()
  
  # residual Analysis
  for (j in seq_len(object$K)) {
    cases <- object$analysis$last.stage >= j
    obsK   <- object$analysis$fitted.values[[j]]
    fitK <- object$analysis$outcome.fit[[j]]$fitted.values
    resids <- object$analysis$residuals[[j]]
    
    # helper to generate the standardized diagnostic plot
    make_resid_plot <- function(x_val, y_val, x_label, title) {
      df <- data.frame(x = x_val, y = y_val)
      ggplot(df, aes(.data[["x"]], .data[["y"]])) +
        geom_point(alpha = 0.5, shape = 1) +
        geom_hline(yintercept = 0, linetype = "dashed", color = "blue") +
        geom_smooth(method = "loess", color = "red", se = FALSE, linewidth = 0.8) +
        labs(title = title, x = x_label, y = "Residuals") +
        theme_minimal()
    }
    
    # plot: Residuals vs Fitted
    name_fitted <- paste0("Stage", j, "_Fitted")
    plot_list[[name_fitted]] <- make_resid_plot(
      fitK, resids, "Fitted values", paste("Stage", j, "Residuals vs Fitted")
    )
    
    # plots: Blip Covariates
    blip <- object$setup$models[[j]]$blip
    blip_mm <- model.matrix(blip, object$analysis$blip.data[[j]])
    if (attr(terms(blip), "intercept") == 1L) blip_mm <- blip_mm[, -1L, drop = FALSE]
    
    if (ncol(blip_mm) > 0L) {
      for (i in seq_len(ncol(blip_mm))) {
        col_name <- colnames(blip_mm)[i]
        name_cov <- paste0("Stage", j, "_", col_name)
        plot_list[[name_cov]] <- make_resid_plot(
          blip_mm[, i], resids, col_name, paste("Stage", j, "Residuals vs", col_name)
        )
      }
    }
  }
  
  # Sub-Model Diagnostics (Treatment, Censoring, etc.)
  # use ggplotify or ggfortify if possible to convert them.
  add_submodel_plots <- function(plot_list, mod, prefix) {
    if (is.logical(mod)) return(plot_list)
    
    tryCatch({
      if (inherits(mod, c("glm", "lm"))) {
        # handle directly - don't rely on ggfortify being loaded
        df  <- data.frame(fitted   = fitted(mod),
                          residuals = stats::residuals(mod, type = "deviance"))
        p   <- ggplot2::ggplot(df, ggplot2::aes(.data[["fitted"]], .data[["residuals"]])) +
          ggplot2::geom_point(alpha = 0.5, shape = 1) +
          ggplot2::geom_hline(yintercept = 0, linetype = "dashed",
                              color = "blue") +
          ggplot2::geom_smooth(method = "loess", color = "red",
                               se = FALSE, linewidth = 0.8) +
          ggplot2::labs(title    = paste(prefix, "Residuals vs Fitted"),
                        x = "Fitted values", y = "Deviance Residuals") +
          ggplot2::theme_minimal()
        plot_list[[prefix]] <- p
      } else {
        p <- ggplot2::autoplot(mod)
        plot_list[[prefix]] <- p
      }
    }, error = function(e1) {
      if (requireNamespace("ggplotify", quietly = TRUE)) {
        tryCatch({
          plot_list[[prefix]] <<- ggplotify::as.ggplot(function() {
            old_ask <- graphics::par(ask = FALSE)
            on.exit(graphics::par(old_ask))
            plot(mod)
          })
        }, error = function(e2) {
          # silently skip
        })
      }
    })
    
    plot_list
  }  
  for (j in seq_len(object$K)) {
    plot_list <- add_submodel_plots(plot_list,
                                    mod = object$analysis$tx.mod.fitted[[j]], 
                                    paste0("Stage", j, "_Treatment"))
    
    if ("DWSurv" %in% class(object)) {
      plot_list <- add_submodel_plots(plot_list,
                                      object$analysis$cens.mod.fitted[[j]], 
                                      paste0("Stage", j, "_Censoring"))
    } else {
      plot_list <- add_submodel_plots(plot_list,
                                      object$analysis$cc.mod.fitted[[j]], 
                                      paste0("Stage", j, "_CompleteCases"))
    }
  }
  
  plot_list
}

# Shared setup - one trusted DTRreg result reused across all tests

data(twoStageCont)
n <- nrow(twoStageCont)

# DWOLS, no variance
mod_none <- suppressMessages(
  DTRreg(twoStageCont$Y,
         blip.mod  = list(~X1, ~X2),
         treat.mod = list(A1 ~ X1, A2 ~ X2),
         tf.mod    = list(~X1, ~X2),
         data      = twoStageCont,
         method    = "dwols",
         var.estim = "none")
)

# DWOLS, sandwich variance
mod_sw <- suppressMessages(
  DTRreg(twoStageCont$Y,
         blip.mod  = list(~X1, ~X2),
         treat.mod = list(A1 ~ X1, A2 ~ X2),
         tf.mod    = list(~X1, ~X2),
         data      = twoStageCont,
         method    = "dwols",
         var.estim = "sandwich")
)

# DWOLS, bootstrap variance
withr::with_seed(42L, {
  mod_boot <- suppressMessages(
    DTRreg(twoStageCont$Y,
           blip.mod       = list(~X1, ~X2),
           treat.mod      = list(A1 ~ X1, A2 ~ X2),
           tf.mod         = list(~X1, ~X2),
           data           = twoStageCont,
           method         = "dwols",
           var.estim      = "bootstrap",
           bootstrap.controls  = list(B = 20L))
  )
})

# GEST, sandwich variance
mod_gest <- suppressMessages(
  DTRreg(twoStageCont$Y,
         blip.mod  = list(~X1, ~X2),
         treat.mod = list(A1 ~ X1, A2 ~ X2),
         tf.mod    = list(~X1, ~X2),
         data      = twoStageCont,
         method    = "gest",
         var.estim = "sandwich")
)


# print.DTRreg

test_that("print.DTRreg: produces output without error (var.estim='none')", {
  expect_output(print.DTRreg(mod_none))
})

test_that("print.DTRreg: produces output without error (sandwich)", {
  expect_output(print.DTRreg(mod_sw))
})

test_that("print.DTRreg: produces output without error (bootstrap)", {
  expect_output(print.DTRreg(mod_boot))
})

test_that("print.DTRreg: output contains 'Stage 1'", {
  expect_output(print.DTRreg(mod_none), "Stage 1")
})

test_that("print.DTRreg: output contains 'Stage 2'", {
  expect_output(print.DTRreg(mod_none), "Stage 2")
})

test_that("print.DTRreg: no-variance output contains psi names", {
  expect_output(print.DTRreg(mod_none), "A1")
  expect_output(print.DTRreg(mod_none), "A2")
})

test_that("print.DTRreg: sandwich output contains Std_Error column", {
  expect_output(print.DTRreg(mod_sw), "Std_Error")
})

test_that("print.DTRreg: var.estim='none' does not print Std_Error", {
  expect_no_match(capture.output(print.DTRreg(mod_none)), "Std_Error")
})

test_that("print.DTRreg: DTR type prints treatment rule", {
  expect_output(print.DTRreg(mod_none), "treat if")
})


# summary.DTRreg

test_that("summary.DTRreg: produces same output as print.DTRreg", {
  expect_equal(capture.output(summary.DTRreg(mod_none)),
               capture.output(print.DTRreg(mod_none)))
})

test_that("summary.DTRreg: produces output without error", {
  expect_output(summary.DTRreg(mod_sw))
})


# coef.DTRreg

test_that("coef.DTRreg: returns a list of length K", {
  res <- coef.DTRreg(mod_none)
  expect_true(is.list(res))
  expect_length(res, mod_none$K)
})

test_that("coef.DTRreg: list names are stage_1, stage_2, ...", {
  res <- coef.DTRreg(mod_none)
  expect_equal(names(res), paste0("stage_", seq_len(mod_none$K)))
})

test_that("coef.DTRreg: each element matches corresponding psi", {
  res <- coef.DTRreg(mod_none)
  for (k in seq_len(mod_none$K)) {
    expect_equal(res[[k]], mod_none$psi[[k]])
  }
})

test_that("coef.DTRreg: psi names preserved within each stage", {
  res <- coef.DTRreg(mod_none)
  expect_true(all(grepl("A1", names(res[["stage_1"]]))))
  expect_true(all(grepl("A2", names(res[["stage_2"]]))))
})

test_that("coef.DTRreg: works for GEST object", {
  res <- coef.DTRreg(mod_gest)
  expect_length(res, mod_gest$K)
  expect_equal(names(res), paste0("stage_", seq_len(mod_gest$K)))
})


# confint.DTRreg

# --- var.estim = "none" early return -----------------------------------------

test_that("confint.DTRreg: returns invisible NULL when var.estim='none'", {
  res <- confint.DTRreg(mod_none)
  expect_null(res)
})

test_that("confint.DTRreg: prints message when var.estim='none'", {
   expect_message(confint.DTRreg(mod_none), "only available")
})

# --- se type (sandwich) ------------------------------------------------------

local({
  res <- confint.DTRreg(mod_sw)
  
  test_that("confint.DTRreg se: returns list of length K", {
    expect_true(is.list(res))
    expect_length(res, mod_sw$K)
  })
  
  test_that("confint.DTRreg se: list names are stage_1, stage_2, ...", {
    expect_equal(names(res), paste0("stage_", seq_len(mod_sw$K)))
  })
  
  test_that("confint.DTRreg se: each element is a matrix with 2 columns", {
    for (k in seq_len(mod_sw$K)) {
      expect_true(is.matrix(res[[k]]))
      expect_equal(ncol(res[[k]]), 2L)
    }
  })
  
  test_that("confint.DTRreg se: row count matches psi length per stage", {
    for (k in seq_len(mod_sw$K)) {
      expect_equal(nrow(res[[k]]), length(mod_sw$psi[[k]]))
    }
  })
  
  test_that("confint.DTRreg se: row names match psi names", {
    for (k in seq_len(mod_sw$K)) {
      expect_equal(rownames(res[[k]]), names(mod_sw$psi[[k]]))
    }
  })
  
  test_that("confint.DTRreg se: lower CI < upper CI for all parameters", {
    for (k in seq_len(mod_sw$K)) {
      expect_true(all(res[[k]][, 1L] < res[[k]][, 2L]))
    }
  })
  
  test_that("confint.DTRreg se: column names reflect default level 0.95", {
    expect_equal(colnames(res[[1L]]), c("2.5 %", "97.5 %"))
  })
  
  test_that("confint.DTRreg se: oracle - CI contains psi estimate", {
    # psi should be within its own CI
    for (k in seq_len(mod_sw$K)) {
      expect_true(all(mod_sw$psi[[k]] >= res[[k]][, 1L] &
                        mod_sw$psi[[k]] <= res[[k]][, 2L]))
    }
  })
  
  test_that("confint.DTRreg se: wider CI with lower level", {
    res_90 <- confint.DTRreg(mod_sw, level = 0.90)
    for (k in seq_len(mod_sw$K)) {
      width_95 <- res[[k]][, 2L]    - res[[k]][, 1L]
      width_90 <- res_90[[k]][, 2L] - res_90[[k]][, 1L]
      expect_true(all(width_95 > width_90))
    }
  })
})

# --- percentile type (bootstrap) ---------------------------------------------

local({
  res <- confint.DTRreg(mod_boot, type = "percentile")
  
  test_that("confint.DTRreg percentile: returns list of length K", {
    expect_length(res, mod_boot$K)
  })
  
  test_that("confint.DTRreg percentile: each element is a matrix with 2 columns", {
    for (k in seq_len(mod_boot$K)) {
      expect_true(is.matrix(res[[k]]))
      expect_equal(ncol(res[[k]]), 2L)
    }
  })
  
  test_that("confint.DTRreg percentile: lower CI < upper CI", {
    for (k in seq_len(mod_boot$K)) {
      expect_true(all(res[[k]][, 1L] < res[[k]][, 2L]))
    }
  })
})

test_that("confint.DTRreg percentile: stops when var.estim is not bootstrap", {
  expect_error(confint.DTRreg(mod_sw, type = "percentile"),
               "Percentile confidence intervals only available with bootstrap")
})


# predict.DTRreg

local({
  # With newdata
  res_new <- predict.DTRreg(mod_none, newdata = twoStageCont)
  # Without newdata (uses training data)
  res_trn <- predict.DTRreg(mod_none)
  
  test_that("predict.DTRreg: returns numeric vector", {
    expect_true(is.numeric(res_new))
    expect_true(is.numeric(res_trn))
  })
  
  test_that("predict.DTRreg: length equals nrow(newdata)", {
    expect_length(res_new, n)
  })
  
  test_that("predict.DTRreg: length equals n when no newdata", {
    expect_length(res_trn, n)
  })
  
  test_that("predict.DTRreg: newdata and training data give same result for same data", {
    expect_equal(res_new, res_trn)
  })
  
  test_that("predict.DTRreg: stops when newdata has missing values", {
    dat_na      <- twoStageCont
    dat_na$X1[1L] <- NA_real_
    expect_error(predict.DTRreg(mod_none, newdata = dat_na),
                 "must be complete")
  })
  
  test_that("predict.DTRreg: treat.range list is accepted without error", {
    expect_no_error(
      predict.DTRreg(mod_none, newdata = twoStageCont,
                     treat.range = NULL)
    )
  })
  
  test_that("predict.DTRreg: stops when treat.range is invalid", {
    expect_error(
      predict.DTRreg(mod_none, newdata = twoStageCont,
                     treat.range = c(0, 1, 2)),
      "treat.range"
    )
  })
})

test_that("predict.DTRreg: works for GEST object", {
  res <- predict.DTRreg(mod_gest, newdata = twoStageCont)
  expect_true(is.numeric(res))
  expect_length(res, n)
})


data(twoStageCont)

mod_dwols <- suppressMessages(
  DTRreg(twoStageCont$Y,
         blip.mod  = list(~X1, ~X2),
         treat.mod = list(A1 ~ X1, A2 ~ X2),
         tf.mod    = list(~X1, ~X2),
         data      = twoStageCont,
         method    = "dwols",
         var.estim = "none")
)

mod_gest <- suppressMessages(
  DTRreg(twoStageCont$Y,
         blip.mod  = list(~X1, ~X2),
         treat.mod = list(A1 ~ X1, A2 ~ X2),
         tf.mod    = list(~X1, ~X2),
         data      = twoStageCont,
         method    = "gest",
         var.estim = "none")
)

# Expected number of residual plots per stage:
# 1 (Fitted) + n_blip_covariates
# blip = ~X1 or ~X2 has 1 covariate → 2 plots per stage → 4 total for 2 stages
n_resid_plots_per_stage <- 2L  # Fitted + X covariate
n_stages <- 2L
n_resid_plots_total <- n_resid_plots_per_stage * n_stages


# autoplot.DTRreg - list structure

local({
  plots <- autoplot.DTRreg(mod_dwols)
  
  test_that("autoplot.DTRreg: returns a list", {
    expect_true(is.list(plots))
  })
  
  test_that("autoplot.DTRreg: contains at least K*2 residual plots", {
    expect_true(length(plots) >= n_resid_plots_total)
  })
  
  test_that("autoplot.DTRreg: names contain stage identifiers", {
    nms <- names(plots)
    expect_true(any(grepl("Stage1", nms)))
    expect_true(any(grepl("Stage2", nms)))
  })
  
  test_that("autoplot.DTRreg: Fitted plots named correctly", {
    expect_true("Stage1_Fitted" %in% names(plots))
    expect_true("Stage2_Fitted" %in% names(plots))
  })
  
  test_that("autoplot.DTRreg: blip covariate plots named correctly", {
    expect_true("Stage1_X1" %in% names(plots))
    expect_true("Stage2_X2" %in% names(plots))
  })
})


# autoplot.DTRreg - individual plot content

local({
  plots <- autoplot.DTRreg(mod_dwols)
  p_fit <- plots[["Stage1_Fitted"]]
  p_cov <- plots[["Stage1_X1"]]
  
  test_that("autoplot.DTRreg: Fitted plot is a ggplot", {
    expect_true(inherits(p_fit, "ggplot"))
  })
  
  test_that("autoplot.DTRreg: covariate plot is a ggplot", {
    expect_true(inherits(p_cov, "ggplot"))
  })
  
  test_that("autoplot.DTRreg: Fitted plot data has correct number of rows", {
    n_stage1 <- sum(mod_dwols$analysis$last.stage >= 1L)
    expect_equal(nrow(p_fit$data), n_stage1)
  })
  
  test_that("autoplot.DTRreg: Fitted plot data has x and y columns", {
    expect_true(all(c("x", "y") %in% names(p_fit$data)))
  })
  
  test_that("autoplot.DTRreg: Fitted plot x-axis label is 'Fitted values'", {
    expect_equal(p_fit$labels$x, "Fitted values")
  })
  
  test_that("autoplot.DTRreg: Fitted plot y-axis label is 'Residuals'", {
    expect_equal(p_fit$labels$y, "Residuals")
  })
  
  test_that("autoplot.DTRreg: Fitted plot title contains stage number", {
    expect_match(p_fit$labels$title, "Stage 1")
  })
  
  test_that("autoplot.DTRreg: covariate plot x-axis label is covariate name", {
    expect_equal(p_cov$labels$x, "X1")
  })
  
  test_that("autoplot.DTRreg: plots contain point layer", {
    layer_geoms <- sapply(p_fit$layers, function(l) class(l$geom)[1L])
    expect_true("GeomPoint" %in% layer_geoms)
  })
  
  test_that("autoplot.DTRreg: plots contain hline layer", {
    layer_geoms <- sapply(p_fit$layers, function(l) class(l$geom)[1L])
    expect_true("GeomHline" %in% layer_geoms)
  })
  
  test_that("autoplot.DTRreg: plots contain smooth layer", {
    layer_geoms <- sapply(p_fit$layers, function(l) class(l$geom)[1L])
    expect_true("GeomSmooth" %in% layer_geoms)
  })
  
  test_that("autoplot.DTRreg: x values in Fitted plot match fitted.values", {
    fitK <- mod_dwols$analysis$outcome.fit[[1L]]$fitted.values
    expect_equal(p_fit$data$x, unname(fitK))
  })
  
  test_that("autoplot.DTRreg: y values in Fitted plot are residuals oracle", {
    obsK  <- mod_dwols$analysis$Y[[1L]]
    fitK  <- mod_dwols$analysis$outcome.fit[[1L]]$fitted.values
    expected_resids <- obsK - fitK
    expect_equal(p_fit$data$y, unname(expected_resids))
  })
})


# autoplot.DTRreg - stage 2 correctness

local({
  plots   <- autoplot.DTRreg(mod_dwols)
  p_fit2  <- plots[["Stage2_Fitted"]]
  n_stage2 <- sum(mod_dwols$analysis$last.stage >= 2L)
  
  test_that("autoplot.DTRreg: Stage 2 Fitted plot has correct number of rows", {
    expect_equal(nrow(p_fit2$data), n_stage2)
  })
  
  test_that("autoplot.DTRreg: Stage 2 title contains 'Stage 2'", {
    expect_match(p_fit2$labels$title, "Stage 2")
  })
  
  test_that("autoplot.DTRreg: Stage 2 residual oracle matches Y - fitted", {
    obsK     <- mod_dwols$analysis$Y[[2L]]
    fitK     <- mod_dwols$analysis$outcome.fit[[2L]]$fitted.values
    expected <- obsK - fitK
    expect_equal(p_fit2$data$y, unname(expected))
  })
})


# autoplot.DTRreg - intercept-only blip produces no covariate plots

test_that("autoplot.DTRreg: intercept-only blip produces no covariate plot for that stage", {
  mod_int <- suppressMessages(
    DTRreg(twoStageCont$Y,
           blip.mod = list(~1, ~X2),
           treat.mod = list(A1 ~ X1, A2 ~ X2),
           tf.mod = list(~X1, ~X2),
           data = twoStageCont,
           method = "dwols",
           var.estim = "none")
  )
  plots <- autoplot.DTRreg(mod_int)
  # Stage 1 has intercept-only blip - no X covariate plot
  expect_false(any(grepl("Stage1_X", names(plots))))
  # Stage 1 still has Fitted plot
  expect_true("Stage1_Fitted" %in% names(plots))
  # Stage 2 still has X2 covariate plot
  expect_true("Stage2_X2" %in% names(plots))
})


# autoplot.DTRreg - GEST object

test_that("autoplot.DTRreg: works for GEST object", {
  plots <- autoplot.DTRreg(mod_gest)
  expect_true(is.list(plots))
  expect_true("Stage1_Fitted" %in% names(plots))
  expect_true(inherits(plots[["Stage1_Fitted"]], "ggplot"))
})


# plot.DTRreg

test_that("plot.DTRreg: returns invisibly", {
  pdf(NULL)
  on.exit(dev.off())
  res <- withVisible(plot.DTRreg(mod_dwols))
  expect_false(res$visible)
})

test_that("plot.DTRreg: invisible return value is a list", {
  pdf(NULL)
  on.exit(dev.off())
  res <- plot.DTRreg(mod_dwols)
  expect_true(is.list(res))
})

test_that("plot.DTRreg: return value matches autoplot.DTRreg output", {
  pdf(NULL)
  on.exit(dev.off())
  res_plot <- plot.DTRreg(mod_dwols)
  res_autoplot <- autoplot.DTRreg(mod_dwols)
  expect_equal(names(res_plot), names(res_autoplot))
})

test_that("plot.DTRreg: each returned element is a ggplot", {
  pdf(NULL)
  on.exit(dev.off())
  res <- plot.DTRreg(mod_dwols)
  ggplot_elements <- sapply(res, function(p) inherits(p, "ggplot"))
  # At minimum the residual plots should all be ggplot objects
  resid_plots <- res[grepl("Fitted|_X", names(res))]
  expect_true(all(sapply(resid_plots, function(p) inherits(p, "ggplot"))))
})

test_that("plot.DTRreg: works for GEST object without error", {
  pdf(NULL)
  on.exit(dev.off())
  expect_no_error(plot.DTRreg(mod_gest))
})


# autoplot.DTRreg - submodel plots (treatment, complete cases)

local({
  plots <- autoplot.DTRreg(mod_dwols)
  nms   <- names(plots)
  
  test_that("autoplot.DTRreg: treatment model plots present for each stage", {
    expect_true("Stage1_Treatment" %in% nms)
    expect_true("Stage2_Treatment" %in% nms)
  })
  
  test_that("autoplot.DTRreg: no CompleteCases plots when censoring not modeled", {
    expect_false(any(grepl("CompleteCases", names(plots))))
  })
  
  test_that("autoplot.DTRreg: treatment plots are ggplot objects", {
    tx_plots <- plots[grepl("_Treatment$", nms)]
    expect_true(all(sapply(tx_plots, inherits, what = "ggplot")))
  })
  
  n_treatment_plots_total <- n_stages   # 1 per stage
  n_total_plots <- n_resid_plots_total + n_treatment_plots_total  # 6

  test_that("autoplot.DTRreg: total plot count equals residual plus treatment plots", {
    expect_equal(length(plots), n_total_plots)
  })
})

test_that("autoplot.DTRreg: no CompleteCases plots for DWSurv object", {
  data(twoStageCens)
  mod_surv <- suppressMessages(
    DWSurv(time = list(~T1, ~T2),
           blip.mod = list(~X11, ~X21),
           treat.mod = list(A1 ~ X11, A2 ~ X21),
           tf.mod = list(~X11, ~X21),
           cens.mod = list(delta~X11, delta~X21),
           var.estim = "none",
           data = twoStageCens)
  )
  plots <- autoplot.DTRreg(mod_surv)
  nms   <- names(plots)
  expect_false(any(grepl("CompleteCases", nms)))
  expect_true(any(grepl("Censoring", nms)))
})

.getY <- function(obj, isSurvival) {
  if (isSurvival) {
    # numeric vector of 0's
    Y <- obj$data |> nrow() |> numeric()
  } else {
    # provided outcome
    Y <- obj$outcome
  }
  Y
}

.getCompleteCaseProbability <- function(obj, quiet, isSurvival) {
  ### Complete case weights
  # object will be a list containing `$last.stage`, `$prob.complete.case`, and
  # `$fitted.censor.models`.
  # `$last.stage` is an integer vector object of length n containing the 
  #   final dp for which each participant has complete data; 
  # `$prob.complete.case` is a numeric matrix object containing the  
  #   weight for each participant at each stage. Note that 
  #   prob.complete.case = NA_real_ indicates that the participant has been lost 
  #   to follow-up.
  # `$cens.mod.fitted` is a list containing NA_character_ values if censoring
  #    was not fit and/or `glm` objects.
  complete_case_info <- .completeCaseProbability(obj, quiet, isSurvival)
  
  if (!isSurvival) {
    # using cumulative probabilities when not survival
    complete_case_info$prob.complete.case <-
      apply(complete_case_info$prob.complete.case, 1L, cumprod) |> t() |>
      matrix(ncol = obj$K)
  }
  complete_case_info
}

.getTxInfo <- function(obj) {
  # Need to ensure that treatments align with user specification and
  # are in expected form (0/1, factor, numeric)

  # process stage treatment information
  # elements include A, A.hat, cts, cts.obj, and treat.mod.fitted
  # Note that we keep A here because we modify `data` to be the optimal
  # treatment as we go along; i.e., A is always the observed treatment
  .processTreatment(stage.models = obj$models, 
                    data = obj$data, 
                    tx.var = obj$dependent.vars$treat, 
                    tx.range = obj$tx.range, 
                    tx.type = obj$tx.type,
                    tx.family = obj$tx.family)
  
}

.addYdelta <- function(obj, step.obj, Y, isSurvival) {
  
  if (isSurvival) {
    # survival outcome is log(t_k + sum_{j = k+1}^{K} t_hat)
    step.obj$Y <- obj$data[, obj$dependent.vars$time]
    step.obj$Y <- log(step.obj$Y + Y)
    
    # status will be NA if interactive and no censoring indicated
    # if not interactive, the user has to specify a status that is
    # expected to contain only 1s
    if (!is.na(obj$dependent.vars$status)) {
      step.obj$delta <- obj$data[, obj$dependent.vars$status]
    } else {
      step.obj$delta <- rep(1.0, step.obj$n)
    }
  } else {
    step.obj$Y <- Y
    step.obj$delta <- rep(1.0, step.obj$n)
  }
  step.obj
}


.addTxWgt <- function(obj, step.obj) {
  
  if (obj$tx.weight == "none" || obj$method == "gest") {
    step.obj$tx.wgt <- rep(1.0, step.obj$n)
  } else if (obj$tx.weight %in% c("manual")) {
    step.obj$tx.wgt <- obj$tx.wgt.man
  } else if (obj$tx.weight %in% c("manual.with.censor")) {
    step.obj$tx.cens.wgt <- obj$tx.wgt.man
    step.obj$cens.wgt <- NULL
  } else {
    step.obj$tx.wgt <- .treatmentWeights(A = step.obj$A, 
                                         A.hat = step.obj$A.hat,
                                         tx.weight = obj$tx.weight,
                                         tx.mod.fitted = step.obj$tx.mod.fitted,
                                         cts.obj = step.obj$cts.obj,
                                         n.bins = obj$n.bins,
                                         data = obj$data)
  }
  step.obj
}

.addWgts <- function(obj, step.obj, complete.case.info) {
  
  step.obj$cens.wgt <- complete.case.info$prob.complete.case
  wgts <- step.obj$cens.wgt
  
  step.obj <- .addTxWgt(obj, step.obj)
  
  if (obj$tx.weight %in% c("manual.with.censor")) {
    wgts <- step.obj$tx.cens.wgt
  } else if (obj$tx.weight != "none" && obj$method != "gest") {
    wgts <- wgts * step.obj$tx.wgt
  }
  wgts <- wgts * step.obj$delta
  step.obj$wgts <- wgts
  step.obj
}

.performMethod <- function(obj, step.obj) {
  
  if (obj$method == "dwols") {
    delta_is_1 <- step.obj$delta > 0.5
    .dwols(Y = step.obj$Y[delta_is_1],
           A = step.obj$A[delta_is_1], 
           data = obj$data[delta_is_1, , drop = FALSE],
           wgts = step.obj$wgts[delta_is_1],
           cts.obj = step.obj$cts.obj, 
           tx.var = step.obj$tx.var)
  } else if (obj$method == "gest") {
    .gest(A = step.obj$A, 
          wgts = step.obj$wgts,
          data = obj$data,
          Y = step.obj$Y, 
          A.hat = step.obj$A.hat,
          tx.var = step.obj$tx.var,
          cts.obj = step.obj$cts.obj, 
          treat.mod.fitted = step.obj$tx.mod.fitted)
  }
}

.getStage <- function(obj, k, quiet, complete.case.info, isSurvival, Y) {
  
  if (!quiet) message("Stage ", k, " analysis")
  
  stage_cases <- complete.case.info$last.stage >= k
  
  # Subset all vectors/matrices to stage_cases and k
  obj$data <- obj$data[stage_cases, , drop = FALSE]
  obj$tx.wgt.man <- if (is.list(obj$tx.wgt.man)) obj$tx.wgt.man[[k]][stage_cases] else obj$tx.wgt.man
  obj$dependent.vars$time <- obj$dependent.vars$time[k]
  obj$dependent.vars$treat <- obj$dependent.vars$treat[k]
  obj$models <- obj$models[[k]]
  obj$tx.range <- obj$tx.range[[k]]
  
  Y <- Y[stage_cases]
  
  complete.case.info$last.stage <- complete.case.info$last.stage[stage_cases]
  complete.case.info$prob.complete.case <- complete.case.info$prob.complete.case[stage_cases, k]
  complete.case.info$d.hat <- complete.case.info$d.hat[stage_cases, k]
  complete.case.info$cens.mod.fitted <- complete.case.info$cens.mod.fitted[[k]]
  
  tx_info <- .getTxInfo(obj)
  obj$data[, obj$dependent.vars$treat] <- tx_info$A
  
  # initialize step specific result list
  step_obj <- c(
    list(dp = k,
         tx.var = obj$dependent.vars$treat,
         models = obj$models,
         n = nrow(obj$data)),
    tx_info
  )
  
  step_obj <- .addYdelta(obj, step_obj, Y, isSurvival)

  step_obj <- .addWgts(obj, step_obj, complete.case.info)
  
  step_obj <- c(step_obj, .performMethod(obj, step_obj))
  
  if (!quiet) {
    message("beta: ", paste(step_obj$beta, collapse = ", "))
    message("psi: ", paste(step_obj$psi, collapse = ", "))
  }
  
  step_obj$fitted.values <- fitted(step_obj$outcome.fit)
  step_obj$residuals <- step_obj$Y[step_obj$delta > 0.5] - step_obj$fitted.values
  step_obj$blip.data <- obj$data[step_obj$delta > 0.5, , drop = FALSE]
  
  step_obj  
}

#' Main procedure for estimating parameters across all decision points for
#'   DTRreg
#'
#' @noRd
#' @param obj A list object. Analysis settings. Expected to include
#'   K: An integer object. The number of decision points (dp); 
#'   outcome: A numeric vector object. The outcome of interest; only
#'     present if not survival
#'   models: A list of length K, each element a list of dp specific models; 
#'   data: A data.frame object. The full covariate and treatment data; 
#'   type: A character object. If "DTR", DTR estimation using regret; otherwise
#'     effect estimation using blip; 
#'   dependent.vars: A list of dependent variables. If not survival, this 
#'     includes only 
#'     treat.vars: A character vector object of length K. The treatment variable 
#'       for each dp; 
#'     if survival, also inclust status.vars and time.vars
#'   tx.range: NULL or a list of numeric vector objects of length 2. The range of
#'     allowed treatment values when tx is continuous.
#'   method: A character object. Must be one of "dwols" or "gest"; 
#'   var.estim: A character object. The covariance estimation method; must be
#'     one of "bootstrap", "sandwich", or "none".
#'   tx.weight: A character object. The type of treatment weighting.
#'   tx.wgt.man: A list of user specified treatment weights. 
#'   tx.type: A character indicating if treatment is binary, multi-nomial, or cont
#'   n.bins: An integer required when tx is continuous and weight is qpom or pom
#'   tx.family: A family object
#'   censoring.modeled: A logical object. If TRUE, element `$models` must 
#'     include censoring models.
#'   boot.controls: A list of control variables when bootstrap variance
#'     is requested.
#'   manual.censor.weight: A logical object. If TRUE, censoring weights are
#'     included in the user provided treatment weights. Provided only if
#'     procedure being called from DWsurv.
#' @param quiet A logical. If TRUE, screen prints are suppressed. Used to quiet
#'   call made by calls from bootstrap
#' @param isSurvival A logical. If TRUE analysis is for survival data.
#'   
#' @return A list object containing
#' 
#' @importFrom stats complete.cases model.matrix model.response
#' @include Ahat.R bootstrapStep.R dwols.R gest.R sandwich.R utils.R weights.R
#' @keywords internal
.dtrProcedure <- function(obj, quiet, isSurvival = FALSE) {

  # Confirm that everything is provided as expected
  obj_musts <- c("K", "dependent.vars", "models", "tx.range", 
                 "var.estim", "method", "data", "type", "tx.weight",
                 "censoring.modeled", "tx.wgt.man",
                 "tx.type", "n.bins",
                 "tx.family", "boot.controls", "full.cov")
  if (!isSurvival) {
    obj_musts <- c("outcome", obj_musts)
  } else {
    obj_musts <- c("manual.censor.weight", obj_musts)
  }

  stopifnot(
    "`obj` must be a list" = is.list(obj) && all(obj_musts %in% names(obj)),
    "`quiet` must be a logical" = is.logical(quiet) && is.vector(quiet) &&
      length(quiet) == 1L,
    "`isSurvival` must be a logical" = is.logical(isSurvival) && 
      is.vector(isSurvival) && length(isSurvival) == 1L
  )
  
  obj <- obj[obj_musts]

  # make sure that we aren't calling gest for multinomial treatment
  if (obj$method == "gest" && obj$tx.type == "multi") {
    stop("G-estimation for multinomial treatments is not yet implemented",
         call. = FALSE)
  }
  
  # get initial Y value
  Y <- .getY(obj, isSurvival)

  # Complete case weights
  complete_case_info <- .getCompleteCaseProbability(obj, quiet, isSurvival)
  
  # stages will contain the analysis results for each decision point
  # in the order of stages (NOT steps)
  obj$stages <- vector("list", length = obj$K)

  for (k in obj$K:1L) {
    
    obj$stages[[k]] <- .getStage(obj, k, quiet, complete_case_info, isSurvival, Y)
    
    # There is the possibility that the optimal treatment identified in this
    # stage will change the optimal treatment in later stages that have already
    # been evaluated. Need to re-evaluate the optimal treatment in each later 
    # stage and include regret/blip adjustment at the (potentially) new optimal

    obj$stages <- .stageTx(obj$stages, complete_case_info, obj$data, quiet)

    if (isSurvival) {
      Y <- .stageYSurvival(obj$dependent.vars$time,
                           obj$stages[k:obj$K],
                           complete.case.info = complete_case_info, 
                           data = obj$data, 
                           type = obj$type)
    } else {
      Y <- .stageY(obj$outcome, obj$stages[k:obj$K], 
                   complete.case.info = complete_case_info, 
                   data = obj$data, 
                   type = obj$type)
    }
  }
  
  obj <- c(obj, complete_case_info)
  
  obj$regret <- .stageRegret(obj$stages, complete_case_info, obj$data, obj$type)

  # Y if all optimal treatments followed
  obj$opt.Y <- Y
  
  obj$treat.vars <- obj$dependent.vars$treat
  
  obj
}

data(twoStageCont)
n <- nrow(twoStageCont)

# Minimal valid obj for non-survival, two-stage, binary treatment, DWOLS
make_obj <- function(method = "dwols",
                     var.estim = "none",
                     tx.weight = "ipw",
                     type = "DTR") {
  list(
    K = 2L,
    outcome = twoStageCont$Y,
    data = twoStageCont,
    models = list(
      list(blip = ~X1, treat = A1 ~ X1, tf = ~X1, cens = ~1),
      list(blip = ~X2, treat = A2 ~ X2, tf = ~X2, cens = ~X1)
    ),
    dependent.vars = list(treat = c("A1", "A2")),
    tx.range = list(NA_real_, NA_real_),
    method = method,
    var.estim = var.estim,
    type = type,
    tx.weight = tx.weight,
    censoring.modeled = FALSE,
    tx.wgt.man = NA,
    tx.type = "bin",
    n.bins = NA_integer_,
    tx.family = NA,
    boot.controls = list(),
    full.cov = FALSE
  )
}

test_that(".getY: non-survival returns obj$outcome", {
  obj <- make_obj()
  expect_equal(.getY(obj, isSurvival = FALSE), obj$outcome)
})

test_that(".getY: survival returns numeric zero vector of length n", {
  obj <- make_obj()
  res <- .getY(obj, isSurvival = TRUE)
  expect_true(is.numeric(res))
  expect_length(res, n)
  expect_true(all(res == 0.0))
})

local({
  obj <- make_obj()
  res <- .getCompleteCaseProbability(obj, quiet = TRUE, isSurvival = FALSE)
  
  test_that(".getCompleteCaseProbability: returns list with required names", {
    expect_true(all(c("last.stage", "prob.complete.case",
                      "d.hat", "cens.mod.fitted") %in% names(res)))
  })
  
  test_that(".getCompleteCaseProbability: prob.complete.case is n x K matrix", {
    expect_true(is.matrix(res$prob.complete.case))
    expect_equal(dim(res$prob.complete.case), c(n, 2L))
  })
  
  test_that(".getCompleteCaseProbability: non-survival applies cumprod across stages", {
    # Oracle: for complete data, cumprod of all-1 weights stays 1
    # For stage 1: prob[,1] should equal the raw weight (no previous stages)
    # For stage 2: prob[,2] should be product of stage 1 and stage 2 weights
    raw <- .completeCaseProbability(obj, quiet = TRUE, isSurvival = FALSE)
    expected <- apply(raw$prob.complete.case, 1L, cumprod) |> t() |> matrix(ncol = 2L)
    expect_equal(res$prob.complete.case, expected)
  })
  
  test_that(".getCompleteCaseProbability: last.stage is integer vector of length n", {
    expect_true(is.integer(res$last.stage))
    expect_length(res$last.stage, n)
  })
})

local({
  obj <- make_obj()
  # Subset to stage 1 as .getStage would
  cc <- .getCompleteCaseProbability(obj, quiet = TRUE, isSurvival = FALSE)
  sc <- cc$last.stage >= 1L
  obj$data <- obj$data[sc, , drop = FALSE]
  obj$dependent.vars$treat <- obj$dependent.vars$treat[1L]
  obj$models <- obj$models[[1L]]
  obj$tx.range <- obj$tx.range[[1L]]
  
  res <- .getTxInfo(obj)
  
  test_that(".getTxInfo: returns list with required element names", {
    expect_named(res, c("A", "A.hat", "cts", "cts.obj", "tx.mod.fitted"),
                 ignore.order = TRUE)
  })
  
  test_that(".getTxInfo: cts is 'bin' for binary treatment", {
    expect_equal(res$cts, "bin")
  })
  
  test_that(".getTxInfo: cts.obj is Binary R6 object", {
    expect_true(inherits(res$cts.obj, "Binary"))
  })
  
  test_that(".getTxInfo: A is 0/1 integer vector", {
    expect_true(all(res$A %in% c(0L, 1L)))
  })
})

local({
  n_s <- 50L
  obj_ns <- list(
    data = data.frame(X = rnorm(n_s)),
    dependent.vars = list(status = NA_character_, time = NA_character_)
  )
  step_obj <- list(n = n_s)
  Y <- rnorm(n_s)
  
  test_that(".addYdelta: non-survival sets Y to provided Y", {
    res <- .addYdelta(obj_ns, step_obj, Y, isSurvival = FALSE)
    expect_equal(res$Y, Y)
  })
  
  test_that(".addYdelta: non-survival sets delta to all 1s", {
    res <- .addYdelta(obj_ns, step_obj, Y, isSurvival = FALSE)
    expect_equal(res$delta, rep(1.0, n_s))
  })
  
  test_that(".addYdelta: survival with NA status sets delta to all 1s", {
    obj_surv <- obj_ns
    obj_surv$data$T1 <- abs(rnorm(n_s)) + 0.1
    obj_surv$dependent.vars$time <- "T1"
    obj_surv$dependent.vars$status <- NA_character_
    Y_surv <- abs(rnorm(n_s))
    res <- .addYdelta(obj_surv, step_obj, Y_surv, isSurvival = TRUE)
    expect_equal(res$delta, rep(1.0, n_s))
  })
  
  test_that(".addYdelta: survival with status column sets delta from data", {
    obj_surv <- obj_ns
    delta_true <- rbinom(n_s, 1L, 0.7)
    obj_surv$data$T1 <- abs(rnorm(n_s)) + 0.1
    obj_surv$data$status <- delta_true
    obj_surv$dependent.vars$time <- "T1"
    obj_surv$dependent.vars$status <- "status"
    Y_surv <- abs(rnorm(n_s))
    res <- .addYdelta(obj_surv, step_obj, Y_surv, isSurvival = TRUE)
    expect_equal(res$delta, delta_true)
  })
  
  test_that(".addYdelta: survival Y is log(time + Y)", {
    obj_surv <- obj_ns
    T1 <- abs(rnorm(n_s)) + 0.1
    obj_surv$data$T1 <- T1
    obj_surv$dependent.vars$time <- "T1"
    obj_surv$dependent.vars$status <- NA_character_
    Y_surv <- abs(rnorm(n_s))
    res <- .addYdelta(obj_surv, step_obj, Y_surv, isSurvival = TRUE)
    expect_equal(res$Y, log(T1 + Y_surv))
  })
})

local({
  n_s <- 30L
  set.seed(42L)
  dat_s <- data.frame(X = rnorm(n_s), A = rbinom(n_s, 1L, 0.5))
  cts_b <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  fit_tx <- glm(A ~ X, dat_s, family = "binomial")
  Ahat <- predict(fit_tx, type = "response")
  
  step_s <- list(
    n = n_s,
    A = as.numeric(dat_s$A),
    A.hat = Ahat,
    tx.mod.fitted = fit_tx,
    cts.obj = cts_b
  )
  
  make_obj_s <- function(tx.weight, tx.wgt.man = NA, method = "dwols") {
    list(tx.weight = tx.weight, method = method,
         tx.wgt.man = tx.wgt.man, n.bins = NA_integer_,
         data = dat_s)
  }
  
  test_that(".addTxWgt: 'none' sets tx.wgt to all 1s", {
    res <- .addTxWgt(make_obj_s("none"), step_s)
    expect_equal(res$tx.wgt, rep(1.0, n_s))
  })
  
  test_that(".addTxWgt: 'gest' method sets tx.wgt to all 1s", {
    res <- .addTxWgt(make_obj_s("ipw", method = "gest"), step_s)
    expect_equal(res$tx.wgt, rep(1.0, n_s))
  })
  
  test_that(".addTxWgt: 'manual' sets tx.wgt to tx.wgt.man", {
    wgt <- runif(n_s, 0.5, 2.0)
    res <- .addTxWgt(make_obj_s("manual", tx.wgt.man = wgt), step_s)
    expect_equal(res$tx.wgt, wgt)
  })
  
  test_that(".addTxWgt: 'manual.with.censor' sets tx.cens.wgt and nulls cens.wgt", {
    wgt <- runif(n_s, 0.5, 2.0)
    step_mc <- step_s
    step_mc$cens.wgt <- rep(1.0, n_s)
    res <- .addTxWgt(make_obj_s("manual.with.censor", tx.wgt.man = wgt), step_mc)
    expect_equal(res$tx.cens.wgt, wgt)
    expect_null(res$cens.wgt)
  })
  
  test_that(".addTxWgt: computed path sets tx.wgt of length n", {
    res <- .addTxWgt(make_obj_s("ipw"), step_s)
    expect_length(res$tx.wgt, n_s)
    expect_true(is.numeric(res$tx.wgt))
  })
})

local({
  n_s <- 20L
  cens <- rep(1.0, n_s)
  delta <- rep(1.0, n_s)
  cc <- list(prob.complete.case = cens)
  
  step_base <- list(n = n_s, delta = delta, A = rep(1.0, n_s))
  
  test_that(".addWgts: 'none' — wgts equals cens.wgt * delta", {
    obj_s <- list(tx.weight = "none", method = "dwols",
                  tx.wgt.man = NA, n.bins = NA_integer_,
                  data = data.frame(X = rnorm(n_s)))
    res <- .addWgts(obj_s, step_base, cc)
    expect_equal(res$wgts, cens * delta)
  })
  
  test_that(".addWgts: 'manual.with.censor' — wgts equals tx.cens.wgt * delta", {
    wgt <- runif(n_s, 0.5, 2.0)
    obj_s <- list(tx.weight = "manual.with.censor", method = "dwols",
                  tx.wgt.man = wgt, n.bins = NA_integer_,
                  data = data.frame(X = rnorm(n_s)))
    step_mc <- step_base
    step_mc$cens.wgt <- cens
    res <- .addWgts(obj_s, step_mc, cc)
    expect_equal(res$wgts, wgt * delta)
  })
})

local({
  obj <- make_obj(method = "dwols")
  cc <- .getCompleteCaseProbability(obj, quiet = TRUE, isSurvival = FALSE)
  # Get a realistic step_obj via .getStage
  step <- .getStage(obj, k = 2L, quiet = TRUE, complete.case.info = cc,
                    isSurvival = FALSE, Y = obj$outcome)
  
  # Build stage-specific obj as .getStage would
  sc <- cc$last.stage >= 2L
  obj_s <- obj
  obj_s$data <- obj$data[sc, , drop = FALSE]
  obj_s$dependent.vars$treat <- obj$dependent.vars$treat[2L]
  obj_s$models <- obj$models[[2L]]
  obj_s$tx.range <- obj$tx.range[[2L]]
  
  test_that(".performMethod DWOLS: returns list with outcome.fit, psi, beta", {
    res <- .performMethod(obj_s, step)
    expect_named(res, c("outcome.fit", "beta", "psi"), ignore.order = TRUE)
  })
  
  test_that(".performMethod DWOLS: psi names contain tx.var", {
    res <- .performMethod(obj_s, step)
    expect_true(all(grepl("A2", names(res$psi))))
  })
  
  test_that(".performMethod DWOLS: beta names do not contain tx.var", {
    res <- .performMethod(obj_s, step)
    expect_false(any(grepl("^A2", names(res$beta))))
  })
  
  test_that(".performMethod DWOLS: psi and beta account for all coefficients", {
    res <- .performMethod(obj_s, step)
    expect_setequal(c(names(res$psi), names(res$beta)),
                    names(coef(res$outcome.fit)))
  })
})

local({
  obj <- make_obj(method = "gest")
  cc <- .getCompleteCaseProbability(obj, quiet = TRUE, isSurvival = FALSE)
  step <- .getStage(obj, k = 2L, quiet = TRUE, complete.case.info = cc,
                    isSurvival = FALSE, Y = obj$outcome)
  
  sc <- cc$last.stage >= 2L
  obj_s <- obj
  obj_s$data <- obj$data[sc, , drop = FALSE]
  obj_s$dependent.vars$treat <- obj$dependent.vars$treat[2L]
  obj_s$models <- obj$models[[2L]]
  obj_s$tx.range <- obj$tx.range[[2L]]
  
  test_that(".performMethod GEST: returns list with outcome.fit, psi, beta", {
    res <- .performMethod(obj_s, step)
    expect_named(res, c("outcome.fit", "beta", "psi"), ignore.order = TRUE)
  })
  
  test_that(".performMethod GEST: outcome.fit has class GEST", {
    res <- .performMethod(obj_s, step)
    expect_true(inherits(res$outcome.fit, "GEST"))
  })
})


local({
  obj <- make_obj()
  cc <- .getCompleteCaseProbability(obj, quiet = TRUE, isSurvival = FALSE)
  
  res_k2 <- .getStage(obj, k = 2L, quiet = TRUE, complete.case.info = cc,
                      isSurvival = FALSE, Y = obj$outcome)
  res_k1 <- .getStage(obj, k = 1L, quiet = TRUE, complete.case.info = cc,
                      isSurvival = FALSE, Y = obj$outcome)
  
  test_that(".getStage: returns list with core structural elements", {
    expect_true(all(c("dp", "tx.var", "models", "n",
                      "A", "A.hat", "cts", "cts.obj",
                      "Y", "delta", "wgts",
                      "outcome.fit", "psi", "beta") %in% names(res_k2)))
  })
  
  test_that(".getStage: dp matches k", {
    expect_equal(res_k2$dp, 2L)
    expect_equal(res_k1$dp, 1L)
  })
  
  test_that(".getStage: tx.var matches dependent.vars$treat[k]", {
    expect_equal(res_k2$tx.var, obj$dependent.vars$treat[2L])
    expect_equal(res_k1$tx.var, obj$dependent.vars$treat[1L])
  })
  
  test_that(".getStage: n equals sum of stage cases", {
    expect_equal(res_k2$n, sum(cc$last.stage >= 2L))
    expect_equal(res_k1$n, sum(cc$last.stage >= 1L))
  })
  
  test_that(".getStage: A is 0/1 integer vector of length n", {
    expect_true(all(res_k2$A %in% c(0L, 1L)))
    expect_length(res_k2$A, res_k2$n)
  })
  
  test_that(".getStage: psi names contain tx.var", {
    expect_true(all(grepl(res_k2$tx.var, names(res_k2$psi))))
  })
  
  test_that(".getStage: beta names do not contain tx.var", {
    expect_false(any(grepl(paste0("^", res_k2$tx.var), names(res_k2$beta))))
  })
  
  test_that(".getStage: delta is all 1s for non-survival", {
    expect_equal(res_k2$delta, rep(1.0, res_k2$n))
  })
  
  test_that(".getStage: wgts is numeric vector of length n", {
    expect_true(is.numeric(res_k2$wgts))
    expect_length(res_k2$wgts, res_k2$n)
  })
})

test_that(".dtrProcedure: stops when required element missing from obj", {
  obj <- make_obj()
  obj$K <- NULL

  expect_error(.dtrProcedure(obj, quiet = TRUE), "`obj`")
  expect_error(.dtrProcedure(make_obj(), quiet = "yes"), "`quiet`")
  expect_error(.dtrProcedure(make_obj(), quiet = TRUE, isSurvival = 0L),
               "`isSurvival`")
  obj <- make_obj(method = "gest")
  obj$tx.type <- "multi"
  expect_error(.dtrProcedure(obj, quiet = TRUE), "multinomial")
})


local({
  obj <- make_obj(method = "dwols")
  res <- suppressMessages(.dtrProcedure(obj, quiet = TRUE))
  
  test_that(".dtrProcedure DWOLS: returns a list", {
    expect_true(is.list(res))
  })
  
  test_that(".dtrProcedure DWOLS: stages element contains K results", {
    expect_length(res$stages, obj$K)
  })
  
  test_that(".dtrProcedure DWOLS: opt.Y is numeric vector of length n", {
    expect_true(is.numeric(res$opt.Y))
    expect_length(res$opt.Y, n)
  })
  
  test_that(".dtrProcedure DWOLS: last.stage is integer vector of length n", {
    expect_true(is.integer(res$last.stage))
    expect_length(res$last.stage, n)
  })
  
  test_that(".dtrProcedure DWOLS: treat.vars matches dependent.vars$treat", {
    expect_equal(res$treat.vars, obj$dependent.vars$treat)
  })
  
  test_that(".dtrProcedure DWOLS: regret is list of length K", {
    expect_length(res$regret, obj$K)
  })
})

local({
  obj <- make_obj(method = "gest")
  res <- suppressMessages(.dtrProcedure(obj, quiet = TRUE))
  
  test_that(".dtrProcedure GEST: opt.Y is numeric vector of length n", {
    expect_true(is.numeric(res$opt.Y))
    expect_length(res$opt.Y, n)
  })
})

#' Estimate parameters using g-estimation
#'   
#' @noRd
#' @param A A numeric vector object of length n. Must contain only 0/1 values.
#' @param wgts A numeric vector object of length n. The complete case + tx weights.
#' @param Y A numeric vector object of length n. The outcome of interest.
#' @param A.hat A numeric vector object of length n. The estimated treatment.
#' @param tx.var A character object. The treatment variable name.
#' @param cts.obj An R6 object.
#' @param treat.mod.fitted A fitted regression object.
#' @param data A data.frame object with n rows. The covariate and treatment data.
#' 
#' @return A list object. Element `$outcome.fit` a list of class GEST providing 
#'   the full outcome model and the estimated parameters; element
#'   `$beta` is the parameter estimates of the treatment-free component of
#'   the model; and `$psi` is the parameter estimates of the blip component.
#'
#' @importFrom R6 is.R6
#' @importFrom stats terms
#'   
#' @include treatmentClasses.R
#' @keywords internal
.gest <- function(A, wgts, data, Y, A.hat, tx.var, cts.obj, treat.mod.fitted, ...) {
  
  stopifnot(
    "`A` must be a numeric vector object" = !missing(A) &&
      is.numeric(A) && is.vector(A) && length(A) > 1L,
    "`wgts` must be a numeric vector object with length equivalent to that of `A`" = 
      !missing(wgts) && is.numeric(wgts) && is.vector(wgts) && 
      length(wgts) == length(A),
    "`Y` must be a numeric vector object" = !missing(Y) &&
      is.numeric(Y) && is.vector(Y) && length(Y) == length(A),
    "`data` must be a data.frame object" = !missing(data) &&
      is.data.frame(data) && nrow(data) == length(A),
    "`A.hat` must be a numeric vector object" = !missing(A.hat) &&
      is.numeric(A.hat) && is.vector(A.hat) && length(A.hat) == length(A),
    "`tx.var` must be a character" = !missing(tx.var) &&
      is.character(tx.var) && is.vector(tx.var) && length(tx.var) == 1L,
    "`cts.obj` must be an R6 object" = !missing(cts.obj) && R6::is.R6(cts.obj),
    "`treat.mod.fitted` must be NULL a fitted lm regression object" = 
      !missing(treat.mod.fitted) && 
      {is.list(treat.mod.fitted) || inherits(treat.mod.fitted, "glm")}
  )
  
  Hd <- cts.obj$Hd(data, A)
  
  Hw <- cts.obj$Hw(data = data, A = A, A.hat = A.hat, wgt = wgts, 
                   treat.mod.fitted = treat.mod.fitted)

  # get estimates
  est <- tryCatch(solve(crossprod(Hw, Hd), crossprod(Hw, Y)) |> drop(),
                  error = function(e) {
                    stop("unable to invert matrix\n\t", e$message,
                         call. = FALSE)
                    })
  names(est) <- colnames(Hd)
  
  # creating a fake regression object to allow for a new S3 prediction method
  outcome.fit <- list("formula" = cts.obj$full.model,
                      "fitted.values" = drop(Hd %*% est),
                      "coefficients" = est)
  class(outcome.fit) <- c("GEST", class(outcome.fit))
  
  blip_vars <- cts.obj$blip_params(est)

  psi <- est[blip_vars]
  names(psi) <- names(blip_vars)

  list("outcome.fit" = outcome.fit,
       "beta" = est[-blip_vars],
       "psi" = psi)
}

#' Internal prediction method for parameters obtained by G-estimation
#'
#' @noRd
#' @param object An S3 object. The full outcome model and estimated parameters.
#' @param newdata A data.frame object. The data on which to make predictions#
#' 
#' @return A numeric vector. The predicted outcome
#' 
#' @importFrom stats model.matrix predict
#' @keywords internal
predict.GEST <- function(object, newdata, ...) {
  mm <- stats::model.matrix(object$formula, newdata)
  mm %*% object$coefficients |> drop()
}

set.seed(42L)
n <- 30L

dat <- data.frame(
  X = cos(seq(0.0, pi, length.out = n)),
  A = as.numeric(rbinom(n, 1L, 0.5)),
  Y = rnorm(n)
)

models <- list(blip = ~X, treat = A ~ X, tf = ~X)

cts_bin <- Binary$new(tf.model = models$tf,
                      blip.model = models$blip,
                      tx.var = "A")

# Fitted treatment model
fit_tx <- glm(models$treat, dat, family = "binomial")
Ahat <- predict(fit_tx, type = "response")
wgts <- rep(1.0, n)

# Fitted outcome models via trusted Layer 4 functions
dwols_fit <- .dwols(Y = dat$Y,
                    A = dat$A,
                    data = dat,
                    wgts = wgts,
                    cts.obj = cts_bin,
                    tx.var = "A")

gest_fit <- .gest(A = dat$A,
                  wgts = wgts,
                  data = dat,
                  Y = dat$Y,
                  A.hat = Ahat,
                  tx.var = "A",
                  cts.obj = cts_bin,
                  treat.mod.fitted = fit_tx)

n_theta <- length(coef(dwols_fit$outcome.fit))


test_that(".sandwich DWOLS: returns a matrix", {
  res <- .sandwich(cts.obj = cts_bin,
                   outcome.fit = dwols_fit$outcome.fit,
                   Y = dat$Y,
                   A = dat$A,
                   A.hat = Ahat,
                   wgts = wgts,
                   tx.var = "A",
                   data = dat,
                   treat.mod.fitted = fit_tx,
                   method = "dwols")
  expect_true(is.matrix(res))
})

test_that(".sandwich DWOLS: dimensions are n_theta x n_theta", {
  res <- .sandwich(cts.obj = cts_bin,
                   outcome.fit = dwols_fit$outcome.fit,
                   Y = dat$Y,
                   A = dat$A,
                   A.hat = Ahat,
                   wgts = wgts,
                   tx.var = "A",
                   data = dat,
                   treat.mod.fitted = fit_tx,
                   method = "dwols")
  expect_equal(dim(res), c(n_theta, n_theta))
})

test_that(".sandwich DWOLS: result is symmetric", {
  res <- .sandwich(cts.obj = cts_bin,
                   outcome.fit = dwols_fit$outcome.fit,
                   Y = dat$Y,
                   A = dat$A,
                   A.hat = Ahat,
                   wgts = wgts,
                   tx.var = "A",
                   data = dat,
                   treat.mod.fitted = fit_tx,
                   method = "dwols")
  expect_equal(res, t(res))
})

test_that(".sandwich GEST: dimensions are n_theta x n_theta", {
  res <- .sandwich(cts.obj = cts_bin,
                   outcome.fit = gest_fit$outcome.fit,
                   Y = dat$Y,
                   A = dat$A,
                   A.hat = Ahat,
                   wgts = wgts,
                   tx.var = "A",
                   data = dat,
                   treat.mod.fitted = fit_tx,
                   method = "gest")
  expect_equal(dim(res), c(n_theta, n_theta))
})

test_that(".sandwich GEST: result is symmetric", {
  res <- .sandwich(cts.obj = cts_bin,
                   outcome.fit = gest_fit$outcome.fit,
                   Y = dat$Y,
                   A = dat$A,
                   A.hat = Ahat,
                   wgts = wgts,
                   tx.var = "A",
                   data = dat,
                   treat.mod.fitted = fit_tx,
                   method = "gest")
  expect_equal(res, t(res))
})


local({
  # DWOLS oracle - write the six steps explicitly
  X_theta <- cts_bin$Hd(dat, dat$A)
  HW <- wgts * X_theta
  dat_pred <- dat
  dat_pred[, "A"] <- dat$A
  dat_pred <- cts_bin$prep(dat_pred, dat$A)
  Yhat <- drop(predict(dwols_fit$outcome.fit, dat_pred))
  U <- drop(dat$Y - Yhat) * HW
  dtheta <- {-1.0 / n} * crossprod(HW, X_theta)
  inv_mat <- solve(dtheta, t(U))
  expected <- {1.0 / n} * var(t(inv_mat))
  
  test_that(".sandwich DWOLS: matches explicit oracle computation", {
    res <- .sandwich(cts.obj = cts_bin,
                     outcome.fit = dwols_fit$outcome.fit,
                     Y = dat$Y,
                     A = dat$A,
                     A.hat = Ahat,
                     wgts = wgts,
                     tx.var = "A",
                     data = dat,
                     treat.mod.fitted = fit_tx,
                     method = "dwols")
    expect_equal(res, expected)
  })
})

local({
  # GEST oracle - Hw replaces wgts * Hd
  X_theta <- cts_bin$Hd(dat, dat$A)
  HW <- cts_bin$Hw(data = dat,
                   A = dat$A,
                   A.hat = Ahat,
                   wgt = wgts,
                   treat.mod.fitted = fit_tx)
  dat_pred <- dat
  dat_pred[, "A"] <- dat$A
  dat_pred <- cts_bin$prep(dat_pred, dat$A)
  Yhat <- drop(predict.GEST(gest_fit$outcome.fit, newdata = dat_pred))
  U <- drop(dat$Y - Yhat) * HW
  dtheta <- {-1.0 / n} * crossprod(HW, X_theta)
  inv_mat <- solve(dtheta, t(U))
  expected <- {1.0 / n} * var(t(inv_mat))
  
  test_that(".sandwich GEST: matches explicit oracle computation", {
    res <- .sandwich(cts.obj = cts_bin,
                     outcome.fit = gest_fit$outcome.fit,
                     Y = dat$Y,
                     A = dat$A,
                     A.hat = Ahat,
                     wgts = wgts,
                     tx.var = "A",
                     data = dat,
                     treat.mod.fitted = fit_tx,
                     method = "gest")
    expect_equal(res, expected)
  })
})


test_that(".sandwich: stops when cts.obj is not R6", {
  expect_error(
    .sandwich(cts.obj = list(),
              outcome.fit = dwols_fit$outcome.fit,
              Y = dat$Y,
              A = dat$A,
              A.hat = Ahat,
              wgts = wgts,
              tx.var = "A",
              data = dat,
              treat.mod.fitted = fit_tx,
              method = "dwols"),
    "`cts.obj`"
  )
  expect_error(
    .sandwich(cts.obj = cts_bin,
              outcome.fit = dwols_fit$outcome.fit,
              Y = as.character(dat$Y),
              A = dat$A,
              A.hat = Ahat,
              wgts = wgts,
              tx.var = "A",
              data = dat,
              treat.mod.fitted = fit_tx,
              method = "dwols"),
    "`Y`"
  )
  expect_error(
    .sandwich(cts.obj = cts_bin,
              outcome.fit = dwols_fit$outcome.fit,
              Y = dat$Y,
              A = dat$A[1L:5L],
              A.hat = Ahat,
              wgts = wgts,
              tx.var = "A",
              data = dat,
              treat.mod.fitted = fit_tx,
              method = "dwols"),
    "`A`"
  )
  expect_error(
    .sandwich(cts.obj = cts_bin,
              outcome.fit = dwols_fit$outcome.fit,
              Y = dat$Y,
              A = dat$A,
              A.hat = Ahat,
              wgts = wgts[1L:5L],
              tx.var = "A",
              data = dat,
              treat.mod.fitted = fit_tx,
              method = "dwols"),
    "`wgts`"
  )
  expect_error(
    .sandwich(cts.obj = cts_bin,
              outcome.fit = dwols_fit$outcome.fit,
              Y = dat$Y,
              A = dat$A,
              A.hat = Ahat,
              wgts = wgts,
              tx.var = c("A", "B"),
              data = dat,
              treat.mod.fitted = fit_tx,
              method = "dwols"),
    "`tx.var`"
  )
  expect_error(
    .sandwich(cts.obj = cts_bin,
              outcome.fit = dwols_fit$outcome.fit,
              Y = dat$Y,
              A = dat$A,
              A.hat = Ahat,
              wgts = wgts,
              tx.var = "A",
              data = dat[1L:5L, ],
              treat.mod.fitted = fit_tx,
              method = "dwols"),
    "`data`"
  )
  expect_error(
    .sandwich(cts.obj = cts_bin,
              outcome.fit = dwols_fit$outcome.fit,
              Y = dat$Y,
              A = dat$A,
              A.hat = Ahat,
              wgts = wgts,
              tx.var = "A",
              data = dat,
              treat.mod.fitted = fit_tx,
              method = 1L),
    "`method`"
  )
  dat_sing <- dat
  dat_sing$X <- 0.0
  cts_sing <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  fit_sing <- .dwols(Y = dat_sing$Y,
                     A = dat_sing$A,
                     data = dat_sing,
                     wgts = wgts,
                     cts.obj = cts_sing,
                     tx.var = "A")
  expect_error(
    .sandwich(cts.obj = cts_sing,
              outcome.fit = fit_sing$outcome.fit,
              Y = dat_sing$Y,
              A = dat_sing$A,
              A.hat = Ahat,
              wgts = wgts,
              tx.var = "A",
              data = dat_sing,
              treat.mod.fitted = fit_tx,
              method = "dwols"),
    "unable to invert matrix"
  )
})

# Minimal formula and coefficients matching what .gest() would produce
gest_formula <- ~X + A + A:X
gest_coefs <- c("(Intercept)" = 1.0, "X" = 0.5, "A" = 2.0, "A:X" = -0.5)

# Match the structure that .gest() assigns to outcome.fit
gest_obj <- structure(
  list(formula = gest_formula,
       coefficients = gest_coefs),
  class = c("GEST", "list")
)

dat <- data.frame(
  X = c(-1.0, 0.0, 0.5, 1.0, 2.0),
  A = c( 0.0, 1.0, 0.0, 1.0, 0.0)
)

test_that("predict.GEST: matches oracle of explicit model.matrix %*% coefficients", {
  expected <- drop(model.matrix(gest_formula, dat) %*% gest_coefs)
  expect_equal(predict.GEST(gest_obj, newdata = dat), expected)
})

test_that("predict.GEST: returns a vector, not a matrix", {
  result <- predict.GEST(gest_obj, newdata = dat)
  expect_true(is.vector(result))
  expect_false(is.matrix(result))
})

test_that("predict.GEST: result has length equal to nrow(newdata)", {
  result <- predict.GEST(gest_obj, newdata = dat)
  expect_length(result, nrow(dat))
})

test_that("predict.GEST: reflects overwritten treatment column in newdata", {
  dat_opt <- dat
  dat_opt$A <- 1.0  # set all to optimal treatment
  result_obs <- predict.GEST(gest_obj, newdata = dat)
  result_opt <- predict.GEST(gest_obj, newdata = dat_opt)
  # predictions must differ when treatment column differs
  expect_false(isTRUE(all.equal(result_obs, result_opt)))
})

test_that("predict.GEST: works correctly for a single-row newdata", {
  single <- dat[1L, , drop = FALSE]
  expected <- drop(model.matrix(gest_formula, single) %*% gest_coefs)
  result <- predict.GEST(gest_obj, newdata = single)
  expect_equal(result, expected)
  expect_length(result, 1L)
})

n <- 30L
set.seed(42L)

dat <- data.frame(
  X = cos(seq(0.0, pi, length.out = n)),
  A = as.numeric(rbinom(n, 1L, 0.5)),
  Y = rnorm(n)
)

models_bin <- list(blip = ~X, treat = A ~ X, tf = ~X)

cts_bin <- Binary$new(tf.model = models_bin$tf,
                      blip.model = models_bin$blip,
                      tx.var = "A")

fit_tx <- glm(models_bin$treat, dat, family = "binomial")
Ahat <- predict(fit_tx, type = "response")
wgts <- rep(1.0, n)

local({
  res <- .gest(A = dat$A,
               wgts = wgts,
               data = dat,
               Y = dat$Y,
               A.hat = Ahat,
               tx.var = "A",
               cts.obj = cts_bin,
               treat.mod.fitted = fit_tx)
  
  test_that(".gest: returns list with required element names", {
    expect_named(res, c("outcome.fit", "beta", "psi"), ignore.order = TRUE)
  })
  
  test_that(".gest: outcome.fit has class GEST", {
    expect_true(inherits(res$outcome.fit, "GEST"))
  })
  
  test_that(".gest: outcome.fit contains formula, fitted.values, coefficients", {
    expect_named(res$outcome.fit,
                 c("formula", "fitted.values", "coefficients"),
                 ignore.order = TRUE)
  })
  
  test_that(".gest: outcome.fit$formula matches cts.obj$full.model", {
    expect_equal(deparse(res$outcome.fit$formula),
                 deparse(cts_bin$full.model))
  })
  
  test_that(".gest: outcome.fit$fitted.values has length n", {
    expect_length(res$outcome.fit$fitted.values, n)
  })
  
  test_that(".gest: coefficients match oracle of explicit crossprod/solve", {
    Hd <- cts_bin$Hd(dat, dat$A)
    Hw <- cts_bin$Hw(data = dat, A = dat$A, A.hat = Ahat,
                     wgt = wgts, treat.mod.fitted = fit_tx)
    expected <- drop(solve(crossprod(Hw, Hd), crossprod(Hw, dat$Y)))
    names(expected) <- colnames(Hd)
    expect_equal(res$outcome.fit$coefficients, expected)
  })
  
  test_that(".gest: psi contains only blip-related terms (contains tx.var)", {
    expect_true(all(grepl("A", names(res$psi))))
  })
  
  test_that(".gest: beta contains no blip-related terms", {
    expect_false(any(grepl("^A", names(res$beta))))
  })
  
  test_that(".gest: psi and beta together account for all coefficients", {
    expect_setequal(c(names(res$psi), names(res$beta)),
                    names(res$outcome.fit$coefficients))
  })
  
  test_that(".gest: fitted.values match oracle of Hd %*% coefficients", {
    Hd <- cts_bin$Hd(dat, dat$A)
    expected <- drop(Hd %*% res$outcome.fit$coefficients)
    expect_equal(res$outcome.fit$fitted.values, expected)
  })
})

test_that(".gest: errors as expected", {
  expect_error(
    .gest(A = as.character(dat$A), wgts = wgts, data = dat, Y = dat$Y,
          A.hat = Ahat, tx.var = "A", cts.obj = cts_bin,
          treat.mod.fitted = fit_tx),
    "`A`"
  )
  expect_error(
    .gest(A = 1.0, wgts = 1.0, data = dat[1L, ], Y = dat$Y[1L],
          A.hat = Ahat[1L], tx.var = "A", cts.obj = cts_bin,
          treat.mod.fitted = fit_tx),
    "`A`"
  )
  expect_error(
    .gest(A = dat$A, wgts = wgts[1L:5L], data = dat, Y = dat$Y,
          A.hat = Ahat, tx.var = "A", cts.obj = cts_bin,
          treat.mod.fitted = fit_tx),
    "`wgts`"
  )
  expect_error(
    .gest(A = dat$A, wgts = wgts, data = dat,
          Y = as.character(dat$Y),
          A.hat = Ahat, tx.var = "A", cts.obj = cts_bin,
          treat.mod.fitted = fit_tx),
    "`Y`"
  )
  expect_error(
    .gest(A = dat$A, wgts = wgts, data = as.matrix(dat), Y = dat$Y,
          A.hat = Ahat, tx.var = "A", cts.obj = cts_bin,
          treat.mod.fitted = fit_tx),
    "`data`"
  )
  expect_error(
    .gest(A = dat$A, wgts = wgts, data = dat[1L:5L, ], Y = dat$Y,
          A.hat = Ahat, tx.var = "A", cts.obj = cts_bin,
          treat.mod.fitted = fit_tx),
    "`data`"
  )
  expect_error(
    .gest(A = dat$A, wgts = wgts, data = dat, Y = dat$Y,
          A.hat = Ahat[1L:5L], tx.var = "A", cts.obj = cts_bin,
          treat.mod.fitted = fit_tx),
    "`A.hat`"
  )
  expect_error(
    .gest(A = dat$A, wgts = wgts, data = dat, Y = dat$Y,
          A.hat = Ahat, tx.var = c("A", "B"), cts.obj = cts_bin,
          treat.mod.fitted = fit_tx),
    "`tx.var`"
  )
  expect_error(
    .gest(A = dat$A, wgts = wgts, data = dat, Y = dat$Y,
          A.hat = Ahat, tx.var = "A", cts.obj = list(),
          treat.mod.fitted = fit_tx),
    "`cts.obj`"
  )
  expect_error(
    .gest(A = dat$A, wgts = wgts, data = dat, Y = dat$Y,
          A.hat = Ahat, tx.var = "A", cts.obj = cts_bin,
          treat.mod.fitted = "not a model"),
    "`treat.mod.fitted`"
  )
  # Constant X makes design matrix rank-deficient
  dat_sing <- dat
  dat_sing$X <- 0.0
  cts_sing <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
  fit_sing <- glm(A ~ X, dat_sing, family = "binomial")
  Ahat_sing  <- predict(fit_sing, type = "response")
  expect_error(
    .gest(A = dat_sing$A, wgts = wgts, data = dat_sing, Y = dat_sing$Y,
          A.hat = Ahat_sing, tx.var = "A", cts.obj = cts_sing,
          treat.mod.fitted = fit_sing),
    "unable to invert matrix"
  )
})

#' Adaptive Choice of the Bootstrap Resample Size M
#' 
#' 
#' Implementation of a double-bootstrap algorithm for choosing the bootstrap 
#'   resample size \eqn{m}{m} in a data-adaptive manner. The function returns the 
#'   resample size to be used to apply the m-out-of-n bootstrap with \link{DTRreg}.
#'
#' The m-out-of-n bootstrap is an adequate tool for constructing valid 
#'   confidence intervals for the first stage parameters in \link{DTRreg}. The 
#'   resample size \eqn{m}{m} is: 
#'   \eqn{m = n^{\frac{1 + \alpha(1-\hat{p})}{1+\alpha}}}{%
#'   m = n^[1+alpha(1-pHat)/(1+alpha)]}. The estimated non-regularity level is 
#'   computed by \link{DTRreg}. The double-bootstrap algorithm is a cross-validation 
#'   tool for choosing the tuning parameter \eqn{\alpha}{alpha} in a data-driven way.
#'  
#' The current implementation is valid for a two-stage DTR. Moreover, the 
#'   current implementation may be unstable when there are many missing data.
#'   
#' @inheritParams DTRreg
#' @param B1 Number of first-level bootstrap resamples.
#' @param B2 Number of second-level bootstrap resamples.
#' 
#' @return A list with a single element 
#'   \item{m }{Resample size for using in the m-out-of-n bootstrap.}
#'   
#' @references
#'   Chakraborty, B., Moodie, E. E. M. (2013) \emph{Statistical Methods for 
#'   Dynamic Treatment Regimes}. New York: Springer.
#'   
#'   Efron B., Tibshirani R. J. (1994) An Introduction to the Bootstrap. 
#'   \emph{CRC press}.
#'   
#'   Wallace, M. P., Moodie, E. M. (2015) Doubly-Robust Dynamic Treatment 
#'   Regimen Estimation Via Weighted Least Squares. \emph{Biometrics} 
#'   \bold{71}(3), 636--644 (doi:10.1111/biom.12306.)
#' % Eventually update with reference to my paper
#' @author Gabrielle Simoneau
#' 
#' @examples 
#' data(twoStageCont)
#' @examples
#' 
#' # models to be passed to DTRreg
#' # blip model
#' blip.mod <- list(~ X1, ~ X2)
#' # treatment model (correctly specified)
#' treat.mod <- list(A1 ~ X1, A2 ~ 1)
#' # treatment-free model (incorrectly specified)
#' tf.mod <- list(~ X1, ~ X2)
#' 
#' # perform dWOLS without calculating confidence intervals
#' mod1 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
#'                data = twoStageCont, method = "dwols")
#'   
#' # choose m adaptively for that model
#' \dontrun{
#'   m <- chooseM(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
#'                data = twoStageCont, method = "dwols",
#'                B1 = 200, B2 = 200)$m
#' }
#' m <- 94
#'   
#' # dWOLS with confidence intervals from the m-out-of-n bootstrap
#' mod2 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
#'                data = twoStageCont, method = "dwols",
#'                var.estim = "bootstrap", 
#'                bootstrap.controls = list(M = m))
#' @concept dynamic treatment regimens
#' @concept adaptive treatment strategies
#' @concept personalized medicine
#' @concept g-estimation
#' @concept dynamic weighted ordinary least squares
#' @concept m-out-of-n bootstrap
#' @concept double bootstrap
#' @export
chooseM <- function(outcome, 
                    blip.mod, treat.mod, tf.mod, 
                    data = NULL, 
                    method = c("gest", "dwols", "qlearn"), 
                    treat.type = c("bin", "multi", "cont"),
                    treat.fam = gaussian(link = "identity"),
                    weight = c("overlap", "ipw", "cipw", "qpom", "none", "manual"), 
                    n.bins = 3L, 
                    treat.wgt.man = NULL,
                    treat.range = NULL, 
                    missing = c("drop", "ipw"), 
                    missing.mod = NULL,
                    B1 = 500, B2 = 500) {
  
  # soft deprecation — remap before match.arg validates
  if (!missing(weight)) {
    if (identical(weight, "abs")) {
      warning("'abs' has been renamed 'overlap'. Please update your code. ",
              "'abs' will be removed in a future version.", call. = FALSE)
      weight <- "overlap"
    } else if (identical(weight, "wo")) {
      warning("'wo' has been renamed 'overlap'. Please update your code. ",
              "'wo' will be removed in a future version.", call. = FALSE)
      weight <- "overlap"
    }
  }
  
  # now validate — "abs" and "wo" are not in this list
  weight <- match.arg(weight, c("overlap", "ipw", "cipw", "qpom", "none", "manual"))

  stopifnot(
    "`outcome`, `blip.mod`, `treat.mod`, and `tf.mod` must be provided" =
      !missing(outcome) && !missing(blip.mod) && !missing(treat.mod) &&
      !missing(tf.mod)
  )
  
  Y <- outcome
  
  # fit initial DTRreg model to the data to get original estimates + phat
  mod1 <- DTRreg(outcome = outcome, 
                 blip.mod = blip.mod, 
                 treat.mod = treat.mod, 
                 tf.mod = tf.mod, 
                 data = data, 
                 method = method, 
                 weight = weight, 
                 n.bins = n.bins,
                 missing = missing, 
                 missing.mod = missing.mod,
                 treat.type = treat.type,
                 treat.fam = treat.fam,
                 treat.wgt.man = treat.wgt.man, 
                 treat.range = treat.range,
                 var.estim = "bootstrap", 
                 bootstrap.controls = list(M = 0, truncate = 0.0, B = 50))
  
  stopifnot("`chooseM` is only appropriate for K = 2" = mod1$K == 2L)
  
  blip.psi.1 <- mod1$psi[[1L]]
  p <- mod1$nonreg[2L]
  
  # deal with missing data here
  
  # initialize
  alpha <- 0.025
  n <- nrow(mod1$training_data$data)
  
  # loop until alpha reaches 0.5, with break when the coverage reaches the 
  # nominal 95% rate
  while (alpha <= 0.5) { 
    # reset coverage
    coverage <- 0
    
    # reset matrix to save estimates
    est <- matrix(NA, ncol = B2 + 3L, nrow = 1L)
    
    # loop over B1 first stage bootstrap samples
    for (j in seq_len(B1)) {
      
      # draw a n-out-of-n bootstrap sample
      index <- sample(seq_len(n), n, replace = TRUE)
      boot1 <- mod1$training_data$data[index, , drop = FALSE]
      Y <- outcome[index]
      
      if (is.list(treat.wgt.man)) {
        tx.wgt.man <- lapply(treat.wgt.man, "[", index)
      } else {
        tx.wgt.man <- NULL
      }
      
      # fit the model to b1-th bootstrap sample
      res1 <- tryCatch(DTRreg(outcome = Y, 
                              blip.mod = blip.mod, 
                              treat.mod = treat.mod, 
                              tf.mod = tf.mod, 
                              data = boot1, 
                              method = method, 
                              weight = weight, 
                              n.bins = n.bins,
                              missing = missing, 
                              missing.mod = missing.mod,
                              treat.type = treat.type,
                              treat.fam = treat.fam,
                              treat.wgt.man = tx.wgt.man, 
                              treat.range = treat.range,
                              var.estim = "bootstrap"),
                       error = function(e) {
                         stop("unable to complete DTRreg\n\t",
                              e$message, call. = FALSE)
                       })
      
      esb1 <- res1$psi[[1L]][1L] # only consider main effect of treatment
      est[1L] <- esb1
      
      # estimate m for each b1 bootstrap sample
      phat <- res1$nonreg[2L]
      est[2L] <- phat
      
      # resampling size
      m <- n^({1.0 + alpha * {1.0 - phat}}/{1.0 + alpha})
      est[3L] <- m
      
      # loop over B2 second stage bootstrap samples
      for (k in seq_len(B2) ) {
        # resample with replacement
        index <- sample(seq_len(n), floor(m), replace = TRUE)

        if (is.list(treat.wgt.man)) {
          tx_wgt_man <- lapply(tx.wgt.man, "[", index)
        } else {
          tx_wgt_man <- NULL
        }
        
        # fit the model to bootstrap sample
        res2 <- tryCatch(DTRreg(outcome = Y[index], 
                                blip.mod = blip.mod, 
                                treat.mod = treat.mod, 
                                tf.mod = tf.mod, 
                                data = boot1[index, , drop = FALSE], 
                                method = method, 
                                weight = weight, 
                                n.bins = n.bins,
                                missing = missing, 
                                missing.mod = missing.mod,
                                treat.type = treat.type,
                                treat.fam = treat.fam,
                                treat.wgt.man = tx_wgt_man, 
                                treat.range = treat.range,
                                var.estim = "none"),
                         error = function(e) {
                           stop("unable to complete DTRreg\n\t",
                                e$message, call. = FALSE)
                         })
        
        # save the (b1,b2) bootstrap estimates in the (k+3) column
        est[k + 3L] <- res2$psi[[1L]][1L]
      }
      quan <- quantile(sqrt(m) * {est[-c(1:3)] - esb1}, 
                       probs = c(0.025, 0.975))
      coverage <- coverage + {{esb1 - quan[2L] / sqrt(m) <= blip.psi.1[1L]} & 
                              {esb1 - quan[1L] / sqrt(m) >= blip.psi.1[1L]}}
    }
    
    coverage <- coverage / B1
    
    # print coverage and alpha
    message("alpha= ", format(alpha, digits = 4L), 
            ", coverage = ", format(coverage, digits = 4L))
    
    if ( coverage >= 0.95 ) {
      message("final value of alpha = ", format(alpha, digits = 4L))
      m <- n^({1.0 + alpha * {1.0 - p}} / {1.0 + alpha})
      message("selected subsample size m = ", floor(m))
      break
    } else {
      alpha <- alpha + 0.025
    }
  }
  list("m" = as.numeric(floor(m)))
}

data(twoStageCont)
n <- nrow(twoStageCont)

blip.mod <- list(~X1, ~X2)
treat.mod <- list(A1 ~ X1, A2 ~ X2)
tf.mod <- list(~X1, ~X2)


# Output structure


local({
  skip_on_cran()
  
  withr::with_seed(42L, {
    res <- suppressMessages(
      chooseM(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
              data = twoStageCont,
              method = "dwols",
              B1 = 3L,
              B2 = 3L)
    )
  })
  
  test_that("chooseM: returns a list", {
    expect_true(is.list(res))
  })
  
  test_that("chooseM: returns list with single element 'm'", {
    expect_named(res, "m")
  })
  
  test_that("chooseM: m is numeric", {
    expect_true(is.numeric(res$m))
  })
  
  test_that("chooseM: m is between 1 and n", {
    expect_true(res$m >= 1.0)
    expect_true(res$m <= n)
  })
  
  test_that("chooseM: m is a whole number", {
    expect_equal(res$m, floor(res$m))
  })
})



# m formula bounds


test_that("chooseM: m formula gives n when p=0 and alpha=0.5", {
  skip_on_cran()
  # When p=0 (maximally non-regular): m = n^((1 + 0.5) / (1 + 0.5)) = n^1 = n
  # When p=1 (regular): m = n^((1 + 0) / (1 + 0.5)) = n^(2/3)
  # So m should always be <= n
  withr::with_seed(42L, {
    res <- suppressMessages(
      chooseM(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
              data = twoStageCont,
              method = "dwols",
              B1 = 3L,
              B2 = 3L)
    )
  })
  expect_true(res$m <= n)
})



# Messages produced during run


test_that("chooseM: produces alpha/coverage messages during run", {
  skip_on_cran()
  withr::with_seed(42L,
                   expect_message(
                     chooseM(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
                             data = twoStageCont,
                             method = "dwols",
                             B1 = 3L,
                             B2 = 3L),
                     "alpha"
                   )
  )
})



# Stopping conditions


test_that("chooseM: stops when outcome is missing", {
  skip_on_cran()
  expect_error(
    chooseM(blip.mod = blip.mod, treat.mod = treat.mod, tf.mod = tf.mod,
            data = twoStageCont, B1 = 3L, B2 = 3L),
    "must be provided"
  )
})

test_that("chooseM: stops when blip.mod is missing", {
  skip_on_cran()
  expect_error(
    chooseM(twoStageCont$Y, treat.mod = treat.mod, tf.mod = tf.mod,
            data = twoStageCont, B1 = 3L, B2 = 3L),
    "must be provided"
  )
})

test_that("chooseM: stops when K != 2", {
  skip_on_cran()
  # Single stage - K=1
  expect_error(
    suppressMessages(
      chooseM(twoStageCont$Y,
              blip.mod = list(~X1),
              treat.mod = list(A1 ~ X1),
              tf.mod = list(~X1),
              data = twoStageCont,
              B1 = 3L,
              B2 = 3L)
    ),
    "K = 2"
  )
})



# Reproducibility


test_that("chooseM: same seed produces same m", {
  skip_on_cran()
  withr::with_seed(42L,
                   res1 <- suppressMessages(
                     chooseM(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
                             data = twoStageCont,
                             method = "dwols",
                             B1 = 3L,
                             B2 = 3L)
                   )
  )
  withr::with_seed(42L,
                   res2 <- suppressMessages(
                     chooseM(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
                             data = twoStageCont,
                             method = "dwols",
                             B1 = 3L,
                             B2 = 3L)
                   )
  )
  expect_equal(res1$m, res2$m)
})

#' Toy Two-Stage Trial Datasets
#' 
#' These datasets are provided only to facilitate examples. They are not based 
#'   on or representative of any real-world applications.
#' 
#' @name data
#' @rdname data
#' 
#' @usage data(twoStageCont)
#'
#' @format twoStageCont is a dataset generated to mimic a simple two-stage
#'   trial. The data.frame contains 1000 observations with 5 columns: 
#' \describe{
#' \item{X1}{The first stage covariate. A normally distributed continuous variable.}
#' \item{A1}{The first stage treatment. A binary variable.}
#' \item{X2}{The second stage covariate. A normally distributed continuous variable.}
#' \item{A2}{The second stage treatment. A binary variable.}
#' \item{Y}{The outcome. A continuous variable.}
#' }
#'
#' @keywords datasets
"twoStageCont"



#' @noRd
#' @keywords internal
.generateTwoStageCont <- function(seed = 1234L, n = 1000L) {
  set.seed(seed)
 
  # expit function
  expit <- function(x) { 1.0 / (1.0 + exp(-x)) }

  # variables (X = patient information, A = treatment)
  X1 <- rnorm(n)
  A1 <- rbinom(n, 1, expit(X1))
  X2 <- rnorm(n)
  A2 <- rbinom(n, 1, expit(X2))
   
  # blip functions
  gamma1 <- A1 * (1 + X1)
  gamma2 <- A2 * (1 + X2)

  # observed outcome: treatment-free outcome plus blip functions
  Y <- exp(X1) + exp(X2) + gamma1 + gamma2 + rnorm(n)
  
  data.frame("X1" = X1, "A1" = A1,
             "X2" = X2, "A2" = A2, "Y" = Y)
}

#' @rdname data
#' 
#' @usage data(twoStageCens)
#'
#' @format twoStageCens is a dataset generated to mimic a simple two-stage
#'   trial with right-censoring. The data.frame contains 1000 observations 
#'   with 9 columns: 
#' \describe{
#' \item{X11}{A first stage covariate. A normally distributed continuous variable.}
#' \item{X12}{A first stage covariate. A continuous variable = X11^4.}
#' \item{A1}{The first stage treatment. A binary variable.}
#' \item{T1}{The time from the beginning of the first stage to the event or to 
#'           stage 2 entry, whichever comes first.}
#' \item{X21}{A second stage covariate. A normally distributed continuous variable.}
#' \item{X22}{A second stage covariate. A continuous variable = X21^3.}
#' \item{A2}{The second stage treatment. A binary variable.}
#' \item{T2}{The time from the beginning of the second stage to the event 
#'           defined only for subjects who enter the second stage.}
#' \item{delta}{Event indicator.}
#' }
#' 
#' Note: For participants who experienced the event during stage 1, i.e., did
#'   not continue to stage 2, the "survival time" is T1. For participants that
#'   entered stage 2, the "survival time" is T1 + T2.
#' 
"twoStageCens"

#' Generate toy data for two-stage survival with censoring
#' @noRd
#' @importFrom stats rbinom rnorm runif
#' 
#' @keywords internal
.generateTwoStageCensored <- function(seed = 1234L, n = 1000L) {
  
  set.seed(seed)
  
  # expit function
  expit <- function(x) {1.0 / (1.0 + exp(-x))}
  
  theta1 <- c(6.3, 1.5, -0.8, 0.1, 0.1)
  theta2 <- c(4, 1.1, -0.2, -0.9, 0.6, -0.1)
  lambda <- 1/300
  p <- 0.9
  beta <- 2
  # covariates and treatment (X = patient information, A = treatment)
  X1 <- runif(n, 0.1, 1.29)
  X14 <- X1^4
  A1 <- rbinom(n, size = 1, prob = expit(2*X1 - 1))
  X2 <- runif(n, 0.9, 2)
  X23 <- X2^3
  A2 <- rbinom(n, size = 1, prob = expit(-2*X2 + 2.8))
  delta <- rbinom(n, size = 1, prob = expit(2*X1 - 0.4))
  eta2 <- rbinom(n, 1, prob = 0.8)
  delta2 <- delta[eta2 == 1]
  # survival time
  logY2 <- logT2 <- theta2[1] + theta2[2]*X2[eta2 == 1] + theta2[3]*X23[eta2 == 1] 
  + theta2[4]*A2[eta2 == 1] + theta2[5]*A2[eta2 == 1]*X2[eta2 == 1] 
  + theta2[6]*X1[eta2 == 1] + rnorm(sum(eta2), sd = 0.3)
  trueA2opt <- ifelse(theta2[4]*A2[eta2 == 1] 
                      + theta2[5]*A2[eta2 == 1]*X2[eta2 == 1] > 0, 1, 0)
  logT2opt <- logT2 + (trueA2opt - A2[eta2 == 1])*(theta2[4]*A2[eta2 == 1] 
                                                   + theta2[5]*A2[eta2 == 1]*X2[eta2 == 1])
  logT <- theta1[1] + theta1[2]*X1 + theta1[3]*X14 + theta1[4]*A1 
  + theta1[5]*A1*X1 + rnorm(n, sd = 0.3)
  T1 <- exp(logT[eta2 == 1 & delta == 1]) - exp(logT2opt[delta2 == 1])
  logT[eta2 == 1 & delta == 1] <- log(T1 + exp(logT2[delta2 == 1]))
  # censoring time
  C <- (- log(runif(n - sum(delta), 0, 1))/(lambda * exp(beta * X1[delta == 0])))^(1/p)
  eta2d0 <- eta2[delta == 0]
  C1 <- rep(NA, length(C))
  C2 <- rep(NA, length(C))
  for(i in 1:length(C))
  {
    if(eta2d0[i] == 0){
      C1[i] <-  C[i]
      C2[i] <- 0
    }else{
      C1[i] <- runif(1, 0, C[i])
      C2[i] <- C[i] - C1[i]
    }
  }
  # observed survival time
  Y2 <- rep(NA, n)
  Y1 <- rep(NA, n)
  Y2[delta == 0] <- C2
  Y1[delta == 0] <- C1
  Y1[delta == 1 & eta2 == 1] <- T1
  Y1[delta == 1 & eta2 == 0] <- exp(logT[delta == 1 & eta2 == 0])
  Y2[delta == 1 & eta2 == 0] <- 0
  Y2[delta == 1 & eta2 == 1] <- exp(logT2[delta2 == 1])
  logY <- log(Y1 + Y2)
  logY2 <- log(Y2[eta2 == 1])

  data.frame("X11" = X1, "X12" = X14, "A1" = A1, 
             "X21" = X2, "X22" = X23, "A2" = A2, 
             "delta" = delta, "T1" = Y1, "T2" = Y2)
  
}

#' @rdname data
#' 
#' @usage data(twoStageSurv)
#'
#' @format twoStageSurv is a dataset generated to mimic a simple two-stage
#'   trial without censoring. The data.frame contains 1000 observations 
#'   with 9 columns: 
#' \describe{
#' \item{X11}{A first stage covariate. A normally distributed continuous variable.}
#' \item{X12}{A first stage covariate. A continuous variable = X11^4.}
#' \item{A1}{The first stage treatment. A binary variable.}
#' \item{T1}{The time from the beginning of the first stage to the event or to 
#'           stage 2 entry, whichever comes first.}
#' \item{X21}{A second stage covariate. A normally distributed continuous variable.}
#' \item{X22}{A second stage covariate. A continuous variable = X21^3.}
#' \item{A2}{The second stage treatment. A binary variable.}
#' \item{T2}{The time from the beginning of the second stage to the event 
#'           non-zero only for subjects who did not have an event in Stage I.}
#' }
#' 
#' Note: The "survival time" is T1 + T2.
#' 
"twoStageSurv"

#' Generate toy data for two-stage survival with censoring
#' @noRd
#' @importFrom stats rbinom rnorm runif
#' @keywords internal
.generateTwoStageSurv <- function(seed = 1234L, n = 1000L) {
  
  set.seed(seed)
  
  # expit function
  expit <- function(x) {1.0 / (1.0 + exp(-x))}
  
  theta1 <- c(6.3, 1.5, -0.8, 0.1, 0.1)
  theta2 <- c(4, 1.1, -0.2, -0.9, 0.6, -0.1)
  lambda <- 1/300
  p <- 0.9
  beta <- 2
  # covariates and treatment (X = patient information, A = treatment)
  X1 <- runif(n, 0.1, 1.29)
  X14 <- X1^4
  A1 <- rbinom(n, size = 1, prob = expit(2*X1 - 1))
  X2 <- runif(n, 0.9, 2)
  X23 <- X2^3
  A2 <- rbinom(n, size = 1, prob = expit(-2*X2 + 2.8))
  delta <- rbinom(n, size = 1, prob = expit(2*X1 - 0.4))
  eta2 <- rbinom(n, 1, prob = 0.8)
  delta2 <- delta[eta2 == 1]
  # survival time
  logY2 <- logT2 <- theta2[1] + theta2[2]*X2[eta2 == 1] + theta2[3]*X23[eta2 == 1] 
  + theta2[4]*A2[eta2 == 1] + theta2[5]*A2[eta2 == 1]*X2[eta2 == 1] 
  + theta2[6]*X1[eta2 == 1] + rnorm(sum(eta2), sd = 0.3)
  trueA2opt <- ifelse(theta2[4]*A2[eta2 == 1] 
                      + theta2[5]*A2[eta2 == 1]*X2[eta2 == 1] > 0, 1, 0)
  logT2opt <- logT2 + (trueA2opt - A2[eta2 == 1])*(theta2[4]*A2[eta2 == 1] 
                                                   + theta2[5]*A2[eta2 == 1]*X2[eta2 == 1])
  logT <- theta1[1] + theta1[2]*X1 + theta1[3]*X14 + theta1[4]*A1 
  + theta1[5]*A1*X1 + rnorm(n, sd = 0.3)

  delta <- rep(1,n)
  delta2 <- delta[eta2 == 1]
  T1 <- exp(logT[eta2 == 1 & delta == 1]) - exp(logT2opt[delta2 == 1])
  logT[eta2 == 1 & delta == 1] <- log(T1 + exp(logT2[delta2 == 1]))
  # observed survival time
  Y2 <- rep(NA, n)
  Y1 <- rep(NA, n)
  Y1[delta == 1 & eta2 == 1] <- T1
  Y1[delta == 1 & eta2 == 0] <- exp(logT[delta == 1 & eta2 == 0])
  Y2[delta == 1 & eta2 == 0] <- 0
  Y2[delta == 1 & eta2 == 1] <- exp(logT2[delta2 == 1])
  logY <- log(Y1 + Y2)
  logY2 <- log(Y2[eta2 == 1])
  
  data.frame("X11" = X1, "X12" = X14, "A1" = A1, 
             "X21" = X2, "X22" = X23, "A2" = A2, 
             "delta" = rep(1L, n), "T1" = Y1, "T2" = Y1)
}

#' DTR Estimation and Inference via G-estimation, Dynamic WOLS, or Q-learning
#'
#' Dynamic treatment regimen estimation and inference via G-estimation and 
#'   dynamic WOLS.  Estimation of blip model parameters for multi-stage data.
#' 
#' \code{DTRreg()} allows the estimation of optimal dynamic treatment regimens 
#'   (DTRs, also known as adaptive treatment strategies) from multi-stage 
#'   trials using G-estimation, dynamic weighted ordinary least squares 
#'   (dWOLS), and genearlized dWOLS. All methods focus on estimating the 
#'   parameters of the blip: a 
#'   model of the difference in expected outcome under the observed treatment 
#'   and some reference treatment (usually a control) at a given stage, assuming
#'   identical histories and optimal treatment thereafter. The reader is 
#'   referred to Chakraborty and Moodie (2013) for a thorough introduction and 
#'   review of DTR methods. The dWOLS method may be used to obtain parameter 
#'   estimates identical to those from Q-learning (by setting \code{weight = "none"}).
#'   This option is intended primarily for exploratory purposes; the authors 
#'   note that there is a dedicated R package for Q-learning (qLearn), although 
#'   it is limited to the 2-stage setting; multi-stage settings are available
#'   in R package DynTxRegime.
#'   
#'   This implementation assumes an outcome regression model of the form
#'   E(Y|X=x,A=a) = tf.mod + a blip.mod. That is -- the input \code{blip.mod} 
#'   formula should include the treatment variable \emph{ONLY} if it is quadratic. 
#'   For example, if the full blip model is linear in the treatment variable 
#'   \deqn{\sim a \psi_0 + a x \psi_1,}{~ a psi_0 + a x psi_1,} then the input
#'   should model should be \code{blip.mod = ~ x}. 
#'   If the full blip model is quadratic in the treatment variable
#'   \deqn{\sim a \psi_0 + a^2 \psi_1 + a x \psi_2 + a^2 x \psi_3,}{
#'   ~ a psi_0 + a^2 psi_1 + a x psi_2 + a^2 x psi_3,} \code{blip.mod =
#'   ~ a*x}. For continuous treatments, only quadratic blip
#'   functions are supported.
#'   
#'   All methods require the specification of three models for each 
#'   stage of the analysis: a treatment model (conditional mean of the treatment 
#'   variable), a treatment-free model (conditional mean of outcome assuming 
#'   only reference treatments are used), and a blip model. Only the blip model
#'   must be correctly specified (or over-specified), with consistent parameter 
#'   estimates obtainable if at least one of the other two models is correctly 
#'   specified. Note that all of these must be specified as lists of formula 
#'   objects, even if only one stage of treatment is considered.
#'   
#'   Note that as is conventional, it is assumed a larger value of the outcome 
#'   is preferred (which can be easily achieved via transformation of your data 
#'   if necessary).
#'   
#'   When treatment is binary, if confidence intervals are computed (via 
#'   specification of \code{var.estim} other than "none"), then DTRreg will calculate 
#'   the proportion of subjects at each stage for whom optimal treatment is 
#'   non-unique. If this proportion exceeds 0.05 a non-regularity warning will 
#'   be displayed, along with the proportion of subjects for whom this is the 
#'   case. Note that this warning is only displayed if a variance estimation 
#'   option is selected.
#'   
#'   Several treatment weight function options have been implemented within the 
#'   package:
#'   \itemize{ 
#'     \item "none": No treatment weights applied. If \code{method = "dWOLS"}, this 
#'       selection results in the implementation of Q-learning, modified 
#'       slightly to use the G-estimation or dWOLS style pseudo-outcome 
#'       (computed using the observed outcome modified by the estimated 
#'        treatment effect) rather than the traditional Q-learning outcome 
#'       (predicted based on model only, rather than observed outcome with 
#'       treatment effect).   
#'     \item "ipw": weights based on the inverse probability of 
#'       treatment. For binary treatments, a logistic regression is used.
#'       For multinomial, a multinomial log-linear model is fit using 
#'       \code{\link[nnet]{multinom}}. For continuous treatments, a GLM with the specified 
#'       family and link function provided in the \code{treat.fam} argument is used.
#'     \item "cipw": inverse probability of treatment weights as described for
#'       "ipw" and capped at the 99th percentile of the observed weights.
#'     \item "qpom": weights based on the stabilized inverse 
#'       probability of treatment applied to the categorized (into n.bins bins) 
#'       continuous doses or multinomial treatments; probabilities are 
#'       calculated using a proportional odds model.
#'       This weight is appropriate only for continuous and multinomial treatments.
#'     \item "overlap": overlap weights for the categorized continuous doses 
#'       or multinomial treatments (Li and Li, 2019) and the absolute difference 
#'       \eqn{|A - E[A|...]|}{|A - E[A|...]|} for binary treatments.
#'     \item "manual": User provides weights through input \code{treat.wgt.man}.
#'       Manual treatments are only used in dwols.
#'    }
#'    
#'  Note that the overlap weights were initially termed "absolute value weights" 
#'    in Wallace and Moodie's (2015, Biometrics, Doubly-robust dynamic 
#'    treatment regimen estimation via weighted least squares. 71:636-644) and 
#'    abbreviated \code{abs} in earlier versions of this package. In addition, the 
#'    overlap weights for continuous or multinomial treatments were termed 
#'    \code{wo} in previous versions, which corresponds to their labeling in 
#'    Schulz & Moodie (2021, Journal of the American Statistical Association, 
#'    Doubly robust estimation of optimal dosing strategies 116:256-268.).
#'    Since version 2.4, these have been united under \code{overlap}, and
#'    \code{abs} and \code{wo} have been deprecated.
#'   
#'
#' @param outcome The outcome variable. Missing data will result in a stopping
#'   error.
#' @param blip.mod A list of formula objects specifying covariates of the
#'   blip function for each stage in order. No dependent variable should be 
#'   specified. Note that this input should include the treatment variable 
#'   ONLY if the blip model is quadratic in treatment. See Details for further 
#'   clarification.
#' @param treat.mod A list of formula objects specifying the treatment model for
#'   each stage in order. Treatment variable should be included as the dependent
#'   variable. If treatment is binary \code{\link[stats]{glm}(family = binomial)}
#'   will be 
#'   used to obtain parameter estimates; if treatment is multi-nomial, 
#'   \code{\link[nnet]{multinom}()} will be used to obtain parameter estimates; and if 
#'   treatment is continuous, \code{\link[stats]{glm}()} with the family and
#'   link specified by \code{treat.fam} will be used to obtain parameter estimates.
#' @param tf.mod A list of formula objects specifying covariates of the
#'   treatment-free model for each stage in order. No dependent variable should 
#'   be specified.
#' @param data A data frame containing all necessary covariates and treatments
#'   contained in the models. Missing data should be coded as \code{NA}.
#' @param method The DTR method to be used, choose "dwols" for dynamic WOLS, 
#'   "gest" for G-estimation, or "qlearn" for Q-learning.
#' @param treat.type A character object. Must be one of \{"bin", "multi", "cont"\}
#'   indicating that the treatments at each stage are binary, multinomial,
#'   or continuous, respectively. Each stage must have the same treatment type.
#' @param weight The form of the treatment weight. See details.
#' @param n.bins An integer object. The number of bins (levels) to be used for 
#'   categorizing continuous doses. This input is required only when
#'   \code{treat.type = "cont"} and \code{weight = "overlap"} or \code{weight = "qpom"}.
#' @param var.estim Covariance matrix estimation method, either "bootstrap"
#'   or "sandwich" for sandwich estimation.
#' @param full.cov A logical. If \code{TRUE}, the full covariance matrix will be
#'   returned. If \code{FALSE}, only the terms pertaining to the blip parameters
#'   are returned.
#' @param bootstrap.controls A named list specifying control parameters of the
#'   bootstrap if \code{var.estim = "bootstrap"}. Available controls are:
#'   \describe{
#'     \item{B: }{The number of bootstrap samples.} 
#'     \item{M: }{The subsample size for m out of n bootstrap.}
#'     \item{type: }{The type of bootstrap. Must be one of \{"standard", 
#'     "empirical", "normal"\}. The last two are parametric bootstraps.}
#'     \item{truncate: }{A number between 0 and 0.5. The lowest and highest 
#'     specified proportion of parameter estimates will be replaced by the
#'     relevant quantiles affording some robustness to extreme values 
#'     when estimating covariance.}
#'     \item{verbose: }{If TRUE, estimated time to completion will be printed 
#'     to the console every ~30 seconds.}
#'     \item{interrupt: }{If TRUE then user will be given the option 
#'     to abort the bootstrap without error if estimated time to completion
#'     exceeds 10 minutes.}
#'   }
#' @param treat.range For continuous treatments. Specify the maximum/minimum 
#'   value that treatments can be take. If unspecified then the minimum/maximum 
#'   value of observed treatments is used. If you wish to have unrestricted 
#'   treatments set this option to \code{c(-Inf, +Inf)}. If each stage has its own
#'   range, provide as a list, the ith element providing the min and max
#'   for the ith stage treatment.
#' @param missing A character object. Must be one of \{"drop", "ipw"\}.
#'   If set to "ipw" and covariate or treatment data are missing then inverse 
#'   probability 
#'   weights are used. The complete case probability is estimated 
#'   via logistic regression. If set to "drop" and data are missing, participants
#'   with missing data are dropped for all stage analyses.
#' @param missing.mod An optional list of formula objects specifying the model
#'   for the inverse probability of weights for each stage in order.
#'   No dependent variable should be specified. If \code{missing = "ipw"} and 
#'   \code{missing.mod = NULL}, then the models are assumed to be linear comprising
#'   the full covariate history derived from all of the previous stage models.   
#' @param interactive If \code{TRUE} on-screen prompts will guide the user through 
#'   the specification of blip, treatment, and treatment-free models.
#' @param treat.wgt.man NULL or a list of vectors of known treatment weights can be 
#'   specified to be used instead of hard-coded treatment weight options.
#'   The \eqn{i^{th}}{ith} element of the list contains the multiplicative weights 
#'   for the \eqn{i^{th}}{ith} stage. Each vector must be of length \eqn{n}{n}, 
#'   the number of participants. Used only for \code{method = "dwols"}.
#' @param dtr A logical object. If \code{TRUE}, use the DTR estimation approach, which
#'   estimates the stage pseudo-outcome by adding a regret function. If \code{FALSE},
#'   use an 'effect estimation' approach, which treats the observed outcome
#'   as being equal to an outcome assuming no treatment is received at any
#'   stage, plus a blip component at each stage; each stage pseudo-outcome is 
#'   generated by subtracting a blip function. Note that most of the 
#'   DTR-specific output will either be suppressed or irrelevant.
#' @param treat.fam A character or family object. 
#'   The description of the dose distribution along with the link 
#'   function to be used in the treatment model for computing weights; should be 
#'   specified in a similar format as that used in \code{\link[stats]{glm}()}. 
#'   If character object, must be one of \{"gaussian", "Gamma"\}, for which
#'   \code{\link[stats]{gaussian}(link = "identity")} or 
#'   \code{\link[stats]{Gamma}(link = "log")} 
#'   will be used,
#'   respectively. Input is ignored for \code{treat.type = "bin"} and 
#'   \code{treat.type = "multi"}. If \code{NULL} and \code{treat.type = "cont"},
#'   defaults to \code{gaussian(link = "identity")}.
#'   
#' @return An object of class \code{DTRreg}, a list including elements
#'     \item{K: }{The number of decision points.}
#'     \item{beta: }{A list. The ith element contains the parameter estimates of
#'       the ith stage treatment-free model.}
#'     \item{psi: }{A list. The ith element contains the parameter estimates of
#'       the ith stage blip model.}
#'     \item{covmat: }{A list. The ith element contains covariance matrix of 
#'       the ith stage blip parameter estimates.}
#'     \item{nonreg: }{Non-regularity estimates.}
#'     \item{setup: }{A list detailing the input parameter settings used for the
#'       analysis}
#'       \describe{
#'         \item{models: }{A list of the models used for the analysis.}
#'         \item{method: }{The parameter estimation method.}
#'         \item{var.estim: }{The variance estimation method.}
#'         \item{cc.modeled: }{If TRUE, missing data was modeled. If FALSE, cases
#'            with missing data were removed from the analysis.}
#'         \item{tx.weight: }{The treatment weighting used for the analysis.}
#'         \item{tx.type: }{Treatment was binary, multinomial, or continuous.}
#'         \item{n.bins: }{The number of bins (levels) used for categorizing
#'           continuous doses when \code{tx.weight = "wo"} or 
#'           \code{tx.weight = "qpom"}.}
#'         \item{tx.wgt.man: }{Any user provided treatment weights.}
#'         \item{tx.range: }{For continuous treatments, the range of allowed
#'           treatment values.}
#'         \item{tx.family: }{The description of the dose distribution along 
#'           with the link function used in the continuous treatment model.}
#'         \item{boot.controls: }{A list of the bootstrap controls.}
#'         \item{type: }{The type of effect. Dynamic treatment regime or treatment
#'           effect.}
#'       }
#'     \item{training_data: }{A list containing the training data.}
#'       \describe{
#'         \item{data: }{The covariates and treatment data.}
#'         \item{outcome: }{The outcome of interest.}
#'         \item{A: }{The treatment variables, possibly recoded to adhere to internal
#'           code requirements.}
#'       }
#'     \item{analysis: }{A list containing the primary results of each stage analysis.}
#'       \describe{
#'         \item{n: }{The number of participants included in the stage analysis.}
#'         \item{last.stage: }{The last stage each participant was included in 
#'           the analysis.}
#'         \item{prob.cc: }{The complete case probabilities.}
#'         \item{cc.mod.fitted: }{The regression objects returned for estimating
#'           the complete case probabilities.}
#'         \item{cc.wgt: }{The complete case weights.}
#'         \item{cts: }{The treatment type at each stage.}
#'         \item{tx.mod.fitted: }{The regression objects returned for estimating
#'           the treatment probabilities.}
#'         \item{A.hat: }{The estimated or provided treatment probabilities.}
#'         \item{tx.wgt: }{The treatment weights.}
#'         \item{outcome.fit: }{The regression objects returned for each stage outcome
#'           regression.}
#'         \item{Y: }{The pseudo-outcomes.}
#'         \item{regret: }{Estimates of the regret for each subject based on observed 
#'           treatment and blip parameter estimates.}
#'         \item{opt.treat: }{Optimal treatment decisions for each subject at each 
#'           stage of treatment.}
#'         \item{opt.Y: }{Predicted optimal outcome under recommended regimen.}
#'      }
#'     \item{call: }{The original function call.}
#'   The functions \code{coef()}, \code{predict()} and 
#'   \code{confint()} may be used with such 
#'   model objects. The first two have specific help files for their 
#'   implementation, while \code{confint()} is used in the same way as 
#'   the standard 
#'   \code{\link[stats]{confint}()} command, with the exception of the \code{parm} 
#'   option, which is not available.
#'
#' @references
#'   Chakraborty, B., Moodie, E. E. M. (2013) \emph{Statistical Methods for 
#'   Dynamic Treatment Regimes}. New York: Springer.
#'   
#'   Robins, J. M. (2004) \emph{Optimal structural nested models for optimal 
#'   sequential decisions}. In Proceedings of the Second Seattle Symposium on 
#'   Biostatistics, D. Y. Lin and P. J. Heagerty (eds), 189--326. New York: 
#'   Springer.
#'   
#'   Wallace, M. P., Moodie, E. E. M. (2015) Doubly-Robust Dynamic Treatment 
#'   Regimen Estimation Via Weighted Least Squares. \emph{Biometrics} 
#'   \bold{71}(3), 636--644 (doi:10.1111/biom.12306.)
#'   
#'   Simoneau, G., Moodie, E. E. M., Nijjar, J. S., and Platt, R. W. (2020)
#'   Finite Sample Variance Estimation for Optimal Dynamic Treatment
#'   Regimes of Survival Outcomes. \emph{Statistics in Medicine} \bold{39},
#'   4466-4479.
#'   
#'   Efron, B., and Tibshirani, R. (1986)
#'   Bootstrap Methods for Standard Errors, Confidence Intervals, and Other 
#'   Measures of Statistical Accuracy \emph{Source: Statistical Science} \bold{1},
#'   54-75.
#' 
#'   Schulz, J. and Moodie, E.E.M. (2021) 
#'   Doubly robust estimation of optimal dosing strategies. 
#'   \emph{Journal of the American Statistical Association}, \bold{116}, 
#'   256–268.
#'   
#'   Li, F. and Li, F. (2019) 
#'   Propensity score weighting for causal inference with multiple treatments. 
#'   \emph{The Annals of Applied Statistics}, \bold{13}(4), 
#'   2389–2415. 
#'   
#' @author Michael Wallace
#' @author Shannon T. Holloway
#' 
#' @examples 
#' data(twoStageCont)
#' @examples
#' # models to be passed to DTRreg
#' # blip model
#' blip.mod <- list(~ X1, ~ X2)
#' # treatment model (correctly specified)
#' treat.mod <- list(A1 ~ X1, A2 ~ 1)
#' # treatment-free model (incorrectly specified)
#' tf.mod <- list(~ X1, ~ X2)
#' 
#' # perform G-estimation
#' mod1 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
#'               data = twoStageCont, method = "gest")
#'               
#' mod1
#' @concept dynamic treatment regimens
#' @concept adaptive treatment strategies
#' @concept personalized medicine
#' @concept g-estimation
#' @concept dynamic weighted ordinary least squares
#' 
#' @importFrom stats as.formula family Gamma gaussian get_all_vars model.frame terms
#' @importFrom stats na.pass
#' @include dtrProcedure.R inputProcessing.R interactive.R variance.R
#' @export
DTRreg <- function(outcome, 
                   blip.mod, treat.mod, tf.mod, 
                   data = NULL, 
                   method = c("gest", "dwols", "qlearn"), 
                   interactive = FALSE,
                   treat.type = c("bin", "multi", "cont"),
                   treat.fam = gaussian(link = "identity"),
                   weight = c("overlap", "ipw", "cipw", "qpom", "none", "manual"), 
                   n.bins = 3L, 
                   treat.range = NULL, 
                   treat.wgt.man = NULL,
                   var.estim = c("none", "bootstrap", "sandwich"), 
                   full.cov = FALSE,
                   bootstrap.controls = list(B = 100L, 
                                             M = nrow(data), 
                                             type = "standard", 
                                             truncate = 0.0, 
                                             verbose = FALSE, 
                                             interrupt = FALSE),
                   missing = c("drop", "ipw"), 
                   missing.mod = NULL,
                   dtr = TRUE) {
  
  # soft deprecation - remap before match.arg validates
  if (!missing(weight)) {
    if (identical(weight, "abs")) {
      warning("'abs' has been renamed 'overlap'. Please update your code. ",
              "'abs' will be removed in a future version.", call. = FALSE)
      weight <- "overlap"
    } else if (identical(weight, "wo")) {
      warning("'wo' has been renamed 'overlap'. Please update your code. ",
              "'wo' will be removed in a future version.", call. = FALSE)
      weight <- "overlap"
    }
  }
  
  # now validate - "abs" and "wo" are not in this list
  weight <- match.arg(weight, c("overlap", "ipw", "cipw", "qpom", "none", "manual"))
  
  

  if (interactive) {
    # if interactive mode chosen, build standardized input from user input
    
    # returned list contains 
    # $models A list. The models (blip, tf, treat) grouped according to 
    #   decision point
    # $method A character. The DTR method (dwols, gest).
    # $var.estim A character. The covariance matrix estimation method 
    #   ("none", "bootstrap", "sandwich")
    # $outcome A numeric vector. The outcome of interest.
    # $K An integer. The number of decision points.
    missing <- match.arg(missing)
    obj <- .interactive(isSurvival = FALSE, 
                        prompt.cens = missing == "ipw",
                        data = data)
  } else {
    
    stopifnot(
      "`blip.mod`, `treat.mod`, and `tf.mod` must be provided" =
        !missing(blip.mod) & !missing(treat.mod) & !missing(tf.mod),
      "`data` must be NULL or a data.frame" = is.null(data) || is.data.frame(data)
    )
    
    # primary analysis object carried around to simplify function
    # definitions
    obj <- list()
    
    # Verify model structure and group by stage
    obj$models <- .modelsTest(blip.mod, treat.mod, tf.mod)

    obj$method <- match.arg(method)
    obj$var.estim <- match.arg(var.estim)
    
    if (is.numeric(outcome)) {
      obj$outcome <- outcome
    } else if (is.character(outcome)) {
      obj$outcome <- .getVariables(outcome, data)
    } else {
      stop("`outcome` must be a numeric vector or character variable name", 
           call. = FALSE)
    }
    
    obj$K <- length(obj$models)
  }
  
  if (any(is.na(obj$outcome))) {
    stop("missing values in the outcome are not allowed.", call. = FALSE)
  }
  
  if (obj$method == "qlearn") {
    method <- "qlearn"
    weight <- "none"
    obj$method <- "dwols"
  } else {
    method <- obj$method
  }
  
  # create missingness models if appropriate
  missing <- match.arg(missing)
  obj$censoring.modeled <- missing == "ipw"
  
  if (obj$censoring.modeled) {
    obj$models <- .constructMissingness(obj$models, missing.mod)
  } 
  
  # get all vars from the models
  data_vars <- obj$models |> 
    unlist() |> 
    lapply(FUN = all.vars) |>
    unlist() |> 
    unique() |>
    paste(collapse = " + ")
  data_vars <- stats::as.formula(paste("~", data_vars))
  
  if (is.null(data)) {
    # if no data set, create one from environment using the models
    
    obj$data <- tryCatch(stats::get_all_vars(data_vars),
                         error = function(e) {
                           stop("unable to retrieve all covariates from the ",
                                "environment\n\t", e$message, call. = FALSE)
                         })
  } else if (is.data.frame(data)) {
    obj$data <- tryCatch(stats::model.frame(data_vars, data, na.action = stats::na.pass),
                         error = function(e) {
                           stop("unable to retrieve all covariates from the ",
                                "`data`\n\t", e$message, call. = FALSE)
                         })
    rm(data, data_vars)
  } else {
    stop("if provided, `data` must be a data.frame", call. = FALSE)
  }
  
  rownames(obj$data) <- seq_len(nrow(obj$data))
  
  
  ### Treatment related inputs
  
  obj$tx.weight <- match.arg(weight)
  obj$tx.type <- match.arg(treat.type)
  
  if (obj$method == "gest" && obj$tx.weight == "manual") {
    warning("treatment weights are not used in g-estimation", call. = FALSE)
  }
  
  obj <- c(obj,
           .treatmentTest(weight = obj$tx.weight, 
                          treat.type = obj$tx.type, 
                          n.bins = n.bins, 
                          treat.wgt.man = treat.wgt.man,
                          treat.range = treat.range, 
                          treat.fam = treat.fam, 
                          K = obj$K, 
                          models = obj$models,
                          data = obj$data))

  if (obj$var.estim != "none" && is.list(obj$tx.wgt.man)) {
    warningMsg <- paste("Treatment weights were supplied as an argument",
                        "(generated outside the DTRreg function);",
                        "standard errors do not account for estimation of the",
                        "treatment score model.\n")
    warning(warningMsg, call. = FALSE)
    addWarning <- TRUE
  } else {
    addWarning <- FALSE
  }
  
  
  ### Variance related inputs
  
  obj$boot.controls <- .varestimTest(var.estim = obj$var.estim, 
                                     method = obj$method, 
                                     bootstrap.controls = bootstrap.controls, 
                                     n = nrow(obj$data))
  
  stopifnot("`full.cov` must be logical" = is.logical(full.cov))
  obj$full.cov <- full.cov
  
  ### Analysis details
  
  if (is.logical(dtr)) {
    obj$type <- ifelse(dtr, "DTR", "effect")
  } else {
    stop("`dtr` must be a logical", call. = FALSE)
  }
  
  obj <- .dtrProcedure(obj = obj, quiet = FALSE, isSurvival = FALSE)

  obj <- .computeVariance(obj, isSurvival = FALSE)
  
  # To conform to original return object, shift list of individual results to 
  # individual results as lists
  # Don't want to keep 
  #   dp, tx.var, models, or cts.obj
  obj$stages <- lapply(obj$stages,
                       FUN = function(stg) {
                         stg[c("dp", "tx.var", "models", "cts.obj")] <- NULL
                         stg
                       })
  
  obj <- c(obj, do.call(mapply, c(obj$stages, FUN = "list", SIMPLIFY = FALSE)))
  obj$stages <- NULL
  
  # non-regularity
  if (obj$cts[[1L]] == "bin") {
    obj$nonreg <- .varest(obj)
  }
  
  obj$call <- match.call()

  analysis_settings <- c("models" = "models", 
                         "method" = "method", 
                         "var.estim" = "var.estim", 
                         "boot.controls" = "boot.controls", 
                         "cc.modeled" = "censoring.modeled",
                         "tx.weight" = "tx.weight", 
                         "tx.type" = "tx.type", 
                         "n.bins" = "n.bins", 
                         "tx.wgt.man" = "tx.wgt.man",
                         "tx.range" = "tx.range", 
                         "tx.family" = "tx.family", 
                         "type" = "type",
                         "tx.vars" = "treat.vars")
  setup <- obj[analysis_settings]
  obj[analysis_settings] <- NULL
  names(setup) <- names(analysis_settings)
  obj$setup <- setup

  training_data <- c("data", "outcome", "A")
  training <- obj[training_data]
  obj[training_data] <- NULL
  obj$training_data <- training
  
  analysis_names <- c("n" = "n", 
                      "last.stage" = "last.stage", 
                      "prob.cc" = "prob.complete.case", 
                      "cc.mod.fitted" = "cens.mod.fitted",
                      "cc.wgt" = "cens.wgt", 
                      "cts" = "cts", 
                      "tx.mod.fitted" = "tx.mod.fitted", 
                      "A.hat" = "A.hat", 
                      "tx.wgt" = "tx.wgt",
                      "outcome.fit" = "outcome.fit", 
                      "Y" = "Y",
                      "regret" = "regret",
                      "opt.treat" = "opt.treat", 
                      "opt.Y" = "opt.Y",
                      "fitted.values" = "fitted.values",
                      "residuals" = "residuals",
                      "blip.data" = "blip.data")
  analysis <- obj[analysis_names]
  obj[analysis_names] <- NULL
  names(analysis) <- names(analysis_names)
  obj$analysis <- analysis
  
  if (is.null(obj$covmat)) obj$covmat <- NA
  if (is.null(obj$nonreg)) obj$nonreg <- NA
  
  for_return <- c("K", "beta", "psi", "covmat", "psi.boot", "nonreg", 
                  "setup", "training_data", "analysis", "call")
  obj <- obj[for_return %in% names(obj)]
  
  if (addWarning) obj$warn <- warningMsg
  
  obj$setup$method <- method

  # return
  class(obj) <- c("DTRreg", class(obj))
  obj
}

# Shared data

data(twoStageCont)
n <- nrow(twoStageCont)
K <- 2L

blip.mod <- list(~X1, ~X2)
treat.mod <- list(A1 ~ X1, A2 ~ X2)
tf.mod <- list(~X1, ~X2)
qblip.mod <- list(~X1+A1, ~X2+A2)


# Output structure - DWOLS, var.estim = "none"


local({
  mod <- suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, method = "dwols", var.estim = "none")
  )
  
  test_that("DTRreg DWOLS: class is 'DTRreg'", {
    expect_true(inherits(mod, "DTRreg"))
  })
  
  test_that("DTRreg DWOLS: K matches number of stages", {
    expect_equal(mod$K, K)
  })
  
  test_that("DTRreg DWOLS: top-level names contain required elements", {
    expect_true(all(c("K", "beta", "psi", "covmat", "nonreg",
                      "setup", "training_data", "analysis", "call") %in% names(mod)))
  })
  
  # - psi / beta -
  
  test_that("DTRreg DWOLS: psi is a list of length K", {
    expect_true(is.list(mod$psi))
    expect_length(mod$psi, K)
  })
  
  test_that("DTRreg DWOLS: psi names contain treatment variable at each stage", {
    expect_true(all(grepl("A1", names(mod$psi[[1L]]))))
    expect_true(all(grepl("A2", names(mod$psi[[2L]]))))
  })
  
  test_that("DTRreg DWOLS: beta is a list of length K", {
    expect_true(is.list(mod$beta))
    expect_length(mod$beta, K)
  })
  
  test_that("DTRreg DWOLS: beta names do not contain treatment variable", {
    expect_false(any(grepl("^A1", names(mod$beta[[1L]]))))
    expect_false(any(grepl("^A2", names(mod$beta[[2L]]))))
  })
  
  test_that("DTRreg DWOLS: covmat is NA when var.estim='none'", {
    expect_true(is.na(mod$covmat))
  })
  
  test_that("DTRreg DWOLS: psi.boot absent when var.estim='none'", {
    expect_null(mod$psi.boot)
  })
  
  # - setup --
  
  test_that("DTRreg DWOLS: setup contains required fields", {
    expect_true(all(c("models", "method", "var.estim", "tx.weight",
                      "tx.type", "type", "tx.vars") %in% names(mod$setup)))
  })
  
  test_that("DTRreg DWOLS: setup$method matches requested method", {
    expect_equal(mod$setup$method, "dwols")
  })
  
  test_that("DTRreg DWOLS: setup$var.estim matches requested var.estim", {
    expect_equal(mod$setup$var.estim, "none")
  })
  
  test_that("DTRreg DWOLS: setup$tx.vars matches treatment variable names", {
    expect_equal(mod$setup$tx.vars, c("A1", "A2"))
  })
  
  test_that("DTRreg DWOLS: setup$models has length K", {
    expect_length(mod$setup$models, K)
  })
  
  # - training_data -
  
  test_that("DTRreg DWOLS: training_data contains data, outcome, A", {
    expect_named(mod$training_data, c("data", "outcome", "A"), ignore.order = TRUE)
  })
  
  test_that("DTRreg DWOLS: training_data$data has n rows", {
    expect_equal(nrow(mod$training_data$data), n)
  })
  
  test_that("DTRreg DWOLS: training_data$outcome matches supplied outcome", {
    expect_equal(mod$training_data$outcome, twoStageCont$Y)
  })
  
  # - analysis -
  
  test_that("DTRreg DWOLS: analysis contains required elements", {
    expect_true(all(c("n", "last.stage", "prob.cc", "cc.mod.fitted",
                      "cc.wgt", "cts", "tx.mod.fitted", "A.hat", "tx.wgt",
                      "outcome.fit", "Y", "regret", "opt.treat",
                      "opt.Y", "fitted.values", "residuals",
                      "blip.data") %in% names(mod$analysis)))
  })
  
  test_that("DTRreg DWOLS: analysis$n is list of length K", {
    expect_length(mod$analysis$n, K)
  })
  
  test_that("DTRreg DWOLS: analysis$last.stage is integer of length n", {
    expect_true(is.integer(mod$analysis$last.stage))
    expect_length(mod$analysis$last.stage, n)
  })
  
  test_that("DTRreg DWOLS: analysis$opt.Y is numeric of length n", {
    expect_true(is.numeric(mod$analysis$opt.Y))
    expect_length(mod$analysis$opt.Y, n)
  })
  
  test_that("DTRreg DWOLS: analysis$cts[[1]] is 'bin' for binary treatment", {
    expect_equal(mod$analysis$cts[[1L]], "bin")
  })
  
  test_that("DTRreg DWOLS: analysis$outcome.fit is list of K fitted models", {
    expect_length(mod$analysis$outcome.fit, K)
    expect_true(inherits(mod$analysis$outcome.fit[[1L]], "lm"))
  })
  
  test_that("DTRreg DWOLS: analysis$opt.treat is list of length K", {
    expect_length(mod$analysis$opt.treat, K)
  })
})



# GEST method


local({
  mod <- suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, method = "gest", var.estim = "none")
  )
  
  test_that("DTRreg GEST: class is 'DTRreg'", {
    expect_true(inherits(mod, "DTRreg"))
  })
  
  test_that("DTRreg GEST: psi names contain treatment variable", {
    expect_true(all(grepl("A1", names(mod$psi[[1L]]))))
    expect_true(all(grepl("A2", names(mod$psi[[2L]]))))
  })
  
  test_that("DTRreg GEST: outcome.fit has class GEST", {
    expect_true(inherits(mod$analysis$outcome.fit[[1L]], "GEST"))
    expect_true(inherits(mod$analysis$outcome.fit[[2L]], "GEST"))
  })
  
  test_that("DTRreg GEST: setup$method is 'gest'", {
    expect_equal(mod$setup$method, "gest")
  })
})



# Sandwich variance


local({
  mod <- suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, method = "dwols", var.estim = "sandwich")
  )
  
  test_that("DTRreg sandwich: covmat is a list of length K", {
    expect_true(is.list(mod$covmat))
    expect_length(mod$covmat, K)
  })
  
  test_that("DTRreg sandwich: each covmat element is a square matrix", {
    for (k in seq_len(K)) {
      expect_true(is.matrix(mod$covmat[[k]]))
      expect_equal(nrow(mod$covmat[[k]]), ncol(mod$covmat[[k]]))
    }
  })
  
  test_that("DTRreg sandwich: covmat row/col names contain psi names", {
    for (k in seq_len(K)) {
      expect_true(all(names(mod$psi[[k]]) %in% rownames(mod$covmat[[k]])))
    }
  })
  
  test_that("DTRreg sandwich: nonreg is numeric of length K for binary treatment", {
    expect_true(is.numeric(mod$nonreg))
    expect_length(mod$nonreg, K)
  })
  
  test_that("DTRreg sandwich: psi.boot absent for sandwich", {
    expect_null(mod$psi.boot)
  })
})



# Bootstrap variance


local({
  withr::with_seed(42L, {
    mod <- suppressMessages(
      DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
             data = twoStageCont,
             method = "dwols",
             var.estim = "bootstrap",
             bootstrap.controls = list(B = 10L))
    )
  })
  
  test_that("DTRreg bootstrap: covmat is list of length K", {
    expect_true(is.list(mod$covmat))
    expect_length(mod$covmat, K)
  })
  
  test_that("DTRreg bootstrap: psi.boot is list of length K", {
    expect_true(is.list(mod$psi.boot))
    expect_length(mod$psi.boot, K)
  })
  
  test_that("DTRreg bootstrap: each psi.boot matrix has B rows", {
    for (k in seq_len(K)) {
      expect_equal(nrow(mod$psi.boot[[k]]), 10L)
    }
  })
})



# type = "effect" (blip rather than DTR)


test_that("DTRreg effect type: setup$type is 'effect'", {
  mod <- suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, var.estim = "none", dtr = FALSE)
  )
  expect_equal(mod$setup$type, "effect")
})



# Missing data - IPW


test_that("DTRreg missing='ipw': runs without error", {
  dat_na <- twoStageCont
  dat_na$X2[1L:10L] <- NA_real_
  expect_no_error(suppressMessages(
    DTRreg(dat_na$Y, blip.mod, treat.mod, tf.mod,
           data = dat_na,
           missing = "ipw",
           var.estim = "none")
  ))
})

test_that("DTRreg missing='drop': runs without error", {
  dat_na <- twoStageCont
  dat_na$X2[1L:10L] <- NA_real_
  expect_no_error(suppressMessages(
    DTRreg(dat_na$Y, blip.mod, treat.mod, tf.mod,
           data = dat_na,
           missing = "drop",
           var.estim = "none")
  ))
})



# Weight options


test_that("DTRreg weight='none': runs without error", {
  expect_no_error(suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, weight = "none", var.estim = "none")
  ))
})

test_that("DTRreg weight='cipw': runs without error", {
  expect_no_error(suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, weight = "cipw", var.estim = "none")
  ))
})

test_that("DTRreg weight='overlap': runs without error for binary treatment", {
  expect_no_error(suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, weight = "overlap", var.estim = "none")
  ))
})

test_that("DTRreg weight='abs': runs with message", {
  expect_warning(suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, weight = "abs", var.estim = "none")
  ), "renamed 'overlap'")
})

test_that("DTRreg weight='manual': runs and psi matches ipw result", {
  mod_ipw <- suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, method = "dwols",
           data = twoStageCont, weight = "ipw", var.estim = "none")
  )
  # Extract computed weights and pass as manual
  tx_man <- list(
    mod_ipw$analysis$tx.wgt[[1L]],
    mod_ipw$analysis$tx.wgt[[2L]]
  )
  mod_man <- suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, weight = "manual", method = "dwols",
           treat.wgt.man = tx_man, var.estim = "none")
  )
  expect_equal(mod_man$psi, mod_ipw$psi)
})



# full.cov = TRUE


test_that("DTRreg full.cov=TRUE: covmat larger than full.cov=FALSE", {
  mod_full <- suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, var.estim = "sandwich", full.cov = TRUE)
  )
  mod_blip <- suppressMessages(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, var.estim = "sandwich", full.cov = FALSE)
  )
  for (k in seq_len(K)) {
    expect_true(nrow(mod_full$covmat[[k]]) >= nrow(mod_blip$covmat[[k]]))
  }
})



# Stopping conditions


test_that("DTRreg: stops when outcome has missing values", {
  Y_na <- twoStageCont$Y
  Y_na[1L] <- NA_real_
  expect_error(
    DTRreg(Y_na, blip.mod, treat.mod, tf.mod, data = twoStageCont),
    "outcome"
  )
})

test_that("DTRreg: stops when data is not a data.frame", {
  expect_error(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = as.matrix(twoStageCont)),
    "data.frame"
  )
})

test_that("DTRreg: stops when method is unrecognised", {
  expect_error(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, method = "qbayesian"),
    "should be one of"
  )
})

test_that("DTRreg: deprecated 'abs' weight triggers warning", {
  expect_warning(
    tryCatch(
      suppressMessages(
        DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
               data = twoStageCont, weight = "abs", var.estim = "none")
      ),
      error = function(e) NULL
    ),
    "renamed.*overlap|overlap.*renamed"
  )
})

test_that("DTRreg: deprecated 'wo' weight triggers warning", {
  expect_warning(
    tryCatch(
      suppressMessages(
        DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
               data = twoStageCont, weight = "wo", var.estim = "none")
      ),
      error = function(e) NULL
    ),
    "renamed.*overlap|overlap.*renamed"
  )
})
test_that("DTRreg: stops when dtr is not logical", {
  expect_error(
    DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
           data = twoStageCont, dtr = "yes"),
    "dtr"
  )
})

test_that("DTRreg: stops when blip.mod and treat.mod have different lengths", {
  expect_error(
    DTRreg(twoStageCont$Y, list(~X1), treat.mod, tf.mod,
           data = twoStageCont),
    "same length"
  )
})

test_that("DTRreg: warns when gest used with manual weights", {
  expect_warning(
    suppressMessages(
      DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod,
             data = twoStageCont,
             method = "gest",
             weight = "manual",
             treat.wgt.man = list(rep(1.0, n), rep(1.0, n)),
             var.estim = "none")
    ),
    "not used in g-estimation"
  )
})

# testing 12/31/2023 updates

test_that("intercept only is not an issue; censored data", {
  skip_on_cran()
  data("twoStageCens")
  
  for (method in c("dwols", "qlearn")) {
    for (weight in c("overlap", "ipw", "cipw", "none")) {
      for (var.estim in c("none", "bootstrap", "sandwich")) {
        for (full.cov in c(TRUE, FALSE)) {
          expect_no_error(DWSurv(time = list(~ T1, ~ T2), 
                                 blip.mod = list(~ 1, ~ 1), 
                                 treat.mod = list(A1 ~ 1, A2 ~ 1), 
                                 tf.mod = list(~ 1, ~ 1), 
                                 cens.mod = list(delta ~ 1, delta ~ 1), 
                                 data = twoStageCens, 
                                 method = method,
                                 var.estim = var.estim,
                                 weight = weight))
        }
      }
    }
  }
  
})

test_that("intercept only is not an issue; censored multinomial data", {
  skip_on_cran()
  data("twoStageCens")
  twoStageCens$A1 <- withr::with_seed(1234, sample(1:3, nrow(twoStageCens), replace = TRUE))
  twoStageCens$A2[!is.na(twoStageCens$A2)] <- withr::with_seed(2345, 
                                                               sample(1:3, 
                                                                      sum(!is.na(twoStageCens$A2)), 
                                                                      replace = TRUE))
  
  for (method in c("dwols", "qlearn")) {
    for (weight in c("ipw", "cipw", "qpom", "overlap", "none")) {
      for (var.estim in c("none", "bootstrap", "sandwich")) {
        for (full.cov in c(TRUE, FALSE)) {
          expect_no_error(DWSurv(time = list(~ T1, ~ T2), 
                                 blip.mod = list(~ 1, ~ 1), 
                                 treat.mod = list(A1 ~ 1, A2 ~ 1), 
                                 tf.mod = list(~ 1, ~ 1), 
                                 cens.mod = list(delta ~ 1, delta ~ 1), 
                                 treat.type = "multi",
                                 data = twoStageCens, 
                                 method = method,
                                 var.estim = var.estim,
                                 weight = weight))
        }
      }
    }
  }
  
})

test_that("intercept only is not an issue; censored continuous data", {
  skip_on_cran()
  data("twoStageCens")
  twoStageCens$A1 <- withr::with_seed(1234, stats::runif(nrow(twoStageCens)))
  twoStageCens$A2[!is.na(twoStageCens$A2)] <- 
    withr::with_seed(2345, 
                     stats::runif(sum(!is.na(twoStageCens$A2))))
  
  for (method in c("dwols", "qlearn")) {
    for (weight in c("ipw", "cipw", "qpom", "overlap", "none")) {
      for (var.estim in c("none", "bootstrap", "sandwich")) {
        for (full.cov in c(TRUE, FALSE)) {
          expect_no_error(DWSurv(time = list(~ T1, ~ T2), 
                                 blip.mod = list(~ 1 + A1, ~ 1 + A2), 
                                 treat.mod = list(A1 ~ 1, A2 ~ 1), 
                                 tf.mod = list(~ 1, ~ 1), 
                                 cens.mod = list(delta ~ 1, delta ~ 1), 
                                 treat.type = "cont",
                                 data = twoStageCens, 
                                 method = method,
                                 var.estim = var.estim,
                                 weight = weight))
        }
      }
    }
  }
  
})
  
test_that("intercept only is not an issue; survival data", {
  skip_on_cran()
  data("twoStageSurv")
  
  for (method in c("dwols", "qlearn")) {
    for (weight in c("overlap", "ipw", "cipw", "none")) {
      for (var.estim in c("none", "bootstrap", "sandwich")) {
        for (full.cov in c(TRUE, FALSE)) {
          expect_no_error(DWSurv(time = list(~ T1, ~ T2), 
                                 blip.mod = list(~ 1, ~ 1), 
                                 treat.mod = list(A1 ~ 1, A2 ~ 1), 
                                 tf.mod = list(~ 1, ~ 1), 
                                 cens.mod = list( ~ delta, ~ delta), 
                                 data = twoStageSurv, 
                                 method = method,
                                 var.estim = var.estim,
                                 weight = weight))
        }
      }
    }
  }
  
})

test_that("intercept only is not an issue; survival multinomial data", {
  skip_on_cran()
  data("twoStageSurv")
  twoStageSurv$A1 <- withr::with_seed(1234, sample(1:3, nrow(twoStageSurv), replace = TRUE))
  twoStageSurv$A2[!is.na(twoStageSurv$A2)] <- withr::with_seed(2345, 
                                                               sample(1:3, 
                                                                      sum(!is.na(twoStageSurv$A2)), 
                                                                      replace = TRUE))
  
  for (method in c("dwols", "qlearn")) {
    for (weight in c("ipw", "cipw", "qpom", "overlap", "none")) {
      for (var.estim in c("none", "bootstrap", "sandwich")) {
        for (full.cov in c(TRUE, FALSE)) {
          expect_no_error(DWSurv(time = list(~ T1, ~ T2), 
                                 blip.mod = list(~ 1, ~ 1), 
                                 treat.mod = list(A1 ~ 1, A2 ~ 1), 
                                 tf.mod = list(~ 1, ~ 1), 
                                 cens.mod = list(~ delta, ~ delta), 
                                 treat.type = "multi",
                                 data = twoStageSurv, 
                                 method = method,
                                 var.estim = var.estim,
                                 weight = weight))
        }
      }
    }
  }
  
})

test_that("intercept only is not an issue; survival continuous data", {
  skip_on_cran()
  data("twoStageSurv")
  twoStageSurv$A1 <- withr::with_seed(1234, stats::runif(nrow(twoStageSurv)))
  twoStageSurv$A2[!is.na(twoStageSurv$A2)] <- 
    withr::with_seed(2345, 
                     stats::runif(sum(!is.na(twoStageSurv$A2))))
  
  for (method in c("dwols", "qlearn")) {
    for (weight in c("ipw", "cipw", "qpom", "overlap", "none")) {
      for (var.estim in c("none", "bootstrap", "sandwich")) {
        for (full.cov in c(TRUE, FALSE)) {
          expect_no_error(DWSurv(time = list(~ T1, ~ T2), 
                                 blip.mod = list(~ 1 + A1, ~ 1 + A2), 
                                 treat.mod = list(A1 ~ 1, A2 ~ 1), 
                                 tf.mod = list(~ 1, ~ 1), 
                                 cens.mod = list(~ delta, ~ delta), 
                                 treat.type = "cont",
                                 data = twoStageSurv, 
                                 method = method,
                                 var.estim = var.estim,
                                 weight = weight))
        }
      }
    }
  }
  
})

test_that("intercept only is not an issue; cont data", {
  skip_on_cran()
  data("twoStageCont")

  for (method in c("gest", "dwols", "qlearn")) {
    for (weight in c("overlap", "ipw", "cipw", "none")) {
      for (var.estim in c("none", "bootstrap", "sandwich")) {
        for (full.cov in c(TRUE, FALSE)) {
          for (miss in c("drop", "ipw")) {
            expect_no_error(DTRreg(twoStageCont$Y, 
                                   blip.mod = list(~ 1, ~ 1), 
                                   treat.mod = list(A1 ~ 1, A2 ~ 1), 
                                   tf.mod = list(~ 1, ~ 1), 
                                   weight = weight, 
                                   method = method,
                                   var.estim = var.estim, 
                                   data = twoStageCont,
                                   missing = miss,
                                   missing.mod = list(~ 1, ~ 1)))
          }
        }
      }
    }
  }
  
})

test_that("intercept only is not an issue; cont multinomial data", {
  skip_on_cran()
  data("twoStageCont", package = "DTRreg")
  twoStageCont$A1 <- withr::with_seed(1234, sample(1:3, nrow(twoStageCont), replace = TRUE))
  twoStageCont$A2[!is.na(twoStageCont$A2)] <- withr::with_seed(2345, 
                                                               sample(1:3, 
                                                                      sum(!is.na(twoStageCont$A2)), 
                                                                      replace = TRUE))
  
  for (method in c("dwols", "qlearn")) {
    for (weight in c("ipw", "cipw", "qpom", "overlap", "none")) {
      for (var.estim in c("none", "bootstrap", "sandwich")) {
        for (full.cov in c(TRUE, FALSE)) {
          for (miss in c("drop", "ipw")) {
            expect_no_error(DTRreg(twoStageCont$Y, 
                                   blip.mod = list(~ 1, ~ 1), 
                                   treat.mod = list(A1 ~ 1, A2 ~ 1), 
                                   treat.type = "multi",
                                   tf.mod = list(~ 1, ~ 1), 
                                   method = method,
                                   weight = weight,
                                   var.estim = var.estim, 
                                   data = twoStageCont,
                                   missing = miss,
                                   missing.mod = list(~ 1, ~ 1)))
          }
        }
      }
    }
  }
  
})

test_that("intercept only is not an issue; cont continuous data", {
  skip_on_cran()
  data("twoStageCont", package = "DTRreg")
  twoStageCont$A1 <- withr::with_seed(1234, stats::runif(nrow(twoStageCont)))
  twoStageCont$A2[!is.na(twoStageCont$A2)] <- withr::with_seed(2345, 
                                                               stats::runif(!is.na(twoStageCont$A2)))
  
  for (method in c("gest", "dwols", "qlearn")) {
    for (weight in c("ipw", "cipw", "qpom", "overlap", "none")) {
      for (var.estim in c("none", "bootstrap", "sandwich")) {
        for (full.cov in c(TRUE, FALSE)) {
          for (miss in c("drop", "ipw")) {
            expect_no_error(DTRreg(twoStageCont$Y, 
                                   blip.mod = list(~ 1 + A1, ~ 1 + A2), 
                                   treat.mod = list(A1 ~ 1, A2 ~ 1), 
                                   treat.type = "cont",
                                   tf.mod = list(~ 1, ~ 1), 
                                   method = method,
                                   weight = weight,
                                   var.estim = var.estim, 
                                   data = twoStageCont,
                                   missing = miss,
                                   missing.mod = list(~ 1, ~ 1)))
          }
        }
      }
    }
  }
  
})

test_that("testing that qlearn in DWsurv is just setting treatment weights to 1.0", {
  data("twoStageCens")
  
  original <- DWSurv(time = list(~ T1, ~ T2), 
                     blip.mod = list(~ X11, ~ X21), 
                     treat.mod = list(A1 ~ X11, A2 ~ X21), 
                     tf.mod = list(~ X11 + X12, ~ X21), 
                     cens.mod = list(delta ~ X11, delta ~ X21),
                     data = twoStageCens,
                     method = "qlearn")
  
  manual <- DWSurv(time = list(~ T1, ~ T2), 
                   blip.mod = list(~ X11, ~ X21), 
                   treat.mod = list(A1 ~ X11, A2 ~ X21), 
                   tf.mod = list(~ X11 + X12, ~ X21), 
                   cens.mod = list(delta ~ X11, delta ~ X21),
                   data = twoStageCens,
                   weight = "manual",
                   treat.wgt.man = list(rep(1.0, nrow(twoStageCens)),
                                        rep(1.0, nrow(twoStageCens))),
                   method = "dwols")
  
  expect_equal(original$beta, manual$beta)
  expect_equal(original$psi, manual$psi)
  expect_equal(original$analysis$opt.treat, manual$analysis$opt.treat)
  expect_equal(original$analysis$opt.Y, manual$analysis$opt.Y)
  
  # use this to extract the manual weights
  censor_wgts <- original$analysis$cens.wgt
  treat_wgts <- original$analysis$tx.wgt
  
  tx_man <- list()
  tx_man[[1L]] <- censor_wgts[[1L]]
  tx_man[[2L]] <- rep(NA_real_, nrow(twoStageCens))
  tx_man[[2L]][!is.na(twoStageCens$T2)] <- censor_wgts[[2L]]
  
  manual <- DWSurv(time = list(~ T1, ~ T2), 
                   blip.mod = list(~ X11, ~ X21), 
                   treat.mod = list(A1 ~ X11, A2 ~ X21), 
                   tf.mod = list(~ X11 + X12, ~ X21), 
                   cens.mod = list( ~ delta, ~ delta),
                   data = twoStageCens,
                   weight = "manual.with.censor",
                   treat.wgt.man = tx_man,
                   method = "dwols")
  
  expect_equal(original$beta, manual$beta)
  expect_equal(original$psi, manual$psi)
  expect_equal(original$analysis$opt.treat, manual$analysis$opt.treat)
  expect_equal(original$analysis$opt.Y, manual$analysis$opt.Y)
})


test_that("testing that using manual overrides any treatment weight specification", {
  data("twoStageCens")
  
  original <- DWSurv(time = list(~ T1, ~ T2), 
                     blip.mod = list(~ X11, ~ X21), 
                     treat.mod = list(A1 ~ X11, A2 ~ X21), 
                     tf.mod = list(~ X11 + X12, ~ X21), 
                     cens.mod = list(delta ~ X11, delta ~ X21),
                     data = twoStageCens,
                     method = "dwols",
                     weight = "manual",
                     treat.wgt.man = list(rep(1.0, nrow(twoStageCens)),
                                          rep(1.0, nrow(twoStageCens))))
  
  manual <- DWSurv(time = list(~ T1, ~ T2), 
                   blip.mod = list(~ X11, ~ X21), 
                   treat.mod = list(A1 ~ X11, A2 ~ X21), 
                   tf.mod = list(~ X11 + X12, ~ X21), 
                   cens.mod = list(delta ~ X11, delta ~ X21),
                   data = twoStageCens,
                   method = "qlearn")
  
  expect_equal(original$beta, manual$beta)
  expect_equal(original$psi, manual$psi)
  expect_equal(original$analysis$opt.treat, manual$analysis$opt.treat)
  expect_equal(original$analysis$opt.Y, manual$analysis$opt.Y)
  
})

test_that("testing the ability to specify censoring and treatment through manual input", {
  data("twoStageCens")
  
  expect_no_error(DWSurv(time = list(~ T1, ~ T2), 
                         blip.mod = list(~ 1, ~ 1), 
                         treat.mod = list(A1 ~ 1, A2 ~ 1), 
                         tf.mod = list(~ 1, ~ 1), 
                         cens.mod = list(delta ~ 1, delta ~ 1), 
                         var.estim = "sandwich", 
                         data = twoStageCens))
  
  expect_no_error(DWSurv(time = list(~ T1, ~ T2), 
                         blip.mod = list(~ 1, ~ 1), 
                         treat.mod = list(A1 ~ 1, A2 ~ 1), 
                         tf.mod = list(~ 1, ~ 1), 
                         cens.mod = list(delta ~ 1, delta ~ 1), 
                         var.estim = "boot", 
                         data = twoStageCens))
  
  original <- DWSurv(time = list(~ T1, ~ T2), 
                     blip.mod = list(~ X11, ~ 1), 
                     treat.mod = list(A1 ~ X11, A2 ~ 1), 
                     tf.mod = list(~ X11 + X12, ~ 1), 
                     cens.mod = list(delta ~ 1, delta ~ X11), 
                     var.estim = "sandwich", 
                     data = twoStageCens)
  
  # use this to extract the manual weights
  censor_wgts <- original$analysis$cens.wgt
  treat_wgts <- original$analysis$tx.wgt
  
  n_total    <- nrow(twoStageCens)
  stage1_idx <- original$analysis$last.stage >= 1L
  stage2_idx <- original$analysis$last.stage >= 2L
  
  tx_man <- list()
  tx_man[[1L]] <- rep(NA_real_, n_total)
  tx_man[[1L]][stage1_idx] <- censor_wgts[[1L]] * treat_wgts[[1L]]
  
  tx_man[[2L]] <- rep(NA_real_, n_total)
  tx_man[[2L]][stage2_idx] <- censor_wgts[[2L]]
  
  manual <- DWSurv(time = list(~ T1, ~ T2), 
                   blip.mod = list(~ X11, ~ 1), 
                   treat.mod = list(A1 ~ X11, A2 ~ 1), 
                   tf.mod = list(~ X11 + X12, ~ 1), 
                   cens.mod = list(~ delta, ~ delta), 
                   var.estim = "sandwich", 
                   weight = "manual.with.censor",
                   treat.wgt.man = tx_man,
                   data = twoStageCens)
  
  expect_equal(original$beta, manual$beta)
  expect_equal(original$psi, manual$psi)
  expect_equal(original$analysis$opt.treat, manual$analysis$opt.treat)
  expect_equal(original$analysis$opt.Y, manual$analysis$opt.Y)
  
})

test_that("testing the ability to specify censoring and treatment through manual input doesn't change DTRreg", {
  data(twoStageCont)
  
  # models to be passed to DTRreg
  # blip model
  blip.mod <- list(~ X1, ~ X2)
  # treatment model (correctly specified)
  treat.mod <- list(A1 ~ X1, A2 ~ X2)
  # treatment-free model (incorrectly specified)
  tf.mod <- list(~ X1, ~ X2)
  
  # perform G-estimation
  expect_no_error(original <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
                         data = twoStageCont, method = "dwols"))
  
  treat_wgts <- original$analysis$tx.wgt
  tx_man <- list()
  tx_man[[1L]] <- treat_wgts[[1L]]
  tx_man[[2L]] <- treat_wgts[[2L]]

  expect_no_error(manual <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
                                   data = twoStageCont, method = "dwols",
                                   weight = "manual", treat.wgt.man = tx_man))
  
  expect_equal(original$beta, manual$beta)
  expect_equal(original$psi, manual$psi)
  expect_equal(original$analysis$opt.treat, manual$analysis$opt.treat)
  expect_equal(original$analysis$opt.Y, manual$analysis$opt.Y)
  
  expect_error(DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
                      data = twoStageCont, method = "dwols",
                      weight = "manual.with.censor", treat.wgt.man = tx_man))
})

#' Extract Model Matrix
#' 
#' @noRd
#' @param model A formula object.
#' @param data A data.frame object.
#'
#' @return A matrix. The design matrix
#' 
#' @keywords internal
.getModelMatrix <- function(model, data) {
  tryCatch(stats::model.frame(model, data, na.action = "na.pass") |>
             stats::model.matrix(object = model),
           error = function(e) {
             stop("unable to extract model frame\n\t", e$message, call. = FALSE)
           })
}

#' Recalculate optimal treatment for all analyzed stages
#'
#' @noRd
#' @param steps A list object. All evaluated stage results.
#' @param complete.case.info A list object. The complete cases information,
#'   `$last.stage` and `$prob.complete.case`
#' @param data A data.frame object. The full covariate and treatment dataset.
#'   Must contain all data for all participants.
#' 
#' @return A numeric vector object.
#' @keywords internal
.stageTx <- function(steps, complete.case.info, data, quiet) {
  
  stopifnot(
    "`steps` must be a list" = is.list(steps),
    "`complete.case.info` must be a list" = is.list(complete.case.info) &&
      all(c("last.stage", "prob.complete.case") %in% names(complete.case.info)),
    "`data` must be a data.frame" = is.data.frame(data)
  )
  
  # NOTE this moves in the FORWARD direction
  for (i in seq_len(length(steps))) {
    if (is.null(steps[[i]])) next
    
    # identify the complete cases for the stage
    stage_cases <- complete.case.info$last.stage >= steps[[i]]$dp

    # estimate optimal treatment
    opt.tx <- steps[[i]]$cts.obj$opt(steps[[i]]$outcome.fit,
                                     data[stage_cases, , drop = FALSE], quiet) |> 
      drop()
    if (steps[[i]]$cts == "multi") {
      tmp <- factor(rep(NA, nrow(data)), levels = levels(opt.tx))
    } else {
      tmp <- rep(NA_real_, nrow(data))
    }
    tmp[stage_cases] <- opt.tx
    
    steps[[i]]$opt.treat <- tmp
    data[, steps[[i]]$tx.var] <- steps[[i]]$opt.treat
  }
  steps
}

#' Recalculate Y based using all model estimates
#'
#' @noRd
#' @param Y A numeric vector object of length N. The observed outcome of 
#'   interest for all participants.
#' @param steps A list object. All evaluated stage results.
#' @param complete.case.info A list object. The complete cases information,
#'   `$last.stage` and `$prob.complete.case`
#' @param data A data.frame object. The full covariate and treatment dataset.
#'   Must contain all data for all participants.
#' @param type A character object. The shift (effect/optimal).
#' @param isSurvival A logical object. TRUE if survival outcome.
#' 
#' @return A numeric vector object.
#' @keywords internal
.stageY <- function(Y, steps, complete.case.info, data, type) {
  
  stopifnot(
    "`Y` must be a numeric vector" = is.numeric(Y) && is.vector(Y),
    "`steps` must be a list" = is.list(steps),
    "`complete.case.info` must be a list" = is.list(complete.case.info) &&
      all(c("last.stage", "prob.complete.case") %in% names(complete.case.info)),
    "`data` must be a data.frame" = is.data.frame(data) && nrow(data) == length(Y),
    "`type` must be a character" = is.character(type) && is.vector(type) &&
      length(type) == 1L
  )

  # NOTE this moves in the FORWARD direction
  for (stp in steps) {
    # identify the complete cases for the stage
    stage_cases <- complete.case.info$last.stage >= stp$dp
    
    shift <- stp$cts.obj$shiftY(type = type, 
                                outcome.fit = stp$outcome.fit,
                                data = data[stage_cases, , drop = FALSE],
                                opt = stp$opt.treat[stage_cases], 
                                A = stp$A)
    Y[stage_cases] <- Y[stage_cases] + shift
  }
  Y
}

#' Recalculate Y based using all model estimates
#'
#' @noRd
#' @param times A character vector. The time variables
#' @param steps A list object. All evaluated stage results.
#' @param complete.case.info A list object. The complete cases information,
#'   `$last.stage` and `$prob.complete.case`
#' @param data A data.frame object. The full covariate and treatment dataset.
#'   Must contain all data for all participants.
#' @param type A character object. The shift (effect/optimal).
#' @param isSurvival A logical object. TRUE if survival outcome.
#' 
#' @return A numeric vector object.
#' @keywords internal
.stageYSurvival <- function(times, steps, complete.case.info, data, type) {
  
  stopifnot(
    "`times` must be a character vector" = is.character(times) && is.vector(times),
    "`steps` must be a list" = is.list(steps),
    "`complete.case.info` must be a list" = is.list(complete.case.info) &&
      all(c("last.stage", "prob.complete.case") %in% names(complete.case.info)),
    "`data` must be a data.frame" = is.data.frame(data),
    "`type` must be a character" = is.character(type) && is.vector(type) &&
      length(type) == 1L
  )
  
  Y <- numeric(nrow(data))
  
  # NOTE this moves in the FORWARD direction
  for (stp in steps) {
    # identify the complete cases for the stage
    stage_cases <- complete.case.info$last.stage >= stp$dp
    
    shift <- stp$cts.obj$shiftY(type = type, 
                                outcome.fit = stp$outcome.fit,
                                data = data[stage_cases, , drop = FALSE],
                                opt = stp$opt.treat[stage_cases], 
                                A = stp$A)
    
    Y[stage_cases] <- Y[stage_cases] + data[stage_cases, times[stp$dp]] * exp(shift)
  }
  Y
}

#' Recalculate regret using all model estimates
#'
#' @noRd
#' @param Y A numeric vector object of length N. The observed outcome of 
#'   interest for all participants.
#' @param steps A list object. All evaluated stage results.
#' @param complete.case.info A list object. The complete cases information,
#'   `$last.stage` and `$prob.complete.case`
#' @param data A data.frame object. The full covariate and treatment dataset.
#'   Must contain all data for all participants.
#' @param type A character object. The shift (effect/optimal).
#' @param isSurvival A logical object. TRUE if survival outcome.
#' 
#' @return A numeric vector object.
#' @keywords internal
.stageRegret <- function(steps, complete.case.info, data, type) {
  
  stopifnot(
    "`steps` must be a list" = is.list(steps),
    "`complete.case.info` must be a list" = is.list(complete.case.info) &&
      all(c("last.stage", "prob.complete.case") %in% names(complete.case.info)),
    "`data` must be a data.frame" = is.data.frame(data),
    "`type` must be a character" = is.character(type) && is.vector(type) &&
      length(type) == 1L
  )
  
  regret <- list()
  
  # NOTE this moves in the FORWARD direction
  for (i in 1L:length(steps)) {
    stp <- steps[[i]]
    # identify the complete cases for the stage
    stage_cases <- complete.case.info$last.stage >= stp$dp
    
    regret[[i]] <- stp$cts.obj$shiftY(type = type, 
                                      outcome.fit = stp$outcome.fit,
                                      data = data[stage_cases, , drop = FALSE],
                                      opt = stp$opt.treat[stage_cases], 
                                      A = stp$A)
    
  }
  regret
}

dat <- data.frame(
  X = c(1.0, 2.0, 3.0, 4.0, 5.0),
  Z = c(2.0, 4.0, 6.0, 8.0, 10.0),
  Y = c(0.1, 0.2, 0.3, 0.4, 0.5)
)

test_that(".getModelMatrix: returns same matrix as direct model.matrix call", {
  model <- ~X + Z
  expected <- model.matrix(model, dat)
  result <- .getModelMatrix(model, dat)
  expect_equal(result, expected)
})

test_that(".getModelMatrix: column names match formula terms", {
  result <- .getModelMatrix(~X + Z, dat)
  expect_true("X" %in% colnames(result))
  expect_true("Z" %in% colnames(result))
  expect_true("(Intercept)" %in% colnames(result))
})

test_that(".getModelMatrix: returns matrix with correct dimensions", {
  result <- .getModelMatrix(~X + Z, dat)
  expect_true(is.matrix(result))
  expect_equal(nrow(result), nrow(dat))
  expect_equal(ncol(result), 3L) # intercept + X + Z
})

test_that(".getModelMatrix: interaction terms are constructed correctly", {
  model <- ~X + Z + X:Z
  expected <- model.matrix(model, dat)
  result <- .getModelMatrix(model, dat)
  expect_equal(result, expected)
})

test_that(".getModelMatrix: preserves rows with NA rather than dropping them", {
  dat_na <- dat
  dat_na$X[3L] <- NA_real_
  result <- .getModelMatrix(~X + Z, dat_na)
  expect_equal(nrow(result), nrow(dat_na))
  expect_true(any(is.na(result)))
})

test_that(".getModelMatrix: stops with informative message when column is absent", {
  expect_error(.getModelMatrix(~X + W, dat), "unable to extract model frame")
})

local({
  
  set.seed(42L)
  n_u <- 20L
  
  dat_u <- data.frame(
    X = cos(seq(0.0, pi, length.out = n_u)),
    A1 = as.integer(rbinom(n_u, 1L, 0.5)),
    A2 = as.integer(rbinom(n_u, 1L, 0.5)),
    T1 = abs(rnorm(n_u)) + 0.5,
    T2 = abs(rnorm(n_u)) + 0.5,
    Y = rnorm(n_u)
  )
  
  cts1 <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A1")
  cts2 <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A2")
  
  wgts_u <- rep(1.0, n_u)
  fit1 <- .dwols(Y = dat_u$Y, A = dat_u$A1, data = dat_u,
                 wgts = wgts_u, cts.obj = cts1, tx.var = "A1")
  fit2 <- .dwols(Y = dat_u$Y, A = dat_u$A2, data = dat_u,
                 wgts = wgts_u, cts.obj = cts2, tx.var = "A2")
  
  # All participants complete at both stages
  cc_u <- list(
    last.stage = rep(2L, n_u),
    prob.complete.case = matrix(1.0, nrow = n_u, ncol = 2L)
  )
  
  # Minimal steps — opt.treat pre-set so .stageY and .stageRegret can run
  make_steps <- function(opt1 = dat_u$A1, opt2 = dat_u$A2) {
    list(
      list(dp = 1L,
           tx.var = "A1",
           cts = "bin",
           cts.obj = cts1,
           outcome.fit = fit1$outcome.fit,
           opt.treat = opt1,
           A = dat_u$A1),
      list(dp = 2L,
           tx.var = "A2",
           cts = "bin",
           cts.obj = cts2,
           outcome.fit = fit2$outcome.fit,
           opt.treat = opt2,
           A = dat_u$A2)
    )
  }
  
  test_that(".stageTx: returns list of same length as input steps", {
    res <- .stageTx(make_steps(), cc_u, dat_u, quiet = TRUE)
    expect_length(res, 2L)
  })
  
  test_that(".stageTx: sets opt.treat on each stage", {
    res <- .stageTx(make_steps(), cc_u, dat_u, quiet = TRUE)
    expect_false(is.null(res[[1L]]$opt.treat))
    expect_false(is.null(res[[2L]]$opt.treat))
  })
  
  test_that(".stageTx: opt.treat for binary contains only 0/1/NA", {
    res <- .stageTx(make_steps(), cc_u, dat_u, quiet = TRUE)
    expect_true(all(res[[1L]]$opt.treat %in% c(0L, 1L, NA_integer_)))
    expect_true(all(res[[2L]]$opt.treat %in% c(0L, 1L, NA_integer_)))
  })
  
  test_that(".stageTx: oracle — opt.treat at stage_cases matches cts.obj$opt", {
    res <- .stageTx(make_steps(), cc_u, dat_u, quiet = TRUE)
    stage_cases <- cc_u$last.stage >= 1L
    expected <- cts1$opt(fit1$outcome.fit, dat_u[stage_cases, , drop = FALSE],
                         quiet = TRUE) |> drop()
    expect_equal(res[[1L]]$opt.treat[stage_cases], expected)
  })
  
  test_that(".stageTx: NULL steps element is skipped without error", {
    steps_with_null <- make_steps()
    steps_with_null[[1L]] <- NULL
    expect_no_error(.stageTx(steps_with_null, cc_u, dat_u, quiet = TRUE))
  })
  
  test_that(".stageTx: stops when steps is not a list", {
    expect_error(.stageTx("not a list", cc_u, dat_u, quiet = TRUE), "`steps`")
  })
  
  test_that(".stageTx: stops when complete.case.info missing required names", {
    bad_cc <- list(last.stage = cc_u$last.stage)
    expect_error(.stageTx(make_steps(), bad_cc, dat_u, quiet = TRUE),
                 "`complete.case.info`")
  })
  
  test_that(".stageTx: stops when data is not a data.frame", {
    expect_error(.stageTx(make_steps(), cc_u, as.matrix(dat_u), quiet = TRUE),
                 "`data`")
  })
  
  test_that(".stageY: returns numeric vector of length n", {
    res <- .stageY(dat_u$Y, make_steps(), cc_u, dat_u, type = "DTR")
    expect_true(is.numeric(res))
    expect_length(res, n_u)
  })
  
  test_that(".stageY: oracle — when opt equals observed A, regret is 0 and Y unchanged", {
    # opt.treat == A for all participants means regret = 0, Y += 0
    steps_same <- make_steps(opt1 = dat_u$A1, opt2 = dat_u$A2)
    res <- .stageY(dat_u$Y, steps_same, cc_u, dat_u, type = "DTR")
    expect_equal(res, dat_u$Y)
  })
  
  test_that(".stageY: Y changes when opt differs from observed A", {
    steps_diff <- make_steps(opt1 = as.integer(!dat_u$A1),
                             opt2 = as.integer(!dat_u$A2))
    res <- .stageY(dat_u$Y, steps_diff, cc_u, dat_u, type = "DTR")
    expect_false(isTRUE(all.equal(res, dat_u$Y)))
  })
  
  test_that(".stageY: stops when Y is not numeric", {
    expect_error(
      .stageY(as.character(dat_u$Y), make_steps(), cc_u, dat_u, type = "DTR"),
      "`Y`"
    )
  })
  
  test_that(".stageY: stops when steps is not a list", {
    expect_error(
      .stageY(dat_u$Y, "not a list", cc_u, dat_u, type = "DTR"),
      "`steps`"
    )
  })
  
  test_that(".stageY: stops when data nrow != length(Y)", {
    expect_error(
      .stageY(dat_u$Y, make_steps(), cc_u, dat_u[1L:5L, ], type = "DTR"),
      "`data`"
    )
  })
  
  test_that(".stageY: stops when type is not a character", {
    expect_error(
      .stageY(dat_u$Y, make_steps(), cc_u, dat_u, type = 1L),
      "`type`"
    )
  })
  
  test_that(".stageYSurvival: returns numeric vector of length n", {
    times <- c("T1", "T2")
    res <- .stageYSurvival(times, make_steps(), cc_u, dat_u, type = "DTR")
    expect_true(is.numeric(res))
    expect_length(res, n_u)
  })
  
  test_that(".stageYSurvival: oracle — when opt equals observed A, Y = sum of times", {
    # shift = 0 when opt == A, exp(0) = 1, so Y[sc] += time[sc]
    # With all participants complete at both stages:
    # Y = T1 * exp(shift1) + T2 * exp(shift2) = T1 + T2
    times <- c("T1", "T2")
    steps_same <- make_steps(opt1 = dat_u$A1, opt2 = dat_u$A2)
    res <- .stageYSurvival(times, steps_same, cc_u, dat_u, type = "DTR")
    expected <- dat_u$T1 + dat_u$T2
    expect_equal(res, expected)
  })
  
  test_that(".stageYSurvival: stops when times is not a character vector", {
    expect_error(
      .stageYSurvival(1L:2L, make_steps(), cc_u, dat_u, type = "DTR"),
      "`times`"
    )
  })
  
  test_that(".stageYSurvival: stops when steps is not a list", {
    expect_error(
      .stageYSurvival(c("T1", "T2"), "not a list", cc_u, dat_u, type = "DTR"),
      "`steps`"
    )
  })
  
  test_that(".stageYSurvival: stops when type is not a length-1 character", {
    expect_error(
      .stageYSurvival(c("T1", "T2"), make_steps(), cc_u, dat_u, type = 1L),
      "`type`"
    )
  })
  
  test_that(".stageRegret: returns list of length K", {
    res <- .stageRegret(make_steps(), cc_u, dat_u, type = "DTR")
    expect_length(res, 2L)
  })
  
  test_that(".stageRegret: each element is numeric vector", {
    res <- .stageRegret(make_steps(), cc_u, dat_u, type = "DTR")
    expect_true(all(sapply(res, is.numeric)))
  })
  
  test_that(".stageRegret: oracle — when opt equals observed A, all regrets are zero", {
    steps_same <- make_steps(opt1 = dat_u$A1, opt2 = dat_u$A2)
    res <- .stageRegret(steps_same, cc_u, dat_u, type = "DTR")
    expect_equal(unname(res[[1L]]), rep(0.0, n_u))
    expect_equal(unname(res[[2L]]), rep(0.0, n_u))
  })
  
  test_that(".stageRegret: regret is non-zero when opt differs from observed A", {
    steps_diff <- make_steps(opt1 = as.integer(!dat_u$A1),
                             opt2 = as.integer(!dat_u$A2))
    res <- .stageRegret(steps_diff, cc_u, dat_u, type = "DTR")
    expect_false(all(res[[1L]] == 0.0))
  })
  
  test_that(".stageRegret: stops when steps is not a list", {
    expect_error(.stageRegret("not a list", cc_u, dat_u, type = "DTR"), "`steps`")
  })
  
  test_that(".stageRegret: stops when complete.case.info missing required names", {
    bad_cc <- list(last.stage = cc_u$last.stage)
    expect_error(.stageRegret(make_steps(), bad_cc, dat_u, type = "DTR"),
                 "`complete.case.info`")
  })
  
  test_that(".stageRegret: stops when data is not a data.frame", {
    expect_error(
      .stageRegret(make_steps(), cc_u, as.matrix(dat_u), type = "DTR"),
      "`data`"
    )
  })
  
  test_that(".stageRegret: stops when type is not a character", {
    expect_error(.stageRegret(make_steps(), cc_u, dat_u, type = 1L), "`type`")
  })
  
}) # end local

#' Test to see if user requested to exit interactive mode
#'
#' @noRd
#' @param x A character object.
#' @returns The unmodified input
#' @keywords internal
.test_quit <- function(x) {
  if (tolower(x) == "quit") stop("interactive mode cancelled", call. = FALSE)
  x
}

.getVariables <- function(x, data) {
  if (is.null(data)) {
    x <- tryCatch(get(x),
                  error = function(e) {
                    stop("unable to retreive ", x, " from environment; ",
                         "`data` must be provided as input\n\t",
                         e$message, call. = FALSE)
                  })
  } else {
    x <- tryCatch(data[, x],
                  error = function(e) {
                    stop("unable to retreive ", x, " from `data`\n\t",
                         e$message, call. = FALSE)
                  })
  }
  x
}

#' Function obtains required inputs from user through prompted interactions
#'
#' @noRd
#' @param isSurvival A logical. If TRUE messaging changes a bit.
#' @param prompt.cens A logical. If TRUE prompt for missingness/censoring model.
#'
#' @return A list containing the following input information
#' \itemize{
#' \item{models }{A list object. The ith element contains the models for the ith
#'   decision point. Models are "blip", "tf", and "treat"}
#' \item{method }{A character object. The DTR method to be used, one of 
#'   "dwols", "gest", or "qlearn"}
#' \item{var.estim }{A character object. Covariance matrix estimation method, 
#'   one of "none", "bootstrap", or "sandwich".}
#' \item{outcome }{A character object. The outcome of interest.}
#' }
#' @keywords internal
.interactive <- function(isSurvival = FALSE, prompt.cens = TRUE, data = NULL) {
  
  cat("DTR estimation interactive mode!\n\nEnter 'quit' at anytime to cancel.\n\n")
  cat("Note: Do not use quotation marks.\n")
  
  if (isSurvival) {
    has_censored_data <- readline("Is the data censored? (y/n) ") |> .test_quit()
    prompt.cens <- tolower(has_censored_data) == "y"
    if (prompt.cens) {
      status_var <- readline("What is the status variable? ")
    } else {
      status_var <- NA
    }
  } else {
    # prompt.cens is set by calling function
    # if `missing` == "ipw", prompt.cens = TRUE and user will be
    # prompted for a model
    outcome <- readline("Enter outcome variable: ") |> .test_quit()
    outcome <- .getVariables(outcome, data)
    status_var <- NA
    
    if (prompt.cens) {
      # if `missing` == "ipw", allow user to take default models
      message("You have set `missing` = 'ipw'.")
      use_default_missingness <- readline("Use default missingness models? (y/n) ")
      prompt.cens <- tolower(use_default_missingness) == "n"
    }
  }
  
  while (TRUE) {
    K <- readline("Enter number of stages: ") |> .test_quit() |> as.integer()
    if (K > 0L) break
    cat("K must be a positive integer\n")
  }
  
  model_type <- c("blip model", "treatment-free model", "treatment model",
                  ifelse(isSurvival, "censoring model", "missingness model"))
  models <- vector(mode = "list", length = K)
  model_names <- c("blip", "tf", "treat", "cens")

  time <- list()
  tx.var <- character(K)
  
  for (j in 1:K) {

    models[[j]] <- list()

    if (isSurvival) {
      out_time <- readline(paste("Enter stage", j, "time variable: ")) |> 
        .test_quit()
      time[[j]] <- paste("~", out_time) |> as.formula()
    }
    
    out_treat <- readline(paste("Enter stage", j, "treatment variable: ")) |> 
      .test_quit()
    tx.var[j] <- out_treat
    
    for (m in seq_along(model_names)) {

      if (model_names[m] == "treat") {
        # only the treatment model requires a dependent variable
        out <- tx.var[j]
      } else if (model_names[m] == "cens") {
        # if survival, `cens.mod` model has dependent variable only if 
        # censored is to be modeled. if not survival, and missingness is
        # to be modeled, there is no dependent variable
        if (!prompt.cens) next
        out <- ifelse(isSurvival, status_var, "")
      } else {
        out <- ""
      }
      
      while (TRUE) {
        new.var <- readline(prompt = paste("Enter all stage", j, model_type[m],
                                           "covariates separated by a comma;",
                                           "press RETURN when finished: "))
        .test_quit(new.var)
        new.var <- strsplit(new.var, ",") |> unlist() |> trimws()
        new.var <- new.var[nchar(new.var) > 0L]
        if (length(new.var) == 0L) new.var <- "1"
        form <-  paste(out, " ~ ", paste(new.var, collapse = " + "))
        cat(model_type[m], ": ", form, "\n")
        move_on <- readline(prompt = "Is this correct (y/n)? ")
        if (tolower(move_on) == "y") break
      }
      
      models[[j]][[model_names[m]]] <- form |> trimws() |> as.formula()
    }
  }
  
  while (TRUE) {
    method <- readline(paste("Dynamic WOLS (dwols), G-estimation (gest), or Q-learning (qlearn)? ")) |> 
      .test_quit()
    if (method %in% c("dwols", "gest", "qlearn")) break
    cat("method must be 'dwols', 'gest', or 'qlearn' (do not use quotation marks)\n")
  }
  
  while (TRUE) {
    var_estim <- readline("Variance estimation: none, bootstrap or sandwich? ") |> 
      .test_quit()
    if (var_estim %in% c("none", "bootstrap", "sandwich")) break
    cat("variance must be 'none', 'bootstrap', or 'sandwich' (do not use quotation marks)\n")
  }
  
  res <- list("models" = models, 
              "method" = method, 
              "var.estim" = var_estim,
              "K" = K)
  
  if (isSurvival) {
    res$time <- time
  } else {
    res$outcome <- outcome
  }
  res
}

#' @noRd
#' @param obj A list object. The analysis results and settings.
#' @param isSurvival A logical. If TRUE, survival end point.
#' 
#' @include dtrProcedure.R utils.R
#' @keywords internal
.resampleObj <- function(obj, boot_sample, isSurvival) {
  bobj <- obj
  if (!isSurvival) bobj$outcome <- obj$outcome[boot_sample]
  bobj$data <- obj$data[boot_sample, , drop = FALSE]
  bobj$var.estim <- "none"
  bobj$boot.controls <- list()
  if (is.list(obj$tx.wgt.man)) {
    bobj$tx.wgt.man <- lapply(obj$tx.wgt.man, "[", boot_sample)
  }
  bobj
}

#' @noRd
#' @keywords internal
.trunc <- function(x,p) {
  l <- quantile(x, p)
  u <- quantile(x, 1.0 - p)
  x[x < l] <- l
  x[x > u] <- u
  x
}

#' @noRd
#' @keywords internal
.stackPsiBoot <- function(psi_boot) {
  do.call(function(...) mapply(rbind, ..., SIMPLIFY = FALSE), 
          psi_boot)
}

.truncPsiBoot <- function(truncate, psi_boot) {
  if (truncate > 0) {
    lapply(psi_boot, FUN = function(x) {
      apply(x, 2L, .trunc, truncate)
    })
  } else {
    psi_boot
  }
}

#' @noRd
#' @keywords internal
.computeCovmat <- function(psi_boot) {
  lapply(psi_boot, var)
}

.reportProgress <- function(ptm, i, B, cont, verbose, interrupt) {
  if (!verbose || i < 10L) return(cont)
  
  eta <- {B - i} * {{proc.time() - ptm}[3L] / i}
  
  if (i > 50L && eta > 600 && cont == "y" && interrupt) {
    last_val <- attr(cont, "last")
    response <- readline(paste("Estimated run time", round(eta / 60L),
                               "minutes. Continue? y/n: "))
    if (response %in% c("n", "no", "N", "NO")) return("abort")
    cont <- "asked" 
    attr(cont, "last") <- last_val
  }
  
  if (i == 10L) attr(cont, "last") <- eta + 31.0
  last <- attr(cont, "last") %||% Inf
  
  if (eta > 30.0 && eta < {last - 30.0}) {
    message("Approximately ", round(eta), " seconds remaining.")
    attr(cont, "last") <- eta
  }
  
  cont
}

#' @noRd
#' @param obj A list object. The analysis results and settings.
#' @param isSurvival A logical. If TRUE, survival end point.
#' 
#' @include dtrProcedure.R utils.R
#' @keywords internal
.bootstrap <- function(obj, isSurvival) {
  
  n_cases <- nrow(obj$data)
  
  # default # of samples in each bootstrap iteration is the number of participants
  if (is.null(obj$boot.controls$M) || obj$boot.controls$M == 0L) {
    obj$boot.controls$M <- n_cases
  }
  
  # note time for progress reports
  ptm <- proc.time()
  psi_boot <- list()
  
  # continue indicator
  last <- Inf
  cont <- "y"
  
  bobj <- obj

  for (i in seq_len(obj$boot.controls$B)) {
    boot_sample <- sample(seq_len(n_cases), obj$boot.controls$M, replace = TRUE)
    bobj <- .resampleObj(obj, boot_sample, isSurvival)

    result <- .dtrProcedure(obj = bobj, quiet = TRUE, isSurvival)
    psi_boot[[i]] <- lapply(result$stages, "[[", "psi")

    cont <- .reportProgress(ptm, i, obj$boot.controls$B, cont,
                            obj$boot.controls$verbose, obj$boot.controls$interrupt)
    if (identical(cont, "abort")) {
      message("bootstrap aborted")
      return(list("covmat" = "aborted", "psi.boot" = "aborted"))
    }
  }
  
  psi_boot <- .stackPsiBoot(psi_boot)
  
  if (any(sapply(psi_boot, function(x) any(is.na(x))))) {
    stop("psi contains NA values", call. = FALSE)
  }
  
  # truncation function for truncate option
  # if requested, truncate estimates based on specified percentile
  psi_boot <- .truncPsiBoot(obj$boot.controls$truncate, psi_boot)
  
  covmat <- .computeCovmat(psi_boot)
  
  list("covmat" = covmat, "psi.boot" = psi_boot)
}

#' @noRd
#' @keywords internal
.computeHoldY <- function(step.obj, data, isSurvival) {
  do.call(ifelse(isSurvival, exp, c),
          list(predict(step.obj$outcome.fit, data)))
}

#' @noRd
#' @keywords internal
.buildSingleStageObj <- function(obj, step.obj, tx.var, status.var, time.var, 
                                 data, tx.wgt.man, isSurvival) {
  single_stage_obj <- obj
  single_stage_obj$K <- 1L
  single_stage_obj$models <- list(step.obj$models)
  single_stage_obj$var.estim <- "none"
  single_stage_obj$dependent.vars$treat <- tx.var
  if (isSurvival) {
    single_stage_obj$dependent.vars$status <- status.var
    single_stage_obj$dependent.vars$time <- time.var
  }
  single_stage_obj$data <- data
  if (is.null(tx.wgt.man)) {
    single_stage_obj["tx.wgt.man"] <- list(NULL)
  } else {
    single_stage_obj$tx.wgt.man <- tx.wgt.man
  }
  single_stage_obj$boot.controls <- list()
  single_stage_obj
}

#' @noRd
#' @keywords internal
.sampleResiduals <- function(type, reshist, mean.res, sd.res, n, n_sample) {
  if (type == "empirical") {
    bins <- with(reshist,
                  sample(length(mids), n_sample, prob = density, replace = TRUE))
    newres <- stats::runif(n, reshist$breaks[bins], reshist$breaks[bins + 1])
  } else if (type == "normal") {
    newres <- stats::rnorm(n, mean = mean.res, sd = sd.res)
  }
  newres
}

#' Non-parametric Bootstrap Algorithm for Survival Data
#' 
#' @noRd
#' @param step.obj A list. The step specific models and parameter estimates
#'   as defined at the end of each stage analysis in .step1Surv(). Must contain
#'   psi, beta, A, n, D.hat, models, optimization, K, time.vars, 
#'   status.vars, treat.vars, data, type, drop)
#' @param obj A list. The analysis controls.
#' @param residuals A numeric vector. The residuals of the current model.
#' @param status.var A character object. The column header of `data` containing
#'   the 0/1 status indicator.
#' @param time.var A character object. The column header of `data` containing 
#'   the current time variable.
#' @param tx.wgt.man NULL or a vector object. User specified treatment weights
#'   for complete cases.
#' @param data A data.frame. The full data for complete cases.
#' @param isSurvival A logical. If TRUE, survival analysis.
#' @param d.hat A numeric vector. The estimated event probability.
#' 
#' @importFrom graphics hist
#' @importFrom stats rbinom rnorm runif sd var
.bootstrap_nonpar <- function(step.obj, obj, residuals, status.var, tx.var, time.var,
                              tx.wgt.man, data, isSurvival, d.hat) {
  
  # Matrix to hold psi estimates for each bootstrap sample
  psi_boot <- matrix(NA, ncol = obj$boot.controls$B, nrow = length(step.obj$psi))

  # To avoid repeated calculations of the outcome
  # Ensure that data contains the observed treatment
  data[[step.obj$tx.var]] <- step.obj$A

  data <- step.obj$cts.obj$prep(data, step.obj$A)
  
  hold_Y <- .computeHoldY(step.obj, data, isSurvival)
  
  # mean and sd of residuals; used to sample N(mean, sd) if bootstrap option is 
  # "normal"
  mean.res <- mean(residuals)
  sd.res <- stats::sd(residuals)
  
  # binned residuals; used to sample U() if bootstrap option is "empirical"
  reshist <- graphics::hist(residuals, plot = FALSE)
  
  # create single step version of main object
  single_stage_obj <- .buildSingleStageObj(obj, step.obj, tx.var, status.var,
                                           time.var, data, tx.wgt.man, isSurvival)
  
  for (k in seq_len(obj$boot.controls$B)) {
    # generate new status variable based on fitted probabilities
    if (isSurvival) {
      single_stage_obj$data[, status.var] <- stats::rbinom(step.obj$n, 1, d.hat)
    }
    
    n_sample <- if (isSurvival) sum(single_stage_obj$data[, status.var]) else step.obj$n
    
    newres <- .sampleResiduals(obj$boot.controls$type, reshist,
                               mean.res, sd.res, step.obj$n, n_sample)
    
    if (isSurvival) {
      single_stage_obj$data[, time.var] <- hold_Y*exp(newres)
    } else {
      single_stage_obj$outcome <- hold_Y + newres
    }

    psi_boot[, k] <- .dtrProcedure(single_stage_obj, quiet = TRUE, isSurvival)$stages[[1L]]$psi
  }
  
  psi_boot <- t(psi_boot)
  list("covmat" = stats::var(psi_boot), "psi.boot" = psi_boot)
}

# Shared minimal data


data(twoStageCont)
n <- nrow(twoStageCont)

withr::with_seed(42L, {
  n_s <- 30L
  dat_s <- data.frame(
    X = cos(seq(0.0, pi, length.out = n_s)),
    A = as.numeric(rbinom(n_s, 1L, 0.5)),
    Y = rnorm(n_s),
    T = abs(rnorm(n_s)) + 0.5
  )
})

# Trusted Layer 0 / Layer 4 objects
cts_b <- Binary$new(tf.model = ~X, blip.model = ~X, tx.var = "A")
fit_dw <- .dwols(Y = dat_s$Y, A = dat_s$A, data = dat_s,
                 wgts = rep(1.0, n_s), cts.obj = cts_b, tx.var = "A")

# Minimal step.obj as .getStage would produce
step_obj_s <- list(
  n = n_s,
  tx.var = "A",
  A = dat_s$A,
  psi = fit_dw$psi,
  cts = "bin",
  cts.obj = cts_b,
  outcome.fit = fit_dw$outcome.fit,
  models = list(blip = ~X, treat = A ~ X, tf = ~X, cens = ~1)
)

# Minimal obj for single-stage
make_ss_obj <- function(isSurvival = FALSE) {
  obj <- list(
    K = 1L,
    outcome = dat_s$Y,
    data = dat_s,
    models = list(list(blip = ~X, treat = A ~ X, tf = ~X, cens = ~1)),
    dependent.vars = list(treat = "A",
                          status = if (isSurvival) "delta" else NA_character_,
                          time = if (isSurvival) "T" else NA_character_),
    tx.range = list(NA_real_),
    method = "dwols",
    var.estim = "none",
    type = "DTR",
    tx.weight = "ipw",
    censoring.modeled = FALSE,
    tx.wgt.man = NA,
    tx.type = "bin",
    n.bins = NA_integer_,
    tx.family = NA,
    boot.controls = list(B = 5L, M = n_s, type = "standard",
                         truncate = 0.0, verbose = FALSE, interrupt = FALSE),
    full.cov = FALSE
  )
  if (isSurvival) {
    obj$manual.censor.weight <- FALSE
    obj$outcome = NULL
  }
  obj
}

# Full obj for .bootstrap() smoke tests — built from trusted .dtrProcedure
make_boot_obj <- function(B = 5L,
                          truncate = 0.0,
                          method = "dwols") {
  obj <- list(
    K = 2L,
    outcome = twoStageCont$Y,
    data = twoStageCont,
    models = list(
      list(blip = ~X1, treat = A1 ~ X1, tf = ~X1, cens = ~1),
      list(blip = ~X2, treat = A2 ~ X2, tf = ~X2, cens = ~X1)
    ),
    dependent.vars = list(treat = c("A1", "A2")),
    tx.range = list(NA_real_, NA_real_),
    method = method,
    var.estim = "bootstrap",
    type = "DTR",
    tx.weight = "ipw",
    censoring.modeled = FALSE,
    tx.wgt.man = NA,
    tx.type = "bin",
    n.bins = NA_integer_,
    tx.family = NA,
    full.cov = FALSE,
    boot.controls = list(B = B, M = n, type = "standard",
                         truncate = truncate, verbose = FALSE,
                         interrupt = FALSE)
  )
  result <- suppressMessages(.dtrProcedure(obj, quiet = TRUE))
  obj$stages <- result$stages
  obj$last.stage <- result$last.stage
  obj$prob.complete.case <- result$prob.complete.case
  obj$d.hat <- result$d.hat
  obj$cens.mod.fitted <- result$cens.mod.fitted
  obj$opt.Y <- result$opt.Y
  obj$treat.vars <- result$treat.vars
  obj$regret <- result$regret
  obj
}



# .resampleObj


local({
  boot_sample <- sample(seq_len(n_s), n_s, replace = TRUE)
  obj <- make_ss_obj()
  
  test_that(".resampleObj: data is subsampled to boot_sample rows", {
    res <- .resampleObj(obj, boot_sample, isSurvival = FALSE)
    expect_equal(res$data, obj$data[boot_sample, , drop = FALSE])
  })
  
  test_that(".resampleObj: outcome is subsampled for non-survival", {
    res <- .resampleObj(obj, boot_sample, isSurvival = FALSE)
    expect_equal(res$outcome, obj$outcome[boot_sample])
  })
  
  test_that(".resampleObj: outcome not modified for survival", {
    obj_surv <- make_ss_obj(isSurvival = TRUE)
    res <- .resampleObj(obj_surv, boot_sample, isSurvival = TRUE)
    expect_null(res$outcome)
  })
  
  test_that(".resampleObj: var.estim set to 'none'", {
    obj$var.estim <- "bootstrap"
    res <- .resampleObj(obj, boot_sample, isSurvival = FALSE)
    expect_equal(res$var.estim, "none")
  })
  
  test_that(".resampleObj: boot.controls reset to empty list", {
    res <- .resampleObj(obj, boot_sample, isSurvival = FALSE)
    expect_identical(res$boot.controls, list())
  })
  
  test_that(".resampleObj: list tx.wgt.man is subsampled per stage", {
    obj$tx.wgt.man <- list(seq_len(n_s) / n_s, rev(seq_len(n_s) / n_s))
    res <- .resampleObj(obj, boot_sample, isSurvival = FALSE)
    expect_equal(res$tx.wgt.man[[1L]], obj$tx.wgt.man[[1L]][boot_sample])
    expect_equal(res$tx.wgt.man[[2L]], obj$tx.wgt.man[[2L]][boot_sample])
  })
  
  test_that(".resampleObj: non-list tx.wgt.man is left unchanged", {
    res <- .resampleObj(obj, boot_sample, isSurvival = FALSE)
    expect_identical(res$tx.wgt.man, obj$tx.wgt.man)
  })
})



# .trunc


test_that(".trunc: oracle — values clamped to [quantile(x,p), quantile(x,1-p)]", {
  x <- c(-3.0, -1.0, 0.0, 1.0, 3.0, 5.0, 10.0)
  p <- 0.1
  expected <- pmax(pmin(x, quantile(x, 1.0 - p)), quantile(x, p))
  expect_equal(.trunc(x, p), expected)
})

test_that(".trunc: p=0 returns x unchanged", {
  x <- rnorm(20L)
  expect_equal(.trunc(x, 0.0), x)
})

test_that(".trunc: output length equals input length", {
  x <- rnorm(20L)
  expect_length(.trunc(x, 0.1), length(x))
})

test_that(".trunc: max of truncated <= max of original", {
  x <- rnorm(50L)
  expect_true(max(.trunc(x, 0.1)) <= max(x))
})

test_that(".trunc: min of truncated >= min of original", {
  x <- rnorm(50L)
  expect_true(min(.trunc(x, 0.1)) >= min(x))
})



# .stackPsiBoot


local({
  # Hand-crafted: B=3 iterations, K=2 stages, 2 psi per stage
  psi_iter <- function(seed) {
    withr::with_seed(seed,
    list(c(A1 = rnorm(1L), `A1:X` = rnorm(1L)),
         c(A2 = rnorm(1L), `A2:X` = rnorm(1L))))
  }
  psi_boot <- list(psi_iter(1L), psi_iter(2L), psi_iter(3L))
  res <- .stackPsiBoot(psi_boot)
  
  test_that(".stackPsiBoot: returns list of length K", {
    expect_length(res, 2L)
  })
  
  test_that(".stackPsiBoot: each element is a matrix with B rows", {
    expect_equal(nrow(res[[1L]]), 3L)
    expect_equal(nrow(res[[2L]]), 3L)
  })
  
  test_that(".stackPsiBoot: each element has n_psi columns", {
    expect_equal(ncol(res[[1L]]), 2L)
    expect_equal(ncol(res[[2L]]), 2L)
  })
  
  test_that(".stackPsiBoot: oracle — row i of stage k matches psi_boot[[i]][[k]]", {
    for (i in seq_along(psi_boot)) {
      expect_equal(unname(res[[1L]][i, ]), unname(psi_boot[[i]][[1L]]))
      expect_equal(unname(res[[2L]][i, ]), unname(psi_boot[[i]][[2L]]))
    }
  })
})



# .truncPsiBoot


local({
  # Known psi_boot — 10 rows (B), 2 cols (psi) per stage
  withr::with_seed(42L,
  pb <- list(
    matrix(rnorm(20L), nrow = 10L, ncol = 2L),
    matrix(rnorm(20L), nrow = 10L, ncol = 2L)
  ))
  
  test_that(".truncPsiBoot: truncate=0 returns psi_boot unchanged", {
    expect_identical(.truncPsiBoot(0.0, pb), pb)
  })
  
  test_that(".truncPsiBoot: truncate>0 narrows range of each column", {
    res <- .truncPsiBoot(0.1, pb)
    for (k in seq_along(pb)) {
      for (j in seq_len(ncol(pb[[k]]))) {
        expect_true(diff(range(res[[k]][, j])) <= diff(range(pb[[k]][, j])))
      }
    }
  })
  
  test_that(".truncPsiBoot: oracle — matches column-wise .trunc application", {
    p <- 0.1
    res <- .truncPsiBoot(p, pb)
    for (k in seq_along(pb)) {
      expected <- apply(pb[[k]], 2L, .trunc, p)
      expect_equal(res[[k]], expected)
    }
  })
  
  test_that(".truncPsiBoot: dimensions preserved after truncation", {
    res <- .truncPsiBoot(0.1, pb)
    for (k in seq_along(pb)) {
      expect_equal(dim(res[[k]]), dim(pb[[k]]))
    }
  })
})



# .computeCovmat


local({
  withr::with_seed(42L,
  pb <- list(
    matrix(rnorm(20L), nrow = 10L, ncol = 2L),
    matrix(rnorm(20L), nrow = 10L, ncol = 2L)
  ))
  
  test_that(".computeCovmat: returns list of length K", {
    res <- .computeCovmat(pb)
    expect_length(res, 2L)
  })
  
  test_that(".computeCovmat: each element is a square matrix", {
    res <- .computeCovmat(pb)
    for (k in seq_along(pb)) {
      expect_true(is.matrix(res[[k]]))
      expect_equal(nrow(res[[k]]), ncol(res[[k]]))
    }
  })
  
  test_that(".computeCovmat: oracle — matches lapply(psi_boot, var)", {
    expect_equal(.computeCovmat(pb), lapply(pb, var))
  })
})



# .computeHoldY


local({
  dat_pred <- cts_b$prep(dat_s, dat_s$A)
  
  test_that(".computeHoldY: non-survival oracle — equals predict(outcome.fit, data)", {
    expected <- drop(predict(fit_dw$outcome.fit, dat_pred))
    expect_equal(.computeHoldY(step_obj_s, dat_pred, isSurvival = FALSE),
                 expected)
  })
  
  test_that(".computeHoldY: survival oracle — equals exp(predict(outcome.fit, data))", {
    expected <- exp(drop(predict(fit_dw$outcome.fit, dat_pred)))
    expect_equal(.computeHoldY(step_obj_s, dat_pred, isSurvival = TRUE),
                 expected)
  })
  
  test_that(".computeHoldY: returns numeric vector of length n", {
    res <- .computeHoldY(step_obj_s, dat_pred, isSurvival = FALSE)
    expect_true(is.numeric(res))
    expect_length(res, n_s)
  })
  
  test_that(".computeHoldY: survival values are strictly positive", {
    res <- .computeHoldY(step_obj_s, dat_pred, isSurvival = TRUE)
    expect_true(all(res > 0.0))
  })
})



# .buildSingleStageObj


local({
  obj <- make_ss_obj()
  dat_ss <- dat_s
  res_ns <- .buildSingleStageObj(obj, step_obj_s, "A", NA_character_,
                                 NA_character_, dat_ss, NULL, isSurvival = FALSE)
  
  test_that(".buildSingleStageObj: K is set to 1", {
    expect_equal(res_ns$K, 1L)
  })
  
  test_that(".buildSingleStageObj: var.estim is 'none'", {
    expect_equal(res_ns$var.estim, "none")
  })
  
  test_that(".buildSingleStageObj: boot.controls is empty list", {
    expect_identical(res_ns$boot.controls, list())
  })
  
  test_that(".buildSingleStageObj: models is list of one element", {
    expect_length(res_ns$models, 1L)
  })
  
  test_that(".buildSingleStageObj: dependent.vars$treat is tx.var", {
    expect_equal(res_ns$dependent.vars$treat, "A")
  })
  
  test_that(".buildSingleStageObj: data is set to provided data", {
    expect_identical(res_ns$data, dat_ss)
  })
  
  test_that(".buildSingleStageObj: NULL tx.wgt.man stored as NULL list element", {
    expect_null(res_ns$tx.wgt.man[[1L]])
  })
  
  test_that(".buildSingleStageObj: non-NULL tx.wgt.man stored directly", {
    wgt <- runif(n_s)
    res <- .buildSingleStageObj(obj, step_obj_s, "A", NA_character_,
                                NA_character_, dat_ss, wgt, isSurvival = FALSE)
    expect_equal(res$tx.wgt.man, wgt)
  })
  
  test_that(".buildSingleStageObj: survival sets dependent.vars$status and $time", {
    obj_surv <- make_ss_obj(isSurvival = TRUE)
    res_surv <- .buildSingleStageObj(obj_surv, step_obj_s, "A", "delta",
                                     "T", dat_ss, NULL, isSurvival = TRUE)
    expect_equal(res_surv$dependent.vars$status, "delta")
    expect_equal(res_surv$dependent.vars$time, "T")
  })
  
  test_that(".buildSingleStageObj: non-survival does not overwrite status/time", {
    expect_equal(res_ns$dependent.vars$status, NA_character_)
    expect_equal(res_ns$dependent.vars$time, NA_character_)
  })
})



# .sampleResiduals


local({
  residuals <- rnorm(n_s)
  reshist <- hist(residuals, plot = FALSE)
  mean.res <- mean(residuals)
  sd.res <- sd(residuals)
  
  test_that(".sampleResiduals: normal returns numeric vector of length n", {
    res <- .sampleResiduals("normal", reshist, mean.res, sd.res,
                            n = n_s, n_sample = n_s)
    expect_true(is.numeric(res))
    expect_length(res, n_s)
  })
  
  test_that(".sampleResiduals: empirical returns numeric vector of length n", {
    res <- .sampleResiduals("empirical", reshist, mean.res, sd.res,
                            n = n_s, n_sample = n_s)
    expect_true(is.numeric(res))
    expect_length(res, n_s)
  })
  
  test_that(".sampleResiduals: empirical values fall within histogram breaks range", {
    res <- .sampleResiduals("empirical", reshist, mean.res, sd.res,
                            n = n_s, n_sample = n_s)
    expect_true(all(res >= min(reshist$breaks)))
    expect_true(all(res <= max(reshist$breaks)))
  })
  
  test_that(".sampleResiduals: normal produces different results with different seeds", {
    r1 <- withr::with_seed(1L, .sampleResiduals("normal", reshist, mean.res, sd.res, n_s, n_s))
    r2 <- withr::with_seed(2L, .sampleResiduals("normal", reshist, mean.res, sd.res, n_s, n_s))
    expect_false(isTRUE(all.equal(r1, r2)))
  })
})



# .reportProgress
#
# readline() and proc.time() are untestable directly. We test all logic
# branches that don't depend on those — early returns, last initialisation,
# message suppression when eta is small, and the attr(last) update path.
# We mock ptm by setting it far in the past to force large eta values.


local({
  # ptm set 1000 seconds in the past so eta is always large
  ptm_past <- proc.time() - c(0, 0, 1000, 0, 0)
  
  test_that(".reportProgress: returns cont unchanged when verbose=FALSE", {
    res <- .reportProgress(proc.time(), i = 1L, B = 100L, cont = "y",
                           verbose = FALSE, interrupt = FALSE)
    expect_equal(res, "y")
  })
  
  test_that(".reportProgress: returns cont unchanged when i < 10", {
    res <- .reportProgress(proc.time(), i = 5L, B = 100L, cont = "y",
                           verbose = TRUE, interrupt = FALSE)
    expect_equal(res, "y")
  })
  
  test_that(".reportProgress: initialises last attribute when i == 10", {
    res <- .reportProgress(ptm_past, i = 10L, B = 100L, cont = "y",
                           verbose = TRUE, interrupt = FALSE)
    expect_false(is.null(attr(res, "last")))
  })
  
  test_that(".reportProgress: last attribute > 30 when set at i=10", {
    res <- .reportProgress(ptm_past, i = 10L, B = 100L, cont = "y",
                           verbose = TRUE, interrupt = FALSE)
    expect_true(attr(res, "last") > 30.0)
  })
  
  test_that(".reportProgress: issues message when eta > 30 and within last threshold", {
    cont_msg <- "y"
    attr(cont_msg, "last") <- 1e8
    expect_message(
      .reportProgress(ptm_past, i = 15L, B = 100L, cont = cont_msg,
                      verbose = TRUE, interrupt = FALSE),
      "seconds remaining"
    )
  })
  
  test_that(".reportProgress: no message when eta < 30", {
    expect_no_message(
      .reportProgress(proc.time(), i = 15L, B = 100L, cont = "y",
                      verbose = TRUE, interrupt = FALSE)
    )
  })
  
  test_that(".reportProgress: updates last attribute after issuing message", {
    cont_msg <- "y"
    attr(cont_msg, "last") <- 1e8
    res <- suppressMessages(
      .reportProgress(ptm_past, i = 15L, B = 100L, cont = cont_msg,
                      verbose = TRUE, interrupt = FALSE)
    )
    expect_true(attr(res, "last") < 1e8)
  })
})



# .bootstrap — integration smoke test


local({
  withr::with_seed(42L, {
    obj <- make_boot_obj(B = 5L)
    res <- suppressMessages(.bootstrap(obj, isSurvival = FALSE))
  })
  
  test_that(".bootstrap: returns list with covmat and psi.boot", {
    expect_named(res, c("covmat", "psi.boot"), ignore.order = TRUE)
  })
  
  test_that(".bootstrap: covmat is list of length K", {
    expect_true(is.list(res$covmat))
    expect_length(res$covmat, obj$K)
  })
  
  test_that(".bootstrap: each covmat element is a square matrix", {
    for (k in seq_len(obj$K)) {
      expect_true(is.matrix(res$covmat[[k]]))
      expect_equal(nrow(res$covmat[[k]]), ncol(res$covmat[[k]]))
    }
  })
  
  test_that(".bootstrap: covmat dimensions match psi length at each stage", {
    for (k in seq_len(obj$K)) {
      n_psi <- length(obj$stages[[k]]$psi)
      expect_equal(dim(res$covmat[[k]]), c(n_psi, n_psi))
    }
  })
  
  test_that(".bootstrap: psi.boot is list of length K", {
    expect_length(res$psi.boot, obj$K)
  })
  
  test_that(".bootstrap: each psi.boot element is a matrix with B rows", {
    for (k in seq_len(obj$K)) {
      expect_true(is.matrix(res$psi.boot[[k]]))
      expect_equal(nrow(res$psi.boot[[k]]), obj$boot.controls$B)
    }
  })
  
  test_that(".bootstrap: covmat oracle — matches var(psi.boot) per stage", {
    for (k in seq_len(obj$K)) {
      expect_equal(res$covmat[[k]], var(res$psi.boot[[k]]))
    }
  })
})

# --- truncation --------------------------------------------------------------

test_that(".bootstrap: truncation narrows psi.boot range vs no truncation", {
  withr::with_seed(42L,
                   res_plain <- suppressMessages(
                     .bootstrap(make_boot_obj(B = 20L, truncate = 0.0), isSurvival = FALSE)
                   )
  )
  withr::with_seed(42L,
                   res_trunc <- suppressMessages(
                     .bootstrap(make_boot_obj(B = 20L, truncate = 0.1), isSurvival = FALSE)
                   )
  )
  for (k in seq_len(2L)) {
    range_plain <- apply(res_plain$psi.boot[[k]], 2L, function(x) diff(range(x)))
    range_trunc <- apply(res_trunc$psi.boot[[k]], 2L, function(x) diff(range(x)))
    expect_true(all(range_trunc <= range_plain))
  }
})

# --- manual weights ----------------------------------------------------------

test_that(".bootstrap: manual tx.wgt.man is handled without error", {
  withr::with_seed(42L, {
    obj_man <- make_boot_obj(B = 5L)
    obj_man$tx.weight <- "manual"
    obj_man$tx.wgt.man <- list(rep(1.0, n), rep(1.0, n))
    expect_no_error(suppressMessages(.bootstrap(obj_man, isSurvival = FALSE)))
  })
})

#' Censoring or Complete Case Probabilities
#' 
#' @noRd
#' @param model A formula object. The censoring or complete case model.
#' @param delta An integer object. Indicator of censored or complete case.
#' @param data A data.frame object. All covariate and treatments.
#' @param censoring.modeled A logical object. If TRUE, censoring is modeled.
#' @param dp An integer. The stage under consideration.
#' @param quiet A logical. If TRUE messaging is suppressed.
#' 
#' @return A list containing the fitted model `$cens.model.fitted` (if estimated),
#'   censoring weight `$wt.cen`, and estimated probabilities `$D.hat`.
#'   
#' @importFrom stats fitted glm update.formula
#' @keywords internal
.processCensoring <- function(model, delta, data, censoring.modeled, dp, quiet) {
  res <- list()
  
  res$cens.model.fitted <- NA

  # censoring model: binary logistic
  if (any(delta == 0L) && censoring.modeled) {
    model <- stats::update.formula(l__delta__l ~ ., model)
    res$cens.model.fitted <- stats::glm(model, cbind(data, "l__delta__l" = delta), 
                                        family = "binomial")
    d.hat <- stats::fitted(res$cens.model.fitted)
  } else if (all(delta == 1L) && censoring.modeled) {
    
    d.hat <- delta
    if (!quiet) {
      message("no censoring in stage", dp, ", censoring model will be ignored.")
    }
    
  } else {
    d.hat <- delta
  }
  res$D.hat <- d.hat
  res$wt.cen <- 1.0 / {delta * d.hat + {1.0 - delta} * {1.0 - d.hat}}

  res
  
}

.initStillIn <- function(obj, isSurvival) {
  still_in <- rep(TRUE, nrow(obj$data))
  
  # if complete cases or censoring is not modeled, we need to remove all data 
  # for cases with missing data. Do not want to require complete data when 
  # censoring is provided through the manual treatments
  if ({!obj$censoring.modeled && !isSurvival} ||
      {!obj$censoring.modeled && isSurvival && !obj$manual.censor.weight}) {
    still_in <- complete.cases(obj$data)
  }
  still_in
}

.dependentVarsCompleteCases <- function(obj, K) {
  # Pull treatment and possibly status and time variables
  dependent_vars <- list()
  cnt <- 1L
  for (dp in c("treat", "time", "status")) {
    if (is.null(obj$dependent.vars[[dp]])) next
    dep_vars <- obj$dependent.vars[[dp]]
    for (k in seq_len(min(K, length(dep_vars)))) {
      dependent_vars[[cnt]] <- tryCatch(obj$data[, dep_vars[k]],
                                        error = function(e) {
                                          stop("unable to extract variable ", 
                                               dep_vars[k], "\n\t", e$message, 
                                               call. = FALSE)
                                        })
      cnt <- cnt + 1L
    }
  }
  dependent_vars <- do.call(what = cbind, dependent_vars)
  # identify the individuals with complete data
  complete.cases(dependent_vars)
}

.modelVarsCompleteCases <- function(obj, K) {
  lapply(obj$models[[K]], .getModelMatrix, data = obj$data) |> complete.cases()
}

.completeCases <- function(obj, K) {
  dependent_vars_complete_cases <- .dependentVarsCompleteCases(obj, K)
  
  model_matrices_complete_cases <- .modelVarsCompleteCases(obj, K)
  
  # indicator is 1 if all stage covariates are present
  model_matrices_complete_cases & dependent_vars_complete_cases
  
}

.getDelta <- function(still.in, complete.cases, status) {
  # delta is an indicator of length # participants still under analysis
  # for non-survival analyses, it is 0 is not a complete case but still in
  # for survival, it is the status for only the complete cases still in
  if (is.null(status)) {
    delta <- rep(1L, length(still.in))
    delta[!complete.cases & still.in] <- 0L
    delta <- delta[still.in]
  } else {
    still_in <- still.in & complete.cases
    delta <- status[still_in]
  }
  delta
}

.updateCensorMatrices <- function(matrix.list, still.in, cc, k, cen.fit) {
  # for the participants that remain in the cohort (no missing data yet)
  # change last_stage to indicate their inclusion in this stage
  matrix.list$last.stage[cc] <- k
  
  matrix.list$prob.complete.case[still.in, k] <- cen.fit$wt.cen
  matrix.list$d.hat[still.in, k] <- cen.fit$D.hat
  matrix.list$cens.mod.fitted[[k]] <- cen.fit$cens.model.fitted
  
  # for those that are lost at or before this stage, set as NA
  matrix.list$prob.complete.case[!cc, k] <- NA_real_
  matrix.list
}

#' Calculate weights for censoring or complete cases for all decision points.
#'
#' @noRd
#' @param obj A list object. Analysis settings. Must include
#'   K: An integer object. The number of decision points (dp); 
#'   outcome: A numeric vector object. The outcome of interest;
#'   models: A list of length K, each element a list of dp specific models; 
#'   data: A data.frame object. The full covariate and treatment data; 
#'   dependent.vars: A list of dependent variables;
#'   censoring.modeled: A logical. If TRUE, element `$models` must include
#'     censoring models
#'   manual.censor.weight: A logical. If TRUE, censoring weights were provided
#'     by the user through treat.wgt.man.
#' @param quiet A logical. If TRUE messaging is suppressed.
#'  
#' @return A list object containing `$last.stage`, an integer vector of length
#'   n indicating the last stage for which each participant had complete data;
#'   `$prob.complete.case`, an n x K matrix of inverse weights; and
#'   `$cens.mod.fitted` a list of `glm` objects and/or NA_character_ (indicating
#'   that a fit was not performed for the stage).
#'   
#' @importFrom stats complete.cases
#' @include utils.R
.completeCaseProbability <- function(obj, quiet, isSurvival) {
  rqrd_elements <- c("K", "models", "data", "censoring.modeled",
                     "dependent.vars")
  if (isSurvival) rqrd_elements <- c(rqrd_elements, "manual.censor.weight")
  
  stopifnot(
    "`obj` does not contain the required information" = 
      is.list(obj) && all(rqrd_elements %in% names(obj))
  )

  n_cases <- nrow(obj$data)
  
  # this last_stage vector will indicate the last stage at which each participant
  # has complete data
  # the prob_complete matrix will contain the inverse probabilities
  # for each decision point. "missing" cases will be NA
  result_list <- list(
    last.stage = integer(n_cases),
    prob.complete.case = matrix(NA_real_, nrow = n_cases, ncol = obj$K),
    d.hat = matrix(NA_real_, nrow = n_cases, ncol = obj$K),
    cens.mod.fitted = list()
  )
  
  # temporary vector to track cases that are still in the analysis at the
  # stage under consideration
  still_in <- .initStillIn(obj, isSurvival)

  # NOTE this is in the FORWARD direction
  # This ensures that once someone has missing data that are removed from 
  # all subsequent stages
  for (k in 1L:obj$K) {
    # indicator is 1 if all stage covariates are present and the participant is
    # still in the cohort
    cc <- .completeCases(obj, k) & still_in

    # delta is an indicator of length # participants still under analysis
    # for non-survival analyses, it is 0 is not a complete case but still in
    # for survival, it is the status for only the complete cases still in
    if (!is.null(obj$dependent.vars$status))
      status_tmp <- obj$data[[obj$dependent.vars$status]]
    else
      status_tmp <- NULL
    delta <- .getDelta(still_in, cc, status_tmp)
    
    if (isSurvival) still_in <- still_in & cc

    cen_fit <- .processCensoring(obj$models[[k]]$cens, 
                                 delta = delta, 
                                 data = obj$data[still_in, , drop = FALSE], 
                                 censoring.modeled = obj$censoring.modeled, 
                                 dp = k,
                                 quiet = quiet)
    
    result_list <- .updateCensorMatrices(result_list, still_in, cc, k, cen_fit)

    # adjust cases that are still under consideration
    if (!isSurvival) still_in <- still_in & cc
  }
  
  result_list
}

#' Calculate treatment weights
#' 
#' @noRd
#' @param A A numeric or factor vector object. The observed treatments.
#' @param Ahat A numeric vector or matrix object of length/rows equivalent to `A`. 
#'   The estimated treatment probabilities.
#' @param tx.weight A character object. If "iptw", the inverse probability of
#'   treatment is used; others, abs(A-Ahat).
#' @param cts.obj An R6 object. The treatment-type specific definitions.
#' @param n.bins An integer object. The number of bins (levels) to be used for 
#'   categorizing continuous doses. This input is required only when
#'   `treat.type` = "cont" and `weight` = "overlap" or `weight` = "qpom".
#' 
#' @return A numeric vector object of length equivalent to input `A`
#'
#' @import R6
#'
#' @include treatmentClasses.R
#'
#' @keywords internal
.treatmentWeights <- function(A, A.hat, tx.weight, tx.mod.fitted, cts.obj, n.bins, data) {

  stopifnot(
    "`A` must be a numeric or factor vector" = !missing(A) &&
      {{is.numeric(A) && is.vector(A)} || is.factor(A)} && length(A) > 1L,
    "`A.hat` must be a numeric vector or matrix with same length/nrow as `A`" = 
      !missing(A.hat) &&
      is.numeric(A.hat) && {{is.vector(A.hat) && length(A.hat) == length(A)} ||
          {is.matrix(A.hat) && nrow(A.hat) == length(A)}},
    "`tx.weight` must be a character object" = !missing(tx.weight) &&
      is.character(tx.weight) && is.vector(tx.weight) && length(tx.weight) == 1L,
    "`tx.mod.fitted` must be provided" = !missing(tx.mod.fitted),
    "`cts.obj` must be an R6 object" = !missing(cts.obj) && R6::is.R6(cts.obj),
    "`n.bins` must be an integer" = !missing(n.bins) && is.integer(n.bins)
  )    
  if ("MultiNom" %in% class(cts.obj)) {
    n.bins <- length(levels(A))
  }

  if (tx.weight == "ipw") {
    tx_wgt <- cts.obj$ipw(A, A.hat = A.hat, tx.mod.fitted = tx.mod.fitted)
  } else if (tx.weight == "cipw") {
    weights <- cts.obj$ipw(A, A.hat = A.hat, tx.mod.fitted = tx.mod.fitted)
    cap <- quantile(weights, 0.99)
    tx_wgt <- pmin(weights, cap)
  } else if (tx.weight == "qpom") {
    tx_wgt <- q.pom(A, tx.mod.fitted, data, n.bins, cts.obj)
  } else if (tx.weight == "overlap") {
    if ("Binary" %in% class(cts.obj)) {
      tx_wgt <- abs(A - A.hat)
    } else {
      tx_wgt <- bin.o.fcn(A, tx.mod.fitted, data, n.bins, cts.obj)
    }
  } else {
    tx_wgt <- rep(1.0, length(A))
  }

  tx_wgt
}




#########################################################
### Different weight functions for two stage function ###
#########################################################

# different weight functions
# @noRd
# quantile binning based on POM
q.pom <- function(A, tx.mod.fitted, data, m, cts.obj) {
  pom <- cts.obj$.pom(A, tx.mod.fitted, data, m)
  {1.0 / m} / pom$prob
}

# @noRd
# overlap weights
bin.o.fcn <- function(A, tx.mod.fitted, data, m, cts.obj) {
  pom <- cts.obj$.pom(A, tx.mod.fitted, data, m)
  {1.0 / pom$prob} / pom$sum.prob
  
}

set.seed(42L)
n <- 20L

# Complete data - two stage, no missingness
dat_complete <- data.frame(
  X1 = rnorm(n),
  A1 = rbinom(n, 1L, 0.5),
  X2 = rnorm(n),
  A2 = rbinom(n, 1L, 0.5),
  Y = rnorm(n)
)

# Data with one NA in stage 2 covariate
dat_na <- dat_complete
dat_na$X2[3L] <- NA_real_

models_2stage <- list(
  list(blip = ~X1, treat = A1 ~ X1, tf = ~X1, cens = ~1),
  list(blip = ~X2, treat = A2 ~ X2, tf = ~X2, cens = ~X1)
)

# Minimal obj for non-survival
make_obj <- function(data = dat_complete,
                     censoring.modeled = FALSE,
                     isSurvival = FALSE) {
  obj <- list(
    K = 2L,
    data = data,
    models = models_2stage,
    dependent.vars = list(treat = c("A1", "A2")),
    censoring.modeled = censoring.modeled
  )
  if (isSurvival) obj$manual.censor.weight <- FALSE
  obj
}


test_that(".initStillIn: all TRUE when censoring.modeled=TRUE, non-survival", {
  obj <- make_obj(censoring.modeled = TRUE)
  res <- .initStillIn(obj, isSurvival = FALSE)
  expect_true(all(res))
  expect_length(res, n)
})

test_that(".initStillIn: all TRUE when censoring.modeled=TRUE, survival", {
  obj <- make_obj(censoring.modeled = TRUE, isSurvival = TRUE)
  res <- .initStillIn(obj, isSurvival = TRUE)
  expect_true(all(res))
})

test_that(".initStillIn: uses complete.cases when censoring.modeled=FALSE, non-survival", {
  obj <- make_obj(data = dat_na, censoring.modeled = FALSE)
  res <- .initStillIn(obj, isSurvival = FALSE)
  expect_equal(res, complete.cases(dat_na))
  expect_false(res[3L])
})

test_that(".initStillIn: uses complete.cases when censoring.modeled=FALSE, survival, no manual weight", {
  obj <- make_obj(data = dat_na, censoring.modeled = FALSE, isSurvival = TRUE)
  obj$manual.censor.weight <- FALSE
  res <- .initStillIn(obj, isSurvival = TRUE)
  expect_equal(res, complete.cases(dat_na))
})

test_that(".initStillIn: all TRUE when censoring.modeled=FALSE, survival, manual weight provided", {
  obj <- make_obj(data = dat_na, censoring.modeled = FALSE, isSurvival = TRUE)
  obj$manual.censor.weight <- TRUE
  res <- .initStillIn(obj, isSurvival = TRUE)
  expect_true(all(res))
})


test_that(".dependentVarsCompleteCases: returns logical vector of length n", {
  obj <- make_obj()
  res <- .dependentVarsCompleteCases(obj, K = 1L)
  expect_true(is.logical(res))
  expect_length(res, n)
})

test_that(".dependentVarsCompleteCases: all TRUE when no missing treatment data", {
  obj <- make_obj()
  res <- .dependentVarsCompleteCases(obj, K = 1L)
  expect_true(all(res))
})

test_that(".dependentVarsCompleteCases: FALSE for rows with missing treatment", {
  dat_na_tx <- dat_complete
  dat_na_tx$A1[5L] <- NA_integer_
  obj <- make_obj(data = dat_na_tx)
  res <- .dependentVarsCompleteCases(obj, K = 1L)
  expect_false(res[5L])
  expect_true(all(res[-5L]))
})

test_that(".dependentVarsCompleteCases: stops when treatment column absent from data", {
  obj <- make_obj()
  obj$dependent.vars$treat <- c("A1", "A99")
  expect_error(.dependentVarsCompleteCases(obj, K = 2L), "unable to extract variable")
})


test_that(".modelVarsCompleteCases: returns logical vector of length n", {
  obj <- make_obj()
  res <- .modelVarsCompleteCases(obj, K = 1L)
  expect_true(is.logical(res))
  expect_length(res, n)
})

test_that(".modelVarsCompleteCases: all TRUE when no missing covariate data", {
  obj <- make_obj()
  res <- .modelVarsCompleteCases(obj, K = 1L)
  expect_true(all(res))
})

test_that(".modelVarsCompleteCases: FALSE for rows with missing model covariate", {
  obj <- make_obj(data = dat_na)
  # X2 is NA for row 3 - stage 2 model uses X2
  res <- .modelVarsCompleteCases(obj, K = 2L)
  expect_false(res[3L])
})


test_that(".completeCases: oracle - conjunction of dependent and model complete cases", {
  obj <- make_obj(data = dat_na)
  dep <- .dependentVarsCompleteCases(obj, K = 2L)
  mod <- .modelVarsCompleteCases(obj, K = 2L)
  expect_equal(.completeCases(obj, K = 2L), dep & mod)
})

test_that(".completeCases: all TRUE for complete data", {
  obj <- make_obj()
  expect_true(all(.completeCases(obj, K = 1L)))
  expect_true(all(.completeCases(obj, K = 2L)))
})


test_that(".getDelta: non-survival returns vector of length sum(still.in)", {
  still_in <- c(TRUE, TRUE, FALSE, TRUE, TRUE)
  cc <- c(TRUE, FALSE, FALSE, TRUE, TRUE)
  res <- .getDelta(still_in, cc, status = NULL)
  expect_length(res, sum(still_in))
})

test_that(".getDelta: non-survival oracle - 0 for dropouts, 1 for complete cases", {
  still_in <- c(TRUE, TRUE, TRUE, TRUE, TRUE)
  cc <- c(TRUE, FALSE, TRUE, FALSE, TRUE)
  res <- .getDelta(still_in, cc, status = NULL)
  expected <- c(1L, 0L, 1L, 0L, 1L)
  expect_equal(res, expected)
})

test_that(".getDelta: non-survival excludes participants not still_in", {
  still_in <- c(TRUE, TRUE, FALSE, TRUE, TRUE)
  cc <- c(TRUE, FALSE, FALSE, TRUE, TRUE)
  res <- .getDelta(still_in, cc, status = NULL)
  # row 3 is not still_in so excluded; row 2 is still_in but not cc so delta=0
  expected <- c(1L, 0L, 1L, 1L)
  expect_equal(res, expected)
})

test_that(".getDelta: survival returns status for still_in & cc subset", {
  still_in <- c(TRUE, TRUE, TRUE, TRUE, TRUE)
  cc <- c(TRUE, FALSE, TRUE, TRUE, FALSE)
  status <- c(1L, 0L, 1L, 0L, 1L)
  res <- .getDelta(still_in, cc, status = status)
  # still_in & cc = rows 1, 3, 4
  expect_equal(res, status[still_in & cc])
})


local({
  n_upd <- 6L
  K_upd <- 2L
  
  ml <- list(
    last.stage = integer(n_upd),
    prob.complete.case = matrix(NA_real_, nrow = n_upd, ncol = K_upd),
    d.hat = matrix(NA_real_, nrow = n_upd, ncol = K_upd),
    cens.mod.fitted = list()
  )
  
  still_in <- c(TRUE,  TRUE,  TRUE,  TRUE,  FALSE, FALSE)
  cc <- c(FALSE,  FALSE,  FALSE, FALSE, FALSE, FALSE)
  k <- 1L
  
  cen_fit <- list(
    wt.cen = c(1.0, 1.0, 1.0, 1.0),
    D.hat = c(0.9, 0.8, 0.7, 0.6),
    cens.model.fitted = NA
  ) 
  
  res <- .updateCensorMatrices(ml, still_in, cc, k, cen_fit)

  test_that(".updateCensorMatrices: last.stage updated for cc rows", {
    expect_equal(res$last.stage[cc], rep(k, sum(cc)))
    expect_equal(res$last.stage[!cc], rep(0L, sum(!cc)))
  })
  
  test_that(".updateCensorMatrices: prob.complete.case updated", {
    expect_equal(res$prob.complete.case[, k], rep(NA_real_, length(still_in)))
  })
  
  test_that(".updateCensorMatrices: prob.complete.case is NA for !cc rows", {
    expect_true(all(is.na(res$prob.complete.case[!cc, k])))
  })
  
  test_that(".updateCensorMatrices: d.hat updated for still_in rows", {
    expect_equal(res$d.hat[still_in, k], cen_fit$D.hat)
  })
  
  test_that(".updateCensorMatrices: cens.mod.fitted stored at correct stage index", {
    expect_identical(res$cens.mod.fitted[[k]], cen_fit$cens.model.fitted)
  })
})

local({
  n_upd <- 6L
  K_upd <- 2L
  
  ml <- list(
    last.stage = integer(n_upd),
    prob.complete.case = matrix(NA_real_, nrow = n_upd, ncol = K_upd),
    d.hat = matrix(NA_real_, nrow = n_upd, ncol = K_upd),
    cens.mod.fitted = list()
  )
  
  still_in <- c(TRUE,  TRUE,  TRUE,  TRUE,  FALSE, FALSE)
  cc <- c(TRUE,  TRUE,  TRUE, TRUE, FALSE, FALSE)
  k <- 1L
  
  cen_fit <- list(
    wt.cen = c(1.0, 1.0, 1.0, 1.0),
    D.hat = c(0.9, 0.8, 0.7, 0.6),
    cens.model.fitted = NA
  ) 
  
  res <- .updateCensorMatrices(ml, still_in, cc, k, cen_fit)
  
  test_that(".updateCensorMatrices: last.stage updated for cc rows", {
    expect_equal(res$last.stage[cc], rep(k, sum(cc)))
    expect_equal(res$last.stage[!cc], rep(0L, sum(!cc)))
  })
  
  test_that(".updateCensorMatrices: prob.complete.case updated for still_in rows", {
    expect_equal(res$prob.complete.case[still_in, k], cen_fit$wt.cen)
  })
  
  test_that(".updateCensorMatrices: prob.complete.case is NA for !cc rows", {
    expect_true(all(is.na(res$prob.complete.case[!cc, k])))
  })
  
  test_that(".updateCensorMatrices: d.hat updated for still_in rows", {
    expect_equal(res$d.hat[still_in, k], cen_fit$D.hat)
  })
  
  test_that(".updateCensorMatrices: cens.mod.fitted stored at correct stage index", {
    expect_identical(res$cens.mod.fitted[[k]], cen_fit$cens.model.fitted)
  })
})

# Minimal data for censoring model
dat_cens <- data.frame(X = rnorm(n), A = rbinom(n, 1L, 0.5))

test_that(".processCensoring: returns list with required element names", {
  delta <- rep(1L, n)
  res <- .processCensoring(~1, delta, dat_cens,
                           censoring.modeled = FALSE, dp = 1L, quiet = TRUE)
  expect_named(res, c("cens.model.fitted", "D.hat", "wt.cen"), ignore.order = TRUE)
})

test_that(".processCensoring: path 3 - censoring not modeled, d.hat equals delta", {
  delta <- rep(1L, n)
  res <- .processCensoring(~1, delta, dat_cens,
                           censoring.modeled = FALSE, dp = 1L, quiet = TRUE)
  expect_equal(as.numeric(res$D.hat), as.numeric(delta))
})

test_that(".processCensoring: path 2 - all delta=1, censoring modeled, messages and d.hat=delta", {
  delta <- rep(1L, n)
  expect_message(
    res <- .processCensoring(~X, delta, dat_cens,
                             censoring.modeled = TRUE, dp = 1L, quiet = FALSE),
    "censoring model will be ignored"
  )
  expect_equal(as.numeric(res$D.hat), as.numeric(delta))
  expect_true(is.na(res$cens.model.fitted[[1L]]))
})

test_that(".processCensoring: path 1 - some delta=0, censoring modeled, fits glm", {
  delta <- c(rep(1L, n - 5L), rep(0L, 5L))
  res <- .processCensoring(~X, delta, dat_cens,
                           censoring.modeled = TRUE, dp = 1L, quiet = TRUE)
  expect_true(inherits(res$cens.model.fitted, "glm"))
  expect_length(res$D.hat, n)
})

test_that(".processCensoring: wt.cen oracle - 1/(delta*d.hat + (1-delta)*(1-d.hat))", {
  delta <- rep(1L, n)
  res <- .processCensoring(~1, delta, dat_cens,
                           censoring.modeled = FALSE, dp = 1L, quiet = TRUE)
  d.hat <- as.numeric(res$D.hat)
  delta_n <- as.numeric(delta)
  expected <- 1.0 / (delta_n * d.hat + (1.0 - delta_n) * (1.0 - d.hat))
  expect_equal(res$wt.cen, expected)
})

test_that(".processCensoring: wt.cen oracle - mixed delta with modeled censoring", {
  delta <- c(rep(1L, n - 5L), rep(0L, 5L))
  res <- .processCensoring(~X, delta, dat_cens,
                           censoring.modeled = TRUE, dp = 1L, quiet = TRUE)
  d.hat <- as.numeric(res$D.hat)
  delta_n <- as.numeric(delta)
  expected <- 1.0 / (delta_n * d.hat + (1.0 - delta_n) * (1.0 - d.hat))
  expect_equal(unname(res$wt.cen), expected)
})


A_bin <- as.numeric(rbinom(n, 1L, 0.5))
Ahat_v <- runif(n, 0.3, 0.7)

cts_bin <- Binary$new(tf.model = ~X1,
                      blip.model = ~X1,
                      tx.var = "A1")

fit_tx <- glm(A1 ~ X1, dat_complete, family = "binomial")

# Convenience wrapper
tw <- function(A = A_bin,
               A.hat = Ahat_v,
               tx.weight = "ipw",
               tx.mod.fitted = fit_tx,
               cts.obj = cts_bin,
               n.bins = NA_integer_,
               data = dat_complete) {
  .treatmentWeights(A = A, A.hat = A.hat, tx.weight = tx.weight,
                    tx.mod.fitted = tx.mod.fitted, cts.obj = cts.obj,
                    n.bins = n.bins, data = data)
}

test_that(".treatmentWeights: ipw matches direct cts.obj$ipw call", {
  res <- tw(tx.weight = "ipw")
  expected <- cts_bin$ipw(A = A_bin, A.hat = Ahat_v,
                          tx.mod.fitted = fit_tx)
  expect_equal(res, expected)
})

test_that(".treatmentWeights: cipw oracle - pmin(ipw, quantile(ipw, 0.99))", {
  res <- tw(tx.weight = "cipw")
  weights <- cts_bin$ipw(A = A_bin, A.hat = Ahat_v, tx.mod.fitted = fit_tx)
  cap <- quantile(weights, 0.99)
  expected <- pmin(weights, cap)
  expect_equal(res, expected)
})

test_that(".treatmentWeights: cipw values are <= ipw values", {
  res_ipw <- tw(tx.weight = "ipw")
  res_cipw <- tw(tx.weight = "cipw")
  expect_true(all(res_cipw <= res_ipw))
})

test_that(".treatmentWeights: overlap binary oracle - abs(A - A.hat)", {
  res <- tw(tx.weight = "overlap")
  expected <- abs(A_bin - Ahat_v)
  expect_equal(res, expected)
})

test_that(".treatmentWeights: unrecognised weight returns vector of 1s", {
  res <- tw(tx.weight = "none")
  expect_equal(res, rep(1.0, n))
})

test_that(".treatmentWeights: stops when A has length 1", {
  expect_error(tw(A = 1.0, A.hat = 1.0), "`A`")
  expect_error(tw(A.hat = Ahat_v[1L:5L]), "`A.hat`")
  expect_error(tw(tx.weight = 1L), "`tx.weight`")
  expect_error(tw(cts.obj = list()), "`cts.obj`")
  expect_error(tw(n.bins = 5.0), "`n.bins`")
})

local({
  # ContQuadraticBlip with .pom() implemented
  dat_cont <- data.frame(
    X = cos(seq(0.0, pi, length.out = n)),
    A = runif(n, 0.1, 0.9)
  )
  cts_cont <- ContQuadraticBlip$new(tf.model = ~X,
                                    blip.model = ~A,
                                    tx.var = "A",
                                    treat.range = c(0.1, 0.9))
  fit_cont <- glm(A ~ X, dat_cont, family = gaussian())
  m <- 5L
  
  test_that("q.pom: oracle - (1/m) / pom$prob with named arguments to .pom()", {
    pom <- cts_cont$.pom(A = dat_cont$A,
                         tx.mod.fitted = fit_cont,
                         data = dat_cont,
                         m = m)
    expected <- (1.0 / m) / pom$prob
    result <- q.pom(dat_cont$A, fit_cont, dat_cont, m, cts_cont)
    expect_equal(result, expected)
  })
  
  test_that("q.pom: returns numeric vector of length n", {
    result <- q.pom(dat_cont$A, fit_cont, dat_cont, m, cts_cont)
    expect_true(is.numeric(result))
    expect_length(result, n)
  })
  
  test_that("bin.o.fcn: oracle - (1/pom$prob) / pom$sum.prob with named arguments", {
    pom <- cts_cont$.pom(A = dat_cont$A,
                         tx.mod.fitted = fit_cont,
                         data = dat_cont,
                         m = m)
    expected <- (1.0 / pom$prob) / pom$sum.prob
    result <- bin.o.fcn(dat_cont$A, fit_cont, dat_cont, m, cts_cont)
    expect_equal(result, expected)
  })
  
  test_that("bin.o.fcn: returns numeric vector of length n", {
    result <- bin.o.fcn(dat_cont$A, fit_cont, dat_cont, m, cts_cont)
    expect_true(is.numeric(result))
    expect_length(result, n)
  })
})


test_that(".completeCaseProbability: stops when required element missing from obj", {
  obj_bad <- make_obj()
  obj_bad$K <- NULL
  expect_error(.completeCaseProbability(obj_bad, quiet = TRUE, isSurvival = FALSE),
               "required information")
})

test_that(".completeCaseProbability: returns list with required element names", {
  obj <- make_obj()
  obj$models[[1L]]$cens <- ~1
  obj$models[[2L]]$cens <- ~X1
  res <- .completeCaseProbability(obj, quiet = TRUE, isSurvival = FALSE)
  expect_true(all(c("last.stage", "prob.complete.case",
                    "d.hat", "cens.mod.fitted") %in% names(res)))
})

test_that(".completeCaseProbability: last.stage is integer vector of length n", {
  obj <- make_obj()
  obj$models[[1L]]$cens <- ~1
  obj$models[[2L]]$cens <- ~X1
  res <- .completeCaseProbability(obj, quiet = TRUE, isSurvival = FALSE)
  expect_true(is.integer(res$last.stage))
  expect_length(res$last.stage, n)
})

test_that(".completeCaseProbability: prob.complete.case is n x K matrix", {
  obj <- make_obj()
  obj$models[[1L]]$cens <- ~1
  obj$models[[2L]]$cens <- ~X1
  res <- .completeCaseProbability(obj, quiet = TRUE, isSurvival = FALSE)
  expect_true(is.matrix(res$prob.complete.case))
  expect_equal(dim(res$prob.complete.case), c(n, 2L))
})

test_that(".completeCaseProbability: all last.stage = K for complete data", {
  obj <- make_obj()
  obj$models[[1L]]$cens <- ~1
  obj$models[[2L]]$cens <- ~X1
  res <- .completeCaseProbability(obj, quiet = TRUE, isSurvival = FALSE)
  expect_true(all(res$last.stage == obj$K))
})

test_that(".completeCaseProbability: last.stage < K for participants with missing stage 2 data", {
  obj <- make_obj(data = dat_na)
  obj$models[[1L]]$cens <- ~1
  obj$models[[2L]]$cens <- ~X1
  res <- suppressMessages(
    .completeCaseProbability(obj, quiet = TRUE, isSurvival = FALSE)
  )
  # row 3 has NA in X2 - cannot complete stage 2
  expect_lt(res$last.stage[3L], obj$K)
})

test_that(".treatmentWeights: overlap continuous oracle — matches bin.o.fcn", {
  dat_cont <- data.frame(X = cos(seq(0.0, pi, length.out = n)),
                         A = runif(n, 0.1, 0.9))
  cts_cont <- ContQuadraticBlip$new(tf.model    = ~X,
                                    blip.model  = ~A,
                                    tx.var      = "A",
                                    treat.range = c(0.1, 0.9))
  fit_cont <- glm(A ~ X, dat_cont, family = gaussian())
  m        <- 5L
  res      <- .treatmentWeights(A = dat_cont$A, A.hat = fit_cont$fitted.values,
                                tx.weight = "overlap", tx.mod.fitted = fit_cont,
                                cts.obj = cts_cont, n.bins = m, data = dat_cont)
  expected <- bin.o.fcn(dat_cont$A, fit_cont, dat_cont, m, cts_cont)
  expect_equal(res, expected)
})

test_that(".treatmentWeights: overlap multinomial oracle — matches bin.o.fcn", {
  dat_multi <- data.frame(
    X = rnorm(n),
    A = factor(sample(c("low", "med", "high"), n, replace = TRUE),
               levels = c("low", "med", "high"))
  )
  cts_multi  <- MultiNom$new(tf.model = ~X, blip.model = ~X, tx.var = "A",
                             tx.levels = c("low", "med", "high"))
  fit_multi  <- suppressMessages(nnet::multinom(A ~ X, dat_multi))
  Ahat_mat   <- predict(fit_multi, type = "probs")
  m          <- length(levels(dat_multi$A))
  res        <- .treatmentWeights(A = dat_multi$A, A.hat = Ahat_mat,
                                  tx.weight = "overlap", tx.mod.fitted = fit_multi,
                                  cts.obj = cts_multi, n.bins = m, data = dat_multi)
  expected   <- bin.o.fcn(dat_multi$A, fit_multi, dat_multi, m, cts_multi)
  expect_equal(res, expected)
})

.finalizeCovmat <- function(covmat, blip_vars, full.cov) {
  colnames(covmat)[blip_vars] <- names(blip_vars)
  rownames(covmat)[blip_vars] <- names(blip_vars)
  if (!full.cov) covmat <- covmat[blip_vars, blip_vars, drop = FALSE]
  covmat
}

.getVariance <- function(obj, step.obj, complete.case.info, isSurvival, k) {
  
  stage_cases <- complete.case.info$last.stage >= k
  
  # Subset all vectors/matrices to stage_cases and k
  stage_data <- obj$data[stage_cases, , drop = FALSE]
  stage_data[, step.obj$tx.var] <- step.obj$A
  step_data <- step.obj$cts.obj$prep(stage_data, step.obj$A)
  
  complete.case.info$d.hat <- complete.case.info$d.hat[stage_cases, k]
  
  if (obj$var.estim == "bootstrap" &&
      obj$boot.controls$type %in% c("empirical", "normal")) {
    
    message("starting ", obj$boot.controls$type, " nonparametric bootstrap procedure")
    
    residuals <- {step.obj$Y - 
        predict(step.obj$outcome.fit, step_data)}[step.obj$delta > 0.5]
    
   step.obj <- c(step.obj, 
                  .bootstrap_nonpar(step.obj = step.obj, 
                                    obj = obj, 
                                    residuals = residuals,
                                    status.var = obj$dependent.vars$status,
                                    tx.var = obj$dependent.vars$treat[k],
                                    time.var = obj$dependent.vars$time[k],
                                    tx.wgt.man = if (is.list(obj$tx.wgt.man)) obj$tx.wgt.man[[k]] else NULL,    
                                    data = step_data,
                                    isSurvival = isSurvival,
                                    d.hat = complete.case.info$d.hat))
  }
  
  # Sandwich variance estimator
  if (obj$var.estim == "sandwich") {
    step.obj$covmat <- .sandwich(cts.obj = step.obj$cts.obj, 
                                 outcome.fit = step.obj$outcome.fit, 
                                 Y = step.obj$Y, 
                                 A = step.obj$A, 
                                 A.hat = step.obj$A.hat,
                                 wgts = step.obj$wgts, 
                                 tx.var = step.obj$tx.var, 
                                 data = step_data,
                                 treat.mod.fitted = step.obj$tx.mod.fitted,
                                 method = obj$method)
    
    blip_vars <- step.obj$cts.obj$blip_params(step.obj$outcome.fit$coefficients)
    step.obj$covmat <- .finalizeCovmat(step.obj$covmat, blip_vars, obj$full.cov)    
  }
  step.obj
}

#' @noRd
#' @keywords internal
#' @include bootstrapStep.R sandwich.R varest.R
.computeVariance <- function(obj, isSurvival = FALSE) {
  
  # Complete case weights
  complete_case_info <- obj[c("last.stage", "prob.complete.case", "d.hat", "cens.mod.fitted")]
  
  for (k in obj$K:1L) {
    obj$stages[[k]] <- .getVariance(obj, obj$stages[[k]], complete_case_info, isSurvival, k)
  }
  
  # bootstrap standard errors
  if (obj$var.estim == "bootstrap" && 
      obj$boot.controls$type == "standard") {
    obj <- c(obj, .bootstrap(obj = obj, isSurvival = isSurvival))
  }

  obj
}

# Shared setup

data(twoStageCont)
n <- nrow(twoStageCont)

make_obj <- function(method = "dwols",
                     var.estim = "sandwich",
                     full.cov = FALSE) {
  list(
    K = 2L,
    outcome = twoStageCont$Y,
    data = twoStageCont,
    models = list(
      list(blip = ~X1, treat = A1 ~ X1, tf = ~X1, cens = ~1),
      list(blip = ~X2, treat = A2 ~ X2, tf = ~X2, cens = ~X1)
    ),
    dependent.vars = list(treat = c("A1", "A2")),
    tx.range = list(NA_real_, NA_real_),
    method = method,
    var.estim = var.estim,
    type = "DTR",
    tx.weight = "ipw",
    censoring.modeled = FALSE,
    tx.wgt.man = NA,
    tx.type = "bin",
    n.bins = NA_integer_,
    tx.family = NA,
    boot.controls = list(B = 5L, M = n, type = "standard",
                         truncate = 0.0, verbose = FALSE,
                         interrupt = FALSE),
    full.cov = full.cov
  )
}

# Run .dtrProcedure() once - reused across all tests
base_obj <- make_obj()
dtr_result <- suppressMessages(.dtrProcedure(base_obj, quiet = TRUE))

# Attach required fields from dtr_result back to base_obj for .computeVariance()
enrich_obj <- function(obj, result) {
  obj$stages <- result$stages
  obj$last.stage <- result$last.stage
  obj$prob.complete.case <- result$prob.complete.case
  obj$d.hat <- result$d.hat
  obj$cens.mod.fitted <- result$cens.mod.fitted
  obj$opt.Y <- result$opt.Y
  obj$regret <- result$regret
  obj
}

enriched_obj <- enrich_obj(base_obj, dtr_result)



# .finalizeCovmat


local({
  # Hand-crafted: 4x4 covariance matrix, 2 blip vars at positions 3 and 4
  n_theta <- 4L
  covmat <- matrix(seq_len(n_theta^2), nrow = n_theta, ncol = n_theta)
  colnames(covmat) <- rownames(covmat) <- c("(Intercept)", "X", "A1", "A1:X")
  
  blip_vars <- c("A1" = 3L, "A1:X" = 4L)
  
  test_that(".finalizeCovmat: blip row names set correctly", {
    res <- .finalizeCovmat(covmat, blip_vars, full.cov = TRUE)
    expect_equal(rownames(res)[blip_vars], names(blip_vars))
  })
  
  test_that(".finalizeCovmat: blip col names set correctly", {
    res <- .finalizeCovmat(covmat, blip_vars, full.cov = TRUE)
    expect_equal(colnames(res)[blip_vars], names(blip_vars))
  })
  
  test_that(".finalizeCovmat: non-blip names unchanged", {
    res <- .finalizeCovmat(covmat, blip_vars, full.cov = TRUE)
    expect_equal(rownames(res)[1L:2L], c("(Intercept)", "X"))
    expect_equal(colnames(res)[1L:2L], c("(Intercept)", "X"))
  })
  
  test_that(".finalizeCovmat: full.cov=TRUE returns full matrix", {
    res <- .finalizeCovmat(covmat, blip_vars, full.cov = TRUE)
    expect_equal(dim(res), c(n_theta, n_theta))
  })
  
  test_that(".finalizeCovmat: full.cov=FALSE subsets to blip_vars square matrix", {
    res <- .finalizeCovmat(covmat, blip_vars, full.cov = FALSE)
    expect_equal(dim(res), c(length(blip_vars), length(blip_vars)))
  })
  
  test_that(".finalizeCovmat: full.cov=FALSE row and col names are psi names", {
    res <- .finalizeCovmat(covmat, blip_vars, full.cov = FALSE)
    expect_equal(rownames(res), names(blip_vars))
    expect_equal(colnames(res), names(blip_vars))
  })
  
  test_that(".finalizeCovmat: oracle - values at blip positions are correct", {
    res <- .finalizeCovmat(covmat, blip_vars, full.cov = FALSE)
    expected <- covmat[blip_vars, blip_vars, drop = FALSE]
    expect_equal(unname(res), unname(expected))
  })
})



# .getVariance - var.estim = "none"


test_that(".getVariance: var.estim='none' returns step.obj without covmat", {
  obj_none <- enriched_obj
  obj_none$var.estim <- "none"
  cc <- list(
    last.stage = obj_none$last.stage,
    prob.complete.case = obj_none$prob.complete.case,
    d.hat = obj_none$d.hat,
    cens.mod.fitted = obj_none$cens.mod.fitted
  )
  res <- .getVariance(obj_none, obj_none$stages[[2L]], cc, isSurvival = FALSE, k = 2L)
  expect_null(res$covmat)
})



# .getVariance - var.estim = "sandwich"


local({
  cc <- list(
    last.stage = enriched_obj$last.stage,
    prob.complete.case = enriched_obj$prob.complete.case,
    d.hat = enriched_obj$d.hat,
    cens.mod.fitted = enriched_obj$cens.mod.fitted
  )
  
  res_k2 <- .getVariance(enriched_obj, enriched_obj$stages[[2L]],
                         cc, isSurvival = FALSE, k = 2L)
  res_k1 <- .getVariance(enriched_obj, enriched_obj$stages[[1L]],
                         cc, isSurvival = FALSE, k = 1L)
  
  test_that(".getVariance sandwich: covmat is added to step.obj", {
    expect_false(is.null(res_k2$covmat))
    expect_false(is.null(res_k1$covmat))
  })
  
  test_that(".getVariance sandwich: covmat is a square matrix", {
    expect_true(is.matrix(res_k2$covmat))
    expect_equal(nrow(res_k2$covmat), ncol(res_k2$covmat))
    expect_true(is.matrix(res_k1$covmat))
    expect_equal(nrow(res_k1$covmat), ncol(res_k1$covmat))
  })
  
  test_that(".getVariance sandwich: full.cov=FALSE gives psi-only covmat", {
    n_psi_k2 <- length(enriched_obj$stages[[2L]]$psi)
    n_psi_k1 <- length(enriched_obj$stages[[1L]]$psi)
    expect_equal(dim(res_k2$covmat), c(n_psi_k2, n_psi_k2))
    expect_equal(dim(res_k1$covmat), c(n_psi_k1, n_psi_k1))
  })
  
  test_that(".getVariance sandwich: covmat row and col names contain psi names", {
    expect_true(all(names(enriched_obj$stages[[2L]]$psi) %in% rownames(res_k2$covmat)))
    expect_true(all(names(enriched_obj$stages[[2L]]$psi) %in% colnames(res_k2$covmat)))
  })
  
  test_that(".getVariance sandwich: covmat is symmetric", {
    expect_equal(res_k2$covmat, t(res_k2$covmat))
    expect_equal(res_k1$covmat, t(res_k1$covmat))
  })
  
  test_that(".getVariance sandwich: full.cov=TRUE returns larger covmat", {
    obj_full <- enriched_obj
    obj_full$full.cov <- TRUE
    res_full <- .getVariance(obj_full, obj_full$stages[[2L]],
                             cc, isSurvival = FALSE, k = 2L)
    n_theta <- length(coef(enriched_obj$stages[[2L]]$outcome.fit))
    expect_equal(dim(res_full$covmat), c(n_theta, n_theta))
  })
  
  test_that(".getVariance sandwich: full.cov=TRUE covmat larger than full.cov=FALSE", {
    obj_full <- enriched_obj
    obj_full$full.cov <- TRUE
    res_full <- .getVariance(obj_full, obj_full$stages[[2L]],
                             cc, isSurvival = FALSE, k = 2L)
    expect_true(nrow(res_full$covmat) >= nrow(res_k2$covmat))
  })
  
  test_that(".getVariance sandwich: other step.obj elements preserved", {
    expect_equal(res_k2$psi, enriched_obj$stages[[2L]]$psi)
    expect_equal(res_k2$beta, enriched_obj$stages[[2L]]$beta)
  })
})



# .computeVariance - var.estim = "sandwich"


local({
  res <- suppressMessages(
    .computeVariance(enriched_obj, isSurvival = FALSE)
  )
  
  test_that(".computeVariance sandwich: covmat added to each stage", {
    expect_false(is.null(res$stages[[1L]]$covmat))
    expect_false(is.null(res$stages[[2L]]$covmat))
  })
  
  test_that(".computeVariance sandwich: stages length unchanged", {
    expect_length(res$stages, enriched_obj$K)
  })
  
  test_that(".computeVariance sandwich: psi unchanged after variance computation", {
    for (k in seq_len(enriched_obj$K)) {
      expect_equal(res$stages[[k]]$psi, enriched_obj$stages[[k]]$psi)
    }
  })
  
  test_that(".computeVariance sandwich: each stage covmat is square matrix", {
    for (k in seq_len(enriched_obj$K)) {
      expect_true(is.matrix(res$stages[[k]]$covmat))
      expect_equal(nrow(res$stages[[k]]$covmat), ncol(res$stages[[k]]$covmat))
    }
  })
})



# .computeVariance - var.estim = "none"


test_that(".computeVariance none: no covmat added to stages", {
  obj_none <- enrich_obj(make_obj(var.estim = "none"), dtr_result)
  res <- .computeVariance(obj_none, isSurvival = FALSE)
  expect_true(all(sapply(res$stages, function(s) is.null(s$covmat))))
})



# .computeVariance - var.estim = "bootstrap" standard


local({
  withr::with_seed(42L, {
    obj_boot <- enrich_obj(make_obj(var.estim = "bootstrap"), dtr_result)
    res <- suppressMessages(.computeVariance(obj_boot, isSurvival = FALSE))
  })
  
test_that(".computeVariance bootstrap standard: returns covmat and psi.boot", {
  expect_true("covmat" %in% names(res))
  expect_true("psi.boot" %in% names(res))
})

test_that(".computeVariance bootstrap standard: covmat is list of length K", {
  expect_true(is.list(res$covmat))
  expect_length(res$covmat, enriched_obj$K)
})

})

# STH SHOULD THIS USE OPTIMAL TREATMENT OR OBSERVED?

#' Estimate the confidence intervals
#' 
#' @noRd
#' @param obj A list object. Must contain `$K`, `$models`, `$data`, `$covmat`, 
#'   `$psi`, `$tx.type`
#'   
#' @return A numeric vector object of length `obj$K`
#' 
#' @keywords internal
.varest <- function(obj) {
  
  if (is.null(obj$covmat)) return(NA)
  if (obj$tx.type != "bin") return(NA)
  
  .each_stage <- function(models, data, covmat, psi) {
    
    H_psi <- stats::model.matrix(models$blip, data)
    H_psi <- H_psi[stats::complete.cases(H_psi), , drop = FALSE]

    n_psi <- length(psi)
    if (nrow(covmat) > n_psi) {
      idx    <- seq(nrow(covmat) - n_psi + 1L, nrow(covmat))
      covmat <- covmat[idx, idx, drop = FALSE]
    }
    
    # lower/upper limits on psi
    tmp <- covmat |> diag() |> sqrt()
    psi_l <- psi - 1.96 * tmp
    psi_u <- psi + 1.96 * tmp

    # look at max/min value of blip based on sign of covariates and max/min of 
    # parameter CIs
    H_psi_p <- pmax(H_psi, 0.0)
    H_psi_n <- pmin(H_psi, 0.0)

    # lower blip is positive values * lower CI + negative values * upper CI
    blip_l <- H_psi_p %*% psi_l + H_psi_n %*% psi_u
    blip_u <- H_psi_p %*% psi_u + H_psi_n %*% psi_l
    
    sum(blip_l < 0 & blip_u > 0) / nrow(H_psi)
  }
    
  res <- numeric(obj$K)
  
  for (j in obj$K:1L) {
    res[j] <- .each_stage(obj$models[[j]], obj$data, obj$covmat[[j]], obj$psi[[j]])
  }
  res
}

set.seed(42L)
n_v <- 30L

dat_v <- data.frame(
  X = cos(seq(0.0, pi, length.out = n_v))
)

# Two-stage models
models_v <- list(
  list(blip = ~X),
  list(blip = ~X)
)

# Known psi and covmat for two stages
psi_v <- list(
  c("A1" = 1.0, "A1:X" = 0.5),
  c("A2" = 2.0, "A2:X" = -1.0)
)

# Tight covariance - narrow CIs, few participants straddle zero
covmat_tight <- list(
  diag(c(0.001, 0.001)),
  diag(c(0.001, 0.001))
)

# Wide covariance - CIs so wide everything straddles zero
covmat_wide <- list(
  diag(c(1e6, 1e6)),
  diag(c(1e6, 1e6))
)

make_obj_v <- function(covmat = covmat_tight,
                       psi = psi_v,
                       tx.type = "bin",
                       K = 2L) {
  list(
    K = K,
    data = dat_v,
    tx.type = tx.type,
    models = models_v[seq_len(K)],
    covmat = covmat[seq_len(K)],
    psi = psi[seq_len(K)]
  )
}

test_that(".varest: returns NA when covmat is NULL", {
  obj <- make_obj_v()
  obj$covmat <- NULL
  expect_true(is.na(.varest(obj)))
})

test_that(".varest: returns NA when tx.type is not 'bin'", {
  expect_true(is.na(.varest(make_obj_v(tx.type = "cont"))))
  expect_true(is.na(.varest(make_obj_v(tx.type = "multi"))))
})

test_that(".varest: returns numeric vector of length K", {
  res <- .varest(make_obj_v())
  expect_true(is.numeric(res))
  expect_length(res, 2L)
})

test_that(".varest: returns numeric vector of length 1 for K=1", {
  res <- .varest(make_obj_v(
    covmat = covmat_tight[1L],
    psi = psi_v[1L],
    K = 1L
  ))
  expect_true(is.numeric(res))
  expect_length(res, 1L)
})

test_that(".varest: all values are in [0, 1]", {
  res <- .varest(make_obj_v())
  expect_true(all(res >= 0.0 & res <= 1.0))
})

test_that(".varest: tight CIs produce near-zero non-regularity", {
  # With very small variance, CIs are narrow and unlikely to straddle zero
  # when psi is clearly non-zero - nonreg should be 0 or very small
  res <- .varest(make_obj_v(covmat = covmat_tight))
  expect_true(all(res < 0.1))
})

test_that(".varest: wide CIs produce non-regularity of 1", {
  # With enormous variance, CIs always span zero - every participant
  # is non-regular
  res <- .varest(make_obj_v(covmat = covmat_wide))
  expect_true(all(res == 1.0))
})

test_that(".varest: psi of all zeros with any variance produces nonreg of 1", {
  # If psi = 0, CI always straddles 0 regardless of variance
  psi_zero <- list(c("A1" = 0.0, "A1:X" = 0.0),
                   c("A2" = 0.0, "A2:X" = 0.0))
  res <- .varest(make_obj_v(covmat = covmat_tight, psi = psi_zero))
  expect_true(all(res == 1.0))
})

local({
  # Oracle: explicit .each_stage computation for stage 1
  .each_stage_oracle <- function(models, data, covmat, psi) {
    H_psi <- model.matrix(models$blip, data)
    H_psi <- H_psi[complete.cases(H_psi), , drop = FALSE]
    tmp <- sqrt(diag(covmat))
    psi_l <- psi - 1.96 * tmp
    psi_u <- psi + 1.96 * tmp
    H_psi_p <- pmax(H_psi, 0.0)
    H_psi_n <- pmin(H_psi, 0.0)
    blip_l <- H_psi_p %*% psi_l + H_psi_n %*% psi_u
    blip_u <- H_psi_p %*% psi_u + H_psi_n %*% psi_l
    sum(blip_l < 0 & blip_u > 0) / nrow(H_psi)
  }
  
  # Intermediate covariance - some but not all participants non-regular
  covmat_mid <- list(
    diag(c(0.5, 0.5)),
    diag(c(0.5, 0.5))
  )
  
  obj <- make_obj_v(covmat = covmat_mid)
  res <- .varest(obj)
  
  test_that(".varest: stage 1 matches oracle computation", {
    expected <- .each_stage_oracle(models_v[[1L]], dat_v,
                                   covmat_mid[[1L]], psi_v[[1L]])
    expect_equal(res[1L], expected)
  })
  
  test_that(".varest: stage 2 matches oracle computation", {
    expected <- .each_stage_oracle(models_v[[2L]], dat_v,
                                   covmat_mid[[2L]], psi_v[[2L]])
    expect_equal(res[2L], expected)
  })
  
  test_that(".varest: both stages oracle simultaneously", {
    expected <- c(
      .each_stage_oracle(models_v[[1L]], dat_v, covmat_mid[[1L]], psi_v[[1L]]),
      .each_stage_oracle(models_v[[2L]], dat_v, covmat_mid[[2L]], psi_v[[2L]])
    )
    expect_equal(res, expected)
  })
})

test_that(".varest: NA rows in data are excluded from proportion calculation", {
  dat_na <- dat_v
  dat_na$X[5L] <- NA_real_
  
  obj_na <- make_obj_v(K = 1L, covmat = covmat_wide[1L], psi = psi_v[1L])
  obj_na$data <- dat_na
  obj_clean <- make_obj_v(K = 1L, covmat = covmat_wide[1L], psi = psi_v[1L])
  
  res_na <- .varest(obj_na)
  res_clean <- .varest(obj_clean)
  
  # Wide CIs: both should be 1.0 - NA exclusion doesn't change the proportion
  # when all non-NA participants are non-regular
  expect_equal(res_na, 1.0)
  expect_equal(res_clean, 1.0)
})

#' Sandwich estimator for UNADJUSTED estimating equation
#' 
#' @noRd
#' @param cts.obj An R6 object. The treatment type specific tools.
#' @param outcome.fit A regression object or S3 GEST used to make predictions.
#' @param Y A numeric vector object of length n. The outcome of interest.
#' @param A A numeric vector object of length n. The observed treatment. 
#'   (may be recast as 0/1)
#' @param wgts A numeric vector object of length n. The stage weights.
#' @param tx.var A character object. The treatment variable name.
#' @param data A data.frame object. The complete data for the participants
#'   currently under analysis.
#' 
#' @return A numeric matrix object {n_beta + n_psi} x {n_beta + n_psi}. The 
#'   sandwich estimator of variance.
#' 
#' @keywords internal
.sandwich <- function(cts.obj, outcome.fit, Y, A, A.hat, wgts, tx.var, data,
                      treat.mod.fitted, method) {

  stopifnot(
    "`cts.obj` must be an R6 object" = !missing(cts.obj) && R6::is.R6(cts.obj),
    "`outcome.fit` must be provided" = !missing(outcome.fit),
    "`Y` must be a numeric vector" = is.numeric(Y) && is.vector(Y),
    "`A` must be a numeric vector" = {is.vector(A) || is.factor(A)} &&
      length(A) == length(Y),
    "`wgts` must be a numeric vector" = is.numeric(wgts) && is.vector(wgts) &&
      length(wgts) == length(Y),
    "`tx.var` must be a character object" = !missing(tx.var) && 
      is.character(tx.var) && length(tx.var) == 1L,
    "`data` must be a data.frame" = !missing(data) && is.data.frame(data) &&
      nrow(data) == length(Y),
    "`method` must be a character" = !missing(method) && is.character(method)
  )

  # n x n_theta
  X_theta <- cts.obj$Hd(data, A)

  # n x n_theta
  if (method == "dwols") {
    HW <- wgts * X_theta
  } else {
    HW <- cts.obj$Hw(data, A, A.hat, wgts, treat.mod.fitted = treat.mod.fitted)
  }
  
  n <- nrow(HW)
  
  # n x n_theta
  # (Y - Yhat) wgt (A-Ahat) H_psi
  data[, tx.var] <- A
  data <- cts.obj$prep(data, A)
  
  U <- drop(Y - predict(outcome.fit, data)) * HW

  # n_theta x n_theta
  dtheta <- {-1.0 / n} * crossprod(HW, X_theta)
  
  # derivative with respect to parameter of interest
  inv_mat <- tryCatch(solve(dtheta, t(U)),
                      error = function(e) {
                        stop("unable to invert matrix\n\t", e$message,
                             call. = FALSE)
                      })

  var <- {1.0 / n} * var(t(inv_mat))

  var
}

#' Verify and group models
#'
#' @noRd
#' @param blip.mod A list of formulae.
#' @param treat.mod A list of formulae.
#' @param tf.mod A list of formulae
#'
#' @return A list of the input formula grouped by stage
#' 
#' @keywords internal
.modelsTest <- function(blip.mod, treat.mod, tf.mod) {
  
  # allow for the possibility of providing single decision point analysis
  # as single formula rather than one element lists
  if (!is.list(treat.mod)) treat.mod <- list(treat.mod)
  if (!is.list(blip.mod)) blip.mod <- list(blip.mod)
  if (!is.list(tf.mod)) tf.mod <- list(tf.mod)

  stopifnot(
    "`treat.mod` must be a list of formulae of the form LHS ~ RHS" = 
      is.list(treat.mod) && all(sapply(treat.mod, inherits, what = "formula")) &&
      all(lengths(treat.mod) == 3L),
    "`blip.mod` must be a list of formulae of the form ~ RHS" = 
      is.list(blip.mod) && all(sapply(blip.mod, inherits, what = "formula")) &&
      all(lengths(blip.mod) == 2L),
    "`tf.mod` must be a list of formulae of the form ~ RHS" = 
      is.list(tf.mod) && all(sapply(tf.mod, inherits, what = "formula")) &&
      all(lengths(tf.mod) == 2L)
  )
  
  if (length(treat.mod) != length(blip.mod) || 
      length(treat.mod) != length(tf.mod)) {
    stop("`treat.mod`, `blip.mod`, and `tf.mod` must be of ",
         "the same length", call. = FALSE)
  }
  
  models <- mapply(list, 
                   "blip" = blip.mod,
                   "treat" = treat.mod,
                   "tf" = tf.mod,
                   SIMPLIFY = FALSE)
  models
  
}


#' Create, verify, or incorporate missingness models
#'
#' Note: Call this function only if user specifies that missingness should
#'   be modeled
#' 
#' @noRd
#' @param models A list. One element for each stage. Each element is itself
#'   a list containing "blip", "treat", and "tf". If interactive session
#'   was used to construct this list, it may also contain "cens".
#' @param missing.mod NULL or a list of formula provided as input to dtr
#' 
#' @return Input `models` possibly augmented with missingness model
#' 
#' @importFrom stats as.formula
#'
#' @keywords internal
.constructMissingness <- function(models, missing.mod) {
  
  K <- length(models)
  
  if (is.null(models[[1L]]$cens) && is.null(missing.mod)) {
    # No missingness model provided as input or specified in interactive
    # session, create model from full covariate history
    # NOTE this logic does not maintain interaction terms of models
    
    missing.mod <- vector(mode = "list", length = K)
    missing.mod[[1L]] <- ~ 1
    
    vars <- lapply(models[[1L]], all.vars) |> unlist() |> unique()
    if (K > 1L) {
      for (i in 2L:K) {
        # missing model defined by all previous stage model covariates
        missing.mod[[i]] <- stats::as.formula(paste("~", paste(vars, collapse = "+")))
        message("Stage ", i, " missingness model defined as ",
                deparse(missing.mod[[i]]))
        
        # include ith stage model covariates in vector of covariates
        vars <- c(vars, lapply(models[[i]], all.vars) |> unlist()) |> unique()
      }
    }
  }
  
  # allow for the possibility of a single stage scenario with formula provided
  if (inherits(missing.mod, "formula")) missing.mod <- list(missing.mod)
  
  if (is.list(missing.mod)) {
    # `missing.mod` either constructed above or provided by user
    if (is.null(models[[1L]]$cens)) {
      # if no censoring/missingness model stored in models list add `missing.mod`
      if (length(missing.mod) == K) {
        # ensure that all elements of the list are formulae
        tst <- sapply(missing.mod, inherits, what = "formula") |> all()
        # add to models list
        if (tst) {
          models <- mapply(function(x, y) {x[["cens"]] <- y; x},
                           models,
                           missing.mod, SIMPLIFY = FALSE)
        } else {
          stop("`missing.mod` must be a list of formulae" , call. = FALSE)
        }
      } else {
        stop("`missing.mod` is not of appropriate length, K = ", K, 
             "length(missing.mod) = ", length(missing.mod), call. = FALSE)
      }
    } else {
      warning("missingness models specified in interactive mode;",
              "input `missing.mod` will be ignored", call. = FALSE)
    }
  } else {
    stop("`missing.mod` must be a list of models, 1 for each decision pt",
         call. = FALSE)
  }
  models
}

#' Verify treatment related inputs
#'
#' @noRd
#' @param weight A character. The type of treatment weights
#' @param treat.type A character. The type of treatment variable (binary,
#'   multinomial, continuous).
#' @param n.bins An integer. If continuous treatment at weight function is pom
#'   the number of discrete bins to discretize the treatment.
#' @param treat.wgt.man NULL or a list of numeric vectors. User provided treatment
#'   weights.
#' @param treat.range NULL or a two element vector. If treatment is continuous,
#'   the allowed range of treatment values.
#' @param treat.fam NULL, character, or family object. If treatment is 
#'   continuous, the family object for obtaining parameter estimates
#' @param K An integer. The number of stages
#' @param models A list. The models.
#' @param data A data.frame.
#'
#' @return A list containing all verified treatment related information
#' 
#' @importFrom stats family Gamma gaussian terms
#' 
#' @keywords internal
.treatmentTest <- function(weight, treat.type, n.bins, treat.wgt.man,
                           treat.range, treat.fam, K, models, data) {
  local_obj <- list()

  local_obj$n.bins <- NA_integer_
  if (weight %in% c("qpom")) {
    if (treat.type == "cont") {
      if (is.numeric(n.bins) && is.vector(n.bins) && length(n.bins) == 1L &&
          isTRUE(all.equal(n.bins, round(n.bins)))) {
        local_obj$n.bins <- n.bins
      } else {
        stop("`n.bins` must be an integer", call. = FALSE)
      }
    } else if (treat.type != "multi") {
      stop("binary treatments are not appropriate for `weight` 'qpom'", 
           call. = FALSE)
    }
  }
  
  if (weight %in% c("overlap")) {
    if (treat.type == "cont") {
      if (is.numeric(n.bins) && is.vector(n.bins) && length(n.bins) == 1L &&
          isTRUE(all.equal(n.bins, round(n.bins)))) {
        local_obj$n.bins <- n.bins
      } else {
        stop("`n.bins` must be an integer", call. = FALSE)
      }
    }
  }
  
  if (is.null(treat.wgt.man)) {
    if (weight %in% c("manual", "manual.with.censor")) {
      stop("manual weights must be provided if `weight` = '", weight, "'",
           call. = FALSE)
    }
    local_obj$tx.wgt.man <- NA
  } else if (is.list(treat.wgt.man)) {
    if (!{weight %in% c("manual", "manual.with.censor")}) {
      stop("manual weights cannot be provided if `weight` = '", weight, "'",
           call. = FALSE)
    }
    # must be a list of length K
    if (length(treat.wgt.man) != K || 
        any(!sapply(treat.wgt.man, is.numeric)) ||
        any(!sapply(treat.wgt.man, is.vector, mode = "numeric"))) {
      stop("if provided, `treat.wgt.man` must be a list of K numeric vectors",
           call. = FALSE)
    }
    
    # each element must be a vector of length n
    if (any(lengths(treat.wgt.man) != nrow(data))) {
      stop("all participants must be included in each element of `treat.wgt.man`",
           call. = FALSE)
    }
    
    # all weights must be positive (though they can be NA)
    if (lapply(treat.wgt.man, "<", 1e-8) |> unlist() |> any(na.rm = TRUE)) {
      stop("treatment weights must be positive", call. = FALSE)
    }
    
    local_obj$tx.wgt.man <- treat.wgt.man
  } else {
    stop("`treat.wgt.man` must be NULL or a list", call. = FALSE)
  }

  # this element facilitates blending of survival
  local_obj$dependent.vars <- list()
  local_obj$dependent.vars$treat <- lapply(models,
                                           function(x) {
                                             as.character(x$treat)[2L]
                                             }) |> unlist()
  
  if (is.null(treat.range) && treat.type == "cont") {
    As <- tryCatch(data[, local_obj$dependent.vars$treat, drop = FALSE],
                   error = function(e) {
                     stop("unable to retrieve treatment variables from `data`\n\t",
                          e$message, call. = FALSE)
                   })
    local_obj$tx.range <- lapply(As, range, na.rm = TRUE)
    message("no treatment range specified, [min, max] of observed ",
            "treatments used; \n")
    for (i in 1L:length(local_obj$tx.range)) {
      message("Stage ", i, " [", 
              paste(format(local_obj$tx.range[[i]], digits = 4), collapse = ", "), "]\n")
    }
    
  } else if (treat.type == "cont") {
    if (is.numeric(treat.range) && 
        is.vector(treat.range) &&
        length(treat.range) == 2L) {
      local_obj$tx.range <- vector(mode = "list", length = K)
      local_obj$tx.range <- lapply(local_obj$tx.range, function(x) { treat.range })
    } else if (is.list(treat.range) && length(treat.range) == K && 
               all(lengths(treat.range) == 2L)) {
      local_obj$tx.range <- treat.range
    } else {
      stop("if provided, `treat.range` must be a list of length K each element ",
           "a numeric vector of length 2",
           call. = FALSE)
    }
  } else {
    local_obj$tx.range <- as.list(rep(NA, K))
  }
  
  if (treat.type == "cont") {
    if (is.null(treat.fam)) {
      local_obj$tx.family <- stats::gaussian(link = "identity")
    } else if (is.character(treat.fam) && length(treat.fam) == 1L) {
      if (grepl("gaussian", treat.fam)) {
        local_obj$tx.family <- stats::gaussian(link = "identity")
      } else if (grepl("Gamma", treat.fam)) {
        local_obj$tx.family <- stats:: Gamma(link = "log")
      } else {
        stop("if specificed as a character, `treat.fam` must be one of ",
             "'gaussian' or 'Gamma'", call. = FALSE)
      }
    } else if (inherits(treat.fam, "family")) {
      if (treat.fam$family == "gaussian" && treat.fam$link != "identity") {
        stop("unsupported family specified for treatment model", call. = FALSE)
      } else if (treat.fam$family == "Gamma" && treat.fam$link != "log") {
        stop("unsupported family specified for treatment model", call. = FALSE)
      }
      local_obj$tx.family <- treat.fam
    } else {
      stop("`treat.fam` must be NULL, a character object, or a family object",
           call. = FALSE)
    }
  } else {
    local_obj$tx.family <- NA
  }
  local_obj
}

#' Verify status variable
#' 
#' @noRd
#' @param status A character vector. The status variable name repeated for each
#'   stage or NA if data is not censored
#' @param data A data.frame. The full data.
#' 
#' @return Input `data` possibly updated to ensure that status variable is
#'   binary 0/1.
#'   
#' @keywords internal
.statusTest <- function(status, data) {

  if (all(is.na(status))) {
    message("data is not censored")
    return(data)
  }
  if (!all(status == status[1L])) {
    stop("more than 1 status variable has been provided", call. = FALSE)
  }
  status <- status[1L]
  
  if (is.na(status)) {
    message("data is not censored")
  } else {
    if (any(is.na(data[, status]))) { 
      stop("missing status values are not allowed", call. = FALSE)
    }
    
    d.bin <- {data[, status] |> unique() |> length()} <= 2L
    if (!d.bin) {
      stop("status variables must be binary", call. = FALSE)
    }
      
    if (!all(data[, status] %in% c(0L, 1L))) {
      # recode as 0/1
      data[, status] <- {factor(data[, status]) |> unclass()} - 1L 
    }
  }
  data
}

#' Verify and accept user specified bootstrap controls if provided
#' 
#' @noRd
#' @param bootstrap.controls A list. Optional bootstrap contraols
#' @param n An integer. The number of participants
#'
#' @returns The full bootstrap.controls list
#' 
#' @keywords internal
.bootstrapControlsTest <- function(bootstrap.controls, n) {

  if (is.list(bootstrap.controls)) {
    
    # ensure that list contains only expected elements
    allowed_names <- c("B", "M", "type", "truncate", "verbose", "interrupt")
    if (!all(names(bootstrap.controls) %in% allowed_names)) {
      stop("`bootstrap.control` has unrecognized element names", call. = FALSE)
    }
    
    # define default values for those not provided as input
    default_values <- list("B" = 100L, 
                           "M" = n, 
                           "type" = "standard", 
                           "truncate" = 0.0, 
                           "verbose" = FALSE, 
                           "interrupt" = FALSE)
    
    # replace defaults with provided values
    default_values[names(bootstrap.controls)] <- bootstrap.controls
    
    boot.controls <- default_values
  } else {
    stop("if provided, `bootstrap.controls` must be a named list", 
         call. = FALSE)
  }

  if (is.null(boot.controls$M) || boot.controls$M < 1e-8) boot.controls$M <- n

  stopifnot(
    "`bootstrap.controls$B` must be a positive integer" = 
      is.numeric(boot.controls$B) && 
      isTRUE(all.equal(boot.controls$B, round(boot.controls$B))) &&
      boot.controls$B > 0L,
    "`bootstrap.controls$M` must be a positive integer" = 
      is.numeric(boot.controls$M) && 
      isTRUE(all.equal(boot.controls$M, round(boot.controls$M))) &&
      boot.controls$M > 0L,
    "`bootstrap.controls$type must be one of {'standard', 'empirical', 'normal'}" =
      is.character(boot.controls$type) && length(boot.controls$type) == 1L &&
      boot.controls$type %in% c("standard", "empirical", "normal"),
    "`bootstrap.controls$truncate must be [0.0, 0.5]" = 
      is.numeric(boot.controls$truncate) && 
      length(boot.controls$truncate) == 1L &&
      boot.controls$truncate > -1e-8 && boot.controls$truncate <= 0.5,
    "`bootstrap.controls$verbose` must be logical" = 
      is.logical(boot.controls$verbose),
    "`bootstrap.controls$interrupt` must be logical" = 
      is.logical(boot.controls$interrupt)
  )
  boot.controls
}

#' Verify variance related inputs
#' 
#' @noRd
#' @param var.estim A character. One of "none", "sandwich", "bootstrap"
#' @param method A character. One of "dwols" "gest"
#' @param bootstrap.controls NULL or a list
#' @param n An integer. The number of participants
#'
#' @returns NA if bootstrap not selected; otherwise a list of bootstrap
#'   controls
#'   
#' @keywords internal
.varestimTest <- function(var.estim, method, bootstrap.controls, n) {
  
  if (var.estim == "bootstrap") {
    boot.controls <- .bootstrapControlsTest(bootstrap.controls, n)
  } else {
    boot.controls <- NA
  }
  
  boot.controls
}

# Two-stage model formulae
treat1 <- A1 ~ X1
treat2 <- A2 ~ X2
blip1 <- ~ X1
blip2 <- ~ X2
tf1 <- ~ X1
tf2 <- ~ X2

models_2stage <- list(
  list(blip = blip1, treat = treat1, tf = tf1),
  list(blip = blip2, treat = treat2, tf = tf2)
)

models_1stage <- list(
  list(blip = blip1, treat = treat1, tf = tf1)
)

n <- 50L
dat <- data.frame(
  X1 = rnorm(n),
  A1 = rbinom(n, 1L, 0.5),
  X2 = rnorm(n),
  A2 = rbinom(n, 1L, 0.5),
  Y  = rnorm(n)
)

test_that(".modelsTest: single formulae are coerced to lists", {
  res <- .modelsTest(blip1, treat1, tf1)
  expect_true(is.list(res))
  expect_length(res, 1L)
})

test_that(".modelsTest: output is grouped by stage with correct element names", {
  res <- .modelsTest(list(blip1, blip2), list(treat1, treat2), list(tf1, tf2))
  expect_length(res, 2L)
  expect_named(res[[1L]], c("blip", "treat", "tf"), ignore.order = TRUE)
  expect_named(res[[2L]], c("blip", "treat", "tf"), ignore.order = TRUE)
})

test_that(".modelsTest: formulae are assigned to the correct stage element", {
  res <- .modelsTest(list(blip1, blip2), list(treat1, treat2), list(tf1, tf2))
  expect_identical(res[[1L]]$blip, blip1)
  expect_identical(res[[1L]]$treat, treat1)
  expect_identical(res[[1L]]$tf, tf1)
  expect_identical(res[[2L]]$blip, blip2)
  expect_identical(res[[2L]]$treat, treat2)
  expect_identical(res[[2L]]$tf, tf2)
})

test_that(".modelsTest: errors as expected", {
  expect_error(.modelsTest(blip1, ~X1, tf1), "treat.mod")
  expect_error(.modelsTest(A1 ~ X1, treat1, tf1), "blip.mod")
  expect_error(.modelsTest(blip1, treat1, A1 ~ X1), "tf.mod")
  expect_error(.modelsTest(blip1, list("not a formula"), tf1), "treat.mod")
  expect_error(
    .modelsTest(list(blip1, blip2), list(treat1), list(tf1, tf2)),
    "same length"
  )
})


test_that(".constructMissingness: stage 1 auto-model is intercept-only", {
  res <- suppressMessages(.constructMissingness(models_2stage, NULL))
  expect_equal(deparse(res[[1L]]$cens), "~1")
})

test_that(".constructMissingness: stage 2 auto-model contains stage 1 covariates", {
  res <- suppressMessages(.constructMissingness(models_2stage, NULL))
  vars <- all.vars(res[[2L]]$cens)
  stage1_vars <- lapply(models_2stage[[1L]], all.vars) |> unlist() |> unique()
  expect_true(all(stage1_vars %in% vars))
})

test_that(".constructMissingness: K=1 with both NULL does not error", {
  expect_no_error(suppressMessages(.constructMissingness(models_1stage, NULL)))
})

test_that(".constructMissingness: K=1 auto-model is intercept-only", {
  res <- suppressMessages(.constructMissingness(models_1stage, NULL))
  expect_equal(deparse(res[[1L]]$cens), "~1")
})

test_that(".constructMissingness: adds cens element from user-provided list", {
  mm  <- list(~1, ~X1)
  res <- .constructMissingness(models_2stage, mm)
  expect_identical(res[[1L]]$cens, ~1)
  expect_identical(res[[2L]]$cens, ~X1)
})

test_that(".constructMissingness: single formula is coerced to a list", {
  res <- .constructMissingness(models_1stage, ~1)
  expect_identical(res[[1L]]$cens, ~1)
})

test_that(".constructMissingness: warns when cens already set and missing.mod provided", {
  models_with_cens <- models_2stage
  models_with_cens[[1L]]$cens <- ~1
  models_with_cens[[2L]]$cens <- ~X1
  expect_warning(
    .constructMissingness(models_with_cens, list(~1, ~X1)),
    "will be ignored"
  )
})

test_that(".constructMissingness: errors as expected", {
  expect_error(
    .constructMissingness(models_2stage, list(~1)),
    "appropriate length"
  )
  expect_error(
    .constructMissingness(models_2stage, list(~1, "not a formula")),
    "list of formulae"
  )
  expect_error(
    .constructMissingness(models_2stage, "not a formula or list"),
    "list of models"
  )
})


# Convenience wrapper — fills in the arguments we are not testing
tt <- function(weight = "ipw",
               treat.type = "bin",
               n.bins = NA_integer_,
               treat.wgt.man = NULL,
               treat.range = NULL,
               treat.fam = NULL,
               K = 2L,
               models = models_2stage,
               data = dat) {
  .treatmentTest(weight, treat.type, n.bins, treat.wgt.man,
                 treat.range, treat.fam, K, models, data)
}

test_that(".treatmentTest: returns a list with required elements", {
  res <- tt()
  expect_true(is.list(res))
  expect_true(all(c("n.bins", "tx.wgt.man", "dependent.vars",
                    "tx.range", "tx.family") %in% names(res)))
})

test_that(".treatmentTest: dependent.vars$treat contains treatment variable names", {
  res <- tt()
  expect_equal(res$dependent.vars$treat, c("A1", "A2"))
})

test_that(".treatmentTest: binary treatment sets tx.family to NA", {
  res <- tt(treat.type = "bin")
  expect_true(is.na(res$tx.family))
})

test_that(".treatmentTest: binary treatment sets tx.range to list of NAs", {
  res <- tt(treat.type = "bin")
  expect_true(all(is.na(unlist(res$tx.range))))
})

test_that(".treatmentTest: n.bins stored when weight is qpom and treat.type is cont", {
  res <- tt(weight = "qpom", treat.type = "cont",
            n.bins = 5L, treat.range = c(0, 1))
  expect_equal(res$n.bins, 5L)
})

test_that(".treatmentTest: n.bins stored when weight is overlap and treat.type is cont", {
  res <- tt(weight = "overlap", treat.type = "cont",
            n.bins = 5L, treat.range = c(0, 1))
  expect_equal(res$n.bins, 5L)
})

test_that(".treatmentTest: qpom with multinomial treatment does not require n.bins", {
  expect_no_error(tt(weight = "qpom", treat.type = "multi"))
})

test_that(".treatmentTest: stops when weight is qpom and treat.type is bin", {
  expect_error(tt(weight = "qpom", treat.type = "bin"), "binary treatments")
})

test_that(".treatmentTest: stops when weight is qpom and n.bins is not an integer", {
  expect_error(
    tt(weight = "qpom", treat.type = "cont", n.bins = 2.5, treat.range = c(0, 1)),
    "n.bins"
  )
})

test_that(".treatmentTest: manual weights stored correctly", {
  wgts <- list(rep(1.0, n), rep(1.0, n))
  res  <- tt(weight = "manual", treat.wgt.man = wgts)
  expect_identical(res$tx.wgt.man, wgts)
})

test_that(".treatmentTest: stops when weight is manual but no weights provided", {
  expect_error(tt(weight = "manual", treat.wgt.man = NULL), "must be provided")
})

test_that(".treatmentTest: stops when weights provided but weight is not manual", {
  expect_error(
    tt(weight = "ipw", treat.wgt.man = list(rep(1.0, n), rep(1.0, n))),
    "cannot be provided"
  )
})

test_that(".treatmentTest: stops when treat.wgt.man is wrong length", {
  expect_error(
    tt(weight = "manual", treat.wgt.man = list(rep(1.0, n))),
    "list of K numeric vectors"
  )
})

test_that(".treatmentTest: stops when treat.wgt.man elements are wrong length", {
  expect_error(
    tt(weight = "manual", treat.wgt.man = list(rep(1.0, n - 1L), rep(1.0, n))),
    "all participants"
  )
})

test_that(".treatmentTest: stops when treat.wgt.man contains non-positive values", {
  expect_error(
    tt(weight = "manual", treat.wgt.man = list(rep(-1.0, n), rep(1.0, n))),
    "must be positive"
  )
})

test_that(".treatmentTest: stops when treat.wgt.man is not NULL or a list", {
  expect_error(tt(weight = "manual", treat.wgt.man = 1.0), "NULL or a list")
})

test_that(".treatmentTest: scalar treat.range replicated to list of length K", {
  res <- tt(treat.type = "cont", treat.range = c(0, 1))
  expect_length(res$tx.range, 2L)
  expect_equal(res$tx.range[[1L]], c(0, 1))
  expect_equal(res$tx.range[[2L]], c(0, 1))
})

test_that(".treatmentTest: list treat.range stored as-is", {
  tr  <- list(c(0, 1), c(0.5, 2))
  res <- tt(treat.type = "cont", treat.range = tr)
  expect_identical(res$tx.range, tr)
})

test_that(".treatmentTest: NULL treat.range computed from data for cont treatment", {
  res <- suppressMessages(tt(treat.type = "cont", treat.range = NULL))
  expect_length(res$tx.range, 2L)
  expect_equal(res$tx.range[[1L]], range(dat$A1))
  expect_equal(res$tx.range[[2L]], range(dat$A2))
})

test_that(".treatmentTest: stops when treat.range is invalid for cont treatment", {
  expect_error(
    tt(treat.type = "cont", treat.range = c(0, 1, 2)),
    "treat.range"
  )
})

test_that(".treatmentTest: NULL treat.fam gives gaussian identity for cont", {
  res <- tt(treat.type = "cont", treat.range = c(0, 1), treat.fam = NULL)
  expect_equal(res$tx.family$family, "gaussian")
  expect_equal(res$tx.family$link,   "identity")
})

test_that(".treatmentTest: 'gaussian' character gives gaussian identity", {
  res <- tt(treat.type = "cont", treat.range = c(0, 1), treat.fam = "gaussian")
  expect_equal(res$tx.family$family, "gaussian")
  expect_equal(res$tx.family$link,   "identity")
})

test_that(".treatmentTest: 'Gamma' character gives Gamma log", {
  res <- tt(treat.type = "cont", treat.range = c(0, 1), treat.fam = "Gamma")
  expect_equal(res$tx.family$family, "Gamma")
  expect_equal(res$tx.family$link,   "log")
})

test_that(".treatmentTest: family object is stored directly", {
  fam <- gaussian(link = "identity")
  res <- tt(treat.type = "cont", treat.range = c(0, 1), treat.fam = fam)
  expect_equal(res$tx.family$family, "gaussian")
  expect_equal(res$tx.family$link,   "identity")
})

test_that(".treatmentTest: stops when treat.fam character is unrecognised", {
  expect_error(
    tt(treat.type = "cont", treat.range = c(0, 1), treat.fam = "poisson"),
    "gaussian.*Gamma|Gamma.*gaussian"
  )
})

test_that(".treatmentTest: stops when gaussian family has non-identity link", {
  expect_error(
    tt(treat.type = "cont", treat.range = c(0, 1),
       treat.fam = gaussian(link = "log")),
    "unsupported family"
  )
})

test_that(".treatmentTest: stops when Gamma family has non-log link", {
  expect_error(
    tt(treat.type = "cont", treat.range = c(0, 1),
       treat.fam = Gamma(link = "identity")),
    "unsupported family"
  )
})

test_that(".treatmentTest: stops when treat.fam is not NULL, character, or family", {
  expect_error(
    tt(treat.type = "cont", treat.range = c(0, 1), treat.fam = 42L),
    "treat.fam"
  )
})

dat_status <- data.frame(
  S01  = c(0L, 1L, 0L, 1L, 1L),   # already 0/1 integer
  S12  = c(1L, 2L, 1L, 2L, 1L),   # binary but coded 1/2
  Sna  = c(0L,  NA, 1L, 0L, 1L),  # has NA
  Smul = c(0L, 1L, 2L, 3L, 4L)    # more than 2 values
)

test_that(".statusTest: returns data unchanged when status is already 0/1", {
  res <- .statusTest(rep("S01", 2L), dat_status)
  expect_equal(res$S01, dat_status$S01)
})

test_that(".statusTest: recodes binary non-0/1 status to 0/1", {
  res <- .statusTest(rep("S12", 2L), dat_status)
  expect_true(all(res$S12 %in% c(0L, 1L)))
  expect_equal(as.integer(res$S12), c(0L, 1L, 0L, 1L, 0L))
})

test_that(".statusTest: NA status messages and returns data unchanged", {
  expect_message(
    res <- .statusTest(rep(NA_character_, 2L), dat_status),
    "not censored"
  )
  expect_identical(res, dat_status)
})

test_that(".statusTest: errors as expected", {
  expect_error(.statusTest(c("S01", "S12"), dat_status), "more than 1 status")
  expect_error(.statusTest(rep("Sna", 2L), dat_status), "missing status")
  expect_error(.statusTest(rep("Smul", 2L), dat_status), "binary")
})


n_bc <- 200L

test_that(".bootstrapControlsTest: empty list fills all defaults correctly", {
  res <- .bootstrapControlsTest(list(), n_bc)
  expect_equal(res$B, 100L)
  expect_equal(res$M, n_bc)
  expect_equal(res$type, "standard")
  expect_equal(res$truncate,  0.0)
  expect_false(res$verbose)
  expect_false(res$interrupt)
})

test_that(".bootstrapControlsTest: provided values override defaults", {
  res <- .bootstrapControlsTest(list(B = 500L, type = "empirical"), n_bc)
  expect_equal(res$B, 500L)
  expect_equal(res$type, "empirical")
  expect_equal(res$M, n_bc)   # default still applied
})

test_that(".bootstrapControlsTest: M reset to n when NULL", {
  res <- .bootstrapControlsTest(list(M = NULL), n_bc)
  expect_equal(res$M, n_bc)
})

test_that(".bootstrapControlsTest: M reset to n when zero", {
  res <- .bootstrapControlsTest(list(M = 0L), n_bc)
  expect_equal(res$M, n_bc)
})

test_that(".bootstrapControlsTest: errors as expected", {
  expect_error(.bootstrapControlsTest("not a list", n_bc), "named list")
  expect_error(.bootstrapControlsTest(list(Z = 1L), n_bc), "unrecognized")
  expect_error(.bootstrapControlsTest(list(B = -1L),  n_bc), "B")
  expect_error(.bootstrapControlsTest(list(B = 0L),   n_bc), "B")
  expect_error(.bootstrapControlsTest(list(B = 1.5),  n_bc), "B")
  expect_error(.bootstrapControlsTest(list(M = 1.5),  n_bc), "M")
  expect_error(.bootstrapControlsTest(list(type = "percentile"), n_bc), "type")
  expect_error(.bootstrapControlsTest(list(truncate = -0.1), n_bc), "truncate")
  expect_error(.bootstrapControlsTest(list(truncate =  0.6), n_bc), "truncate")
  expect_error(.bootstrapControlsTest(list(verbose = 1L), n_bc), "verbose")
  expect_error(.bootstrapControlsTest(list(interrupt = 1L), n_bc), "interrupt")
  expect_no_error(.bootstrapControlsTest(list(truncate = 0.0), n_bc))
  expect_no_error(.bootstrapControlsTest(list(truncate = 0.5), n_bc))
})


test_that(".varestimTest: returns NA when var.estim is 'none'", {
  expect_true(is.na(.varestimTest("none", "dwols", list(), n_bc)))
})

test_that(".varestimTest: returns NA when var.estim is 'sandwich'", {
  expect_true(is.na(.varestimTest("sandwich", "dwols", list(), n_bc)))
})

test_that(".varestimTest: returns bootstrap controls list when var.estim is 'bootstrap'", {
  res <- .varestimTest("bootstrap", "dwols", list(B = 200L), n_bc)
  expect_true(is.list(res))
  expect_equal(res$B, 200L)
})

test_that(".varestimTest: propagates stopping conditions from .bootstrapControlsTest", {
  expect_error(
    .varestimTest("bootstrap", "dwols", list(type = "invalid"), n_bc),
    "type"
  )
})

#' DTR estimation and inference for time-to-event data using DWSurv
#' 
#' Dynamic treatment regimen estimation and inference via dynamic weighted 
#'   survival modeling (DWSurv).  Inference for the blip estimators with single- 
#'   and multi-stage data.
#'   
#' The function \code{DWSurv()} allows estimating an optimal dynamic treatment regime
#'   from multi-stage trials or observational data when the outcome of interest
#'   is survival time subject to right-censoring. The dynamic weighted survival 
#'   modeling (DWSurv) algorithm is implemented.  The method focuses on 
#'   estimating the parameters of the blip: a model of the difference in 
#'   expected outcome under the observed treatment and some reference treatment 
#'   (usually a control) at a given stage, assuming identical histories and 
#'   optimal treatment thereafter.
#'   
#' The method requires the specification of four models for each stage of the 
#'   analysis: a treatment model (conditional mean of the treatment variable), 
#'   a censoring model, a treatment-free model (conditional mean of outcome 
#'   assuming only reference treatments are used), and a blip model.  Only the 
#'   blip model must be correctly specified (or over-specified), with consistent 
#'   parameter estimates obtainable if at least one of the treatment-free or the 
#'   treatment and censoring models are correctly specified.  Note that all of 
#'   these must be specified as lists of formula objects, even if only one stage
#'   of treatment is considered.
#'   
#' Note that as is conventional, it is assumed a larger survival time is 
#'   preferred (which can be easily achieved via transformation of your data if 
#'   necessary).
#'   
#' Several treatment weight function options have been implemented within the 
#'   package:
#'   \itemize{ 
#'     \item "none": No treatment weights applied. If \code{method = "dWOLS"}, this 
#'       selection results in the implementation of Q-learning, modified 
#'       slightly to use the dWOLS style pseudo-outcome 
#'       (computed using the observed outcome modified by the estimated 
#'        treatment effect) rather than the traditional Q-learning outcome 
#'       (predicted based on model only, rather than observed outcome with 
#'       treatment effect).  
#'     \item "ipw": weights based on the inverse probability of 
#'       treatment. For binary treatments, a logistic regression is used.
#'       For multinomial, a multinomial log-linear model is fit using 
#'       \code{\link[nnet]{multinom}}. For continuous treatments, a GLM with the specified 
#'       family and link function provided in the \code{treat.fam} argument is used.
#'     \item "cipw": inverse probability of treatment weights as described for
#'       "ipw" and capped at the 99th percentile of the observed weights.
#'     \item "qpom": weights based on the stabilized inverse 
#'       probability of treatment applied to the categorized (into n.bins bins) 
#'       continuous doses or multinomial treatments; probabilities are 
#'       calculated using a proportional odds model.
#'       This weight is appropriate only for continuous and multinomial treatments.
#'     \item "overlap": overlap weights for the categorized continuous doses 
#'       or multinomial treatments (Li and Li, 2019) and the absolute difference 
#'       \eqn{|A - E[A|...]|}{|A - E[A|...]|} for binary treatments.
#'     \item "manual": User provides treatment weights through input 
#'     \code{treat.wgt.man}.
#'     \item "manual.with.censor": User provides combined treatment * censoring
#'       weights through input \code{treat.wgt.man}. Note that `cens.mod` should
#'       be specified with the event indicator on the right-hand side of the 
#'       formula (e.g., \code{~ status}).
#'    }
#'    
#'  Note that the overlap weights were initially termed "absolute value weights" 
#'    in Wallace and Moodie's (2015, Biometrics, Doubly-robust dynamic 
#'    treatment regimen estimation via weighted least squares. 71:636-644) and 
#'    abbreviated \code{abs} in earlier versions of this package. In addition, the 
#'    overlap weights for continuous or multinomial treatments were termed 
#'    \code{wo} in previous versions, which corresponds to their labeling in 
#'    Schulz & Moodie (2021, Journal of the American Statistical Association, 
#'    Doubly robust estimation of optimal dosing strategies 116:256-268.).
#'    Since version 2.4, these have been united under \code{overlap}, and
#'    \code{abs} and \code{wo} have been deprecated.
#'   
#' @inheritParams DTRreg
#' @param method The DTR method to be used, choose "dwols" for dynamic WOLS, 
#'   or "qlearn" for Q-learning.
#' @param time A list of formula specifying the survival time variable for each 
#'   stage in order. The time variable should be specified on the right hand 
#'   side of the formula. No dependent variable should be specified. The list 
#'   should be as long as the number of stages.
#' @param cens.mod A list of formula objects specifying the censoring
#'   model for each stage in order. The event indicator, which takes value 1 if
#'   an event was observed and 0 otherwise, should be included as the dependent
#'   variable and should be the same across stages. In the absence of censoring
#'   or if censoring weights are provided by the user through `treat.wgt.man`, 
#'   (i.e., \code{weight = 'manual.with.censor'})
#'   one still needs to specify an event indicator on the right-hand
#'   side of the formula and leave the left-hand side empty (see example below).
#' @param treat.wgt.man NULL or a list of vectors of known treatment 
#'   (or treatment * censoring) weights can be 
#'   specified to be used instead of hard-coded treatment weight options.
#'   The \eqn{i^{th}}{ith} element of the list contains the multiplicative weights 
#'   for the \eqn{i^{th}}{ith} stage. Each vector must be of length \eqn{n}{n}, 
#'   the number of participants. Used only for \code{method = "dwols"}. If
#'   providing the treatment * censoring weights, \code{cens.mod = NA} must
#'   be used.
#'    
#' @return An object of class \code{DWSurv}, a list including elements
#'     \item{K: }{The number of decision points.}
#'     \item{beta: }{A list. The ith element contains the parameter estimates of
#'       the ith stage treatment-free model.}
#'     \item{psi: }{A list. The ith element contains the parameter estimates of
#'       the ith stage blip model.}
#'     \item{covmat: }{A list. The ith element contains covariance matrix of 
#'       the ith stage blip parameter estimates.}
#'     \item{nonreg: }{Non-regularity estimates.}
#'     \item{setup: }{A list detailing the input parameter settings used for the
#'       analysis}
#'       \itemize{
#'         \item models: A list of the models used for the analysis.
#'         \item method: The parameter estimation method.
#'         \item var.estim: The variance estimation method.
#'         \item cc.modeled: If TRUE, missing data was modeled. If FALSE, cases
#'            with missing data were removed from the analysis.
#'         \item tx.weight: The treatment weighting used for the analysis.
#'         \item tx.type: Treatment was binary, multinomial, or continuous.
#'         \item n.bins: The number of bins (levels) used for categorizing
#'           continuous doses when \code{tx.weight = "overlap"} or 
#'           \code{tx.weight = "qpom"}.
#'         \item tx.wgt.man: Any user provided treatment weights.
#'         \item tx.range: For continuous treatments, the range of allowed
#'           treatment values.
#'         \item tx.family: The description of the dose distribution along 
#'           with the link function used in the continuous treatment model.
#'         \item boot.controls: A list of the bootstrap controls.
#'         \item type: The type of effect. Dynamic treatment regime or treatment
#'           effect.
#'       }
#'     \item{training_data: }{A list containing the training data.}
#'       \itemize{
#'         \item data: The covariates and treatment data.
#'         \item outcome: The outcome of interest.
#'         \item A: The treatment variables, possibly recoded to adhere to internal
#'           code requirements.
#'       }
#'     \item{analysis: }{A list containing the primary results of each stage analysis.}
#'       \itemize{
#'         \item n: The number of participants included in the stage analysis.
#'         \item last.stage: The last stage each participant was included in 
#'           the analysis.
#'         \item prob.cens: The complete case probabilities.
#'         \item cens.mod.fitted: The regression objects returned for estimating
#'           the complete case probabilities.
#'         \item cens.wgt: The complete case weights.
#'         \item cts: The treatment type at each stage.
#'         \item tx.mod.fitted: The regression objects returned for estimating
#'           the treatment probabilities.
#'         \item A.hat: The estimated or provided treatment probabilities.
#'         \item tx.wgt: The treatment weights.
#'         \item outcome.fit: The regression objects returned for each stage outcome
#'           regression.
#'         \item Y: The pseudo-outcomes.
#'         \item regret: Estimates of the regret for each subject based on observed 
#'           treatment and blip parameter estimates.
#'         \item opt.treat: Optimal treatment decisions for each subject at each 
#'           stage of treatment.
#'         \item opt.Y: Predicted optimal outcome under recommended regimen.
#'      }
#'     \item{call: }{The original function call.}
#'     
#'   The functions \code{coef()}, \code{predict()} and 
#'   \code{confint()} may be used with such 
#'   model objects. The first two have specific help files for their 
#'   implementation, while \code{confint()} is used in the same way as 
#'   the standard 
#'   \code{\link[stats]{confint}()} command, with the exception of the \code{parm} 
#'   option, which is not available.
#'
#' @references
#' Simoneau, G., Moodie, E. E. M., Wallace, M.P., Platt, R. W. (2020) Optimal 
#'   Dynamic Treatment Regimes with Survival Endpoints: Introducing DWSurv in the 
#'   R package DTRreg. \emph{Journal of Statistical Computation and Simulation}.
#'   \bold{90}, 2991-3008. (doi:10.1080/00949655.2020.1793341)
#'   
#' Simoneau, G., Moodie, E. E. M., Nijjar, J. S., Platt, R. W. (2019) Estimating 
#'   Optimal Dynamic Treatment with Survival Outcomes. \emph{Journal of the 
#'   American Statistical Association}, \bold{115}, 1531-1539 
#'   (doi:10.1080/01621459.2019.1629939).
#'   
#' Wallace, M. P., Moodie, E. E. M., Stephens, D. A. (2017) Dynamic Treatment 
#'   Regimen Estimation via Regression-Based Techniques: Introducing {R} Package 
#'   {DTRreg}. \emph{Journal of Statistical Software} \bold{80}(2), 1--20 
#'   (doi:10.18637/jss.v080.i02).
#'   
#' Simoneau, G., Moodie, E. E. M., Nijjar, J. S., and Platt, R. W. (2020)
#'   Finite Sample Variance Estimation for Optimal Dynamic Treatment
#'   Regimes of Survival Outcomes. \emph{Statistics in Medicine} \bold{39},
#'   4466-4479.
#'   
#' Efron, B., and Tibshirani, R. (1986)
#'   Bootstrap Methods for Standard Errors, Confidence Intervals, and Other 
#'   Measures of Statistical Accuracy \emph{Source: Statistical Science} \bold{1}
#'   54-75.
#'   
#' @examples
#' #### example single run of a 2-stage DWSurv analysis
#' data(twoStageCens)
#' mod <- DWSurv(time = list(~ T1, ~ T2), 
#'               blip.mod = list(~ X11, ~ X21), 
#'               treat.mod = list(A1 ~ X11, A2 ~ 1), 
#'               tf.mod = list(~ X11 + X12, ~ X21 + X22 + X11), 
#'               cens.mod = list(delta ~ 1, delta ~ X11), 
#'               var.estim = "sandwich", 
#'               data = twoStageCens)
#' mod
#'   
#' #### example in the absence of censoring
#' data(twoStageSurv)
#' mod_nocensoring <- DWSurv(time = list(~ T1, ~ T2), 
#'                           blip.mod = list(~ X11, ~ X21), 
#'                           treat.mod = list(A1 ~ X11, A2 ~ 1), 
#'                           tf.mod = list(~ X11 + X12, ~ X21 + X22 + X11), 
#'                           cens.mod = list(~ delta, ~ delta), 
#'                           var.estim = "sandwich", 
#'                           data = twoStageSurv)
#' mod_nocensoring
#' 
#' @concept dynamic treatment regimens
#' @concept adaptive treatment strategies
#' @concept personalized medicine
#' @concept dynamic weighted survival modeling
#' @concept accelerated failure time
#'
#' @importFrom stats as.formula family Gamma gaussian get_all_vars model.frame terms
#' @importFrom stats na.pass
#' @include dtrProcedure.R inputProcessing.R interactive.R variance.R
#' @export
DWSurv <- function(time, 
                   blip.mod, treat.mod, tf.mod, cens.mod,
                   data = NULL, 
                   method = c("dwols", "qlearn"), 
                   interactive = FALSE,
                   treat.type = c("bin", "multi", "cont"), 
                   treat.fam = gaussian(link = "identity"), 
                   weight = c("overlap", "ipw", "cipw", "qpom", "none", 
                              "manual", "manual.with.censor"),
                   n.bins = 3L, 
                   treat.range = NULL, 
                   treat.wgt.man = NULL,
                   var.estim = c("none", "bootstrap", "sandwich"), 
                   bootstrap.controls = list(B = 100L, 
                                             M = 0L, 
                                             type = "standard", 
                                             truncate = 0.0, 
                                             verbose = FALSE, 
                                             interrupt = FALSE), 
                   dtr = TRUE,
                   full.cov = FALSE) {
  
  # soft deprecation - remap before match.arg validates
  if (!missing(weight)) {
    if (identical(weight, "abs")) {
      warning("'abs' has been renamed 'overlap'. Please update your code. ",
              "'abs' will be removed in a future version.", call. = FALSE)
      weight <- "overlap"
    } else if (identical(weight, "wo")) {
      warning("'wo' has been renamed 'overlap'. Please update your code. ",
              "'wo' will be removed in a future version.", call. = FALSE)
      weight <- "overlap"
    }
  }
  
  # now validate - "abs" and "wo" are not in this list
  weight <- match.arg(weight, c("overlap", "ipw", "cipw", "qpom", "none", "manual", "manual.with.censor"))
  
  
  # if interactive mode chosen, build standardized input from user input
  if (interactive) {
    # returned list contains 
    # $models A list. The models (blip, tf, treat) grouped according to 
    #   decision point
    # $method A character. The DTR method (dwols).
    # $var.estim A character. The covariance matrix estimation method 
    #   ("none", "bootstrap", "sandwich")
    # $time A list of formulae. RHS is the time variable for each stage.
    # $K An integer. The number of decision points.
    obj <- .interactive(isSurvival = TRUE, prompt.cens = TRUE, data = data)
    
    time <- obj$time
    obj$time <- NULL
    
  } else {
    if (inherits(time, "formula")) time <- list(time)
    if (inherits(cens.mod, "formula")) cens.mod <- list(cens.mod)
    stopifnot(
      "`time`, `blip.mod`, `treat.mod`, `tf.mod`, and `cens.mod` must be provided" =
        !missing(time) && !missing(blip.mod) & !missing(treat.mod) & 
        !missing(tf.mod) && !missing(cens.mod),
      "`time` must be a list of formulae of the form ~ RHS" = 
        is.list(time) && all(sapply(time, inherits, what = "formula")) &&
        all(lengths(time) == 2L),
      "`cens.mod` must be a list of formulae" = !is.null(cens.mod) &&
        {is.list(cens.mod) && all(sapply(cens.mod, inherits, what = "formula")) &&
         (all(lengths(cens.mod) == 2L) || all(lengths(cens.mod) == 3L))},
      "`data` must be NULL or a data.frame" = is.null(data) || is.data.frame(data)
    )
    
    obj <- list()
    
    # group models by stage
    # this is the model testing for non-survival, so it is missing tests of 
    # time and censoring
    obj$models <- .modelsTest(blip.mod, treat.mod, tf.mod)

    # tests for time and censoring    
    if (!is.list(time)) time <- list(time)
    if (!is.list(cens.mod)) cens.mod <- list(cens.mod)
    
    if (length(obj$models) != length(time) || 
        length(obj$models) != length(cens.mod)) {
      stop("`treat.mod`, `blip.mod`, `tf.mod`, `cens.mod`, and `time` must be of ",
           "the same length", call. = FALSE)
    }
    
    # add appropriate censoring model to each set of models
    obj$models <- mapply(function(m, cens) {
                           m[["cens"]] <- cens
                           m
                         }, obj$models, cens.mod, SIMPLIFY = FALSE)
    
    obj$method <- match.arg(method)
    obj$var.estim <- match.arg(var.estim)
    
    obj$K <- length(obj$models)
  }
  
  if (obj$method == "qlearn") {
    method <- "qlearn"
    weight <- "none"
    obj$method <- "dwols"
  } else {
    method <- obj$method
  }
  
  # time formula should only contain the survival time as the RHS
  if (any(lengths(lapply(time, all.vars)) != 1L)) {
    stop("the survival time must be supplied as a formula of the form ~ RHS, ",
         "where the RHS contains only the survival time variable", 
         call. = FALSE)
  }
  
  # If all censoring models are of length 2, no dependent variable given and
  # RHS provides the status variable. If of length 3, censoring is to be modeled
  # and LHS is status variable and RHS is model covariates
  # NOTE this is all or nothing. Can't model some stages but not others
  if (!is.null(cens.mod)) {
    obj$censoring.modeled <- length(obj$models[[1L]]$cens) == 3L
  } else {
    obj$censoring.modeled <- FALSE
  }
  
  # get all vars from the models
  data_vars <- obj$models |> 
    unlist() |> 
    lapply(FUN = all.vars) |>
    unlist()
  data_vars <- c(data_vars, 
                 time |> 
                   unlist() |> 
                   lapply(FUN = all.vars) |>
                   unlist())
  data_vars <- data_vars |> unique() |>
    paste(collapse = " + ")
  data_vars <- stats::as.formula(paste("~", data_vars))
  
  if (is.null(data)) {
    # if no data set, create one from environment using the models
    
    # get all vars from the models
    obj$data <- tryCatch(stats::get_all_vars(data_vars),
                         error = function(e) {
                           stop("unable to retrieve all covariates from the ",
                                "environment\n\t", e$message, call. = FALSE)
                         })
  } else if (is.data.frame(data)) {
    obj$data <- tryCatch(stats::model.frame(data_vars, data, 
                                            na.action = stats::na.pass),
                         error = function(e) {
                           stop("unable to retrieve all covariates from the ",
                                "`data`\n\t", e$message, call. = FALSE)
                         })
    rm(data, data_vars)
  } else {
    stop("if provided, `data` must be a data.frame", call. = FALSE)
  }
  
  rownames(obj$data) <- seq_len(nrow(obj$data))

    
  ### Treatment related inputs
  
  obj$tx.weight <- match.arg(weight)
  obj$tx.type <- match.arg(treat.type)
  
  obj$manual.censor.weight <- obj$tx.weight == "manual.with.censor"
  # can't have weight = "manual.with.censor" and specify a full censoring model
  if (obj$censoring.modeled && obj$manual.censor.weight) {
    stop("censoring models should be of the form ~ status when ",
         "`weight` = 'manual.with.censor'",
         call. = FALSE)
  }
  
  obj <- c(obj,
           .treatmentTest(weight = obj$tx.weight, 
                          treat.type = obj$tx.type, 
                          n.bins = n.bins, 
                          treat.wgt.man = treat.wgt.man, 
                          treat.range = treat.range, 
                          treat.fam = treat.fam, 
                          K = obj$K, 
                          data = obj$data,
                          models = obj$models))

  if (obj$var.estim != "none" && is.list(obj$tx.mod.man)) {
    warningMsg <- paste("Treatment weights were supplied as an argument",
                        "(generated outside the DWSurv function);",
                        "standard errors do not account for estimation of the",
                        "treatment score model.\n")
    warning(warningMsg, call. = FALSE)
    addWarning <- TRUE
  } else {
    addWarning <- FALSE
  }

  ### Status and Time variables  
  obj$dependent.vars$time <- lapply(time, 
                                    function(x) {
                                      rownames(attr(stats::terms(x), "factors"))
                                    }) |> unlist()
  
  # Note that if user opted for interactive mode, the censoring model will
  # be NULL if data is not censored. 
  obj$dependent.vars$status <- lapply(obj$models, 
                                      function(x) {
                                        if (is.null(x$cens)) return(NA)
                                        as.character(x$cens)[2L]
                                      }) |> unlist() |> unique()

  # Ensure that status variable is 0/1 if data is censored
  obj$data <- .statusTest(obj$dependent.vars$status, obj$data)
  
  ### Variance related inputs
  
  obj$boot.controls <- .varestimTest(var.estim = obj$var.estim, 
                                     method = obj$method, 
                                     bootstrap.controls = bootstrap.controls, 
                                     n = nrow(obj$data))
  
  stopifnot("`full.cov` must be logical" = is.logical(full.cov))
  obj$full.cov <- full.cov
  
  ### Analysis details
  
  if (is.logical(dtr)) {
    obj$type <- ifelse(dtr, "DTR", "effect")
  } else {
    stop("`dtr` must be a logical", call. = FALSE)
  }
  
  obj <- .dtrProcedure(obj = obj, quiet = FALSE, isSurvival = TRUE)
  
  obj <- .computeVariance(obj, isSurvival = TRUE)
  
  # To conform to original return object, shift list of individual results to 
  # individual results as lists
  # Don't want to keep 
  #   dp, tx.var, models, or cts.obj
  obj$stages <- lapply(obj$stages,
                       FUN = function(stg) {
                         stg[c("dp", "tx.var", "models", "cts.obj")] <- NULL
                         stg
                       })
  obj <- c(obj, do.call(mapply, c(obj$stages, FUN = "list", SIMPLIFY = FALSE)))
  obj$stages <- NULL
  
  # non-regularity
  if (obj$cts[[1L]] == "bin") {
    obj$nonreg <- .varest(obj)
  }
  
  obj$call <- match.call()
  
  analysis_settings <- c("models" = "models", 
                         "method" = "method", 
                         "var.estim" = "var.estim", 
                         "boot.controls" = "boot.controls", 
                         "cens.modeled" = "censoring.modeled",
                         "tx.weight" = "tx.weight", 
                         "tx.type" = "tx.type", 
                         "n.bins" = "n.bins", 
                         "tx.wgt.man" = "tx.wgt.man",
                         "tx.range" = "tx.range", 
                         "tx.family" = "tx.family", 
                         "type" = "type",
                         "tx.vars" = "treat.vars")
  setup <- obj[analysis_settings]
  obj[analysis_settings] <- NULL
  names(setup) <- names(analysis_settings)
  obj$setup <- setup
  
  training_data <- c("data", "outcome", "A")
  training <- obj[training_data]
  obj[training_data] <- NULL
  obj$training_data <- training
  
  analysis_names <- c("n" = "n", 
                      "last.stage" = "last.stage", 
                      "prob.cens" = "prob.complete.case", 
                      "cens.mod.fitted" = "cens.mod.fitted",
                      "cens.wgt" = "cens.wgt", 
                      "cts" = "cts", 
                      "tx.mod.fitted" = "tx.mod.fitted", 
                      "A.hat" = "A.hat", 
                      "tx.wgt" = "tx.wgt",
                      "outcome.fit" = "outcome.fit", 
                      "Y" = "Y",
                      "regret" = "regret",
                      "opt.treat" = "opt.treat", 
                      "opt.Y" = "opt.Y",
                      "fitted.values" = "fitted.values",
                      "residuals" = "residuals",
                      "blip.data" = "blip.data")
  analysis <- obj[analysis_names]
  obj[analysis_names] <- NULL
  names(analysis) <- names(analysis_names)
  obj$analysis <- analysis
  
  if (is.null(obj$covmat)) obj$covmat <- NA
  if (is.null(obj$nonreg)) obj$nonreg <- NA

  for_return <- c("K", "beta", "psi", "covmat", "psi.boot", "nonreg", 
                  "setup", "training_data", "analysis", "call")
  obj <- obj[for_return[for_return %in% names(obj)]]
  
  if (addWarning) obj$warn <- warningMsg
  
  obj$setup$method <- method
  
  # return
  class(obj) <- c("DWSurv", "DTRreg", class(obj))
  obj
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DWsurv.R
\name{DWSurv}
\alias{DWSurv}
\title{DTR estimation and inference for time-to-event data using DWSurv}
\usage{
DWSurv(
  time,
  blip.mod,
  treat.mod,
  tf.mod,
  cens.mod,
  data = NULL,
  method = c("dwols", "qlearn"),
  interactive = FALSE,
  treat.type = c("bin", "multi", "cont"),
  treat.fam = gaussian(link = "identity"),
  weight = c("overlap", "ipw", "cipw", "qpom", "none", "manual", "manual.with.censor"),
  n.bins = 3L,
  treat.range = NULL,
  treat.wgt.man = NULL,
  var.estim = c("none", "bootstrap", "sandwich"),
  bootstrap.controls = list(B = 100L, M = 0L, type = "standard", truncate = 0, verbose =
    FALSE, interrupt = FALSE),
  dtr = TRUE,
  full.cov = FALSE
)
}
\arguments{
\item{time}{A list of formula specifying the survival time variable for each 
stage in order. The time variable should be specified on the right hand 
side of the formula. No dependent variable should be specified. The list 
should be as long as the number of stages.}

\item{blip.mod}{A list of formula objects specifying covariates of the
blip function for each stage in order. No dependent variable should be 
specified. Note that this input should include the treatment variable 
ONLY if the blip model is quadratic in treatment. See Details for further 
clarification.}

\item{treat.mod}{A list of formula objects specifying the treatment model for
each stage in order. Treatment variable should be included as the dependent
variable. If treatment is binary \code{\link[stats]{glm}(family = binomial)}
will be 
used to obtain parameter estimates; if treatment is multi-nomial, 
\code{\link[nnet]{multinom}()} will be used to obtain parameter estimates; and if 
treatment is continuous, \code{\link[stats]{glm}()} with the family and
link specified by \code{treat.fam} will be used to obtain parameter estimates.}

\item{tf.mod}{A list of formula objects specifying covariates of the
treatment-free model for each stage in order. No dependent variable should 
be specified.}

\item{cens.mod}{A list of formula objects specifying the censoring
model for each stage in order. The event indicator, which takes value 1 if
an event was observed and 0 otherwise, should be included as the dependent
variable and should be the same across stages. In the absence of censoring
or if censoring weights are provided by the user through `treat.wgt.man`, 
(i.e., \code{weight = 'manual.with.censor'})
one still needs to specify an event indicator on the right-hand
side of the formula and leave the left-hand side empty (see example below).}

\item{data}{A data frame containing all necessary covariates and treatments
contained in the models. Missing data should be coded as \code{NA}.}

\item{method}{The DTR method to be used, choose "dwols" for dynamic WOLS, 
or "qlearn" for Q-learning.}

\item{interactive}{If \code{TRUE} on-screen prompts will guide the user through 
the specification of blip, treatment, and treatment-free models.}

\item{treat.type}{A character object. Must be one of \{"bin", "multi", "cont"\}
indicating that the treatments at each stage are binary, multinomial,
or continuous, respectively. Each stage must have the same treatment type.}

\item{treat.fam}{A character or family object. 
The description of the dose distribution along with the link 
function to be used in the treatment model for computing weights; should be 
specified in a similar format as that used in \code{\link[stats]{glm}()}. 
If character object, must be one of \{"gaussian", "Gamma"\}, for which
\code{\link[stats]{gaussian}(link = "identity")} or 
\code{\link[stats]{Gamma}(link = "log")} 
will be used,
respectively. Input is ignored for \code{treat.type = "bin"} and 
\code{treat.type = "multi"}. If \code{NULL} and \code{treat.type = "cont"},
defaults to \code{gaussian(link = "identity")}.}

\item{weight}{The form of the treatment weight. See details.}

\item{n.bins}{An integer object. The number of bins (levels) to be used for 
categorizing continuous doses. This input is required only when
\code{treat.type = "cont"} and \code{weight = "overlap"} or \code{weight = "qpom"}.}

\item{treat.range}{For continuous treatments. Specify the maximum/minimum 
value that treatments can be take. If unspecified then the minimum/maximum 
value of observed treatments is used. If you wish to have unrestricted 
treatments set this option to \code{c(-Inf, +Inf)}. If each stage has its own
range, provide as a list, the ith element providing the min and max
for the ith stage treatment.}

\item{treat.wgt.man}{NULL or a list of vectors of known treatment 
(or treatment * censoring) weights can be 
specified to be used instead of hard-coded treatment weight options.
The \eqn{i^{th}}{ith} element of the list contains the multiplicative weights 
for the \eqn{i^{th}}{ith} stage. Each vector must be of length \eqn{n}{n}, 
the number of participants. Used only for \code{method = "dwols"}. If
providing the treatment * censoring weights, \code{cens.mod = NA} must
be used.}

\item{var.estim}{Covariance matrix estimation method, either "bootstrap"
or "sandwich" for sandwich estimation.}

\item{bootstrap.controls}{A named list specifying control parameters of the
bootstrap if \code{var.estim = "bootstrap"}. Available controls are:
\describe{
  \item{B: }{The number of bootstrap samples.} 
  \item{M: }{The subsample size for m out of n bootstrap.}
  \item{type: }{The type of bootstrap. Must be one of \{"standard", 
  "empirical", "normal"\}. The last two are parametric bootstraps.}
  \item{truncate: }{A number between 0 and 0.5. The lowest and highest 
  specified proportion of parameter estimates will be replaced by the
  relevant quantiles affording some robustness to extreme values 
  when estimating covariance.}
  \item{verbose: }{If TRUE, estimated time to completion will be printed 
  to the console every ~30 seconds.}
  \item{interrupt: }{If TRUE then user will be given the option 
  to abort the bootstrap without error if estimated time to completion
  exceeds 10 minutes.}
}}

\item{dtr}{A logical object. If \code{TRUE}, use the DTR estimation approach, which
estimates the stage pseudo-outcome by adding a regret function. If \code{FALSE},
use an 'effect estimation' approach, which treats the observed outcome
as being equal to an outcome assuming no treatment is received at any
stage, plus a blip component at each stage; each stage pseudo-outcome is 
generated by subtracting a blip function. Note that most of the 
DTR-specific output will either be suppressed or irrelevant.}

\item{full.cov}{A logical. If \code{TRUE}, the full covariance matrix will be
returned. If \code{FALSE}, only the terms pertaining to the blip parameters
are returned.}
}
\value{
An object of class \code{DWSurv}, a list including elements
    \item{K: }{The number of decision points.}
    \item{beta: }{A list. The ith element contains the parameter estimates of
      the ith stage treatment-free model.}
    \item{psi: }{A list. The ith element contains the parameter estimates of
      the ith stage blip model.}
    \item{covmat: }{A list. The ith element contains covariance matrix of 
      the ith stage blip parameter estimates.}
    \item{nonreg: }{Non-regularity estimates.}
    \item{setup: }{A list detailing the input parameter settings used for the
      analysis}
      \itemize{
        \item models: A list of the models used for the analysis.
        \item method: The parameter estimation method.
        \item var.estim: The variance estimation method.
        \item cc.modeled: If TRUE, missing data was modeled. If FALSE, cases
           with missing data were removed from the analysis.
        \item tx.weight: The treatment weighting used for the analysis.
        \item tx.type: Treatment was binary, multinomial, or continuous.
        \item n.bins: The number of bins (levels) used for categorizing
          continuous doses when \code{tx.weight = "overlap"} or 
          \code{tx.weight = "qpom"}.
        \item tx.wgt.man: Any user provided treatment weights.
        \item tx.range: For continuous treatments, the range of allowed
          treatment values.
        \item tx.family: The description of the dose distribution along 
          with the link function used in the continuous treatment model.
        \item boot.controls: A list of the bootstrap controls.
        \item type: The type of effect. Dynamic treatment regime or treatment
          effect.
      }
    \item{training_data: }{A list containing the training data.}
      \itemize{
        \item data: The covariates and treatment data.
        \item outcome: The outcome of interest.
        \item A: The treatment variables, possibly recoded to adhere to internal
          code requirements.
      }
    \item{analysis: }{A list containing the primary results of each stage analysis.}
      \itemize{
        \item n: The number of participants included in the stage analysis.
        \item last.stage: The last stage each participant was included in 
          the analysis.
        \item prob.cens: The complete case probabilities.
        \item cens.mod.fitted: The regression objects returned for estimating
          the complete case probabilities.
        \item cens.wgt: The complete case weights.
        \item cts: The treatment type at each stage.
        \item tx.mod.fitted: The regression objects returned for estimating
          the treatment probabilities.
        \item A.hat: The estimated or provided treatment probabilities.
        \item tx.wgt: The treatment weights.
        \item outcome.fit: The regression objects returned for each stage outcome
          regression.
        \item Y: The pseudo-outcomes.
        \item regret: Estimates of the regret for each subject based on observed 
          treatment and blip parameter estimates.
        \item opt.treat: Optimal treatment decisions for each subject at each 
          stage of treatment.
        \item opt.Y: Predicted optimal outcome under recommended regimen.
     }
    \item{call: }{The original function call.}
    
  The functions \code{coef()}, \code{predict()} and 
  \code{confint()} may be used with such 
  model objects. The first two have specific help files for their 
  implementation, while \code{confint()} is used in the same way as 
  the standard 
  \code{\link[stats]{confint}()} command, with the exception of the \code{parm} 
  option, which is not available.
}
\description{
Dynamic treatment regimen estimation and inference via dynamic weighted 
  survival modeling (DWSurv).  Inference for the blip estimators with single- 
  and multi-stage data.
  
The function \code{DWSurv()} allows estimating an optimal dynamic treatment regime
  from multi-stage trials or observational data when the outcome of interest
  is survival time subject to right-censoring. The dynamic weighted survival 
  modeling (DWSurv) algorithm is implemented.  The method focuses on 
  estimating the parameters of the blip: a model of the difference in 
  expected outcome under the observed treatment and some reference treatment 
  (usually a control) at a given stage, assuming identical histories and 
  optimal treatment thereafter.
  
The method requires the specification of four models for each stage of the 
  analysis: a treatment model (conditional mean of the treatment variable), 
  a censoring model, a treatment-free model (conditional mean of outcome 
  assuming only reference treatments are used), and a blip model.  Only the 
  blip model must be correctly specified (or over-specified), with consistent 
  parameter estimates obtainable if at least one of the treatment-free or the 
  treatment and censoring models are correctly specified.  Note that all of 
  these must be specified as lists of formula objects, even if only one stage
  of treatment is considered.
  
Note that as is conventional, it is assumed a larger survival time is 
  preferred (which can be easily achieved via transformation of your data if 
  necessary).
  
Several treatment weight function options have been implemented within the 
  package:
  \itemize{ 
    \item "none": No treatment weights applied. If \code{method = "dWOLS"}, this 
      selection results in the implementation of Q-learning, modified 
      slightly to use the dWOLS style pseudo-outcome 
      (computed using the observed outcome modified by the estimated 
       treatment effect) rather than the traditional Q-learning outcome 
      (predicted based on model only, rather than observed outcome with 
      treatment effect).  
    \item "ipw": weights based on the inverse probability of 
      treatment. For binary treatments, a logistic regression is used.
      For multinomial, a multinomial log-linear model is fit using 
      \code{\link[nnet]{multinom}}. For continuous treatments, a GLM with the specified 
      family and link function provided in the \code{treat.fam} argument is used.
    \item "cipw": inverse probability of treatment weights as described for
      "ipw" and capped at the 99th percentile of the observed weights.
    \item "qpom": weights based on the stabilized inverse 
      probability of treatment applied to the categorized (into n.bins bins) 
      continuous doses or multinomial treatments; probabilities are 
      calculated using a proportional odds model.
      This weight is appropriate only for continuous and multinomial treatments.
    \item "overlap": overlap weights for the categorized continuous doses 
      or multinomial treatments (Li and Li, 2019) and the absolute difference 
      \eqn{|A - E[A|...]|}{|A - E[A|...]|} for binary treatments.
    \item "manual": User provides treatment weights through input 
    \code{treat.wgt.man}.
    \item "manual.with.censor": User provides combined treatment * censoring
      weights through input \code{treat.wgt.man}. Note that `cens.mod` should
      be specified with the event indicator on the right-hand side of the 
      formula (e.g., \code{~ status}).
   }
   
 Note that the overlap weights were initially termed "absolute value weights" 
   in Wallace and Moodie's (2015, Biometrics, Doubly-robust dynamic 
   treatment regimen estimation via weighted least squares. 71:636-644) and 
   abbreviated \code{abs} in earlier versions of this package. In addition, the 
   overlap weights for continuous or multinomial treatments were termed 
   \code{wo} in previous versions, which corresponds to their labeling in 
   Schulz & Moodie (2021, Journal of the American Statistical Association, 
   Doubly robust estimation of optimal dosing strategies 116:256-268.).
   Since version 2.4, these have been united under \code{overlap}, and
   \code{abs} and \code{wo} have been deprecated.
}
\examples{
#### example single run of a 2-stage DWSurv analysis
data(twoStageCens)
mod <- DWSurv(time = list(~ T1, ~ T2), 
              blip.mod = list(~ X11, ~ X21), 
              treat.mod = list(A1 ~ X11, A2 ~ 1), 
              tf.mod = list(~ X11 + X12, ~ X21 + X22 + X11), 
              cens.mod = list(delta ~ 1, delta ~ X11), 
              var.estim = "sandwich", 
              data = twoStageCens)
mod
  
#### example in the absence of censoring
data(twoStageSurv)
mod_nocensoring <- DWSurv(time = list(~ T1, ~ T2), 
                          blip.mod = list(~ X11, ~ X21), 
                          treat.mod = list(A1 ~ X11, A2 ~ 1), 
                          tf.mod = list(~ X11 + X12, ~ X21 + X22 + X11), 
                          cens.mod = list(~ delta, ~ delta), 
                          var.estim = "sandwich", 
                          data = twoStageSurv)
mod_nocensoring

}
\references{
Simoneau, G., Moodie, E. E. M., Wallace, M.P., Platt, R. W. (2020) Optimal 
  Dynamic Treatment Regimes with Survival Endpoints: Introducing DWSurv in the 
  R package DTRreg. \emph{Journal of Statistical Computation and Simulation}.
  \bold{90}, 2991-3008. (doi:10.1080/00949655.2020.1793341)
  
Simoneau, G., Moodie, E. E. M., Nijjar, J. S., Platt, R. W. (2019) Estimating 
  Optimal Dynamic Treatment with Survival Outcomes. \emph{Journal of the 
  American Statistical Association}, \bold{115}, 1531-1539 
  (doi:10.1080/01621459.2019.1629939).
  
Wallace, M. P., Moodie, E. E. M., Stephens, D. A. (2017) Dynamic Treatment 
  Regimen Estimation via Regression-Based Techniques: Introducing {R} Package 
  {DTRreg}. \emph{Journal of Statistical Software} \bold{80}(2), 1--20 
  (doi:10.18637/jss.v080.i02).
  
Simoneau, G., Moodie, E. E. M., Nijjar, J. S., and Platt, R. W. (2020)
  Finite Sample Variance Estimation for Optimal Dynamic Treatment
  Regimes of Survival Outcomes. \emph{Statistics in Medicine} \bold{39},
  4466-4479.
  
Efron, B., and Tibshirani, R. (1986)
  Bootstrap Methods for Standard Errors, Confidence Intervals, and Other 
  Measures of Statistical Accuracy \emph{Source: Statistical Science} \bold{1}
  54-75.
}
\concept{accelerated failure time}
\concept{adaptive treatment strategies}
\concept{dynamic treatment regimens}
\concept{dynamic weighted survival modeling}
\concept{personalized medicine}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-methods.R
\name{plot.DTRreg}
\alias{plot.DTRreg}
\title{Diagnostic Plots for DTR Estimation}
\usage{
\method{plot}{DTRreg}(x, ...)
}
\arguments{
\item{x}{A model object generated by the functions DTRreg and DWSurv.}

\item{...}{Space for additional arguments (not currently used)}
}
\value{
Invisibly, a list of gg objects.
}
\description{
Diagnostic plots for assessment of treatment, treatment-free, and blip models 
  following DTR estimation using DTRreg or DWSurv.
  
DTR estimation using G-estimation and dWOLS requires the specification of 
  three models: the treatment, treatment-free, and blip.  The treatment model 
  may be assessed via standard diagnostics, whereas the treatment-free and 
  blip models may be simultaneously assessed using diagnostic plots 
  introduced by Rich et al.  The plot() function first presents diagnostic 
  plots that assess the latter, plotting fitted values against residuals and 
  covariates following DTR estimation.  If there is any evidence of a 
  relationship between the variables in these plots, this is evidence that at 
  least one of the blip or treatment-free models is mis-specified.
  
Following these plots, the plot() function will present standard diagnostic 
  plots for the treatment model.  These are produced directly by the standard 
  plot() command applied to the models that were fit.  For example, if 
  treatment is binary, the resulting plots are the same as those that are 
  generated by the plot() command applied to a glm object for logistic 
  regression.
}
\examples{
# example single run of a 2-stage g-estimation analysis

set.seed(1)

# expit function
expit <- function(x) { 1.0 / (1.0 + exp(-x)) }

# sample size
n <- 10000

# variables (X = patient information, A = treatment)
X1 <- rnorm(n)
A1 <- rbinom(n, 1, expit(X1))
X2 <- rnorm(n)
A2 <- rbinom(n, 1, expit(X2))

# blip functions
gamma1 <- A1 * (1 + X1)
gamma2 <- A2 * (1 + X2)

# observed outcome: treatment-free outcome plus blip functions
Y <- exp(X1) + exp(X2) + gamma1 + gamma2 + rnorm(n)
 
# models to be passed to DTRreg
# blip model
blip.mod <- list(~ X1, ~ X2)
# treatment model (correctly specified)
treat.mod <- list(A1 ~ X1, A2 ~ 1)
# treatment-free model (incorrectly specified)
tf.mod <- list(~ X1, ~ X2)

# perform G-estimation
mod1 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
              data = twoStageCont, method = "gest")
              
# model diagnostics: note treatment-free model is mis-specified
plot(mod1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data_generate.R
\docType{data}
\name{data}
\alias{data}
\alias{twoStageCont}
\alias{twoStageCens}
\alias{twoStageSurv}
\title{Toy Two-Stage Trial Datasets}
\format{
twoStageCont is a dataset generated to mimic a simple two-stage
  trial. The data.frame contains 1000 observations with 5 columns: 
\describe{
\item{X1}{The first stage covariate. A normally distributed continuous variable.}
\item{A1}{The first stage treatment. A binary variable.}
\item{X2}{The second stage covariate. A normally distributed continuous variable.}
\item{A2}{The second stage treatment. A binary variable.}
\item{Y}{The outcome. A continuous variable.}
}

twoStageCens is a dataset generated to mimic a simple two-stage
  trial with right-censoring. The data.frame contains 1000 observations 
  with 9 columns: 
\describe{
\item{X11}{A first stage covariate. A normally distributed continuous variable.}
\item{X12}{A first stage covariate. A continuous variable = X11^4.}
\item{A1}{The first stage treatment. A binary variable.}
\item{T1}{The time from the beginning of the first stage to the event or to 
          stage 2 entry, whichever comes first.}
\item{X21}{A second stage covariate. A normally distributed continuous variable.}
\item{X22}{A second stage covariate. A continuous variable = X21^3.}
\item{A2}{The second stage treatment. A binary variable.}
\item{T2}{The time from the beginning of the second stage to the event 
          defined only for subjects who enter the second stage.}
\item{delta}{Event indicator.}
}

Note: For participants who experienced the event during stage 1, i.e., did
  not continue to stage 2, the "survival time" is T1. For participants that
  entered stage 2, the "survival time" is T1 + T2.

twoStageSurv is a dataset generated to mimic a simple two-stage
  trial without censoring. The data.frame contains 1000 observations 
  with 9 columns: 
\describe{
\item{X11}{A first stage covariate. A normally distributed continuous variable.}
\item{X12}{A first stage covariate. A continuous variable = X11^4.}
\item{A1}{The first stage treatment. A binary variable.}
\item{T1}{The time from the beginning of the first stage to the event or to 
          stage 2 entry, whichever comes first.}
\item{X21}{A second stage covariate. A normally distributed continuous variable.}
\item{X22}{A second stage covariate. A continuous variable = X21^3.}
\item{A2}{The second stage treatment. A binary variable.}
\item{T2}{The time from the beginning of the second stage to the event 
          non-zero only for subjects who did not have an event in Stage I.}
}

Note: The "survival time" is T1 + T2.
}
\usage{
data(twoStageCont)

data(twoStageCens)

data(twoStageSurv)
}
\description{
These datasets are provided only to facilitate examples. They are not based 
  on or representative of any real-world applications.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-methods.R
\name{confint.DTRreg}
\alias{confint.DTRreg}
\title{Confidence Interval Calculations for DTRs}
\usage{
\method{confint}{DTRreg}(object, parm = NULL, level = 0.95, type = c("se", "percentile"), ...)
}
\arguments{
\item{object}{A model object generated by the function DTRreg.}

\item{parm}{Not available for DTRreg objects.}

\item{level}{The confidence level required.}

\item{type}{Typical Wald-type confidence interval "se" (default) or 
confidence intervals derived with the percentile method "percentile" 
(bootstrap variance estimates only).}

\item{...}{Space for additional arguments (not currently used).}
}
\value{
A list with columns giving lower and upper confidence limits for each
  parameter. These will be labelled as (1-level)/2 and 1 - (1-level)/2 in 
  percentage (by default 2.5\% and 97.5\%).
}
\description{
Confidence intervals for parameters, with the option of 
  constructing the confidence intervals using the percentile method 
  when bootstrap is used.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-methods.R
\name{predict.DTRreg}
\alias{predict.DTRreg}
\title{Optimal Outcome Prediction for DTRs}
\usage{
\method{predict}{DTRreg}(object, newdata, treat.range = NULL, ...)
}
\arguments{
\item{object}{A model object generated by the function `DTRreg()` or `DWSurv()`.}

\item{newdata}{A dataset (usually the data analyzed by DTRreg for which 
predicted outcomes are desired.  If a new dataset is provided, variable 
names should correspond to those presented to `DTRreg()` or `DWSurv()`.}

\item{treat.range}{If treatment is continuous (rather than binary), a 
vectors of the form c(min,max) which specify the minimum and maximum value 
the treatment may take at stage 1.  
If unspecified, this will be inferred from the 
treat.range provided with use of the original DTRreg command.  As such, if 
no treatment range was specified there either, treat.range will be the 
minimum and maximum observed first stage treatment.}

\item{...}{Space for additional arguments (not currently used)}
}
\value{
An \eqn{n}{n} x 1 matrix of predicted outcome values.
}
\description{
Predicted outcome assuming optimal treatment (according to analysis via 
  G-estimation or dWOLS) was followed.  Assumes blip and treatment-free 
  models correctly specified.
 
This function may be used in a similar fashion to more traditional modeling 
  commands (such as lm).  Users are referred to the primary `DTRreg()` 
  and `DTRSurv()` help command
  (and associated literature) for information concerning model specification.
  In particular, we note that the predict function assumes that the 
  treatment-free model has been correctly specified, as the treatment-free 
  parameters are used in the prediction process.
}
\examples{
# example single run of a 2-stage g-estimation analysis

set.seed(1)

# expit function
expit <- function(x) { 1.0 / (1.0 + exp(-x)) }

# sample size
n <- 10000

# variables (X = patient information, A = treatment)
X1 <- rnorm(n)
A1 <- rbinom(n, 1, expit(X1))
X2 <- rnorm(n)
A2 <- rbinom(n, 1, expit(X2))

# blip functions
gamma1 <- A1 * (1 + X1)
gamma2 <- A2 * (1 + X2)

# observed outcome: treatment-free outcome plus blip functions
Y <- exp(X1) + exp(X2) + gamma1 + gamma2 + rnorm(n)
 
# models to be passed to DTRreg
# blip model
blip.mod <- list(~ X1, ~ X2)
# treatment model (correctly specified)
treat.mod <- list(A1 ~ X1, A2 ~ 1)
# treatment-free model (incorrectly specified)
tf.mod <- list(~ X1, ~ X2)

# perform G-estimation
mod1 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
              data = twoStageCont, method = "gest")
              
# predicted Y for optimal treatment
dat <- data.frame(X1, X2, A1, A2)
predict(mod1, newdata = dat)
}
\references{
Chakraborty, B., Moodie, E. E. M. (2013) \emph{Statistical Methods for 
  Dynamic Treatment Regimes}. New York: Springer.
  
Robins, J. M. (2004) \emph{Optimal structural nested models for optimal 
  sequential decisions}. In Proceedings of the Second Seattle Symposium on 
  Biostatistics, D. Y. Lin and P. J. Heagerty (eds), 189-326. 
  New York: Springer.
  
Wallace, M. P., Moodie, E. M. (2015) Doubly-Robust Dynamic Treatment Regimen 
  Estimation Via Weighted Least Squares. \emph{Biometrics} \bold{71}(3), 
  636-644 (doi:10.1111/biom.12306.)
}
\author{
Michael Wallace
}
\concept{adaptive treatment strategies}
\concept{dynamic treatment regimens}
\concept{dynamic weighted ordinary least squares}
\concept{g-estimation}
\concept{personalized medicine}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/s3-methods.R
\name{autoplot.DTRreg}
\alias{autoplot.DTRreg}
\title{Diagnostic Plots for DTR Estimation via autoplot()}
\usage{
\method{autoplot}{DTRreg}(object, ...)
}
\arguments{
\item{object}{A model object generated by the functions DTRreg and DWSurv.}

\item{...}{Space for additional arguments (not currently used)}
}
\value{
A named list of gg objects.
}
\description{
Diagnostic plots for assessment of treatment, treatment-free, and blip models 
  following DTR estimation using DTRreg or DWSurv.
}
\examples{
# example single run of a 2-stage g-estimation analysis

set.seed(1)

# expit function
expit <- function(x) { 1.0 / (1.0 + exp(-x)) }

# sample size
n <- 10000

# variables (X = patient information, A = treatment)
X1 <- rnorm(n)
A1 <- rbinom(n, 1, expit(X1))
X2 <- rnorm(n)
A2 <- rbinom(n, 1, expit(X2))

# blip functions
gamma1 <- A1 * (1 + X1)
gamma2 <- A2 * (1 + X2)

# observed outcome: treatment-free outcome plus blip functions
Y <- exp(X1) + exp(X2) + gamma1 + gamma2 + rnorm(n)
 
# models to be passed to DTRreg
# blip model
blip.mod <- list(~ X1, ~ X2)
# treatment model (correctly specified)
treat.mod <- list(A1 ~ X1, A2 ~ 1)
# treatment-free model (incorrectly specified)
tf.mod <- list(~ X1, ~ X2)

# perform G-estimation
mod1 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
              data = twoStageCont, method = "gest")
              
# model diagnostics: note treatment-free model is mis-specified
gg_list <- ggplot2::autoplot(mod1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DTRreg.R, R/s3-methods.R
\name{DTRreg}
\alias{DTRreg}
\alias{print.DTRreg}
\alias{summary.DTRreg}
\alias{coef.DTRreg}
\title{DTR Estimation and Inference via G-estimation, Dynamic WOLS, or Q-learning}
\usage{
DTRreg(
  outcome,
  blip.mod,
  treat.mod,
  tf.mod,
  data = NULL,
  method = c("gest", "dwols", "qlearn"),
  interactive = FALSE,
  treat.type = c("bin", "multi", "cont"),
  treat.fam = gaussian(link = "identity"),
  weight = c("overlap", "ipw", "cipw", "qpom", "none", "manual"),
  n.bins = 3L,
  treat.range = NULL,
  treat.wgt.man = NULL,
  var.estim = c("none", "bootstrap", "sandwich"),
  full.cov = FALSE,
  bootstrap.controls = list(B = 100L, M = nrow(data), type = "standard", truncate = 0,
    verbose = FALSE, interrupt = FALSE),
  missing = c("drop", "ipw"),
  missing.mod = NULL,
  dtr = TRUE
)

\method{print}{DTRreg}(x, ...)

\method{summary}{DTRreg}(object, ...)

\method{coef}{DTRreg}(object, ...)
}
\arguments{
\item{outcome}{The outcome variable. Missing data will result in a stopping
error.}

\item{blip.mod}{A list of formula objects specifying covariates of the
blip function for each stage in order. No dependent variable should be 
specified. Note that this input should include the treatment variable 
ONLY if the blip model is quadratic in treatment. See Details for further 
clarification.}

\item{treat.mod}{A list of formula objects specifying the treatment model for
each stage in order. Treatment variable should be included as the dependent
variable. If treatment is binary \code{\link[stats]{glm}(family = binomial)}
will be 
used to obtain parameter estimates; if treatment is multi-nomial, 
\code{\link[nnet]{multinom}()} will be used to obtain parameter estimates; and if 
treatment is continuous, \code{\link[stats]{glm}()} with the family and
link specified by \code{treat.fam} will be used to obtain parameter estimates.}

\item{tf.mod}{A list of formula objects specifying covariates of the
treatment-free model for each stage in order. No dependent variable should 
be specified.}

\item{data}{A data frame containing all necessary covariates and treatments
contained in the models. Missing data should be coded as \code{NA}.}

\item{method}{The DTR method to be used, choose "dwols" for dynamic WOLS, 
"gest" for G-estimation, or "qlearn" for Q-learning.}

\item{interactive}{If \code{TRUE} on-screen prompts will guide the user through 
the specification of blip, treatment, and treatment-free models.}

\item{treat.type}{A character object. Must be one of \{"bin", "multi", "cont"\}
indicating that the treatments at each stage are binary, multinomial,
or continuous, respectively. Each stage must have the same treatment type.}

\item{treat.fam}{A character or family object. 
The description of the dose distribution along with the link 
function to be used in the treatment model for computing weights; should be 
specified in a similar format as that used in \code{\link[stats]{glm}()}. 
If character object, must be one of \{"gaussian", "Gamma"\}, for which
\code{\link[stats]{gaussian}(link = "identity")} or 
\code{\link[stats]{Gamma}(link = "log")} 
will be used,
respectively. Input is ignored for \code{treat.type = "bin"} and 
\code{treat.type = "multi"}. If \code{NULL} and \code{treat.type = "cont"},
defaults to \code{gaussian(link = "identity")}.}

\item{weight}{The form of the treatment weight. See details.}

\item{n.bins}{An integer object. The number of bins (levels) to be used for 
categorizing continuous doses. This input is required only when
\code{treat.type = "cont"} and \code{weight = "overlap"} or \code{weight = "qpom"}.}

\item{treat.range}{For continuous treatments. Specify the maximum/minimum 
value that treatments can be take. If unspecified then the minimum/maximum 
value of observed treatments is used. If you wish to have unrestricted 
treatments set this option to \code{c(-Inf, +Inf)}. If each stage has its own
range, provide as a list, the ith element providing the min and max
for the ith stage treatment.}

\item{treat.wgt.man}{NULL or a list of vectors of known treatment weights can be 
specified to be used instead of hard-coded treatment weight options.
The \eqn{i^{th}}{ith} element of the list contains the multiplicative weights 
for the \eqn{i^{th}}{ith} stage. Each vector must be of length \eqn{n}{n}, 
the number of participants. Used only for \code{method = "dwols"}.}

\item{var.estim}{Covariance matrix estimation method, either "bootstrap"
or "sandwich" for sandwich estimation.}

\item{full.cov}{A logical. If \code{TRUE}, the full covariance matrix will be
returned. If \code{FALSE}, only the terms pertaining to the blip parameters
are returned.}

\item{bootstrap.controls}{A named list specifying control parameters of the
bootstrap if \code{var.estim = "bootstrap"}. Available controls are:
\describe{
  \item{B: }{The number of bootstrap samples.} 
  \item{M: }{The subsample size for m out of n bootstrap.}
  \item{type: }{The type of bootstrap. Must be one of \{"standard", 
  "empirical", "normal"\}. The last two are parametric bootstraps.}
  \item{truncate: }{A number between 0 and 0.5. The lowest and highest 
  specified proportion of parameter estimates will be replaced by the
  relevant quantiles affording some robustness to extreme values 
  when estimating covariance.}
  \item{verbose: }{If TRUE, estimated time to completion will be printed 
  to the console every ~30 seconds.}
  \item{interrupt: }{If TRUE then user will be given the option 
  to abort the bootstrap without error if estimated time to completion
  exceeds 10 minutes.}
}}

\item{missing}{A character object. Must be one of \{"drop", "ipw"\}.
If set to "ipw" and covariate or treatment data are missing then inverse 
probability 
weights are used. The complete case probability is estimated 
via logistic regression. If set to "drop" and data are missing, participants
with missing data are dropped for all stage analyses.}

\item{missing.mod}{An optional list of formula objects specifying the model
for the inverse probability of weights for each stage in order.
No dependent variable should be specified. If \code{missing = "ipw"} and 
\code{missing.mod = NULL}, then the models are assumed to be linear comprising
the full covariate history derived from all of the previous stage models.}

\item{dtr}{A logical object. If \code{TRUE}, use the DTR estimation approach, which
estimates the stage pseudo-outcome by adding a regret function. If \code{FALSE},
use an 'effect estimation' approach, which treats the observed outcome
as being equal to an outcome assuming no treatment is received at any
stage, plus a blip component at each stage; each stage pseudo-outcome is 
generated by subtracting a blip function. Note that most of the 
DTR-specific output will either be suppressed or irrelevant.}

\item{x}{An object of class `DTRreg`.}

\item{...}{Ignored.}

\item{object}{An object of class `DTRreg`.}
}
\value{
An object of class \code{DTRreg}, a list including elements
    \item{K: }{The number of decision points.}
    \item{beta: }{A list. The ith element contains the parameter estimates of
      the ith stage treatment-free model.}
    \item{psi: }{A list. The ith element contains the parameter estimates of
      the ith stage blip model.}
    \item{covmat: }{A list. The ith element contains covariance matrix of 
      the ith stage blip parameter estimates.}
    \item{nonreg: }{Non-regularity estimates.}
    \item{setup: }{A list detailing the input parameter settings used for the
      analysis}
      \describe{
        \item{models: }{A list of the models used for the analysis.}
        \item{method: }{The parameter estimation method.}
        \item{var.estim: }{The variance estimation method.}
        \item{cc.modeled: }{If TRUE, missing data was modeled. If FALSE, cases
           with missing data were removed from the analysis.}
        \item{tx.weight: }{The treatment weighting used for the analysis.}
        \item{tx.type: }{Treatment was binary, multinomial, or continuous.}
        \item{n.bins: }{The number of bins (levels) used for categorizing
          continuous doses when \code{tx.weight = "wo"} or 
          \code{tx.weight = "qpom"}.}
        \item{tx.wgt.man: }{Any user provided treatment weights.}
        \item{tx.range: }{For continuous treatments, the range of allowed
          treatment values.}
        \item{tx.family: }{The description of the dose distribution along 
          with the link function used in the continuous treatment model.}
        \item{boot.controls: }{A list of the bootstrap controls.}
        \item{type: }{The type of effect. Dynamic treatment regime or treatment
          effect.}
      }
    \item{training_data: }{A list containing the training data.}
      \describe{
        \item{data: }{The covariates and treatment data.}
        \item{outcome: }{The outcome of interest.}
        \item{A: }{The treatment variables, possibly recoded to adhere to internal
          code requirements.}
      }
    \item{analysis: }{A list containing the primary results of each stage analysis.}
      \describe{
        \item{n: }{The number of participants included in the stage analysis.}
        \item{last.stage: }{The last stage each participant was included in 
          the analysis.}
        \item{prob.cc: }{The complete case probabilities.}
        \item{cc.mod.fitted: }{The regression objects returned for estimating
          the complete case probabilities.}
        \item{cc.wgt: }{The complete case weights.}
        \item{cts: }{The treatment type at each stage.}
        \item{tx.mod.fitted: }{The regression objects returned for estimating
          the treatment probabilities.}
        \item{A.hat: }{The estimated or provided treatment probabilities.}
        \item{tx.wgt: }{The treatment weights.}
        \item{outcome.fit: }{The regression objects returned for each stage outcome
          regression.}
        \item{Y: }{The pseudo-outcomes.}
        \item{regret: }{Estimates of the regret for each subject based on observed 
          treatment and blip parameter estimates.}
        \item{opt.treat: }{Optimal treatment decisions for each subject at each 
          stage of treatment.}
        \item{opt.Y: }{Predicted optimal outcome under recommended regimen.}
     }
    \item{call: }{The original function call.}
  The functions \code{coef()}, \code{predict()} and 
  \code{confint()} may be used with such 
  model objects. The first two have specific help files for their 
  implementation, while \code{confint()} is used in the same way as 
  the standard 
  \code{\link[stats]{confint}()} command, with the exception of the \code{parm} 
  option, which is not available.
}
\description{
Dynamic treatment regimen estimation and inference via G-estimation and 
  dynamic WOLS.  Estimation of blip model parameters for multi-stage data.
}
\details{
\code{DTRreg()} allows the estimation of optimal dynamic treatment regimens 
  (DTRs, also known as adaptive treatment strategies) from multi-stage 
  trials using G-estimation, dynamic weighted ordinary least squares 
  (dWOLS), and genearlized dWOLS. All methods focus on estimating the 
  parameters of the blip: a 
  model of the difference in expected outcome under the observed treatment 
  and some reference treatment (usually a control) at a given stage, assuming
  identical histories and optimal treatment thereafter. The reader is 
  referred to Chakraborty and Moodie (2013) for a thorough introduction and 
  review of DTR methods. The dWOLS method may be used to obtain parameter 
  estimates identical to those from Q-learning (by setting \code{weight = "none"}).
  This option is intended primarily for exploratory purposes; the authors 
  note that there is a dedicated R package for Q-learning (qLearn), although 
  it is limited to the 2-stage setting; multi-stage settings are available
  in R package DynTxRegime.
  
  This implementation assumes an outcome regression model of the form
  E(Y|X=x,A=a) = tf.mod + a blip.mod. That is -- the input \code{blip.mod} 
  formula should include the treatment variable \emph{ONLY} if it is quadratic. 
  For example, if the full blip model is linear in the treatment variable 
  \deqn{\sim a \psi_0 + a x \psi_1,}{~ a psi_0 + a x psi_1,} then the input
  should model should be \code{blip.mod = ~ x}. 
  If the full blip model is quadratic in the treatment variable
  \deqn{\sim a \psi_0 + a^2 \psi_1 + a x \psi_2 + a^2 x \psi_3,}{
  ~ a psi_0 + a^2 psi_1 + a x psi_2 + a^2 x psi_3,} \code{blip.mod =
  ~ a*x}. For continuous treatments, only quadratic blip
  functions are supported.
  
  All methods require the specification of three models for each 
  stage of the analysis: a treatment model (conditional mean of the treatment 
  variable), a treatment-free model (conditional mean of outcome assuming 
  only reference treatments are used), and a blip model. Only the blip model
  must be correctly specified (or over-specified), with consistent parameter 
  estimates obtainable if at least one of the other two models is correctly 
  specified. Note that all of these must be specified as lists of formula 
  objects, even if only one stage of treatment is considered.
  
  Note that as is conventional, it is assumed a larger value of the outcome 
  is preferred (which can be easily achieved via transformation of your data 
  if necessary).
  
  When treatment is binary, if confidence intervals are computed (via 
  specification of \code{var.estim} other than "none"), then DTRreg will calculate 
  the proportion of subjects at each stage for whom optimal treatment is 
  non-unique. If this proportion exceeds 0.05 a non-regularity warning will 
  be displayed, along with the proportion of subjects for whom this is the 
  case. Note that this warning is only displayed if a variance estimation 
  option is selected.
  
  Several treatment weight function options have been implemented within the 
  package:
  \itemize{ 
    \item "none": No treatment weights applied. If \code{method = "dWOLS"}, this 
      selection results in the implementation of Q-learning, modified 
      slightly to use the G-estimation or dWOLS style pseudo-outcome 
      (computed using the observed outcome modified by the estimated 
       treatment effect) rather than the traditional Q-learning outcome 
      (predicted based on model only, rather than observed outcome with 
      treatment effect).   
    \item "ipw": weights based on the inverse probability of 
      treatment. For binary treatments, a logistic regression is used.
      For multinomial, a multinomial log-linear model is fit using 
      \code{\link[nnet]{multinom}}. For continuous treatments, a GLM with the specified 
      family and link function provided in the \code{treat.fam} argument is used.
    \item "cipw": inverse probability of treatment weights as described for
      "ipw" and capped at the 99th percentile of the observed weights.
    \item "qpom": weights based on the stabilized inverse 
      probability of treatment applied to the categorized (into n.bins bins) 
      continuous doses or multinomial treatments; probabilities are 
      calculated using a proportional odds model.
      This weight is appropriate only for continuous and multinomial treatments.
    \item "overlap": overlap weights for the categorized continuous doses 
      or multinomial treatments (Li and Li, 2019) and the absolute difference 
      \eqn{|A - E[A|...]|}{|A - E[A|...]|} for binary treatments.
    \item "manual": User provides weights through input \code{treat.wgt.man}.
      Manual treatments are only used in dwols.
   }
   
 Note that the overlap weights were initially termed "absolute value weights" 
   in Wallace and Moodie's (2015, Biometrics, Doubly-robust dynamic 
   treatment regimen estimation via weighted least squares. 71:636-644) and 
   abbreviated \code{abs} in earlier versions of this package. In addition, the 
   overlap weights for continuous or multinomial treatments were termed 
   \code{wo} in previous versions, which corresponds to their labeling in 
   Schulz & Moodie (2021, Journal of the American Statistical Association, 
   Doubly robust estimation of optimal dosing strategies 116:256-268.).
   Since version 2.4, these have been united under \code{overlap}, and
   \code{abs} and \code{wo} have been deprecated.
}
\examples{
data(twoStageCont)
# models to be passed to DTRreg
# blip model
blip.mod <- list(~ X1, ~ X2)
# treatment model (correctly specified)
treat.mod <- list(A1 ~ X1, A2 ~ 1)
# treatment-free model (incorrectly specified)
tf.mod <- list(~ X1, ~ X2)

# perform G-estimation
mod1 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
              data = twoStageCont, method = "gest")
              
mod1
}
\references{
Chakraborty, B., Moodie, E. E. M. (2013) \emph{Statistical Methods for 
  Dynamic Treatment Regimes}. New York: Springer.
  
  Robins, J. M. (2004) \emph{Optimal structural nested models for optimal 
  sequential decisions}. In Proceedings of the Second Seattle Symposium on 
  Biostatistics, D. Y. Lin and P. J. Heagerty (eds), 189--326. New York: 
  Springer.
  
  Wallace, M. P., Moodie, E. E. M. (2015) Doubly-Robust Dynamic Treatment 
  Regimen Estimation Via Weighted Least Squares. \emph{Biometrics} 
  \bold{71}(3), 636--644 (doi:10.1111/biom.12306.)
  
  Simoneau, G., Moodie, E. E. M., Nijjar, J. S., and Platt, R. W. (2020)
  Finite Sample Variance Estimation for Optimal Dynamic Treatment
  Regimes of Survival Outcomes. \emph{Statistics in Medicine} \bold{39},
  4466-4479.
  
  Efron, B., and Tibshirani, R. (1986)
  Bootstrap Methods for Standard Errors, Confidence Intervals, and Other 
  Measures of Statistical Accuracy \emph{Source: Statistical Science} \bold{1},
  54-75.

  Schulz, J. and Moodie, E.E.M. (2021) 
  Doubly robust estimation of optimal dosing strategies. 
  \emph{Journal of the American Statistical Association}, \bold{116}, 
  256–268.
  
  Li, F. and Li, F. (2019) 
  Propensity score weighting for causal inference with multiple treatments. 
  \emph{The Annals of Applied Statistics}, \bold{13}(4), 
  2389–2415.
}
\author{
Michael Wallace

Shannon T. Holloway
}
\concept{adaptive treatment strategies}
\concept{dynamic treatment regimens}
\concept{dynamic weighted ordinary least squares}
\concept{g-estimation}
\concept{personalized medicine}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/chooseM.R
\name{chooseM}
\alias{chooseM}
\title{Adaptive Choice of the Bootstrap Resample Size M}
\usage{
chooseM(
  outcome,
  blip.mod,
  treat.mod,
  tf.mod,
  data = NULL,
  method = c("gest", "dwols", "qlearn"),
  treat.type = c("bin", "multi", "cont"),
  treat.fam = gaussian(link = "identity"),
  weight = c("overlap", "ipw", "cipw", "qpom", "none", "manual"),
  n.bins = 3L,
  treat.wgt.man = NULL,
  treat.range = NULL,
  missing = c("drop", "ipw"),
  missing.mod = NULL,
  B1 = 500,
  B2 = 500
)
}
\arguments{
\item{outcome}{The outcome variable. Missing data will result in a stopping
error.}

\item{blip.mod}{A list of formula objects specifying covariates of the
blip function for each stage in order. No dependent variable should be 
specified. Note that this input should include the treatment variable 
ONLY if the blip model is quadratic in treatment. See Details for further 
clarification.}

\item{treat.mod}{A list of formula objects specifying the treatment model for
each stage in order. Treatment variable should be included as the dependent
variable. If treatment is binary \code{\link[stats]{glm}(family = binomial)}
will be 
used to obtain parameter estimates; if treatment is multi-nomial, 
\code{\link[nnet]{multinom}()} will be used to obtain parameter estimates; and if 
treatment is continuous, \code{\link[stats]{glm}()} with the family and
link specified by \code{treat.fam} will be used to obtain parameter estimates.}

\item{tf.mod}{A list of formula objects specifying covariates of the
treatment-free model for each stage in order. No dependent variable should 
be specified.}

\item{data}{A data frame containing all necessary covariates and treatments
contained in the models. Missing data should be coded as \code{NA}.}

\item{method}{The DTR method to be used, choose "dwols" for dynamic WOLS, 
"gest" for G-estimation, or "qlearn" for Q-learning.}

\item{treat.type}{A character object. Must be one of \{"bin", "multi", "cont"\}
indicating that the treatments at each stage are binary, multinomial,
or continuous, respectively. Each stage must have the same treatment type.}

\item{treat.fam}{A character or family object. 
The description of the dose distribution along with the link 
function to be used in the treatment model for computing weights; should be 
specified in a similar format as that used in \code{\link[stats]{glm}()}. 
If character object, must be one of \{"gaussian", "Gamma"\}, for which
\code{\link[stats]{gaussian}(link = "identity")} or 
\code{\link[stats]{Gamma}(link = "log")} 
will be used,
respectively. Input is ignored for \code{treat.type = "bin"} and 
\code{treat.type = "multi"}. If \code{NULL} and \code{treat.type = "cont"},
defaults to \code{gaussian(link = "identity")}.}

\item{weight}{The form of the treatment weight. See details.}

\item{n.bins}{An integer object. The number of bins (levels) to be used for 
categorizing continuous doses. This input is required only when
\code{treat.type = "cont"} and \code{weight = "overlap"} or \code{weight = "qpom"}.}

\item{treat.wgt.man}{NULL or a list of vectors of known treatment weights can be 
specified to be used instead of hard-coded treatment weight options.
The \eqn{i^{th}}{ith} element of the list contains the multiplicative weights 
for the \eqn{i^{th}}{ith} stage. Each vector must be of length \eqn{n}{n}, 
the number of participants. Used only for \code{method = "dwols"}.}

\item{treat.range}{For continuous treatments. Specify the maximum/minimum 
value that treatments can be take. If unspecified then the minimum/maximum 
value of observed treatments is used. If you wish to have unrestricted 
treatments set this option to \code{c(-Inf, +Inf)}. If each stage has its own
range, provide as a list, the ith element providing the min and max
for the ith stage treatment.}

\item{missing}{A character object. Must be one of \{"drop", "ipw"\}.
If set to "ipw" and covariate or treatment data are missing then inverse 
probability 
weights are used. The complete case probability is estimated 
via logistic regression. If set to "drop" and data are missing, participants
with missing data are dropped for all stage analyses.}

\item{missing.mod}{An optional list of formula objects specifying the model
for the inverse probability of weights for each stage in order.
No dependent variable should be specified. If \code{missing = "ipw"} and 
\code{missing.mod = NULL}, then the models are assumed to be linear comprising
the full covariate history derived from all of the previous stage models.}

\item{B1}{Number of first-level bootstrap resamples.}

\item{B2}{Number of second-level bootstrap resamples.}
}
\value{
A list with a single element 
  \item{m }{Resample size for using in the m-out-of-n bootstrap.}
}
\description{
Implementation of a double-bootstrap algorithm for choosing the bootstrap 
  resample size \eqn{m}{m} in a data-adaptive manner. The function returns the 
  resample size to be used to apply the m-out-of-n bootstrap with \link{DTRreg}.
}
\details{
The m-out-of-n bootstrap is an adequate tool for constructing valid 
  confidence intervals for the first stage parameters in \link{DTRreg}. The 
  resample size \eqn{m}{m} is: 
  \eqn{m = n^{\frac{1 + \alpha(1-\hat{p})}{1+\alpha}}}{%
  m = n^[1+alpha(1-pHat)/(1+alpha)]}. The estimated non-regularity level is 
  computed by \link{DTRreg}. The double-bootstrap algorithm is a cross-validation 
  tool for choosing the tuning parameter \eqn{\alpha}{alpha} in a data-driven way.
 
The current implementation is valid for a two-stage DTR. Moreover, the 
  current implementation may be unstable when there are many missing data.
}
\examples{
data(twoStageCont)

# models to be passed to DTRreg
# blip model
blip.mod <- list(~ X1, ~ X2)
# treatment model (correctly specified)
treat.mod <- list(A1 ~ X1, A2 ~ 1)
# treatment-free model (incorrectly specified)
tf.mod <- list(~ X1, ~ X2)

# perform dWOLS without calculating confidence intervals
mod1 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
               data = twoStageCont, method = "dwols")
  
# choose m adaptively for that model
\dontrun{
  m <- chooseM(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
               data = twoStageCont, method = "dwols",
               B1 = 200, B2 = 200)$m
}
m <- 94
  
# dWOLS with confidence intervals from the m-out-of-n bootstrap
mod2 <- DTRreg(twoStageCont$Y, blip.mod, treat.mod, tf.mod, 
               data = twoStageCont, method = "dwols",
               var.estim = "bootstrap", 
               bootstrap.controls = list(M = m))
}
\references{
Chakraborty, B., Moodie, E. E. M. (2013) \emph{Statistical Methods for 
  Dynamic Treatment Regimes}. New York: Springer.
  
  Efron B., Tibshirani R. J. (1994) An Introduction to the Bootstrap. 
  \emph{CRC press}.
  
  Wallace, M. P., Moodie, E. M. (2015) Doubly-Robust Dynamic Treatment 
  Regimen Estimation Via Weighted Least Squares. \emph{Biometrics} 
  \bold{71}(3), 636--644 (doi:10.1111/biom.12306.)
% Eventually update with reference to my paper
}
\author{
Gabrielle Simoneau
}
\concept{adaptive treatment strategies}
\concept{double bootstrap}
\concept{dynamic treatment regimens}
\concept{dynamic weighted ordinary least squares}
\concept{g-estimation}
\concept{m-out-of-n bootstrap}
\concept{personalized medicine}
